(*Copyright (C) 2016, Daniel Wyckoff*)

(*This file is part of BooleanAlgebrasIntro2.

BooleanAlgebrasIntro2 is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

BooleanAlgebrasIntro2 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BooleanAlgebrasIntro2.  If not, see <http://www.gnu.org/licenses/>.*)


Require Import FunctionProperties.
Require Import SetUtilities.
Require Import SetUtilities2.
Require Import LogicUtilities.
Require Import BoolAlgBasics.
Require Import FiniteOperations.
Require Import FunctionProperties.
Require Import TypeUtilities.
Require Import Homomorphisms.
Require Import FieldsOfSets.
Require Import ArithUtilities.
Require Import Even.
Require Import Subalgebras.
Require Import DecidableDec.
Require Import ExtensionsOfHomomorphisms.
Require Import FiniteMaps.
Require Import ProofIrrelevance.
Require Import Description.
Require Import SetUtilities2_5.


Definition atomless (B:Bool_Alg) : Prop := 
  forall x:bt B, ~ atom x.


Definition atomless_p {T:Type} (Bp:Bool_Alg_p T) : Prop :=
  forall x:btp Bp, ~atom_p x.

Lemma atomless_p_iff : 
  forall {T:Type} (Bp:Bool_Alg_p T),
    atomless_p Bp <-> atomless (ba_conv Bp).
intros T Bp. 
split; intro h1; red in h1; red.
intros x h2. specialize (h1 _ h2). assumption.
intros x h2.
rewrite atom_p_iff in h2.
specialize (h1 _ h2). assumption.
Qed.


Lemma degen_atomless :
  forall (B:Bool_Alg),
    degen B -> atomless B.
intros B h1.
red. intros x h2. red in h2.
destruct h2 as [h2 h3].
pose proof (degen_eq _ h1 x 0).
subst.
contradict h2. reflexivity.
Qed.


Lemma atomless_impl_infinite : 
  forall (B:Bool_Alg),
    ~degen B ->
    atomless B -> ~finite_ba B.
intros B h0 h1 h2.
assert (h3:forall x:bt B, x <> 0 -> exists y:bt B, lt 0 y /\ lt y x).
  intros x h3. 
  red in h1. specialize (h1 x). 
  apply NNPP.
  intro h4.
  contradict h1.
  pose proof (not_ex_all_not _ (fun y => lt 0 y /\ lt y x) h4) as h5.
  simpl in h5.
  red. split. assumption.
  intros b h6. 
  specialize (h5 b).
  apply not_and_or in h5.
  destruct h5 as [h5 | h7].
  destruct (zero_lt_or _ b) as [h7 | h8].
  right. assumption.
  contradiction.
  left.
  unfold lt in h7.
  apply not_and_or in h7.
  destruct h7 as [h8 | h9].
  contradiction.
  apply NNPP.
  intro h10. 
  apply neq_sym in h10.
  contradiction.  
assert (h3':forall x:bt_pos B, exists y:bt_pos B, lt (proj1_sig y) (proj1_sig x)).
  intro x.
  destruct x as [x h4].
  pose proof (iff2 (neq_zero_lt_iff _ x) h4) as h5.
  apply h3 in h5.
  destruct h5 as [y h5]. destruct h5 as [h5 h6].
  exists (exist _ _ h5). simpl. assumption.
assert (h3'':forall x:sig_set (ba_pos B), 
             exists y:sig_set (ba_pos B), lt (proj1_sig y) (proj1_sig x)).
  intro x.
  destruct x as [x h4]. destruct h4 as [h4]. simpl.
  pose proof (h3' (exist _ _ h4)) as h5.
  destruct h5 as [y h5].
  destruct y as [y h6].
  assert (h7:In (ba_pos B) y). constructor; auto.
  exists (exist _ _ h7). simpl. simpl in h5.
  assumption. 
red in h2.
pose proof (finite_ba_pos B h2) as hf.
assert (hf':Finite (full_sig (ba_pos B))).
  rewrite <- finite_full_sig_iff. assumption.
pose proof (bij_seg _ hf) as h4. 
destruct h4 as [g h4].
assert (h5:Inhabited (ba_pos B)).
  apply Inhabited_intro with 1.
  constructor. red. split. apply zero_min.
  contradict h0.
  red. assumption.
assert (h5':Inhabited (full_sig (ba_pos B))).
  destruct h5 as [x h5]. apply Inhabited_intro with (exist _ _ h5).
  constructor.
pose (fun x:sig_set (ba_pos B) => finite_constructive_definite_description _ hf g h4  (fun y =>lt (proj1_sig y) (proj1_sig x)) (h3'' x)) as psi.
pose (fun x:sig_set (ba_pos B) => proj1_sig (psi x)) as psi'.
assert (h6:Strict_Order (exist_rel (lt (B:=B)) (fun a : bt B => In (ba_pos B) a))).
  apply strict_order_exist_rel. apply strict_order_lt.
assert (h7:(forall x : sig_set (ba_pos B),
        exist_rel (lt (B:=B)) (fun a : bt B => In (ba_pos B) a) (psi' x) x)).
  intro x. red. 
  unfold psi', psi.
  pose proof (proj2_sig (finite_constructive_definite_description 
              (ba_pos B) hf g h4
              (fun y : sig_set (ba_pos B) => lt (proj1_sig y) (proj1_sig x))
              (h3'' x))) as h7.
  simpl in h7.
  assumption.
pose proof (strict_order_decreasing_uo_infinite h5' psi' (exist_rel (@lt B) (fun a=>In (ba_pos B) a)) h6 h7) as h8.
contradict h8.
rewrite <- (finite_full_sig_iff (ba_pos B)).
assumption.
Qed.

Lemma atomless_impl_infinite_p : 
  forall {T:Type} (Bp:Bool_Alg_p T),
    ~degen_p Bp ->
    atomless_p Bp -> ~finite_ba_p Bp.
intros T Bp h1 h2.
rewrite degen_p_iff in h1.
rewrite atomless_p_iff in h2.
pose proof (atomless_impl_infinite _ h1 h2) as h3.
rewrite finite_ba_p_iff.
assumption.
Qed.



Lemma below_non_zero_atomless_ex :
  forall (B:Bool_Alg),
    atomless B ->
    forall (p:bt B),
      p <> 0 ->
      exists a : bt B,
        a <> 0 /\ lt a p.
  intros B h1 x h3.
  red in h1. specialize (h1 x). 
  apply NNPP.
  intro h4.
  contradict h1.
  pose proof (not_ex_all_not _ (fun y => y <> 0 /\ lt y x) h4) as h5.
  simpl in h5.
  red. split. assumption.
  intros b h6.
  specialize (h5 b).
  apply not_and_or in h5.
  destruct h5 as [h5 | h7].
  destruct (zero_lt_or _ b) as [h7 | h8].
  right. assumption.
  left.
  assert (h9 : b = 0). tauto.
  subst.
  red in h8. destruct h8 as [h8 h9]. contradict h9. reflexivity.
  unfold lt in h7.
  apply not_and_or in h7.
  destruct h7 as [h8 | h9].
  contradiction.
  apply NNPP.
  intro h10. 
  apply not_or_and in h10.
  destruct h10 as [h10 h11].
  apply neq_sym in h10.
  contradiction.
Qed.

Lemma below_non_zero_atomless_ex_p :
  forall {T:Type} (Bp:Bool_Alg_p T),
    atomless_p Bp ->
    forall (p:btp Bp),
      p <> %0 ->
      exists a : btp Bp,
        a <> %0 /\ lt_p a p.
  intros T Bp h1 p h2.
  rewrite atomless_p_iff in h1.
  rewrite ba_conv_zero in h2.
  pose proof (below_non_zero_atomless_ex _ h1 p h2) as h3.
  destruct h3 as [a h3].
  exists a; auto.
Qed.


Definition countable_below_non_zero_atomless
           {B:Bool_Alg} (S:Ensemble nat)
           (pfinf: ~Finite S)
           (f:sig_set S -> bt B)
           (pfb:bijective f) (pfa:atomless B) (p:bt B)
           (pfnz:p <> 0) : bt B.
  assert (h1 : exists x : sig_set (Full_set (bt B)), (proj1_sig x) <> 0 /\ lt (proj1_sig x) p).
    pose proof (below_non_zero_atomless_ex B pfa p pfnz) as h1.
    destruct h1 as [x h2].
    exists (exist _ _ (Full_intro _ x)).
    simpl. assumption.
  refine (proj1_sig (proj1_sig (countable_set_constructive_definite_description (full_wrap_fun2 f) _ pfinf (iff1 (bij_full_wrap_fun2_iff f) pfb) h1))).
Defined.


Definition countable_below_non_zero_atomless_p
           {T:Type} {Bp:Bool_Alg_p T} (S:Ensemble nat)
           (pfinf: ~Finite S)
           (f:sig_set S -> btp Bp)
           (pfb:bijective f) (pfa:atomless_p Bp) (p:btp Bp)
           (pfnz:p <> %0) : btp Bp :=
  countable_below_non_zero_atomless S pfinf (ba_conv_fun2 f) pfb pfa p pfnz.



Lemma countable_below_non_zero_atomless_compat : 
  forall {B:Bool_Alg} (S:Ensemble nat)
         (pfinf: ~Finite S)
         (f:sig_set S -> bt B)
         (pfb:bijective f) (pfa:atomless B) (p:bt B)
         (pfnz:p <> 0), 
    let a := countable_below_non_zero_atomless S pfinf f pfb pfa p pfnz in
    a <> 0 /\ lt a p.
  unfold countable_below_non_zero_atomless.
  intros B S pfinf f pfb pfa p pfnz. 
  pose proof (proj2_sig   (countable_set_constructive_definite_description 
           (full_wrap_fun2 f)
           (fun x : sig_set (Full_set (bt B)) =>
            proj1_sig x <> 0 /\ lt (proj1_sig x) p) pfinf
           (iff1 (bij_full_wrap_fun2_iff f) pfb)
           match below_non_zero_atomless_ex B pfa p pfnz with
           | ex_intro x h2 =>
               ex_intro
                 (fun x0 : sig_set (Full_set (bt B)) =>
                  proj1_sig x0 <> 0 /\ lt (proj1_sig x0) p)
                 (exist (In (Full_set (bt B))) x (Full_intro (bt B) x)) h2
           end)) as h1.
  destruct h1; split; auto.
Qed.

Lemma countable_below_non_zero_atomless_compat_p : 
  forall {T:Type} {Bp:Bool_Alg_p T} (S:Ensemble nat)
         (pfinf: ~Finite S)
         (f:sig_set S -> btp Bp)
         (pfb:bijective f) (pfa:atomless_p Bp) (p:btp Bp)
         (pfnz:p <> %0), 
    let a := countable_below_non_zero_atomless_p S pfinf f pfb pfa p pfnz in
    a <> %0 /\ lt_p a p. 
  unfold countable_below_non_zero_atomless_p,
         countable_below_non_zero_atomless.
  intros B S pfinf f pfb pfa pfnz p h1.
  pose proof (proj2_sig  (countable_set_constructive_definite_description
           (full_wrap_fun2 (ba_conv_fun2 pfb))
           (fun x : sig_set (Full_set (bt (ba_conv S))) =>
            proj1_sig x <> 0 /\ lt (proj1_sig x) p) f
           (iff1 (bij_full_wrap_fun2_iff (ba_conv_fun2 pfb)) pfa)
           match below_non_zero_atomless_ex (ba_conv S) pfnz p h1 with
           | ex_intro x h2 =>
               ex_intro
                 (fun x0 : sig_set (Full_set (bt (ba_conv S))) =>
                  proj1_sig x0 <> 0 /\ lt (proj1_sig x0) p)
                 (exist (In (Full_set (bt (ba_conv S)))) x
                    (Full_intro (bt (ba_conv S)) x)) h2
           end)) as h2.
  destruct h2; split; auto.
Qed.


Lemma atomless_int_alg_ba : 
  atomless int_alg_ba.
red.
intros x h1.   
red in h1. 
destruct h1 as [h1 h2]. 
destruct x as [S h3].
assert (h4:S <> Empty_set _).
  intro; subst.
  contradict h1.
  apply proj1_sig_injective; simpl.
  reflexivity. 
apply not_empty_Inhabited in h4. 
pose proof (union_of_ho_invls_ex_strict_included _ h4 h3) as h5. 
destruct h5 as [S' [h5 [h6 h7]]].  
pose proof (h2 (exist _ _ h6)) as h2'.
assert 
  (h8:Strict_Included 
        (proj1_sig 
           (exist (fun S0 : Ensemble QArith_base.Q => union_of_ho_invls S0) S' h6))
         (proj1_sig (exist (fun S0 : Ensemble QArith_base.Q => union_of_ho_invls S0) S h3))).
  simpl. assumption.  
rewrite <- int_alg_ba_lt_iff in h8.
destruct h8 as [h8 h9].
apply h2 in h8.
destruct h8 as [h8 | h8].
pose proof (f_equal (@proj1_sig _ _) h8) as h10.
simpl in h10. subst.
contradict h9. 
apply proj1_sig_injective; auto.
simpl in h8.
pose proof (f_equal (@proj1_sig _ _) h8) as h10. 
simpl in h10.
subst.
assert (h10: le (B:=int_alg_ba)
         (exist (fun S0 : Ensemble QArith_base.Q => union_of_ho_invls S0)
            (Empty_set QArith_base.Q) h6)
         (exist (fun S0 : Ensemble QArith_base.Q => union_of_ho_invls S0) S
            h3)).
  pose proof (zero_min (exist _ _ h3) (B:=int_alg_ba)) as h11.
  simpl in h11.
  pose proof (proof_irrelevance _ h6 (union_of_ho_invls_empty)); subst; auto.
specialize (h2 (exist _ _ h6) h10).
destruct h2 as [h2 | h2].
pose proof (f_equal (@proj1_sig _ _) h2) as h11.
simpl in h11. subst.
contradict h9. 
apply proj1_sig_injective; auto.
apply Inhabited_not_empty in h7.
contradict h7.
reflexivity.
Qed.

