(* Copyright (C) 2018, Daniel Wyckoff *)
(*This file is part of BooleanAlgebrasIntro2.

BooleanAlgebrasIntro2 is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

BooleanAlgebrasIntro2 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BooleanAlgebrasIntro2.  If not, see <http://www.gnu.org/licenses/>.*)

(*Version 0.3 Alpha bool2 -- certified -- Coq 8.4 pl4.*)

(* This brief experimental file uses some definitions 
from Chapter 2 in Part II of Dietlinde Lau's "Function Algebras."  
It's not part of the core development of bool2.*)

(*Relation on E_k := {0, . . . . k - 1} of arity [h]*) 
Require Import NaryFunctions.
Require Import TypeUtilities.
Require Import SetUtilities.
Require Import FunctionAlgebrasBasics.

Definition erelation h k : Type :=
  Ensemble ({m | m < k}^h).

Lemma finite_erelation :
  forall {h k} (R:erelation h k),
    Finite R.
  intros.
  eapply Finite_downward_closed.
  rewrite Finite_FiniteT_iff.
  apply finite_nprod.
  apply finite_nat_initial_segment.
  apply incl_full.
Qed.




(*This is an "elementary relation," (of arity [h]),
 which represents a full abstraction 
of the standard diagonal relation on a segment type.*)
Inductive edelta
          {k} (h:nat)
          (R:Relation {m | m < k}) : erelation h k :=
| edelta_intro : forall (x:{m | m < k}^h),
                   (forall i j
                           (pfik:i < k)
                           (pfjk:j < k),
                      i < h ->
                      j < h ->
                      R (exist (fun m => m < k) _ pfjk)
                        (exist (fun m => m < k) _ pfjk) ->
                      x __ i = 
                      x __ j) ->
                   Inn (edelta h R) x.

(*This is the set of all [edelta]s parameterized by 
equivalence relations.*)
Inductive D'_ (h k:nat) :
  Ensemble (erelation h k) :=
| D'__intro : forall {R},
               Equivalence _ R ->
               Inn (D'_ h k) (edelta h R).

(*The set of all [D'_]s across all arities, including 0*)
Inductive D' (k:nat) : Ensemble {h:nat & (Ensemble (erelation h k))} :=
  D'_intro : 
    forall h,
      Inn (D' k) (existT _ h (D'_ h k)).



