(* Copyright (C) 2016 Daniel Wyckoff *)
(*This file is part of BooleanAlgebrasIntro2.

BooleanAlgebrasIntro2 is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

BooleanAlgebrasIntro2 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BooleanAlgebrasIntro2.  If not, see <http://www.gnu.org/licenses/>.*)
   
Require Import SetUtilities.
Require Import SetUtilities1_5.
Require Import SetUtilities2.
Require Import ListUtilities.
Require Import LogicUtilities.
Require Import ArithUtilities.
Require Import FiniteBags.
Require Import DecidableDec.
Require Import Description.
Require Import FunctionalExtensionality.
Require Import FunctionProperties.
Require Import FunctionProperties2.
Require Import TypeUtilities.
Require Import EqDec.
Require Import FiniteMaps.
Require Import NPeano.

(*Version 0.3 Alpha bool2 -- certified -- Version 8.4 pl4 Coq*)

(*[apply proof_irrelevance]-certified.*)   
(*"Conve"rt 'f'irstn*)
Definition conv_fn {T:Type} {l:list T} {m:(length l) @}
           (r:(S (proj1_sig m)) @) :
  (length (firstn (S (proj1_sig m)) l)) @ :=
  segT_transfer (eq_sym (length_firstn_S _ _ (proj2_sig m))) r.


Lemma subtract_remove_ens_compat :
  forall {T:Type} (pfd:type_eq_dec T) (l:list (Ensemble T))
         (B:Ensemble T)  (pfb:Finite B)
         (pfp:p_mem_l (@Finite _) l),
    Subtract (list_to_set l) B =
    list_to_set (remove_dep (@Finite _) (fin_ens_eq_dec T pfd) l B pfb pfp).
  intros T hdt l. 
  induction l as [|A l ih]; simpl.
  intros; apply sub_empty.
  intros B h1 h2.
  destruct fin_ens_eq_dec as [h3 | h3]; subst.
  rewrite sub_add_eq.
  apply ih.
  simpl.
  rewrite sub_add_comm.
  specialize (ih B h1).
  hfirst ih.
    apply (p_mem_l_cons _ _  _ h2); auto.
  specialize (ih hf).
  rewrite ih.
  f_equal.
  f_equal.
  f_equal.
  apply proof_irrelevance.
  apply neq_sym; auto.
Qed.


Lemma plus_bag_nat_im_linds :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    plus_bag_nat
      (im_set_bag
         nat_eq_dec (list_to_set (linds pfdt l))
         (list_to_set_finite _)
         (@length _)) = length l.
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  unfold linds.
  simpl.
  pose proof (im_set_bag_empty nat_eq_dec (@length T)) as h2.
  erewrite im_set_bag_functional2.
  rewrite im_set_bag_empty.
  apply plus_bag_nat_empty.
  reflexivity.
  reflexivity.
  pose proof (list_to_set_sort_list _ lt_list_nat_sort_dec (linds h1 (a::l))) as h2.
  simpl in h2.
  rewrite (im_set_bag_functional1 nat_eq_dec _ _ (list_to_set_finite _) (list_to_set_finite _) (@length _) h2).
  fold (sorted_linds h1 (a::l)).
  destruct (in_dec h1 a l) as [h3 | h3]. 
  rewrite sorted_linds_cons_in.
  simpl.
  pose proof (im_set_bag_add nat_eq_dec (list_to_set
              (famlS
                 (remove list_nat_eq_dec (linds_occ h1 l a)
                         (sorted_linds h1 l)))) (list_to_set_finite _) (0 :: lS (linds_occ h1 l a))) as h4.
  hfirst h4.
  intro h5.
  rewrite <- list_to_set_in_iff in h5.
  rewrite in_famlS_iff in h5.
  destruct h5 as [h5 h6].
  destruct h6 as [h6 h7].
  pose proof (in_eq 0 (lS (linds_occ h1 l a))) as h8.
  rewrite h7 in h8.
  rewrite in_lS_iff in h8.
  destruct h8 as [? [? ?]]; discriminate. 
  specialize (h4 hf (@length nat)).
  erewrite im_set_bag_functional2.
  rewrite h4 at 1.
  rewrite plus_bag_nat_add. 
  simpl.
  rewrite length_lS.
  pose proof (linds_decompose_remove h1 _ _ h3) as h5.
  pose proof (list_to_set_sort_list _ lt_list_nat_sort_dec (linds h1 l)) as h6.
  simpl in h6.
  unfold sorted_linds in h5.
  rewrite h5 in h6.
  rewrite lt_list_nat_tso_dec_eq in h6.
  rewrite <- list_to_set_sort_list in h6.
  simpl in h6.
  rewrite (im_set_bag_functional1 _ _ _ _ (Add_preserves_Finite _ _ _ (list_to_set_finite _)) (@length _)  h6) in ih.
  rewrite im_set_bag_add, plus_bag_nat_add in ih.
  unfold linds_remove in ih.
  rewrite plus_S_shift_r.
  f_equal.
  rewrite <- ih.
  f_equal.
  symmetry.
  apply plus_bag_nat_im_set_bag_eq_f'.
  exists lS.
  split.
  apply inj_lS.
  split.
  unfold famlS.
  rewrite map_im_compat.
  f_equal.
  unfold sorted_linds.
  rewrite <- subtract_remove_compat.
  rewrite <- subtract_remove_compat.
  f_equal.
  rewrite <- list_to_set_sort_list; auto.
  intros l' h7.
  rewrite length_lS; auto.
  rewrite <- list_to_set_in_iff.
  apply nin_linds_remove.
  assumption.
  reflexivity.
  reflexivity.
  assumption.

  rewrite sorted_linds_cons_nin; auto.
  simpl.
  rewrite im_set_bag_add'.
  rewrite <- ih.
  simpl.
  rewrite plus_bag_nat_add.
  rewrite S_compat.
  f_equal.
  symmetry.
  apply plus_bag_nat_im_set_bag_eq_f'.
  exists lS.
  split.
  apply inj_lS.
  split.
  unfold famlS.
  rewrite map_im_compat.
  f_equal.
  unfold sorted_linds.
  rewrite <- list_to_set_sort_list; auto.
  intros x h4.
  rewrite length_lS; auto.
  rewrite <- list_to_set_in_iff.
  intro h4.
  rewrite in_famlS_iff in h4.
  destruct h4 as [ln [h4 h5]].
  pose proof (in_eq 0 nil) as h6.
  rewrite h5 in h6.
  rewrite in_lS_iff in h6.
  destruct h6 as [? [? ?]]; discriminate.
Qed.

Lemma inv_dep_nil :
  forall (T:Type),
    inv_dep (fun_in_dep_nil T False).
  intros; apply bij_impl_inv_dep; apply bij_dep_nil. 
Qed.

Lemma inv_dep_nil_nil :
  forall T U (f:fun_in_dep (@nil T) U),
    inv_dep f ->
    equi U False.
  intros T U f h1.
  pose proof (inv_dep_sig f h1) as h2.
  pose proof (invertible_impl_inv_invertible _ h2) as h3.
  eapply equi_trans.
  econstructor.
  apply h3.
  apply equi_sym.
  apply equi_sig_false.
Qed.


Lemma inv_from_cons_in :
  forall {T U:Type} {l:list T} {a:T} (f:fun_in_dep (a::l) U),
    In a l ->
    inv_dep f ->
    inv_dep (fun_from_cons f).
  intros ? ? ? ? ? ? h1;
  apply inv_impl_bij_dep in h1;
  apply bij_impl_inv_dep;
  apply bij_from_cons_in; auto.
Qed.


Lemma inv_from_cons_ran :
  forall {T U:Type} {l:list T} {a:T} (f:fun_in_dep (a::l) U)
         (pfnin : ~In a l)
         (pfinv:inv_dep f),
    let pfinj := inv_impl_inj_dep _ pfinv in
    inv_dep (fun_from_cons_ran f pfinj pfnin).
  intros;
  apply bij_impl_inv_dep.
  apply inj_surj_bij_from_cons_ran.
  apply inv_impl_surj_dep; auto.
Qed.

Lemma inv_tl_ran :
  forall {T U} {l:list T}
         (f:fun_in_dep l U)
         (pfn:l <> nil)
         (pfinj:inj_dep f)
         (pfnin : ~In (hd' l pfn) (tl l)),
    inv_dep f ->
    inv_dep (tl_in_dep_ran f pfn pfinj pfnin).
  intros;
  apply bij_impl_inv_dep.
  apply bij_tl_ran.
  apply inv_impl_bij_dep; auto.
Qed.


      
Lemma card_fun1_family_union_linds :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    card_fun1 (FamilyUnion (faml_to_fam (linds pfdt l))) = length l.
  intros T hdt l.
  pose proof (partition_linds hdt l) as h1.
  do 2 red in h1.
  destruct h1 as [h1 h2].
  pose proof (card_pairwise_disjoint (faml_to_fam (linds hdt l)) (finite_faml_to_fam _)) as h3.
  hfirst h3.
  intros S h4.
  apply in_faml_to_fam_impl in h4.
  destruct h4 as [l' [h4 h5]]; subst.
  apply list_to_set_finite.
  specialize (h3 hf h1). 
  rewrite h3.
  pose proof (plus_bag_nat_im_linds hdt l) as h4.
  rewrite <- h4.
  symmetry.
  apply plus_bag_nat_im_set_bag_eq_f_hetero_dep.
  exists (fun l pf => list_to_set l).
  split.
  red.
  intros lm ln h5 h6 h7.
  rewrite <- list_to_set_in_iff in h5, h6.
  eapply in_linds_list_to_set_injable.
  apply h5.
  assumption.
  assumption.
  split.
  apply Extensionality_Ensembles; red; split; red; intros S h5.
  apply in_faml_to_fam_impl in h5.
  destruct h5 as [ln [h5 h6]]; subst.
  econstructor.
  rewrite <- list_to_set_in_iff.
  apply h5.
  reflexivity.
  destruct h5 as [S h5]; subst.
  rewrite <- list_to_set_in_iff in h5.
  apply in_faml_impl; auto.

  intros ln h5. 
  erewrite card_fun1_length_eq.
  rewrite <- list_to_set_in_iff in h5.
  apply no_dup_faml_linds in h5.
  rewrite (list_singularize_no_dup nat_eq_dec) in h5.
  rewrite h5 at 1; auto.
Qed.


Lemma sorted_linds_eq_length :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l':list U),
    sorted_linds pfdt l =
    sorted_linds pfdu l' ->
    length l =
    length l'.
  intros T U h1 h2 l l' h3.
  pose proof (partition_linds h1 l) as h4. 
  do 2 red in h4.
  destruct h4 as [h4 h5].
  pose proof h5 as h5'.
  rewrite <- faml_to_fam_sorted_linds in h5'.
  rewrite h3 in h5'.
  pose proof (f_equal card_fun1 h5') as h6.
  rewrite list_to_set_seg_list_eq, faml_to_fam_sorted_linds,
  card_fun1_seg_eq,
  card_fun1_family_union_linds in h6; auto.
Qed.


Lemma sorted_linds_eq_iff' :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l':list U),
    (sorted_linds pfdt l =
     sorted_linds pfdu l') <->
    (exists pfe:length l = length l',
       forall (m:nat) (pfm:m < length l),
       let pfm' := lt_eq_trans _ _ _ pfm pfe in
       lec pfdt pfm =
       lec pfdu pfm').
  intros T U h1 h2 l l'; split; intro h3.
  unfold sorted_linds in h3.
  pose proof (sorted_linds_eq_length _ _ _ _ h3) as he.
  exists he. 
  intros m h4; simpl. 
  pose proof (partition_linds h1 l) as h5.
  pose proof (partition_linds h2 l') as h6.
  pose proof (in_linds_list_to_set_injable h1 l) as h9. 
  red in h9.
  apply h9.
  apply in_lec.
  rewrite <- (in_sort_list_iff _ lt_list_nat_sort_dec (linds h1 l)).

  rewrite h3, in_sort_list_iff.
  apply in_lec.
  eapply partition_faml_eq.
  apply h5.
  apply in_lec.
  rewrite <- (in_sort_list_iff _ lt_list_nat_sort_dec (linds h1 l)).
  rewrite h3, in_sort_list_iff.
  apply in_lec.
  apply rep_in_lec.
  rewrite in_lec_iff.
  reflexivity.

  destruct h3 as [h3 h4].
  pose proof (list_to_set_injable_increasingR _ lt_list_nat_sort_dec) as h5.
  red in h5.
  apply h5.
  unfold sorted_linds.
  apply weak_incrR_no_dup_impl.
  apply weak_incrR_sort_list.
  apply no_dup_sort_list.
  apply no_dup_linds.

  apply weak_incrR_no_dup_impl.
  apply weak_incrR_sort_list.
  apply no_dup_sort_list.
  apply no_dup_linds.
  apply Extensionality_Ensembles; red; split; red; intros ln h6; rewrite <- list_to_set_in_iff in h6; rewrite <- list_to_set_in_iff; unfold sorted_linds in h6; unfold sorted_linds; rewrite in_sort_list_iff in h6; rewrite in_sort_list_iff; rewrite in_linds_lec_iff in h6; rewrite in_linds_lec_iff; destruct h6 as [i h7]; destruct h7 as [h7]; subst; exists i.
  exists (lt_eq_trans _ _ _ h7 h3).
  apply h4.
  exists (lt_eq_trans _ _ _ h7 (eq_sym h3)).
  rewrite h4.
  f_equal.
  apply proof_irrelevance.
Qed.

Lemma linds_eq_iff :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l':list U),
    (linds pfdt l =
     linds pfdu l') <->
    (exists pfe:length l = length l',
       forall (m:nat) (pfm:m < length l),
       let pfm' := lt_eq_trans _ _ _ pfm pfe in
       lec pfdt pfm =
       lec pfdu pfm').
  intros; rewrite <- sorted_linds_eq_iff; apply sorted_linds_eq_iff'.
Qed.



