(* Copyright (C) 2014-2016, Daniel Wyckoff, except for the portions so labeleled as (*Schepler*) which I got from Daniel Schepler.  
My functions will be labeled (*Wyckoff*)*)
(*This file is part of BooleanAlgebrasIntro2.

BooleanAlgebrasIntro2 is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

BooleanAlgebrasIntro2 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BooleanAlgebrasIntro2.  If not, see <http://www.gnu.org/licenses/>.*)

(*Version 0.3 bool2 -- certified -- Version 8.4 pl4 Coq.*)
 
Require Export EnsemblesImplicit.
Require Export ImageImplicit.
Require Import Classical.
Require Import Image.
Require Import LogicUtilities.
Require Import Setoid.
Require Import FunctionProperties.
Require Import Arith.
Require Import Even.
Require Import Equality.
Require Import Infinite_sets.
Require Import ArithUtilities. 
Require Import Description.
Require Import DecidableDec.
Require Import Bool.
Require Import FunctionalExtensionality.
Require Import EqDec.
Require Import Basics.

(*This notation is to prevent the need to keep on writing 
[Ensembles.In] or [Ensembles.Complement] 
in other files that also import [List].*)
Notation Inn := Ensembles.In.
Notation Comp := Ensembles.Complement.

(*Wyckoff*)
Lemma prop_ext : forall (P Q : Prop), 
                   (P <-> Q) -> P = Q.
  intros P Q h1.
  pose proof (@Extensionality_Ensembles Prop (fun _ => P) (fun _ => Q)) as h2.
  hfirst h2.
  red; split; red; intros x h3; red in h3; red; tauto.
  specialize (h2 hf).
  pose proof (equal_f h2 True) as h3.
  simpl in h3; auto.
Qed.

(*Wyckoff*)
Definition pred_ext : 
  forall {T} (P Q:T->Prop),
    pred_eq P Q ->
    P = Q.
  intros; apply functional_extensionality; intro; apply prop_ext; apply H.
Qed.


(*This relatively "strong" assumption is used in a few 
places in the Boolean code explicitly since there's a vague plan 
to work my way around those arguments, but then some are indispensable, 
Overall, since this development uses LEM, this axiom doesn't really 
hurt, but merely indicates an inclination to want to eliminate some 
instances of it!*)
(*Wyckoff*)
Axiom ens_eq_dec :
  forall (T:Type), type_eq_dec (Ensemble T).


(*Wyckoff*)
Lemma incl_add : 
  forall {T:Type} (A:Ensemble T) (a:T),
    Included A (Add A a). auto with sets.
Qed.

(*Wyckoff*)
Lemma incl_full :
  forall {T} (A:Ensemble T),
    Included A (Full_set _).
  intros; red; intros; constructor.
Qed.

(*Wyckoff*)
Lemma eq_incl :
  forall {T:Type} (A B:Ensemble T),
    A = B -> Included A B.
  intros; subst; auto with sets.
Qed.
 
(*Wyckoff*)
Lemma in_eq_set : 
  forall {T:Type} (A B:Ensemble T) (pf:A = B) (x:T),
    Inn A x -> Inn B x.
intros; subst; auto.
Qed.

(*Wyckoff*)
Lemma in_eq_set_iff : 
  forall {T:Type} (A B:Ensemble T) (pf:A = B) (x:T),
    In A x <-> In B x.
intros; subst; tauto.
Qed.

(*Wyckoff*)
Lemma in_eq_elt_l :
  forall {T:Type} (A:Ensemble T) (x y:T) (pf:x = y),
    In A x -> In A y.
  intros; subst; auto.
Qed.

(*Wyckoff*)
Lemma in_eq_elt_r :
forall {T:Type} (A:Ensemble T) (x y:T) (pf:y = x),
  In A x -> In A y.
  intros; subst; auto.
Qed.
  

(*Wyckoff*)
Lemma in_add_eq : forall {T:Type} (A:Ensemble T) (x:T),
                    In A x -> Add A x = A.
intros T A x h1.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red. intros y h2.
destruct h2 as [y h2l | y h2r]. assumption.
destruct h2r; subst. assumption.
(* >= *)
red. intros y h2.
left. assumption.
Qed.

(*Wyckoff*)
Lemma eq_sing_iff :
  forall {T} A x,
    A = Singleton x <->
    (Inn A x /\
     forall a:T,
       Inn A a -> a = x).
  intros; split; subst; intros; subst.
  split.
  constructor.
  intros a h1.
  destruct h1; auto.
  destruct H as [h1 h2].
  apply Extensionality_Ensembles; red; split; red; intros c h3.
  apply h2 in h3; subst; constructor.
  destruct h3; auto.
Qed.


(*Wyckoff*)
Lemma nin_iff :
  forall {T:Type} (A:Ensemble T) (x:T),
    ~ In A x <-> Intersection A (Singleton x) = Empty_set _.
  intros T A x; split; intro h1.
  apply Extensionality_Ensembles; red; split; red; intros c h2.
  destruct h2 as [c h2 h3].
  destruct h3; contradiction.
  contradiction.
  intro h2.
  pose proof (Intersection_intro _ _ _ _ h2 (In_singleton _ _ )) as h3.
  rewrite h1 in h3; contradiction.
Qed.



(*Wyckoff*)
Lemma nin_neq_nin_add : 
  forall {T:Type} (A:Ensemble T) (a:T),
    ~Inn A a ->
  forall (x:T),
    x <> a ->
    ~Inn A x ->
    ~Inn (Add A a) x.
intros T A a h1 x h2 h3 h4.
destruct h4 as [x h4 | x h5].
contradiction.
destruct h5.
contradict h2; auto.
Qed.

(*Wyckoff*)
Lemma nin_add_or :
  forall {T:Type} (A:Ensemble T) (a x:T),
    ~In (Add A a) x ->
    a <> x /\ ~In A x.
  intros T A a x h1.
  split. intro; subst.
  contradict h1. right; auto. constructor.
  intro h2.
  contradict h1.
  left; auto.
Qed.


(*Wyckoff*)
Lemma incl_add_nin_incl : 
  forall {T:Type} (A B:Ensemble T) (b:T),
    Included A (Add B b) ->
    ~In A b ->
    Included A B.
intros T A B b h1 h3.
red. intros y h4. 
assert (h5:y <> b).
  intro h5. subst.
  contradiction.
apply h1 in h4. 
destruct h4 as [y h4 | y h6]. assumption.
destruct h6.
contradict h5.
reflexivity.
Qed.

(*Wyckoff*)
(*Uses LEM.*)
Lemma not_incl :
  forall {T:Type} (A B:Ensemble T),
    ~Included A B ->
    exists x:T,
      Inn A x /\
      ~Inn B x.
  intros T A B h1.
  apply not_all_not_ex.
  intro h2.
  contradict h1.
  red; intros x h3.
  specialize (h2 x).
  tauto.
Qed.


(*Wyckoff*)
Lemma add_preserves_inclusion : 
  forall {T:Type} (A B:Ensemble T) (x:T), 
    Included A B -> Included (Add A x) (Add B x).
intros T A  B x h1.
red. intros x' h2.
destruct h2 as [x' h2 | x' h3].
left. auto with sets. destruct h3. right. constructor.
Qed.




(*Wyckoff*)
Lemma in_add_nin_dec : 
  forall {T:Type} (A:Ensemble T) (a:T),
    ~Inn A a ->
    forall (x:T), 
      Inn (Add A a) x ->
      {Inn A x} + {x = a}.
intros T A a h1 x h2.
destruct (classic_dec (x = a)) as [h3 | h4].
right; auto.
left. 
apply NNPP.
intro h5.
pose proof (nin_neq_nin_add A a h1 x h4 h5).
contradiction.
Qed.


(*Wyckoff*)
Lemma add_not_empty : forall {T:Type} (A:Ensemble T) (x:T),
                        Add A x <> Empty_set _.
intros T A x. intro h1.
pose proof (Add_intro2 _ A x) as h2.
rewrite h1 in h2.
contradiction.
Qed.


(*Wyckoff*)
Lemma sub_add_same_nin : 
  forall {T:Type} (A:Ensemble T) (x:T),
    ~ In A x ->
    (Subtract (Add A x) x) = A.
intros T A x h1.
apply Extensionality_Ensembles.
red. split.
(* => *)
red.
intros y h2.
destruct h2 as [h2 ?].
destruct h2; try contradiction; auto.
(* <= *)
red.
intros y h2.
constructor. left; auto.
intro h3.
destruct h3; subst.
contradiction.
Qed.                          

(*Wyckoff*)
Lemma sub_add_same_in :
  forall {T:Type} (A:Ensemble T) (x:T),
    In A x ->
    (Subtract (Add A x) x) = Subtract A x.
intros T A x h1.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red.
intros y h2.
destruct h2 as [h3 h4].
destruct h3 as [y h3 | y h5].
constructor; auto. contradiction.
(* >= *)
red.
intros y h2.
destruct h2. 
constructor; try constructor; auto.
Qed.


(*Wyckoff*)
Lemma add_sub_compat_in :
  forall {T:Type} (A:Ensemble T) (x:T),
    In A x ->
    (Add (Subtract A x) x) = A.
intros T A x h1.
apply Extensionality_Ensembles.
red. split.
red. intros x' h2.
destruct h2 as [x' h2 | x' h3].
destruct h2; auto. inversion h3. subst. assumption.
red. intros x' h2.
destruct (classic (x = x')). subst. right. constructor.
left.
constructor. assumption. intro h3.
inversion h3. subst. contradict H. reflexivity.
Qed.


(*Wyckoff*)
Lemma add_nin_sub_compat :
  forall {T:Type} (A B:Ensemble T) (x:T),
    ~Inn A x ->
    Add A x = Add B x ->
    A = Subtract B x.
intros T A B x h1 h2.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red.
intros y h3.
pose proof (Add_intro1 _ A x y h3) as h4.
rewrite h2 in h4.
pose proof (Add_inv _ B x y h4) as h5.
destruct h5 as [h5l | h5r].
constructor; auto.
intro h6.
destruct h6; subst; contradiction.
subst. contradiction.
(* >= *)
red. 
intros y h3.
destruct h3 as [h3 h4].
assert (h5: x <> y).
  intro. subst.
  pose proof (In_singleton _ y). contradiction.
clear h4.
pose proof (Add_intro1 _ B x y h3) as h4.
rewrite <- h2 in h4.
destruct h4 as [y h4l | y h4r]. auto.
destruct h4r; subst. 
contradict h5. auto.
Qed.

(*Wyckoff*)
Lemma add2 : forall {T:Type} (A:Ensemble T) (x:T), 
               Add (Add A x) x = Add A x.
intros T A x.
unfold Add.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red.
intros y h1.
destruct h1 as [y h2 | y h3].
auto.
destruct h3.
apply Union_intror. constructor.
(* >= *)
red. intros y h1.
constructor. auto.
Qed.

(*Wyckoff*)
Lemma add_comml : forall {T:Type} (A:Ensemble T) (x y:T),
                   Included (Add (Add A x) y) (Add (Add A y) x).
intros T A x y.
red. 
intros z h1.
destruct h1 as [z h2|z h3].
destruct h2 as [z h4 | z h5].
left. left. auto.
destruct h5; subst.
right. constructor.
destruct h3; subst.
left. right. constructor.
Qed.

(*Wyckoff*)
Lemma add_comm : forall {T:Type} (A:Ensemble T) (x y:T),
                   Add (Add A x) y = Add (Add A y) x.
intros;
apply Extensionality_Ensembles; red; split;
apply add_comml.
Qed.

(*Wyckoff*)
Lemma union_add_comm : forall {T:Type} (A B:Ensemble T) (x:T),
                    Union (Add A x) B = Add (Union A B) x.
intros T A B x.
apply Extensionality_Ensembles.
red. split.
(* <= *) red.
intros y h1.
destruct h1 as [y h1 | y h2].
destruct h1 as [y h1 | y h3].
left. left. assumption.
destruct h3. subst. right. constructor.
left. right. assumption.
(* >= *)
red. intros y h1.
destruct h1 as [y h1 | y h2].
destruct h1 as [y h1 | y h3].
left. left. assumption.
right. assumption.
destruct h2; subst.
left. right. constructor.
Qed.

(*Wyckoff*)
Lemma add_sing_same :
  forall {T:Type} (a:T),
    Add (Singleton a) a = Singleton a.
  intros T a; apply Extensionality_Ensembles; red; split; intros x h1.
  destruct h1 as [x h1 | x h1]; auto.
  destruct h1. 
  right; auto.
  constructor.
Qed.


(*Wyckoff*)
Lemma couple_add_sing : forall {T:Type} (x y:T),
                         Couple x y = Add (Singleton x) y.
intros T x y.
apply Extensionality_Ensembles.
red. split. 
red.
intros z h1.
destruct h1. left. constructor.
right. constructor.
red.
intros z h1.
destruct h1 as [z h1l | z h1r].
destruct h1l; subst. left.
destruct h1r. right.
Qed.


(*Wyckoff*)
Lemma add_empty_sing : forall {T:Type} (x:T), 
                         Add (Empty_set _) x =
                         Singleton x.
intros T x.
apply Extensionality_Ensembles.
red. split.
red.
intros y h1.
destruct h1; try contradiction; auto.
red.
intros y h2. right; auto.
Qed.

(*Wyckoff*)
Lemma couple_add_add : forall {T:Type} (x y:T),
                         Couple x y = Add (Add (Empty_set _) x) y.
intros T x y.
rewrite couple_add_sing.
rewrite add_empty_sing.
reflexivity.
Qed.

(*Wyckoff*)
Lemma add_nin_inj2 :
  forall {T:Type} (A:Ensemble T) (a a':T),
    ~In A a ->
    Add A a = Add A a' -> a = a'.
  intros T A a a' h0 h1.
  pose proof (Add_intro2 _ A a) as h2.
  rewrite h1 in h2.
  destruct h2 as [a h2 | a h2].
  contradiction.
  destruct h2; auto.
Qed.




(*Wyckoff*)
Lemma add_nin_inj_eq :
  forall {T:Type} (A A':Ensemble T) (a:T),
    ~In A a -> ~In A' a ->
    Add A a = Add A' a -> A = A'.
intros T A A' a h0 h0' h1. 
pose proof (f_equal (fun S => Subtract S a) h1) as h2.
simpl in h2.
do 2 rewrite sub_add_same_nin in h2; auto.
Qed. 

(*Wyckoff*)
Lemma incl_add_nin_inj_eq :
  forall {T:Type} (A A':Ensemble T) (a:T),
    ~In A a -> ~In A' a ->
    Included (Add A a) (Add A' a) ->
    Included A A'.
  intros ? ? ? ? ? ? h1; red; intros.
  pose proof (Add_intro1 _  A a x H1) as h2.
  apply h1 in h2.
  destruct h2; subst; auto.
  destruct H2.
  contradiction.
Qed.


(*Wyckoff*)
Lemma not_inh_empty :
  forall (T:Type),
    ~Inhabited (Empty_set T).
  intros ? h1; destruct h1; contradiction.
Qed.

Inductive well_founded_ens {T} : Ensemble T -> Prop :=
| well_founded_ens0 :
    well_founded_ens (Empty_set _) 
| well_founded_ensS :
    forall A x,
      ~Inn A x ->
      well_founded_ens A ->
      well_founded_ens (Add A x)
| well_founded_ens_lim :
    forall A:Ensemble T,
      well_founded_ens A ->
      forall B:Ensemble T,
        well_founded_ens B ->
        well_founded_ens (Union A B).


Inductive in_closed {T} : Ensemble T -> relation T :=
| in_closed_intro :
    forall A B x y,
      Inn A x ->
      well_founded_ens A ->
      Inn B x ->
      well_founded_ens B ->
      in_closed (Union A B) x y.



Inductive const_on_of {T U} (f:T->U) (A:Ensemble T) : U -> Prop :=
  const_on_of_intro : forall c,
                     (forall x, Inn A x -> f x = c) ->
                     const_on_of f A c.

Inductive const_on {T U} (f:T->U) (A:Ensemble T) : Prop :=
  const_on_intro : forall c,
                     const_on_of f A c ->
                     const_on f A.

Lemma const_on_app_ex :
  forall {T U} (f:T->U) (A:Ensemble T),
    const_on f A ->
    Inhabited A ->
    exists! u,
      forall x:T,
        Inn A x ->
        f x = u.
  intros T U f A h1 h2.
  destruct h2 as [x h2]. 
  destruct h1 as [y h1].
  destruct h1 as [y h1].
  exists y.
  red; split; auto.
  intros y' h3.
  pose proof (h1 x h2); subst; auto.
Qed.

Definition const_on_app 
           {T U} {f:T->U} {A:Ensemble T}
           (pfc:const_on f A)
           (pfinh:Inhabited A) : U :=
  proj1_sig (constructive_definite_description
               _
               (const_on_app_ex f A pfc pfinh)).

Definition const_on_app_compat :
  forall {T U} {f:T->U} {A:Ensemble T}
         (pfc:const_on f A)
         (pfinh:Inhabited A),
    let y := const_on_app pfc pfinh in
    forall x:T,
      Inn A x ->
      f x = y.
  unfold const_on_app; intros; destruct constructive_definite_description; auto.
Qed.


      
(*Wyckoff*)
Lemma in_sub :
  forall {T:Type} (A:Ensemble T) (a x:T),
    In (Subtract A a) x -> In A x.
  intros ? ? ? ? h1.
  destruct h1; auto.
Qed.

(*Wyckoff*)
Lemma inh_subtract_empty :
  forall {T:Type} (pfdt:type_eq_dec T) (A:Ensemble T) (x:T),
    Inhabited A ->
    Subtract A x = Empty_set _ ->
    Inn A x.
  intros T h0 A x h1 h2.
  destruct h1 as [x' h1].
  destruct (h0 x x') as [h3 | h3]; subst; auto.
  pose proof (Subtract_intro _ _ _ _ h1 h3) as h4.
  rewrite h2 in h4.
  contradiction.
Qed.

  
(*Wyckoff*)
Lemma subtract_sing :
  forall {T:Type} (a:T),
    Subtract (Singleton a) a = Empty_set _.
  intros T a.
  apply Extensionality_Ensembles; red; split; auto with sets.
  red; intros x h1.
  destruct h1 as [h1 h2].
  contradict h1; auto.
Qed.

(*Wyckoff*)
Lemma subtract_sing_eq :
  forall {T:Type} (a b:T),
    a = b ->
    Subtract (Singleton a) b = Empty_set _.
  intros; subst; apply subtract_sing.
Qed.



(*Wyckoff*)
Lemma in_sing_iff :
  forall {T:Type} (a b:T),
    In (Singleton a) b <-> a = b.
  intros; split; intro h1; destruct h1; subst; try constructor; auto.
Qed.


(*Wyckoff*)
Lemma sing_not_empty :
  forall {T:Type} (a:T),
    Singleton a <> Empty_set _.
  intros T a h1.
  pose proof (In_singleton _ a) as h2.
  rewrite h1 in h2.
  contradiction.
Qed.

(*Wyckoff*)
Lemma couple_not_empty :
  forall {T:Type} (a b:T),
    Couple a b <> Empty_set _.
  intros T a b h1.
  pose proof (Couple_l _ a b) as h2.
  rewrite h1 in h2.
  contradiction.
Qed.



(*Wyckoff*)
Lemma card_sing : forall {T:Type} (x:T), cardinal _ (Singleton x) 1.
intros T x.
pose proof (add_empty_sing x) as h1.
rewrite <- h1.
constructor.
constructor.
auto with sets.
Qed.



(*Wyckoff*)
Lemma inh_sing :
  forall {T:Type} (a:T),
    Inhabited (Singleton a).
  intros T a.
  apply Inhabited_intro with a.
  constructor.
Qed.

(*Wyckoff*)
Lemma inh_add :
  forall {T:Type} (A:Ensemble T) (a:T),
    Inhabited (Add A a).
  intros; econstructor; right; constructor.
Qed.

(*Wyckoff*)
Lemma inh_couple :
  forall {T:Type} (a b:T),
    Inhabited (Couple a b).
  intros; econstructor; left; constructor.
Qed.

(*psa = power set algebra.*)
(*Wyckoff*)
Lemma abs_sum_psa  : 
  forall {T:Type} (N M:(Ensemble T)), Union N (Intersection N M) = N.
intros T N M.
apply Extensionality_Ensembles.
unfold Same_set.
split; unfold Included.
(*left*)
intros x h1.
inversion h1 as [x' h2 h3 | x'' h4 h5].
assumption.
inversion h4 as [x''' h6 h7].
assumption.
(*right*)
intros x h8.
apply Union_introl. assumption.
Qed.

(*Wyckoff*)
Lemma abs_prod_psa  : 
  forall {T:Type} (N M:(Ensemble T)), Intersection N (Union N M) = N.
intros T N M.
apply Extensionality_Ensembles.
unfold Same_set.
split; unfold Included.
(*left*)
intros x h1.
inversion h1 as [x' h2 h3 h4].
assumption.
(*right*)
intros x h5.
assert (h6:In (Union N M) x).
apply Union_introl; assumption.
apply Intersection_intro; assumption.
Qed.


(*Wyckoff*)
Lemma assoc_prod_psa : forall {T:Type} (N M P:Ensemble T),
  (Intersection N (Intersection M P)) = 
  (Intersection (Intersection N M) P).
intros T N M P.
apply Extensionality_Ensembles.
unfold Same_set.
split; unfold Included.
(*left*)
intros x h1.
inversion h1 as [x' h2 h3 h4].
inversion h3 as [x'' h5 h6 h7].
assert (h8: In (Intersection N M) x).
apply Intersection_intro; assumption.
apply Intersection_intro; assumption.
(*right*)
intros x h9.
inversion h9 as [x' h10 h11 h12].
inversion h10 as [x'' h13 h14 h15].
assert (h16: In (Intersection M P) x).
apply Intersection_intro; assumption.
apply Intersection_intro; assumption.
Qed.


(*Wyckoff*)
Lemma Intersection_associative :
  forall {T:Type} (A B C:Ensemble T),
    Intersection (Intersection A B) C = Intersection A (Intersection B C).
  intros; rewrite <- assoc_prod_psa; auto.
Qed.

  
(*Wyckoff*)
Lemma sub_add_comm : 
  forall {T:Type} (A:Ensemble T) 
         (x y:T),  (x <> y) ->
    (Subtract (Add A x) y) = (Add (Subtract A y) x).
intros T A x y h0.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red. intros z h1.
destruct h1 as [h1 h2].
destruct h1 as [h1 | z h3].
left. constructor; auto.
right; auto.
(* >= *)
red. intros z h1.
destruct h1 as [z h1 | z h2].
destruct h1 as [h1 h3].
constructor. constructor; auto. auto.
destruct h2; subst.
constructor. apply Add_intro2.
intro h1. destruct h1. 
contradict h0. auto.
Qed.

(*Wyckoff*)
Lemma sub_add_eq :
  forall {T:Type} (A:Ensemble T) (a:T),
    (Subtract (Add A a) a) = Subtract A a.
  intros T A a.
  apply Extensionality_Ensembles; red; split; red; intros x h1.
  destruct h1 as [h1 h2].
  rewrite in_sing_iff in h2.
  destruct h1 as [x h1 | x h1].
  constructor; auto.
  rewrite in_sing_iff; auto.
  destruct h1.
  contradict h2; auto.
  destruct h1 as [h1 h2].
  rewrite in_sing_iff in h2.
  constructor.
  left; auto.
  rewrite in_sing_iff.
  assumption.
Qed.

(*Wyckoff*)
Lemma add_sub_eq :
  forall {T:Type} (pfdt:type_eq_dec T) (A:Ensemble T) (a:T),
    Add (Subtract A a) a = Add A a.
  intros T h1 A a.
  apply Extensionality_Ensembles; red; split; red; intros x h2.
  destruct h2 as [x h2 | x h2].
  destruct h2 as [h2 h3].
  left; auto.
  destruct h2.
  right. constructor; auto.
  destruct (h1 a x) as [h3 | h3]; subst.
  right. constructor; auto.

  left.
  destruct h2 as [x h2 | x h2].
  constructor; auto.
  rewrite in_sing_iff; auto.
  destruct h2.
  contradict h3; auto.
Qed.


(*Wyckoff*)
Lemma couple_singleton : forall {T:Type} (x:T), 
                           Couple x x = Singleton x.
intros T x.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red. intros ? h1.
destruct h1; constructor.
(* >= *)
red. intros ? h1.
destruct h1; constructor.
Qed.

(*Wyckoff*)
Lemma couple_eq_sing :
  forall {T:Type} (a b c:T),
    Couple a b = Singleton c ->
    a = b /\ a = c.
  intros T a b c h1.
  pose proof (Couple_l _ a b) as h2.
  rewrite h1 in h2.
  destruct h2.
  pose proof (Couple_r _ c b) as h2.
  rewrite h1 in h2.
  destruct h2.
  split; auto.
Qed.


(*Wyckoff*)
Lemma couple_inv_dec : 
  forall {T:Type} (a b x:T),
    Inn (Couple a b) x ->
    {x = a} + {x = b}.
intros T a b x h1.
apply Couple_inv in h1.
apply or_to_sumbool in h1.
assumption.
Qed.


(*Wyckoff*)
Lemma incl_sing_couple_l :
  forall {T:Type} (a b:T),
    Included (Singleton a) (Couple a b).
  intros; red; intros x h8; destruct h8; left; auto.
Qed.

(*Wyckoff*)
Lemma incl_sing_couple_r :
  forall {T:Type} (a b:T),
    Included (Singleton b) (Couple a b).
  intros; red; intros x h8; destruct h8; right; auto.
Qed.


(*Wyckoff*)
Lemma empty_union : forall {T:Type} (S:Ensemble T),
                      Union (Empty_set T) S = S.
intros T S.
apply Extensionality_Ensembles.
red. split.
  (* <= *)
  red.
  intros x h1.
  destruct h1; [contradiction | assumption].
  (* >= *)
  red.
  intros x h1.
  apply Union_intror; assumption.
Qed.

(*Wyckoff*)
Lemma empty_intersection : forall {T:Type} (S:Ensemble T),
                             Intersection (Empty_set T) S =
                             Empty_set T.
  intros T S.
  apply Extensionality_Ensembles.
  red. split.
  (* <= *)
  red.
  intros x h1.
  destruct h1; contradiction.
  (* >= *)
  auto with sets.
Qed.

(*Wyckoff*)
Lemma Intersection_idempotent : 
  forall {T:Type} (A:Ensemble T),
    Intersection A A = A.
intros T A.
apply Extensionality_Ensembles.
red. split.
red. intros ? h1. destruct h1; auto.
red. intros ? h1. constructor; auto.
Qed.
                                       

(*Wyckoff*)
Lemma singleton_inj : forall {T:Type} (a b:T), 
                        Singleton a = Singleton b ->
                        a = b.
intros T a b h1.
pose proof (In_singleton _ a) as h2.
rewrite h1 in h2.
destruct h2; reflexivity. 
Qed.

  
(*Wyckoff*)
Definition sig_set {T:Type} (A:Ensemble T) :=
  {x:T | In A x}.

(*Wyckoff*)
Lemma sig_set_eq : forall {T:Type} (A B:Ensemble T),
                     A = B ->
                     sig_set A = sig_set B.
intros; subst; auto.
Defined.

(*Wyckoff*)
Lemma f_equal_sig_set :
  forall {T} {A B:Ensemble T}
         (pfe:A = B)
         (x:sig_set A)
         (y:sig_set B),
      proj1_sig x = proj1_sig y ->
      existT id (sig_set A) x =
      existT id (sig_set B) y.
  intros ? ? ? ? ? ? h1; subst; apply proj1_sig_injective in h1;
  subst; auto.
Qed.

  


(*Wyckoff*)
Lemma sig_set_eq_dec :
  forall {T:Type} (pf:type_eq_dec T) (E:Ensemble T),
    type_eq_dec (sig_set E).
  intros; apply sig_eq_dec; auto.
Qed.


(*Wyckoff*)
Definition sig_fun_subst 
           {T U:Type} {A B:Ensemble T}
           (pf:A = B)
           (f:sig_set A->U) : 
  sig_set B->U :=
  (fun c => f (exist  _ (proj1_sig c) (in_eq_set _ _ (eq_sym pf) (proj1_sig c) (proj2_sig c)))).


(*Wyckoff*)
Lemma sig_fun_subst_compat :
  forall {T U:Type} {A B:Ensemble T}
         (pfe:A = B)
         (f:sig_set A->U)
         (x:T)
         (pfx:Inn B x),
    sig_fun_subst pfe f (exist _ _ pfx) = 
    f (exist _ _ (in_eq_set _ _ (eq_sym pfe) x pfx)).
  unfold sig_fun_subst; intros; repeat f_equal.
Qed.


(*Wyckoff*)
Definition full_sig {T:Type} (A:Ensemble T) :=
  Full_set (sig_set A).

(*Wyckoff*)
Definition full_set_sig_conv {T:Type} (x:T) : sig_set (Full_set T) := exist _ _ (Full_intro _ x).

(*Wyckoff*)
Lemma inh_full_sig :
  forall {T:Type} (E:Ensemble T),
    Inhabited E ->
    Inhabited (full_sig E).
  intros T E h1.
  destruct h1 as [x h1].
  apply Inhabited_intro with (exist _ _ h1).
  constructor.
Qed.

(*Wyckoff*)
Lemma in_im :
  forall {T U} x (f:T->U),
    Inn (Im (Full_set _) f) (f x).
  intros; econstructor; constructor.
Qed.


(*Wyckoff*)
Lemma im_singleton : forall {T U:Type} (x:T) (f:T->U),
                       Im (Singleton x) f = Singleton (f x).
intros T U x f.
apply Extensionality_Ensembles.
red. split.
red.
intros y h1.
destruct h1 as [z h2]. subst.
destruct h2; subst.
constructor; auto.
red. 
intros y h1.
destruct h1; subst.
apply Im_intro with x; auto; constructor.
Qed.

(*Wyckoff*)
Lemma full_sig_sing :
  forall {T:Type} (x:T),
    full_sig (Singleton x) =
    Singleton (exist (fun c => In (Singleton x) c) _ (In_singleton _ _)).
  intros T x.
  unfold full_sig.
  apply Extensionality_Ensembles; red; split; red; intros a h2.
  clear h2.
  destruct a as [a h1]. destruct h1.
  constructor.
  constructor.
Qed.


(*Wyckoff*)
Lemma im_couple : 
  forall {T U:Type} (x y:T) (f:T->U),
    Im (Couple x y) f = Couple (f x) (f y).
intros T U x y f.
apply Extensionality_Ensembles; red; split; red; intros a h1.
destruct h1 as [a h1]. subst.
destruct h1; [left | right].
inversion h1; subst; [apply Im_intro with x;
                       auto | apply Im_intro with y; auto].
left. right.
Qed.


(*Wyckoff*)
Lemma im_preserves_inclusion : 
  forall {T U:Type} (A B:Ensemble T) (f:T->U),
    Included A B -> Included (Im A f) (Im B f).
intros T U A B f h1.
red. intros u h2.
destruct h2 as [u h2]. subst.
apply h1 in h2.
apply Im_intro with u; auto.
Qed.



(*Wyckoff*)
Lemma im_im : 
  forall {T U V:Type} (A:Ensemble T) (f:T->U) (g:U->V),
    Im (Im A f) g = Im A (fun x => g (f x)).
intros T U V A f g.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red. intros v h1.
inversion h1 as [x h2 v' h3 h4].  subst.
destruct h2 as [x h2 u h3]. subst.
apply Im_intro with x. assumption. reflexivity.
(* >= *)
red. intros v h1.
destruct h1 as [x h2 v' h3]. subst.
apply Im_intro with (f x). apply Im_intro with x; auto.
reflexivity.
Qed.

(*Wyckoff*)
Lemma incl_im_subtract :
  forall {T U:Type} (A:Ensemble T) (a:T) (f:T->U),
    Included (Subtract (Im A f) (f a))
             (Im (Subtract A a) f).
  intros T U A a f; red; intros x h1.
  destruct h1 as [h1 h2].
  destruct h1 as [x h1]. subst.
  apply Im_intro with x; auto.
  constructor; auto.
  intro h3. destruct h3.
  contradict h2.
  constructor.
Qed.

(*Wyckoff*)
Lemma im_subtract_inj :
  forall {T U:Type} (A:Ensemble T) (a:T) (f:T->U),
    injective f ->
    Im (Subtract A a) f = 
    Subtract (Im A f) (f a).
  intros T U A a f h0.
  apply Extensionality_Ensembles; red; split; red; intros x h1;
  try apply incl_im_subtract; auto.
  destruct h1 as [x h1]. subst.
  destruct h1 as [h1 h2].
  red in h0.
  constructor.
  apply Im_intro with x; auto.
  intro h3. inversion h3 as [h4]. clear h3.
  apply h0 in h4.
  subst.
  contradict h2.
  constructor.
Qed.

     
                                                          

(*Wyckoff*)
Lemma empty_image : forall {T U:Type} (A:Ensemble T) (f:T->U),
                      Im A f = Empty_set _ -> A = Empty_set _.
intros T U A f h1.
apply NNPP.
intro h2. 
pose proof (not_empty_Inhabited _ _ h2) as h3. 
destruct h3 as [x h3].
assert (h4:Inn (Im A f) (f x)).
    apply Im_intro with x. assumption. reflexivity.
  rewrite h1 in h4.
contradiction.
Qed.

(*Wyckoff*)
Lemma inh_im :
  forall {T U:Type} (E:Ensemble T) (f:T->U),
    Inhabited E -> Inhabited (Im E f).
  intros T U E f h1.
  destruct h1 as [x h1].
  apply Inhabited_intro with (f x); auto.
  apply Im_intro with x; auto.
Qed.


(*Wyckoff*)  
Lemma im_const_fun :
  forall {T U:Type} (A:Ensemble T) (u:U),
    Inhabited A ->
    Im A (fun _:T => u) = Singleton u.
  intros T U A u h1.
  apply Extensionality_Ensembles; red; split; red; intros x h2.
  destruct h2 as [x h2]; subst.
  constructor.
  destruct h2.
  destruct h1 as [x h1].
  apply Im_intro with x; auto.
Qed.






(*Wyckoff*)
Lemma im_full_sig_proj1_sig : forall {T:Type} (S:Ensemble T), 
  S = Im (full_sig S) (@proj1_sig _ _).
intros T S.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red. intros x h1.
apply Im_intro with (exist _ x h1).
constructor. simpl. reflexivity.
(* >= *)
red. intros x h1.
inversion h1 as [x' h2 ? h3 h4].
rewrite h3.
apply proj2_sig.
Qed.


 

(*Wyckoff*)
Lemma im_id : forall {T:Type} (S:Ensemble T),
                S = Im S id.
intros T S.
apply Extensionality_Ensembles.
red. split.
red. intros x h1. apply Im_intro with x; auto.
red. intros x h1. destruct h1 as [x h1]. subst. unfold id.
assumption.
Qed.







(*Wyckoff*)
Lemma im_full_sig_im_eq : 
  forall {T U V:Type} (A:Ensemble T) (f:T->U)
         (g:sig_set (Im A f)-> V),
    Im (full_sig (Im A f)) g  = 
    Im (full_sig A)
       (fun x:sig_set A =>
          g (exist _ _
                   (Im_intro _ _ A f _ (proj2_sig x) _ (eq_refl _)))).
intros T U V A f g.
apply Extensionality_Ensembles.
red. split.
red. intros x h1.
destruct h1 as [x h1]. subst. clear h1.
destruct x as [x h2]. destruct h2 as [x h2]. subst.
apply Im_intro with (exist _ _ h2). constructor.
f_equal.
red. intros x h1.
destruct h1 as [x h1]. subst. clear h1.
destruct x as [x h1].
assert (h2:In (Im A f) (f x)). apply Im_intro with x; auto.
apply Im_intro with (exist _ _ h2). constructor.
f_equal.
apply proj1_sig_injective.
simpl.
reflexivity.
Qed.



(*Wyckoff*)
Lemma im_ext_in : 
  forall {T U:Type} (A:Ensemble T) (f g:T->U),
    (forall x:T, In A x -> f x = g x) ->
    Im A f = Im A g.
intros T U A f g h1.
apply Extensionality_Ensembles.
red. split.
red. intros x h2.
destruct h2 as [x h2]. subst.
rewrite h1; auto. apply Im_intro with x; auto.
red. intros x h2.
destruct h2 as [x h2]. subst.
rewrite <- h1; auto. apply Im_intro with x; auto.
Qed.

(*Wyckoff*)
Lemma im_functional1 :
  forall {T U:Type} {A B:Ensemble T} (pf:A = B) (f:T->U),
    Im A f =
    Im B f.
  intros; subst; auto.
Qed.

(*Wyckoff*)
Lemma im_functional2 :
  forall {T U:Type} {f g:T->U} (pf:f = g) (A:Ensemble T),
    Im A f =
    Im A g.
  intros; subst; auto.
Qed.

(*Wyckoff*)
Lemma im_functional :
  forall {T U:Type} {A B:Ensemble T} {f g:T->U}
         (pfab:A = B) (pffg:f = g),
    Im A f =
    Im A g.
  intros; subst; auto.
Qed.


(*Wyckoff*)
Definition im_proj2_sig 
           {T:Type} {P:T->Prop} (A:Ensemble T)  
           (pfx:(forall x:T, Inn A x -> P x)) :
  Ensemble {x | P x} :=
 Im (full_sig A) (fun p => (exist _ _ (pfx _ (proj2_sig p)))).


(*Wyckoff*)
Definition im_in {T U:Type} (A:Ensemble T) 
           (P:T->Prop) (f:{x:T|P x}->U)
           (pfx:(forall x:T, Inn A x -> P x)) : 
  Ensemble U :=
  Im (full_sig A) 
     (fun x:{t:T|In A t}=>f (exist _ (proj1_sig x)
                                   (pfx _ (proj2_sig x)))).  

(*Wyckoff*)
Lemma im_full_sig_subtract_in_inj :
  forall {T U:Type} (E:Ensemble T)  (x:T)
         (f:T->U),
    injective f ->
    let f' := (fun a:sig_set (Subtract E x) => f (proj1_sig a)) in
    In E x ->
    Im (full_sig (Subtract E x)) f' =
    Subtract (Im E f) (f x).
  intros T U E x f h1 f'  h2.
  apply Extensionality_Ensembles; red; split; red; intros a h3.
  destruct h3 as [a h3]. subst. clear h3.
  destruct a as [a h3].
  destruct h3 as [h3 h4].
  constructor.
  apply Im_intro with a; auto.
  intro h5. inversion h5 as [h6]. clear h5.
  red in h1. apply h1 in h6.
  simpl in h6. clear f'. subst.
  contradict h4. constructor.
  destruct h3 as [h3 h4].
  destruct h3 as [a h3]. subst.
  assert (h5:In (Subtract E x) a).
    constructor; auto.
    intro h5. destruct h5. contradict h4. constructor.
  apply Im_intro with (exist _ _ h5).
  constructor.
  unfold f'. simpl.
  reflexivity.
Qed.

  

(*Wyckoff*)
(*assigns image to range*)
Definition sig_im_fun {T U:Type}
           (f:T->U) : T->sig_set (Im (Full_set T) f) :=
  fun x => exist _ _
                 (Im_intro _ _ _
                           f _ (Full_intro _ x) _ (@eq_refl _ _)).


(*Wyckoff*)
Lemma surj_sig_im_fun : 
  forall {T U:Type} (f:T->U),
    surjective (sig_im_fun f).
intros T U f. red. intro y. destruct y as [y h1].
destruct h1 as [x]. subst.
exists x. unfold sig_im_fun.
apply proj1_sig_injective. simpl.
reflexivity.
Qed.


(*Wyckoff*)
Lemma inj_bij_sig_im_fun_iff :
  forall {T U:Type} (f:T->U),
    injective f <-> bijective (sig_im_fun f).
intros T U f.
split.
intro h1. red. split. red in h1. red.
intros a b h2. unfold sig_im_fun in h2. apply exist_injective in h2.
apply h1; auto.
apply surj_sig_im_fun.
intro h1. red in h1. destruct h1 as [h1l h1r].
red in h1l. red.
intros a b h2. specialize (h1l a b).
apply h1l. unfold sig_im_fun. apply proj1_sig_injective.
simpl.
assumption.
Qed.

(*Wyckoff*)
Lemma inj_sig_pred :
  forall (E:Ensemble nat),
    injective (fun n : sig_set (Subtract E 0) => pred (proj1_sig n)).
  intro E.
  red; intros m n h1.
  destruct m as [m h2], n as [n h3].
  simpl in h1.
  destruct h2 as [h2a h2b], h3 as [h3a h3b].
  destruct (zerop m) as [h4 | h4], (zerop n) as [h5 | h5]; subst.
  f_equal. apply proof_irrelevance.
  contradict h2b. constructor; auto.
  contradict h3b. constructor; auto.
  apply pred_inj0 in h1.
  subst.
  f_equal. apply proof_irrelevance.
  assumption. assumption.
Qed.


Section Disjoint.
(*Wyckoff*)
Definition disjoint {T:Type} (A B:Ensemble T) :=
  Intersection A B = Empty_set _.

(*Wyckoff*)
Lemma sym_disjoint :
  forall {T:Type} (A B:Ensemble T),
    disjoint A B -> disjoint B A.
intros T A B h1.
unfold disjoint in h1. 
rewrite Intersection_commutative in h1.
red. assumption.
Qed.

(*Wyckoff*)
Definition disjoint_incl_comp_l :
  forall {T:Type} (A B:Ensemble T),
    disjoint A B ->
    Included A (Comp B).
intros T A B h1.
red. intros x h2. intro h3.
red in h1.
pose proof (Intersection_intro _ _ _ _  h2 h3) as h4.
rewrite h1 in h4.
contradiction.
Qed.

(*Wyckoff*)
Definition disjoint_incl_comp_r :
  forall {T:Type} (A B:Ensemble T),
    disjoint A B ->
    Included B (Comp A).
intros T A B h1.
red. intros x h2. intro h3.
red in h1.
pose proof (Intersection_intro _ _ _ _  h3 h2) as h4.
rewrite h1 in h4.
contradiction.
Qed.

(*Wyckoff*)
Lemma disjoint_setminus_compat :
  forall {T:Type} (A B:Ensemble T),
    disjoint A B ->
    Setminus A B = A.
  intros T A B h1.
  red in h1.
  apply Extensionality_Ensembles; red; split; red; intros x h2.
  destruct h2; auto.
  constructor; auto.
  intro h3.
  pose proof (Intersection_intro _ A B _ h2 h3) as h4.
  rewrite h1 in h4.
  destruct h4.
Qed.

End Disjoint.


Section ImProjSig.

(*Wyckoff*)
Definition im_proj1_sig {T:Type} {A:Ensemble T} 
           (S:Ensemble (sig_set A)) : Ensemble T :=
  Im S (@proj1_sig _ _).



(*Wyckoff*)
Definition im_proj1_sig' {T:Type} {P:T->Prop}
           (S:Ensemble {x:T | P x}) : Ensemble T :=
  Im S (@proj1_sig _ _).

(*Wyckoff*)
Lemma im_proj1_sig_empty :
  forall {T:Type} (A:Ensemble T),
    im_proj1_sig (Empty_set (sig_set A)) = Empty_set _.
  unfold im_proj1_sig; intros; apply image_empty.
Qed.

(*Wyckoff*)
Lemma im_proj1_sig_couple :
  forall {T:Type} {A:Ensemble T}
         (x y:sig_set A),
    im_proj1_sig (Couple x y) = Couple (proj1_sig x) (proj1_sig y).
  intros T A x y.
  unfold im_proj1_sig.
  apply im_couple.
Qed.


(*Wyckoff*)
Lemma im_proj1_sig_injective : 
  forall {T:Type} {A:Ensemble T} (R S:Ensemble (sig_set A)),
    im_proj1_sig R = im_proj1_sig S ->
    R = S.
intros T A R S h1.
apply Extensionality_Ensembles.
red. split.
red. intros x h2. destruct x as [x h3]. 
assert (h4:In (im_proj1_sig R) x).
  apply Im_intro with (exist _ _ h3).
  assumption. simpl. reflexivity.
rewrite h1 in h4.
destruct h4 as [x h4]. subst.
rewrite unfold_sig in h4.
assert (h3 = proj2_sig x). apply proof_irrelevance. subst.
assumption.
red. intros x h2. destruct x as [x h3]. 
assert (h4:In (im_proj1_sig S) x).
  apply Im_intro with (exist _ _ h3).
  assumption. simpl. reflexivity.
rewrite <- h1 in h4.
destruct h4 as [x h4]. subst.
rewrite unfold_sig in h4.
assert (h3 = proj2_sig x). apply proof_irrelevance. subst.
assumption.
Qed.



(*Wyckoff*)
Lemma im_proj1_sig_undoes_im_proj2_sig : 
  forall {T:Type} (A B:Ensemble T)
         (pf:Included A B),
    A = im_proj1_sig (im_proj2_sig A pf).
intros T A B h1.
unfold im_proj1_sig, im_proj2_sig.
rewrite im_im.
rewrite im_full_sig_proj1_sig at 1.
simpl.
apply im_ext_in.
intros; auto.
Qed.


(*Wyckoff*)
Lemma incl_im_proj1_sig_ens_sig_set : 
  forall {T:Type} {A:Ensemble T} (B:Ensemble (sig_set A)),
    Included (im_proj1_sig B) A.
intros T A B.
red. intros x h1.
destruct h1 as [x h1]. subst. apply proj2_sig.
Qed.



(*Wyckoff*)
Lemma im_proj2_sig_undoes_im_proj1_sig : 
  forall {T:Type} {B:Ensemble T} (A:Ensemble (sig_set B)),
    A = im_proj2_sig (im_proj1_sig A) (incl_im_proj1_sig_ens_sig_set A).
intros T B A.
unfold im_proj1_sig, im_proj2_sig.
apply Extensionality_Ensembles.
red. split.
red. intros x h1.
destruct x as [x h2].
assert (h3:In (Im A (@proj1_sig _ _)) x).
  apply Im_intro with (exist _ _ h2). assumption.
  simpl. reflexivity.
apply Im_intro with (exist _ _ h3). constructor.
apply proj1_sig_injective. simpl. reflexivity.
red. intros x h1.
destruct h1 as [x h1]. subst.
destruct x as [x h2]. simpl.
destruct h2 as [x h2]. subst.
pose proof h2 as h2'. rewrite unfold_sig in h2'.
pose proof (proof_irrelevance
              _ 
              (incl_im_proj1_sig_ens_sig_set
                 A (proj1_sig x)
                 (Im_intro (sig_set B) T A
                           (proj1_sig (P:=fun x0 : T => In B x0)) x
                           h2 (proj1_sig x) eq_refl))
              (proj2_sig x)) as h3.
rewrite h3; auto.
Qed.

(*Wyckoff*)
Lemma im_proj2_sig_undoes_im_proj1_sig' : 
  forall {T:Type} (B:Ensemble T) (A:Ensemble (sig_set B))
         (pf:Included (im_proj1_sig A) B),
    A = im_proj2_sig (im_proj1_sig A) pf.
intros T B A h1.
pose proof (im_proj2_sig_undoes_im_proj1_sig A) as h2.
pose proof (proof_irrelevance
              _ h1
              (incl_im_proj1_sig_ens_sig_set A)) as h3.
rewrite h3; auto.
Qed.

(*Wyckoff*)
Lemma in_im_proj1_sig_iff :
  forall {T:Type} (B:Ensemble T) (A:Ensemble (sig_set B))
         (x:sig_set B),
    In (im_proj1_sig A) (proj1_sig x) <-> In A x.
intros T B A x.
split.
intro h1.
inversion h1 as [i h2 ? h3 h4]. subst.
apply proj1_sig_injective in h3.
subst.
assumption.
intro h1.
apply Im_intro with x; auto.
Qed.

(*Wyckoff*)
Lemma in_im_proj1_sig : 
  forall {T:Type} {B:Ensemble T} (A:Ensemble (sig_set B))
         (x:T),
         In (im_proj1_sig A) x ->
         exists pf:In B x,
           In A (exist _ _ pf).
intros T B A x h1.
destruct h1 as [x h1]. 
subst.
exists (proj2_sig _).
rewrite <- unfold_sig.
assumption.
Qed.

(*Wyckoff*)
Lemma in_im_proj1_sig_inv_ex :
  forall {T:Type} (B:Ensemble T) (A:Ensemble (sig_set B)) (x:T),
    In (im_proj1_sig A) x ->
    exists! b:sig_set B,
      x = proj1_sig b.
  intros T B A x h1.
  destruct h1 as [x h1]. subst.
  exists x.
  red. split; auto.
  intros h h2.
  apply proj1_sig_injective in h2.
  assumption.
Qed.

(*Wyckoff*)
Definition in_im_proj1_sig_inv
  {T:Type} (B:Ensemble T) (A:Ensemble (sig_set B)) (x:T)
  (pf:In (im_proj1_sig A) x) :=
  proj1_sig (constructive_definite_description
               _ (in_im_proj1_sig_inv_ex B A x pf)).

(*Wyckoff*)
Lemma in_im_proj1_sig_inv_compat :
  forall {T:Type} (B:Ensemble T) (A:Ensemble (sig_set B)) (x:T)
         (pf:In (im_proj1_sig A) x),
    let b := in_im_proj1_sig_inv B A x pf in
    x = proj1_sig b.
  unfold in_im_proj1_sig_inv; intros;
  destruct constructive_definite_description; auto.
Qed.

(*Wyckoff*)
Lemma in_im_proj1_sig' :
  forall {T:Type} (B:Ensemble T) (A:Ensemble (sig_set B)) (x:T)
         (pf:In (im_proj1_sig A) x),
    In A (in_im_proj1_sig_inv B A x pf).
  intros T B A x h1.
  pose proof (in_im_proj1_sig_inv_compat B A x h1) as h2.
  simpl in h2.
  destruct h1 as [x h1]. subst.
  apply proj1_sig_injective in h2.
  pose proof h1 as h1'.
  rewrite h2 in h1' at 1.
  assumption.
Qed.   


(*Wyckoff*)
Lemma incl_im_proj1_sig_iff : 
  forall {T:Type} {S:Ensemble T} (A B:Ensemble (sig_set S)),
    Included (im_proj1_sig A) (im_proj1_sig B) <->
    Included A B.
intros T S A B. split.
intro h1. red. intros x h2.
rewrite <- in_im_proj1_sig_iff in h2.
apply h1 in h2. inversion h2 as [x' h3 q h4]. subst.
apply proj1_sig_injective in h4. subst.
assumption.
intro h1. red. intros x h2.
destruct h2 as [x h2]. subst.
apply h1 in h2.
apply Im_intro with x; auto.
Qed.




(*Wyckoff*)
Lemma im_proj1_sig_add_eq : 
  forall {T:Type} {E:Ensemble T} (A:Ensemble (sig_set E)) 
         (a:sig_set E),
    im_proj1_sig (Add A a) = Add (im_proj1_sig A) (proj1_sig a).
intros T E A a.
apply Extensionality_Ensembles; red; split.
red; intros x h1.
destruct h1 as [x h1]. subst.
destruct h1 as [x h1 | x h2].
left. apply Im_intro with x; auto.
destruct h2. right. constructor.
red. intros x h1.
destruct h1 as [x h1 | x h2].
destruct h1 as [x h1]. subst.
apply Im_intro with x. left; auto.
reflexivity.
destruct h2.
apply Im_intro with a. right. constructor.
reflexivity.
Qed.



(*Wyckoff*)
Lemma im_proj1_sig_subtract_eq : 
  forall {T:Type} {E:Ensemble T} (A:Ensemble (sig_set E)) 
         (a:sig_set E),
    im_proj1_sig (Subtract A a) =
    Subtract (im_proj1_sig A) (proj1_sig a).
intros T E A a.
apply Extensionality_Ensembles; red; split.
red; intros x h1.
destruct h1 as [x h1]. subst. destruct h1 as [h1 h2].
constructor.
apply Im_intro with x; auto.  
intro h3. inversion h3 as [h4]. clear h3.
apply proj1_sig_injective in h4. subst.
contradict h2.
constructor.
red. intros x h1.
destruct h1 as [h1 h2]. destruct h1 as [x h1]. subst.
apply Im_intro with x.
constructor; auto.
intro h3. destruct h3.
contradict h2. constructor.
reflexivity.
Qed.

(*Wyckoff*)
Lemma im_proj1_sig_sing : 
  forall {T:Type} (E:Ensemble T) (x:sig_set E),
    im_proj1_sig (Singleton x) = Singleton (proj1_sig x).
intros T E x.
unfold im_proj1_sig. 
rewrite im_singleton.
reflexivity.
Qed.

(*Wyckoff*)
Lemma finite_im_proj1_sig_iff : 
  forall {T:Type} {A:Ensemble T} (E:Ensemble (sig_set A)),
    Finite (im_proj1_sig E) <-> Finite E.
intros T A E. split.
intro h1.
apply NNPP.
intro h2.
pose proof (Pigeonhole_bis _ _ E _ h2 h1) as h3.
contradict h3.
rewrite <- inj_image_fun_prop_compat.
red; intros; apply proj1_sig_injective; auto.
intro h1.
apply finite_image; auto.
Qed.

(*Wyckoff*)
Lemma inh_im_proj1_sig_iff :
  forall {T:Type} {A:Ensemble T} (E:Ensemble (sig_set A)),
    Inhabited (im_proj1_sig E) <-> Inhabited E.
  intros T A E. split.
  intro h1. destruct h1 as [x h1]. destruct h1 as [x h1]. subst.
  apply Inhabited_intro with x; auto.
  intro h1. destruct h1 as [x h1].
  apply Inhabited_intro with (proj1_sig x).
  apply Im_intro with x; auto.
Qed.

(*Wyckoff*)
Lemma im_proj1_sig_full_set_sig :
  forall {T:Type} (S:Ensemble T),
    im_proj1_sig (Full_set {x:T | In S x}) = S.
  intros T S.
  apply Extensionality_Ensembles; red; split; red; intros x h1.
  destruct h1 as [x h1]. subst. apply proj2_sig.
  apply Im_intro with (exist _ _ h1). constructor.
  simpl. reflexivity.
Qed.


(*Wyckoff*)
Lemma sig_set_im_proj1_sig_in_aux : 
  forall {T:Type} {A:Ensemble T} {E:Ensemble (sig_set A)} 
         (x:sig_set (im_proj1_sig E)),
    In A (proj1_sig x).
intros T A E x.
destruct x as [x h1]. simpl.
destruct h1 as [x h1]. subst.
apply proj2_sig.
Qed.

(*Wyckoff*)
Lemma sig_set_im_proj1_sig_in : 
  forall {T:Type} {A:Ensemble T} {E:Ensemble (sig_set A)} 
         (x:sig_set (im_proj1_sig E)),
    In E (exist _ _ (sig_set_im_proj1_sig_in_aux x)).
intros T A E x.
destruct x as [x h1]. simpl.
destruct h1 as [x h1]. subst.
destruct x as [x h2]. simpl.
pose proof (proof_irrelevance
              _ h2
              (sig_set_im_proj1_sig_in_aux
                 (exist (fun x0 : T => In (im_proj1_sig E) x0) x
                        (Im_intro (sig_set A) T E
                                  (proj1_sig (P:=fun x0 : T => In A x0))
                                  (exist (fun x0 : T => In A x0) x h2) h1 x eq_refl)))) as h3.
rewrite <- h3; auto.
Qed.


(*Wyckoff*)
Lemma im_proj1_sig_eq_iff :
  forall {T:Type} {S:Ensemble T} (A B:Ensemble (sig_set S)),
    im_proj1_sig A = im_proj1_sig B <-> A = B.
  intros T S A B.
  split.
  intro h1.
  apply Extensionality_Ensembles; red; split; red; intros x h2.
  destruct x as [x h3].
  assert (h4:In (im_proj1_sig A) x).
    apply Im_intro with (exist _ _ h3); auto.
  rewrite h1 in h4.
  destruct h4 as [x h4]. subst.
  destruct x as [x h5]. simpl.
  simpl in h3.
  assert (h5 = h3). apply proof_irrelevance.
  subst.
  assumption.
  destruct x as [x h3].
  assert (h4:In (im_proj1_sig A) x).
    rewrite h1.
    apply Im_intro with (exist _ _ h3); auto.
  destruct h4 as [x h4]. subst.
  destruct x as [x h5].
  simpl in h2. simpl.
  pose proof (proof_irrelevance _ h3 h5); subst; auto.
  intro; subst; auto.
Qed.

(*Wyckoff*)
Lemma intersection_im_proj1_sig_eq :
  forall {T:Type} {S:Ensemble T}
         (A B:Ensemble (sig_set S)),
    Intersection (im_proj1_sig A) (im_proj1_sig B) =
    im_proj1_sig (Intersection A B).
  intros T S A B.
  apply Extensionality_Ensembles; red; split; red; intros x h1.
  destruct h1 as [x h1 h2].
  destruct h1 as [x h1 ? h4], h2 as [x' h2 ? h3].
  subst.
  apply proj1_sig_injective in h3.
  subst.
  apply Im_intro with x'; auto.
  constructor; auto.
  destruct h1 as [x h1]. subst.
  destruct h1 as [x h1 h2].
  constructor.
  apply Im_intro with x; auto.
  apply Im_intro with x; auto.
Qed.

(*Wyckoff*)
Lemma im_proj1_sig_disjoint_iff :
  forall {T:Type} {S:Ensemble T}
         (A B:Ensemble (sig_set S)),
    disjoint A B <-> disjoint (im_proj1_sig A) (im_proj1_sig B).
  intros T S A B.
  split.
  intro h1.
  red in h1. red.
  rewrite intersection_im_proj1_sig_eq.
  rewrite h1.
  unfold im_proj1_sig.
  rewrite image_empty.
  reflexivity.
  intro h1.
  red in h1.
  rewrite intersection_im_proj1_sig_eq in h1.
  apply im_proj1_sig_injective.
  unfold im_proj1_sig at 2.
  rewrite image_empty.
  assumption.
Qed.

Lemma in_proj1_sig_fun :
  forall {T U:Type} {P:U->Prop} (f:T->{x:U | P x})
         (A:Ensemble T) (y:U),
    In (Im A (proj1_sig_fun f)) y ->
    P y.
  intros T U P f A y h1.
  destruct h1; subst.
  rewrite proj1_sig_fun_compat.
  apply proj2_sig.
Qed.


End ImProjSig.


Definition fun_in_ens {T:Type} (A:Ensemble T) (U:Type) : Type :=
  prop_dep_fun (fun x => Inn A x) U.

Definition fun_in_ens' {T U} (A:Ensemble T) (B:Ensemble U) : Type :=
  fun_in_ens A (sig_set B).

Definition fun_in_set {T:Type} (A:Ensemble T) (U:Set) : Type :=
  prop_dep_fun (fun x => Inn A x) U.

(*Like the [transfer] functions in [TypeUtilities], but for 
dependency reasons, we'll have a minor substitute.*)
Definition fun_in_ens_subst 
           {T:Type} {A A':Ensemble T} {U:Type}
           (pfe:A = A')
           (f:fun_in_ens A U) (pfe:A = A') :
  fun_in_ens A' U :=
  fun x (pfx:Inn A' x) =>
    let pfx' := in_eq_set _ _ (eq_sym pfe) _ pfx in
    f x pfx'.

Definition fun_in_set_subst 
           {T:Type} {A A':Ensemble T} {U:Set}
           (pfe:A = A')
           (f:fun_in_set A U) (pfe:A = A') :
  fun_in_set A' U :=
  fun x (pfx:Inn A' x) =>
    let pfx' := in_eq_set _ _ (eq_sym pfe) _ pfx in
    f x pfx'.




Definition fun_in_ens_subst1
           {T U} {A A':Ensemble T} {B:Ensemble U}
           (f:fun_in_ens' A B)
           (pfe:A = A') :
  fun_in_ens' A' B :=
  fun x (pfx:Inn A' x) =>
    let pfx' := in_eq_set _ _ (eq_sym pfe) _ pfx in
    f x pfx'.

Definition fun_in_ens_subst2
           {T U} {A:Ensemble T} {B B':Ensemble U}
           (f:fun_in_ens' A B)
           (pfe:B = B') :
  fun_in_ens' A B' :=
  fun x (pfx:Inn A x) =>
    exist (Inn B') _ (in_eq_set _ _ pfe _ (proj2_sig (f x pfx))).





Definition fun_eq_set {T:Type} (A:Ensemble T) (U:Set) : Type :=
  prop_dep_fun (fun B => A = B) U.


Inductive im_in_ens {T U:Type} {A:Ensemble T} (f:fun_in_ens A U) :
  Ensemble U:=
| im_in_ens_intro : forall (x:T) (pf:In A x) (y:U),
                      y = f x pf -> In (im_in_ens f) y.

Inductive im_in_set {T:Type} {U:Set} {A:Ensemble T} (f:fun_in_ens A U) :
  Ensemble U:=
| im_in_set_intro : forall (x:T) (pf:In A x) (y:U),
                      y = f x pf -> In (im_in_set f) y.

Lemma in_im_in_ens :
  forall {T U:Type} {A:Ensemble T} (f:fun_in_ens A U) (x:T) (pf:Inn A x),
    Inn (im_in_ens f) (f x pf).
  intros; econstructor; reflexivity.
Qed.

Lemma in_im_in_set :
  forall {T:Type} {U:Set} {A:Ensemble T}
         (f:fun_in_ens A U) (x:T) (pf:Inn A x),
    Inn (im_in_set f) (f x pf).
  intros; econstructor; reflexivity.
Qed.



Definition inj_in_ens {T U:Type} {A:Ensemble T} (f:fun_in_ens A U) : Prop :=
  forall (x y:T)
         (pfx:Inn A x)
         (pfy:Inn A y),
    f _ pfx = f _ pfy ->
    x = y.

Definition inj_in_ens' {T U:Type} {A:Ensemble T} {B:Ensemble U}
           (f:fun_in_ens' A B) : Prop :=
  inj_in_ens f.


Definition surj_in_ens {T U:Type} {A:Ensemble T} {B:Ensemble U}
           (f:fun_in_ens' A B) : Prop :=
  forall y (pfy:Inn B y),
  exists x (pfx:Inn A x),
    f x pfx = exist _ y pfy.

Definition bij_in_ens {T U:Type} {A:Ensemble T} {B:Ensemble U}
           (f:fun_in_ens' A B) : Prop :=
  inj_in_ens f /\ surj_in_ens f.


Lemma surj_in_ens_compat :
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U}
         (f:fun_in_ens' A B),
    surj_in_ens f ->
    im_proj1_sig (im_in_ens f) = B.
  intros T U A B f h1.
  red in h1.
  apply Extensionality_Ensembles; red; split; red; intros x h2.
  destruct h2 as [x h2]; subst; apply proj2_sig.
  specialize (h1 x h2).
  destruct h1 as [y [h1 h3]].
  pose proof (f_equal (@proj1_sig _ _) h3) as h4; simpl in h4; subst.
  apply Im_intro with (f y h1);auto.
  apply im_in_ens_intro with y h1; auto.
Qed.

  
(*Wyckoff*)
Inductive inverses_in_ens {T U} {A:Ensemble T} {B:Ensemble U}
          (f:fun_in_ens' A B) : fun_in_ens' B A -> Prop :=
| inverses_in_ens_intro :
    forall g:fun_in_ens' B A,
      (forall x (pfx:Inn A x),
         g (proj1_sig (f x pfx)) (proj2_sig (f x pfx)) = exist _ x pfx) ->
      (forall y (pfy:Inn B y),
         f (proj1_sig (g y pfy)) (proj2_sig (g y pfy)) = exist _ y pfy) ->
      inverses_in_ens f g.

(*Wyckoff*)
Inductive invertible_in_ens
          {T U} {A:Ensemble T} {B:Ensemble U}
          (f:fun_in_ens' A B) : Prop :=
| invertible_in_ens_intro :
    forall g:fun_in_ens' B A,
      inverses_in_ens f g ->
      invertible_in_ens f.

(*Wyckoff*)
Definition inv_in_ens_ex :
  forall {T U} {A:Ensemble T} {B:Ensemble U}
         (f:fun_in_ens' A B),
    invertible_in_ens f ->
    exists! g:fun_in_ens' B A,
      inverses_in_ens f g.
  intros T U A B f h1.
  destruct h1 as [g hinv]. 
  exists g; auto; red; split; auto.
  intros h h1. 
  destruct h1 as [k h2]. 
  rename H into h3.
  destruct hinv as [q h4].
  rename H into h5.
  apply functional_extensionality_dep.
  intro y.
  apply functional_extensionality.
  intro hy.
  specialize (h3 y hy).
  pose proof (f_equal (@proj1_sig _ _) h3) as h6; simpl in h6.
  specialize (h4 (proj1_sig (k y hy)) (proj2_sig (k y hy))).
  revert h4.
  Gen0.
  rewrite h6; intros h7 h8.
  pose proof (proof_irrelevance _ h7 hy) as h9.
  rewrite <- h9, h8.
  apply proj1_sig_injective; simpl.
  repeat f_equal; auto.
Qed.


(*Wyckoff*)
Definition inv_in_ens
           {T U} {A:Ensemble T} {B:Ensemble U}
           (f:fun_in_ens' A B)
           (pfinv:invertible_in_ens f) :
  fun_in_ens' B A :=
  proj1_sig (constructive_definite_description _ (inv_in_ens_ex f pfinv)).


(*Wyckoff*)
Definition inv_in_ens_compat : 
  forall {T U} {A:Ensemble T} {B:Ensemble U}
         (f:fun_in_ens' A B)
         (pfinv:invertible_in_ens f),
    inverses_in_ens f (inv_in_ens f pfinv).
  unfold inv_in_ens; intros; destruct constructive_definite_description;
  simpl; auto.
Qed.

(*Wyckoff*)
Lemma inv_impl_bij_in_ens :
  forall {T U} {A:Ensemble T} {B:Ensemble U}
         (f:fun_in_ens' A B),
    invertible_in_ens f ->
    bij_in_ens f.
  intros T U A B f h1.
  destruct h1 as [g h2].
  destruct h2 as [g h2].
  rename H into h3. 
  red; split.
  red; intros x y hx hy h1.
  pose proof (h2 x hx) as h2x.
  pose proof (h2 y hy) as h2y.
  pose proof (f_equal (@proj1_sig _ _) h2x) as h3x;
    pose proof (f_equal (@proj1_sig _ _) h2y) as h3y;
    simpl in h3x, h3y.
  rewrite h1 in h3x.
  rewrite <- h1 in h3y.
  revert h3y.
  subst.
  intro h4.
  rewrite h2; simpl; auto.
  red; intros y hy.
  specialize (h3 y hy).
  pose proof (f_equal (@proj1_sig _ _) h3) as h4; simpl in h4.
  exists (proj1_sig (g y hy)), (proj2_sig _); apply proj1_sig_injective;
  simpl; auto.
Qed.


(*Wyckoff*)
Definition inj_in_set {T:Type} {U:Set} {A:Ensemble T}
           (f:fun_in_ens A U) : Prop :=
  forall (x y:T)
         (pfx:Inn A x)
         (pfy:Inn A y),
    f _ pfx = f _ pfy ->
    x = y.


(*Wyckoff*)
Definition fun_in_ens_subtract {T U:Type} {A:Ensemble T}
           (f:fun_in_ens A U) (a:T) : fun_in_ens (Subtract A a) U.
  do 2 red. intros x h1.
  destruct h1 as [h1 h2].
  do 2 red in f.
  refine (f _ h1).
Defined.

(*Wyckoff*)
Definition fun_in_set_subtract {T:Type} {U:Set} {A:Ensemble T}
           (f:fun_in_set A U) (a:T) : fun_in_ens (Subtract A a) U.
  do 2 red. intros x h1.
  destruct h1 as [h1 h2].
  do 2 red in f.
  refine (f _ h1).
Defined.

(*Wyckoff*)
Definition fun_from_add {T U:Type} {A:Ensemble T} {a:T}
           (f:fun_in_ens (Add A a) U) : fun_in_ens A U :=
  fun x pf => f x (Add_intro1 _ _ _ _ pf).

(*Wyckoff*)
Definition fun_from_add' {T:Type} {U:Set} {A:Ensemble T} {a:T}
           (f:fun_in_set (Add A a) U) : fun_in_set A U :=
  fun x pf => f x (Add_intro1 _ _ _ _ pf).


(*Wyckoff*)
Lemma inj_fun_from_add_compat :
  forall {T U:Type} {A:Ensemble T} {a:T}
         (f:fun_in_ens (Add A a) U),
    inj_in_ens f ->
    inj_in_ens (fun_from_add f).
  intros T U A a f h1.
  red in h1; red.
  intros x y h2 h3 h4.
  specialize (h1 x y (Add_intro1 _ _ _ _ h2) (Add_intro1 _ _ _ _ h3)).
  unfold fun_from_add in h4.
  apply h1 in h4; auto.
Qed.

(*Wyckoff*)
Lemma inj_fun_from_add_compat' :
  forall {T:Type} {U:Set} {A:Ensemble T} {a:T}
         (f:fun_in_ens (Add A a) U),
    inj_in_set f ->
    inj_in_set (fun_from_add' f).
  intros T U A a f h1.
  red in h1; red.
  intros x y h2 h3 h4.
  specialize (h1 x y (Add_intro1 _ _ _ _ h2) (Add_intro1 _ _ _ _ h3)).
  unfold fun_from_add in h4.
  apply h1 in h4; auto.
Qed.

 
(*Wyckoff*)
Lemma inh_im_in_ens :
  forall {T U:Type} (A:Ensemble T) (f:fun_in_ens A U),
    Inhabited A ->
    Inhabited (im_in_ens f).
  intros T U A f h1.
  destruct h1 as [a h1].
  econstructor.
  apply (im_in_ens_intro _ _ h1).
  reflexivity.
Qed.

(*Wyckoff*)
Lemma inh_im_in_set :
  forall {T:Type} {U:Set} (A:Ensemble T) (f:fun_in_set A U),
    Inhabited A ->
    Inhabited (im_in_set f).
  intros T U A f h1.
  destruct h1 as [a h1].
  econstructor.
  apply (im_in_set_intro _ _ h1).
  reflexivity.
Qed.

(*Wyckoff*)
Lemma im_in_ens_empty :
  forall {T U:Type} (f:fun_in_ens (Empty_set T) U),
    im_in_ens f = Empty_set _.
  intros T U f.
  apply Extensionality_Ensembles; red; split; auto with sets.
  red; intros x h2.
  destruct h2 as [x h2]. contradict h2.
Qed.

(*Wyckoff*)
Lemma fun_in_ens_subtract_compat :
  forall {T U:Type} {A:Ensemble T}
         (f:fun_in_ens A U) (a:T),
    let f' := fun_in_ens_subtract f a in
    forall (x:T) (pf:In (Subtract A a) x),
      f' x pf = f x (in_sub _ _ _ pf).
  intros T U A f. simpl. intros x h1 h2.
  unfold fun_in_ens_subtract.
  destruct h2. f_equal. apply proof_irrelevance.
Qed.


(*Wyckoff*)
Lemma im_in_ens_add :
  forall {T U:Type} {A:Ensemble T} {x:T}
         (f:fun_in_ens (Add A x) U),
    ~ In A x ->
    im_in_ens f = Add (im_in_ens (fun_from_add f))
                      (f x (Add_intro2 _ A x)).
  intros T U A x f hn.
  apply Extensionality_Ensembles; red; split; red; intros c h1.
  destruct h1 as [c h1]; subst.
  destruct h1 as [c h1 | c h1]; subst.
  left. econstructor. unfold fun_from_add. 
  f_equal. apply proof_irrelevance.
  destruct h1.
  right.
  rewrite in_sing_iff. 
  f_equal. apply proof_irrelevance.
  destruct h1 as [c h1 | c h1].
  destruct h1 as [c h1]; subst.
  econstructor.
  unfold fun_from_add.
  reflexivity.
  destruct h1.
  econstructor.
  reflexivity.

  Grab Existential Variables.
  assumption.
Qed.

(*Wyckoff*)
Lemma im_in_set_add :
  forall {T:Type} {U:Set} {A:Ensemble T} {x:T}
         (f:fun_in_set (Add A x) U),
    ~ In A x ->
    im_in_set f = Add (im_in_ens (fun_from_add f))
                      (f x (Add_intro2 _ A x)).
  intros T U A x f hn.
  apply Extensionality_Ensembles; red; split; red; intros c h1.
  destruct h1 as [c h1]; subst.
  destruct h1 as [c h1 | c h1]; subst.
  left. econstructor. unfold fun_from_add. 
  f_equal. apply proof_irrelevance.
  destruct h1.
  right.
  rewrite in_sing_iff. 
  f_equal. apply proof_irrelevance.
  destruct h1 as [c h1 | c h1].
  destruct h1 as [c h1]; subst.
  econstructor.
  unfold fun_from_add.
  reflexivity.
  destruct h1.
  econstructor.
  reflexivity.

  Grab Existential Variables.
  assumption.
Qed.


(*Wyckoff*)
Lemma im_in_ens_add' :
  forall {T U:Type} {A:Ensemble T} {x:T}
         (f:fun_in_ens (Add A x) U),
    ~ In A x ->
    im_in_ens f = Add (im_in_ens (fun_in_ens_subtract f x))
                      (f x (Add_intro2 _ A x)).
  intros T U A x f hn.
  apply Extensionality_Ensembles; red; split; red; intros c h1.
  destruct h1 as [c h1]; subst.
  destruct h1 as [c h1 | c h1]; subst.
  left. econstructor. rewrite fun_in_ens_subtract_compat.
  f_equal. apply proof_irrelevance.
  destruct h1.
  right.
  rewrite in_sing_iff. 
  f_equal. apply proof_irrelevance.
  destruct h1 as [c h1 | c h1].
  destruct h1 as [c h1]; subst.
  destruct h1 as [h1 h2]. pose proof h2 as h2'. rewrite in_sing_iff in h2'.
  destruct h1 as [c h1 | c h1].
  econstructor. rewrite fun_in_ens_subtract_compat. f_equal.
  contradiction.
  destruct h1.
  econstructor. f_equal.
  Grab Existential Variables.
  constructor.
  left; auto.
  intro h2. destruct h2. contradiction.
Qed.

(*Wyckoff*)
Lemma im_in_set_add' :
  forall {T:Type} {U:Set} {A:Ensemble T} {x:T}
         (f:fun_in_ens (Add A x) U),
    ~ In A x ->
    im_in_ens f = Add (im_in_ens (fun_in_ens_subtract f x))
                      (f x (Add_intro2 _ A x)).
  intros T U A x f hn.
  apply Extensionality_Ensembles; red; split; red; intros c h1.
  destruct h1 as [c h1]; subst.
  destruct h1 as [c h1 | c h1]; subst.
  left. econstructor. rewrite fun_in_ens_subtract_compat.
  f_equal. apply proof_irrelevance.
  destruct h1.
  right.
  rewrite in_sing_iff. 
  f_equal. apply proof_irrelevance.
  destruct h1 as [c h1 | c h1].
  destruct h1 as [c h1]; subst.
  destruct h1 as [h1 h2]. pose proof h2 as h2'.
  rewrite in_sing_iff in h2'.
  destruct h1 as [c h1 | c h1].
  econstructor. rewrite fun_in_ens_subtract_compat. f_equal.
  contradiction.
  destruct h1.
  econstructor. f_equal.
  Grab Existential Variables.
  constructor.
  left; auto.
  intro h2. destruct h2. contradiction.
Qed.




  
Lemma finite_im_in_ens :
  forall {T U:Type} {A:Ensemble T} (f:fun_in_ens A U),
    Finite A ->
    Finite (im_in_ens f).
  intros T U A f h1. revert f.
  induction h1 as [|S h2 h3 x h4].
  intro. rewrite im_in_ens_empty. constructor.
  intro f.
  pose (fun_in_ens_subtract f x) as f'.
  pose proof (sub_add_same_nin _ _ h4) as h5.
  assert (h6:Included S (Subtract (Add S x) x)). auto with sets.
  pose (fun (x:T) (pf:In S x) => (f' x (h6 _ pf))) as f''.
  specialize (h3 f'').
  unfold f'', f'  in h3.

  rewrite im_in_ens_add'.
  apply Add_preserves_Finite.
  eapply Finite_downward_closed. apply h3.
  red; intros c h7. destruct h7 as [c h7]. clear h5. subst.
  econstructor; auto. 
  rewrite fun_in_ens_subtract_compat; auto.
  destruct h7 as [h7 h8].
  destruct h7 as [c h7 | c h7]; auto.
  rewrite fun_in_ens_subtract_compat.
  f_equal. apply proof_irrelevance.
  contradiction.
  assumption.
  Grab Existential Variables.
  destruct h7 as [h7 h8].
  destruct h7 as [c h7 | c h7]; auto.
  contradiction.
Qed.

         
Lemma im_in_ens_const_sing :
  forall {T:Type} {U:Set} (A:Ensemble T) (f:fun_in_set A U),
    Inhabited A ->
    (forall (x x':T) (pf:In A x) (pf':In A x'),
       f x pf = f x' pf') ->
    exists! u:U,
      Singleton u = im_in_set f.
  intros T U A f h1 h2.
  destruct h1 as [a h1].
  exists (f _ h1).
  red. split.
  apply Extensionality_Ensembles; red; split; red; intros x h3.
  destruct h3.
  econstructor.
  reflexivity.
  destruct h3. subst.
  erewrite h2.
  constructor.
  intros u h3.
  pose proof (In_singleton _ u) as h4.
  rewrite h3 in h4.
  destruct h4 as [u h4]. subst.
  apply h2; auto.
Qed.

Lemma im_in_ens_eq :
  forall {T U:Type} {A:Ensemble T} (f:fun_in_ens A U),
    im_in_ens f = Im (full_sig A)
                     (fun x => f (proj1_sig x) (proj2_sig x)).
  intros T U A f.
  apply Extensionality_Ensembles; red; split; red; intros x h1.
  destruct h1 as [x h1]; subst.
  apply Im_intro with (exist _ _ h1); auto; constructor.
  destruct h1 as [x h1]; subst.
  econstructor. reflexivity.
Qed.


Lemma im_im_in_ens :
  forall {T U V:Type} {E:Ensemble T}
         (f:fun_in_ens E U)
         (g:U->V),
    Im (im_in_ens f) g =
    im_in_ens (fun x (pf:Inn E x) =>
                 g (f x pf)).
  intros T U V E f g.
  apply Extensionality_Ensembles; red; split; red; intros x h1; destruct h1 as [x h1]; subst.
  destruct h1 as [x h1]; subst.
  econstructor.
  reflexivity.
  econstructor.
  econstructor.
  reflexivity.
  reflexivity.
Qed.


Definition fun_in_ens_transfer {T U:Type} {A A':Ensemble T} (f:fun_in_ens A U)
      (pfe:A = A') : fun_in_ens A' U :=
  fun x (pfi:In A' x) =>
    let pf' :=  in_eq_set _ _ (eq_sym pfe) _ pfi in
    f x pf'.

Lemma fun_in_ens_transfer_compat :
  forall {T U:Type} {A A':Ensemble T} (f:fun_in_ens A U)
         (pfe:A = A') (x:T) (pfi:Inn A' x),
    let pf' := in_eq_set _ _ (eq_sym pfe) _ pfi in
    fun_in_ens_transfer f pfe x pfi =
    f x pf'.
  unfold fun_in_ens_transfer; intros; auto.
Qed.

Definition fun_in_set_transfer {T:Type} {U:Set}
           {A A':Ensemble T} (f:fun_in_ens A U)
           (pfe:A = A') : fun_in_set A' U :=
  fun x (pfi:In A' x) =>
    let pf' := in_eq_set _ _ (eq_sym pfe) _ pfi in
    f x pf'.

Lemma fun_in_set_transfer_compat :
  forall {T:Type} {U:Set} {A A':Ensemble T} (f:fun_in_ens A U)
         (pfe:A = A') (x:T) (pfi:Inn A' x),
    let pf' := in_eq_set _ _ (eq_sym pfe) _ pfi in
    fun_in_set_transfer f pfe x pfi =
    f x pf'.
  unfold fun_in_set_transfer; intros; auto.
Qed.

Lemma fun_in_ens_transfer_eq_refl :
  forall {T U:Type} {A:Ensemble T} (f:fun_in_ens A U),
    fun_in_ens_transfer f eq_refl = f.
  intros; apply functional_extensionality_dep; intro;
  apply functional_extensionality; intro; unfold fun_in_ens_transfer;
  f_equal; apply proof_irrelevance.
Qed.

Lemma fun_in_set_transfer_eq_refl :
  forall {T:Type} {U:Set} {A:Ensemble T} (f:fun_in_ens A U),
    fun_in_ens_transfer f eq_refl = f.
  intros; apply functional_extensionality_dep; intro;
  apply functional_extensionality; intro; unfold fun_in_ens_transfer;
  f_equal; apply proof_irrelevance.
Qed.

Lemma im_in_ens_transfer :
  forall {T U:Type} {A B:Ensemble T} (f:fun_in_ens A U)
         (pfe:A = B),
    im_in_ens f =
    im_in_ens (fun_in_ens_transfer f pfe).
  intros; subst.
  rewrite fun_in_ens_transfer_eq_refl; auto.
Qed.

  
Lemma linv_in_ens_ex :
  forall {T U:Type} {A:Ensemble T} (f:fun_in_ens A U)
         (pfi:inj_in_ens f),
  exists! f':fun_in_ens (im_in_ens f) T,
    forall (x:T) (pf:In A x),
      f' (f x pf) (im_in_ens_intro _  _ pf _ eq_refl) = x.
  intros T U A f h1.
  assert (h2:forall y:U,
               Inn (im_in_ens f) y ->
               exists! x:T,
                 exists pf:Inn A x,
                   f x pf = y).
    intros y h2.
    destruct h2; subst.
    exists x.
    red.
    split.
    exists pf; auto.
    intros x' h2.
    destruct h2 as [h2 h3].
    apply h1 in h3; auto.
  exists (fun y pf => proj1_sig (constructive_definite_description _ (h2 y pf))).
  red.
  split.
  intros x h3.
  destruct constructive_definite_description as [x' h4].
  destruct h4 as [h4 h5].
  pose proof (h1 _ _ _ _ h5); subst.
  simpl; auto.

  intros f' h3.
  apply functional_extensionality_dep.
  intro y.
  apply functional_extensionality.
  intro h4.
  destruct constructive_definite_description as [z' h5].
  destruct h5 as [h5]; subst.
  simpl.
  symmetry.
  specialize (h3 z' h5).
  rewrite <- h3 at 2.
  f_equal.
  apply proof_irrelevance.
Qed.


Definition linv_in_ens {T U:Type} {A:Ensemble T}
           (f:fun_in_ens A U) (pfi:inj_in_ens f) :
  fun_in_ens (im_in_ens f) T :=
  proj1_sig (constructive_definite_description _ (linv_in_ens_ex f pfi)).

Lemma linv_in_ens_compat :
  forall {T U:Type} {A:Ensemble T} (f:fun_in_ens A U)
         (pfi:inj_in_ens f),
    let f' := linv_in_ens f pfi in
    forall (x:T) (pf:In A x),
      f' (f x pf) (im_in_ens_intro _  _ pf _ eq_refl) = x.
  unfold linv_in_ens; intros; destruct constructive_definite_description; auto.
Qed.

Lemma in_linv_in_ens :
  forall {T U:Type} {A:Ensemble T} (f:fun_in_ens A U)
         (pfi:inj_in_ens f) (y:U) (pf:Inn (im_in_ens f) y),
    Inn A (linv_in_ens f pfi y pf).
  intros T U A f h1 y h2.
  destruct h2; subst.
  rewrite linv_in_ens_compat; auto.
Qed.


Lemma bij_impl_inv_in_ens :
  forall {T U} {A:Ensemble T} {B:Ensemble U}
         (f:fun_in_ens' A B),
    bij_in_ens f ->
    invertible_in_ens f.
  intros T U A B f h1. 
  destruct h1 as [h2 h3]. 
  pose proof (surj_in_ens_compat _ h3) as h4. 
  pose proof (linv_in_ens_compat f h2) as h5; simpl in h5.
  assert (h6:forall y (pfy:Inn B y),
               Inn (im_in_ens f) (exist _ y pfy)).
    intros y hy.
    pose proof hy as hy'.
    rewrite <- h4 in hy'.
    destruct hy' as [y h6]; clear h4; subst.
    destruct h6 as [y h6]; subst.
    apply im_in_ens_intro with y h6.
    apply proj1_sig_injective; auto.
  pose (fun y (pfy:Inn B y) =>
          linv_in_ens f h2 (exist _ y pfy) (h6 y pfy)) as g.
  assert (h7:forall y (pfy:Inn B y),
               Inn A (g y pfy)).
    unfold g; intros; apply in_linv_in_ens.
  pose (fun y (pfy:Inn B y) =>
          exist (fun c => Inn A c) (g y pfy) (h7 y pfy)) as g'.
  apply (invertible_in_ens_intro f g').
  unfold g', g.
  constructor.
  intros x hy; apply proj1_sig_injective; simpl.
  specialize (h5 x hy).
  rewrite <- h5.
  apply f_equal_dep.
  apply proj1_sig_injective; auto.
  simpl.
  intros y hy.
  pose proof hy as hy'.
  rewrite <- h4 in hy'.
  destruct hy' as [y h8]; clear h4; subst.
  destruct h8 as [y h8]; subst.
  specialize (h5 y h8).
  apply proj1_sig_injective; simpl.
  f_equal.
  apply f_equal_dep.
  rewrite <- h5.
  apply f_equal_dep.
  apply proj1_sig_injective; simpl; repeat f_equal.
Qed.



Definition fun_from_add_im 
           {T U V:Type} {A:Ensemble T} {a:T}
           {g:fun_in_ens (Add A a) U}
           (f:fun_in_ens (im_in_ens g) V) (pfi:inj_in_ens g) :
  fun_in_ens (im_in_ens (fun_from_add g)) V :=
  fun x pf =>
    let pfi' := inj_fun_from_add_compat _ pfi in
    f (g (linv_in_ens (fun_from_add g) pfi' x pf)
         (Add_intro1 _ _ _ _ (in_linv_in_ens (fun_from_add g) pfi' _ _)))
      (im_in_ens_intro
         g _
         (Add_intro1 _ _ _ _ (in_linv_in_ens (fun_from_add g) pfi' _ _))
         _ eq_refl).

Lemma fun_from_add_im_compat : 
  forall {T U V:Type} {A:Ensemble T} {a:T}
         {g:fun_in_ens (Add A a) U}
         (f:fun_in_ens (im_in_ens g) V) (pfi:inj_in_ens g)
         (x:T) (pf:Inn A x),
    fun_from_add_im f pfi (fun_from_add g x pf)
                    (im_in_ens_intro (fun_from_add g) _ pf _ eq_refl) =
    f (g x (Add_intro1 _ _ _ _ pf))
      (im_in_ens_intro g _ (Add_intro1 _ _ _ _ pf) _ eq_refl).
  unfold fun_from_add_im; intros.
  apply f_equal_dep.
  apply f_equal_dep.
  rewrite linv_in_ens_compat; auto.
Qed.

Lemma fun_from_add_im_functional : 
  forall {T U V:Type} {A:Ensemble T} {a:T}
         {g:fun_in_ens (Add A a) U}
         (f:fun_in_ens (im_in_ens g) V) (pfi:inj_in_ens g)
         (y y':U) (pfeq:y = y')
         (pf: Inn (im_in_ens (fun_from_add g)) y)
         (pf':Inn (im_in_ens (fun_from_add g)) y'),
    fun_from_add_im f pfi y pf =
    fun_from_add_im f pfi y' pf'.
  intros; subst; f_equal; apply proof_irrelevance.
Qed.


Lemma fun_in_set_inh_const : 
  forall {T:Type} {U:Set} (A:Ensemble T)
         (f:fun_in_set A U)
         (pf:forall u u':U, u = u'),
    Inhabited A -> U.
  intros T U A f h0 h1.
  pose proof (im_in_ens_const_sing  _ f h1) as h2.
  hfirst h2.
  intros x x' h3 h4.
  apply h0.
  specialize (h2 hf).
  refine (proj1_sig (constructive_definite_description _ h2)).
Qed.


(*Wyckoff*)
Lemma finite_union_finite_left :
  forall {T:Type} (A B:Ensemble T),
    Finite (Union A B) ->
    Finite A.
  intros T A B h1.
  eapply Finite_downward_closed.
  apply h1.
  auto with sets.
Qed.

(*Wyckoff*)
Lemma finite_union_finite_right :
  forall {T:Type} (A B:Ensemble T),
    Finite (Union A B) ->
    Finite B.
  intros T A B h1.
  eapply Finite_downward_closed.
  apply h1.
  auto with sets.
Qed.


(*Wyckoff*)
Lemma finite_couple : forall {T:Type} (x y:T),
                        Finite (Couple x y).
intros T x y.
destruct (eq_dec x y) as [he | hne].
subst.
rewrite couple_singleton. 
apply Singleton_is_finite.
rewrite couple_add_sing.
apply Add_preserves_Finite.
apply Singleton_is_finite.
Qed.



(*Wyckoff*)
Lemma incl_subtract :
  forall {T:Type} (P:Ensemble T) (x:T),
    Included (Subtract P x) P.
intros T P x. red. intros x' h1.
destruct h1. assumption.
Qed.


(*Wyckoff*)
Lemma subtract_preserves_finite : 
  forall {T:Type} (P:Ensemble T) (x:T),
    Finite P -> Finite (Subtract P x).
intros T P x h1.
assert (h2:Included (Subtract P x) P).  apply incl_subtract.
apply Finite_downward_closed with P; auto.
Qed.

(*Wyckoff*)
Lemma subtract_finite :
  forall {T:Type} (pfdt:type_eq_dec T) (A:Ensemble T) (x:T),
    Finite (Subtract A x) ->
    Finite A.
  intros T h0 A x h1.
  pose proof (Add_preserves_Finite _ _ x h1) as h2.
  eapply Finite_downward_closed.
  apply h2.
  rewrite add_sub_eq; auto.
  auto with sets.
Qed.


(*Wyckoff*)
Lemma couple_comm : 
  forall {T:Type} (a b:T), 
    Couple a b = Couple b a.
  intros T a b.
  apply Extensionality_Ensembles.
  red. split.
  (* <= *)
  red. intros ? h1.
  destruct h1; subst; constructor.
  (* >= *)
  red. intros ? h1.
  destruct h1; subst; constructor.
Qed.

(*Wyckoff*)
Lemma  bool_couple : Full_set bool = Couple true false.
apply Extensionality_Ensembles.
red. split. red. intros x ?. destruct x. left. right.
red. intros; constructor.
Qed.

(*Wyckoff*)
Lemma finite_bool : Finite (Full_set bool).
rewrite bool_couple.
apply finite_couple.
Qed.


(*Wyckoff*)
Lemma finite_inh_or_empty : 
  forall {T:Type} (A:Ensemble T),
    Finite A -> 
    Inhabited A \/ A = Empty_set _.
intros T A h1.
induction h1 as [|A h1 h2 a h3]. right; auto.
left. apply Inhabited_intro with a.
right.
constructor.
Qed.

(*Wyckoff*)
(*uses LEM*)
Lemma infinite_inh :
  forall {T:Type} (A:Ensemble T),
    ~ Finite A -> Inhabited A.
  intros T A h1.
  assert (h2:A <> Empty_set _).
    intro; subst.
    contradict h1. constructor.
    apply not_empty_Inhabited; auto.
Qed.

(*This is the finite anlaogue of the clasical Standard Library function,
[not_empty_Inhabited] . . . but of course without LEM*)
Lemma fin_not_empty_inh : 
  forall {T:Type} (A:Ensemble T),
    Finite A ->
    A <> Empty_set _ ->
    Inhabited A.
  intros T A h1 h2.
  apply finite_inh_or_empty in h1.
  destruct h1; subst; auto.
  contradict h2; auto.
Qed.

(*Wyckoff*)
Lemma fin_in_or :
  forall {T:Type} (pfdt:type_eq_dec T)
         (A:Ensemble T) (x:T),
    Finite A ->
    In A x \/ ~In A x.
  intros T h0 A x h1. revert x.
  induction h1 as [|A h1 h2 a h3].
  intros; right; intro; contradiction.
  intro x.
  specialize (h2 x).
  destruct h2 as [h2 | h2].
  left. left; auto.
  destruct (h0 a x) as [h4 | h4]; simpl; subst.
  left; right; constructor; auto.
  right.
  intro h5.
  destruct h5 as [x h5 | x h5]; try contradiction; subst.
  destruct h5; contradict h4; auto.
Qed.

      
(*Wyckoff*)
Lemma finite_cardinal_unq : 
  forall {T:Type} (S:Ensemble T), 
    Finite  S -> exists! n:nat, cardinal _ S n.
intros T S h1.
pose proof (finite_cardinal _ _ h1) as h2.
destruct h2 as [n h2].
exists n.
red.
split; try assumption.
apply cardinal_unicity; assumption.
Qed.

                          
(*Wyckoff*)
Definition card_fun {T:Type} (S:Ensemble T)
           (pf:Finite S) : nat := 
  proj1_sig (constructive_definite_description
               _
               (finite_cardinal_unq S pf)).

(*Wyckoff*)
Lemma card_fun_compat : 
  forall {T:Type} (S:Ensemble T) (pf:Finite S),
    cardinal _ S (card_fun _ pf).
intros T S h1.
unfold card_fun.
destruct constructive_definite_description.
simpl.
assumption.
Qed.

(*Wyckoff*)
Lemma card_fun_empty : 
  forall T:Type,
    card_fun (Empty_set T) (Empty_is_finite T) = 0.
intro T.
pose proof (card_fun_compat _ (Empty_is_finite T)) as h1.
pose proof (card_empty T) as h2.
eapply cardinal_is_functional; auto. apply h1. assumption.
Qed.

                    


(*Wyckoff*)
(*Uses LEM.*)
Lemma card_fun1_ex : 
  forall {T:Type} (S:Ensemble T),
    exists! n:nat, (Finite S -> cardinal _ S n) /\
                   (~Finite S -> n = 0).
intros T S.
destruct (classic (Finite S)) as [h1 | h2].
exists (card_fun _ h1). red. split.
split. intro h2. apply card_fun_compat.
intro; contradiction.
intros n h2. destruct h2 as [h2l h2r]. specialize (h2l h1).
pose proof (card_fun_compat _ h1) as h3.
pose proof cardinal_is_functional.
pose proof (cardinal_is_functional _ _ _ h2l _ _ h3 (eq_refl _)).
subst. reflexivity.
exists 0. red. split. split.
intros; contradiction. auto.
intros n h3. destruct h3 as [h3l h3r].
symmetry. apply h3r; auto.
Qed.

(*Wyckoff*)
Definition card_fun1 {T:Type} (S:Ensemble T) :=
  proj1_sig (constructive_definite_description _ (card_fun1_ex S)).

(*Wyckoff*)
Lemma card_fun1_compat : 
  forall {T:Type} (S:Ensemble T),
    let n:= card_fun1 S in 
    (Finite S -> cardinal _ S n) /\
    (~Finite S -> n = 0).
  intros T S. unfold card_fun1.
  destruct constructive_definite_description. simpl. assumption.
Qed.

(*Wyckoff*)
Lemma card_fun1_empty : 
  forall (T:Type), card_fun1 (Empty_set T) = 0.
intro T.
pose proof (card_fun1_compat (Empty_set T)) as h1.
destruct h1 as [h1l h1r].
specialize (h1l (Empty_is_finite T)).
pose proof (card_empty T) as h2.
eapply cardinal_is_functional; auto. apply h1l. apply h2.
Qed.


(*Wyckoff*)
(*Uses LEM.*)
Lemma card_fun1_O : 
  forall {T:Type} (A:Ensemble T), card_fun1 A = 0 ->
         A = Empty_set _ \/ ~Finite A.
intros T A h1.
pose proof (card_fun1_compat A) as h2.
destruct h2 as [h2l h2r].
destruct (classic (Finite A)) as [h3 | h4].
apply h2l in h3. rewrite h1 in h3. 
apply cardinalO_empty in h3. left. assumption.
right. assumption.
Qed.

(*Wyckoff*)
Lemma card_fun1_sing : 
  forall {T:Type} (x:T), card_fun1 (Singleton x) = 1.
intros T x.
pose proof (card_fun1_compat (Singleton x)) as h1.
destruct h1 as [h1l h1r].
specialize (h1l (Singleton_is_finite _ x)).
pose proof (card_sing x) as h2.
eapply cardinal_is_functional; auto.
apply h1l. apply h2.
Qed.

(*Wyckoff*)
Lemma lt_card_fun1_finite : 
  forall {T:Type} (A:Ensemble T) 
         (n:nat), n < card_fun1 A -> Finite A.
intros T A n h1.
pose proof (card_fun1_compat A) as h2.
destruct h2 as [h2 h3].
apply NNPP.
intro h4. apply h3 in h4.
rewrite h4 in h1. omega.
Qed.
                                    
(*Wyckoff*)
Lemma card_fun1_1 : 
  forall {T:Type} (A:Ensemble T), 
    card_fun1 A = 1 -> exists x:T, A = Singleton x.
intros T A h1.
assert (h0:0 < card_fun1 A). omega.
apply lt_card_fun1_finite in h0. 
revert h1.
induction h0 as [|A h2 h3 x h4].
intro h1. rewrite card_fun1_empty in h1. omega.
exists x. 
assert (h5:A = Empty_set _).
  apply NNPP. intro h5. apply not_empty_Inhabited in h5.
  destruct h5 as [y h5].
  assert (h6:x <> y). intro h6. subst. contradiction.
pose proof (card_fun1_compat (Add A x)) as h7.
destruct h7 as [h7l h7r]. clear h7r.
pose proof (Add_preserves_Finite _ _ x h2) as h8. apply h7l in h8.
rewrite h1 in h8.
pose proof card_soustr_1.
pose proof (card_soustr_1 _ _ _ h8 _ (Add_intro2 _ A x)) as h9.
simpl in h9.
rewrite sub_add_same_nin in h9.
apply cardinalO_empty in h9. subst. contradiction. assumption.
subst.
apply add_empty_sing; auto.
Qed.

(*Wyckoff*)
Lemma card_fun1_ge_1 : 
  forall {T:Type} (A:Ensemble T),
    Inhabited A -> Finite A ->
    1 <= card_fun1 A.
intros T A h1 h0.
destruct (le_lt_dec 1 (card_fun1 A)) as [|h2]; auto.
assert (h3:card_fun1 A = 0). omega.
contradict h1.
apply card_fun1_O in h3.
destruct h3 as [h3|]; try contradiction.
subst.
intro h3.
destruct h3; contradiction.
Qed.

(*Wyckoff*)
Lemma subtract_eq_empty_card :
  forall {T:Type} (pfdt:type_eq_dec T) (A:Ensemble T) (x:T),
    Subtract A x = Empty_set _ ->
    card_fun1 A <= 1.
  intros T h0 A x h1.
  pose proof (Empty_is_finite T) as h2.
  rewrite <- h1 in h2.
  apply subtract_finite in h2; auto.
  apply finite_inh_or_empty in h2.
  destruct h2 as [h2 | h2].
  pose proof (inh_subtract_empty h0 _ _ h2 h1) as h3.
  pose proof (f_equal (fun S => Add S x) h1) as h4.
  simpl in h4.
  rewrite add_sub_eq in h4.
  rewrite in_add_eq in h4.
  rewrite add_empty_sing in h4.
  subst.
  rewrite card_fun1_sing; auto with arith.
  assumption.
  assumption.
  subst.
  rewrite card_fun1_empty; auto with arith.
Qed.

  

(*Wyckoff*)
Lemma card_fun_card_fun1_compat : 
  forall {T:Type} (A:Ensemble T) (pf:Finite A),
    card_fun A pf = card_fun1 A.
intros T A h1.
pose proof (card_fun_compat A h1) as h2.
pose proof (card_fun1_compat A) as h3.
destruct h3 as [h3l h3r]. specialize (h3l h1).
eapply cardinal_is_functional; auto. apply h2. apply h3l.
Qed.

(*Wyckoff*)
Lemma card_add_nin : 
  forall {T:Type} (A:Ensemble T) (n:nat),
    cardinal _ A n -> 
    forall x:T, ~ In A x -> cardinal _ (Add A x) (S n).
intros T A n h1 x h2.
constructor; auto.
Qed.                                
           
(*Wyckoff*)
Lemma card_add_nin' : 
  forall {T:Type} (A:Ensemble T),
    Finite A -> forall x:T, 
                  ~ In A x -> 
                  card_fun1 (Add A x) = S (card_fun1 A).
intros T A h1 x h2.
pose proof (card_fun1_compat A) as h3. destruct h3 as [h3l h3r].
specialize (h3l h1).
pose proof (card_fun1_compat (Add A x)) as h4. destruct h4 as [h4l h4r].
specialize (h4l (Add_preserves_Finite _ _ _ h1)).
eapply card_add_nin in h3l.
eapply cardinal_is_functional.
apply h4l. apply h3l. apply eq_refl. apply h2.
Qed.


(*Wyckoff*)
Lemma incl_card_fun1 : 
  forall {T:Type} (A B:Ensemble T),
    Finite B ->
    Included A B -> card_fun1 A <= card_fun1 B.
intros T A B h4 h1.
assert (h2:Finite A). 
  eapply Finite_downward_closed; auto.
  apply h4; auto.
  assumption.
pose proof (finite_cardinal _ _  h2) as h5.
pose proof (finite_cardinal _ _ h4) as h6.
destruct h5 as [m h5]. destruct h6 as [n h6].
pose proof (incl_card_le _ _ _ _ _ h5 h6 h1) as h7.
pose proof (card_fun1_compat A) as [h8].
pose proof (card_fun1_compat B) as [h9].
specialize (h8 h2). specialize (h9 h4).
pose proof (cardinal_is_functional _ _ _ h5 _ _ h8 (eq_refl _)).
subst.
pose proof (cardinal_is_functional _ _ _ h9 _ _ h6 (eq_refl _)).
subst.
assumption.
Qed.



(*Wyckoff*)
Lemma finite_inh_or_empty_dec :
  forall {T:Type} (A:Ensemble T),
    Finite A -> {Inhabited A} + {A = Empty_set _}.
  intros T A h1. 
  pose proof (finite_cardinal_unq A h1) as h2.
  pose proof (proj2_sig (constructive_definite_description _ h2)) as h3.
  simpl in h3.
  let P := type of h3 in match P with cardinal _ _ ?n =>
                                      destruct n as [|m] end.
  right.
  apply cardinalO_empty; auto.
  left.
  inversion h3. subst.
  econstructor. right. constructor.
Qed.


(*Wyckoff*)                       
Lemma singleton_inc : forall {T:Type} (a:T) (A:Ensemble T),
                        Included A (Singleton a) ->
                        A = Empty_set _ \/ A = Singleton a.
intros T a A h0.
pose proof (Singleton_is_finite _ a) as h6.
assert (h7:cardinal _ (Singleton a) 1).
assert (h7:Singleton a = Add (Empty_set _) a).
unfold Add.
rewrite empty_union. reflexivity.
rewrite h7.
constructor. constructor. intro. contradiction.
pose proof (Finite_downward_closed _ _ h6 _ h0) as h8.
pose proof (finite_cardinal _ _ h8) as h9.
destruct h9 as [n h9].
pose proof (incl_card_le _ _ _ _ _ h9 h7 h0) as h10.
pose proof (le_pred n 1 h10) as h11.
simpl in h11.
pose proof (le_n_0_eq _ h11) as h12.
unfold pred in h12.

destruct (zerop n) as [? | h13].
subst.        
left.
apply cardinalO_empty. assumption.
assert (h14: n = S O). auto with arith.
subst.
right.
apply Extensionality_Ensembles.
red. split. assumption.
red.
intros x h14.
destruct h9.    
pose proof (lt_irrefl 0).
contradiction.        
assert (h15:a = x0).
unfold Included in h0.
pose proof (Add_intro2 _ A x0) as h15.
specialize (h0 _ h15).
destruct h0; auto.
subst.
destruct h14; subst.
apply Add_intro2.
Qed.


(*Wyckoff*)
Lemma sing_eq_add :
  forall {T:Type} (A:Ensemble T) (x a:T),
    Singleton x = Add A a ->
    a = x /\ (A = Empty_set _ \/ A = Singleton x).
  intros T A x a h1. split.
  pose proof (Add_intro2 _ A a) as h2.
  rewrite <- h1 in h2.
  destruct h2; auto.
  assert (h2:A <> Empty_set _ -> A = Singleton x).
  intro h2.
  apply Extensionality_Ensembles; red; split; red; intros c h3.
  pose proof (Add_intro1 _ A a _ h3) as h4.
  rewrite h1.
  assumption.
  destruct h3.
  pose proof (Add_intro2 _ A a) as h3.
  rewrite <- h1 in h3.
  destruct h3.
  apply not_empty_Inhabited in h2.
  destruct h2 as [c h2].
  pose proof (Add_intro1 _ A x _ h2) as h3.
  rewrite <- h1 in h3.
  destruct h3.
  assumption.
  tauto.
Qed.



(*Wyckoff*)
Lemma singleton_inc_int : 
  forall {T:Type} (a:T) (A:Ensemble T),
    Intersection A (Singleton a) =  Empty_set _ \/ 
    Intersection A (Singleton a) = Singleton a.
intros T a A.
assert (h1:Included (Intersection A (Singleton a)) (Singleton a)); 
auto with sets.
apply singleton_inc; auto.
Qed.



(*Wyckoff*)
Lemma card_sub_in : 
  forall {T:Type} (A:Ensemble T),
    Finite A -> forall x:T,
                  In A x ->
                  card_fun1 A = S (card_fun1 (Subtract A x)).
intros T A h1 x h2.
pose proof (card_fun1_compat A) as h3. destruct h3 as [h3l h3r].
clear h3r. 
specialize (h3l h1).
pose proof (card_fun1_compat (Subtract A x)) as h4.
destruct h4 as [h4l h4r]. clear h4r.
pose proof (subtract_preserves_finite _ x h1) as h5.
specialize (h4l h5). 
rewrite <- (add_sub_compat_in A x) at 1; auto. 
rewrite card_add_nin'. f_equal.
assumption.
intro h6.
destruct h6 as [h6 h7]. contradict h7. constructor.
Qed.

(*Wyckoff*)
Lemma card_fun1_subtract : 
  forall {T:Type} (A:Ensemble T),
    Finite A ->
    forall x:T,
      In A x ->
      card_fun1 (Subtract A x) = pred (card_fun1 A).
  intros T A h1 x h2.
  pose proof (card_sub_in A h1 x h2).
  omega.
Qed.


(*Wyckoff*)
Lemma card_fun1_eq_sub :
  forall {T U:Type} (A:Ensemble T) (B:Ensemble U),
    Finite A -> Finite B ->
    card_fun1 A = card_fun1 B ->
    forall (x:T) (y:U),
      In A x -> In B y ->
      card_fun1 (Subtract A x) = card_fun1 (Subtract B y).
intros T U A B hfa hfb h1 x y h2 h3.
rewrite (card_sub_in A hfa x h2) in h1.
rewrite (card_sub_in B hfb y h3) in h1.
apply S_inj; auto.
Qed.
  

(*Wyckoff*)
Lemma cardinal_couple : forall {T:Type} (x y:T),
                             x <> y ->
                             cardinal _ (Couple x y) 2.
intros T x y h1.
rewrite couple_add_sing.
constructor. apply card_sing. intro h2. destruct h2; auto.
Qed.

(*Wyckoff*)
Lemma inclusion_intersection : forall {T:Type} (P S R : Ensemble T), 
  (Included P S /\ Included P R) <-> Included P (Intersection S R).

intros T P S R.
split.
(*left*)
intro h1.
destruct h1 as [h2 h3].
red.
intros x h4.
constructor; auto with sets.
(*right*)
intro h1.
split; red.
  (*left*)
  intros x h5.
  red in h1.
  pose proof (h1 x h5) as h6.
  inversion h6; trivial.
  (*right*)
  intros x h7.
  pose proof (h1 x h7) as h6. 
  inversion h6; trivial.
Qed.

(*Wyckoff*)
Lemma intersection_preserves_inclusion :
  forall {T:Type} (P Q R:Ensemble T), 
  Included P Q -> Included (Intersection R P) (Intersection R Q).
intros T P Q R h1.
red.
intros x h2.
destruct h2 as [x h2l h2r].
auto with sets.
Qed.

(*Wyckoff*)
Lemma union_preserves_inclusion : forall {T:Type} (P Q R:Ensemble T), 
  Included P Q -> Included (Union R P) (Union R Q).
intros T P Q R h1.
red.
intros x h2.
destruct h2 as [x h2 | x h3].
left; auto.
apply h1 in h3.
right; auto.
Qed.


(*Wyckoff*)
Lemma inclusion_mono : 
  forall {T:Type} (A B C D:Ensemble T),
    Included A B -> Included C D ->
    Included (Intersection A C) (Intersection B D).
intros T A B C D h1 h2.
red. intros x h3.
destruct h3 as [x h3 h4].
apply h1 in h3. apply h2 in h4.
constructor; auto.
Qed.



(*Wyckoff*)
Lemma in_intersection_iff : 
  forall {T:Type} (A B:Ensemble T) (x:T),
    In (Intersection A B) x <-> In A x /\ In B x.
intros T A B x.
split.
intros h1.
destruct h1. split; auto.
intros h1.
destruct h1. constructor; auto.
Qed.

(*Wyckoff*)
Lemma in_union_iff : 
  forall {T:Type} (A B:Ensemble T) (x:T),
    In (Union A B) x <-> In A x \/ In B x.
intros T A B x.
split.
intro h1.
destruct h1. left; auto. right; auto.
intro h1.
destruct h1. left; auto. right; auto.
Qed.



(*Wyckoff*)
Lemma in_match_sumbool_excl_middle :
  forall {T U:Type} {V:U->Type} 
         (P:forall (u:U) (v:V u), Prop)
         (A:forall (u:U) (v:V u) (pf:P u v), Ensemble T)
         (B:forall (u:U) (v:V u) (pf:~ P u v), Ensemble T)
         (u:U) (v:V u)
         (pfd:{P u v} + {~P u v})
         (x:T),
      Inn match pfd with
            | left pfa => A u v pfa
            | right pfb => B u v pfb
          end x <->
      (exists pfp:P u v, Inn (A u v pfp) x) \/
      (exists pfnp: ~P u v, Inn (B u v pfnp) x).
  intros T U V P A B u v h1 x.
  destruct h1 as [h1 | h1].
  split.
  intro h2.
  left. exists h1; auto.
  intro h2. destruct h2 as [h2 | h2]. destruct h2 as [h2 h3].
  assert (h1 = h2). apply proof_irrelevance. subst; auto.
  destruct h2; contradiction.
  split.
  intro h2.
  right. exists h1; auto.
  intro h2. destruct h2 as [h2 | h2]. destruct h2; contradiction.
  destruct h2 as [h2 h3].
  assert (h1 = h2). apply proof_irrelevance. subst; auto.
Qed.
  

(*Wyckoff*)
Lemma intersection_preserves_finite : 
  forall {T:Type} (P Q:Ensemble T),
    Finite P -> Finite (Intersection P Q).
intros T P Q h1.
assert (h2:Included (Intersection P Q) P). auto with sets.
apply Finite_downward_closed with P; auto.
Qed.

(*Wyckoff*)
Lemma add_preserves_infinite : 
  forall {T:Type} (A:Ensemble T) (a:T),
    ~Finite A -> ~Finite (Add A a).
intros T A a h1. 
intro h2.
assert (h5:Included A (Add A a)). auto with sets.
pose proof (Finite_downward_closed _ _ h2 _  h5).
contradiction.
Qed.



(*Wyckoff*)
Lemma in_sub_inj :
  forall {T:Type} (E F:Ensemble T) (x:T),
    In E x -> In F x ->
    Subtract E x = Subtract F x ->
    E = F.
  intros T E F x h1 h2 h3.
  apply Extensionality_Ensembles; red; split; red; intros a h5.
  destruct (eq_dec a x) as [h6 | h6] ;subst; auto.
  assert (h7:In (Subtract E x) a).  constructor; auto.
    intro h7. destruct h7; auto.
    rewrite h3 in h7.
  destruct h7; auto.
  destruct (eq_dec a x) as [h6 | h6] ;subst; auto.
  assert (h7:In (Subtract F x) a).  constructor; auto.
    intro h7. destruct h7; auto.
    rewrite <- h3 in h7.
  destruct h7; auto.
Qed.



(*Wyckoff*)
Lemma inclusion_iff_intersection_eq : forall {T:Type} (P Q:Ensemble T), 
  Included P Q <-> Intersection P Q = P.
intros T P Q.
split.
(* -> *)
intro.
apply Extensionality_Ensembles.
red. split; auto with sets.
(* <= *)
intro h1.
red.
intros x h2.
rewrite <- h1 in h2.
destruct h2; assumption.
Qed.

(*Wyckoff*)
Lemma inclusion_iff_union : forall {T:Type} (P Q:Ensemble T), 
  Included P Q <->  Union P Q = Q.
intros T P Q.
split. intro h1.
apply Extensionality_Ensembles.
red. split.
red. intros x h2. destruct h2; auto with sets.
red. intros; right; auto.
intro h1. red. intros.  rewrite <- h1. auto with sets.
Qed.


(*Wyckoff*)
Lemma incl_empty : forall {T:Type} (S:Ensemble T),
                     Included (Empty_set _) S.
intros T S.
red. intros; contradiction.
Qed.



(*Wyckoff*)
Lemma eq_empty :
  forall {T:Type} (A:Ensemble T),
    Included A (Empty_set _) ->
    A = Empty_set _.
  intros T A h1.
  pose proof (Finite_downward_closed _ _ (Empty_is_finite T) _ h1) as h2.
  revert h1 .
  induction h2; auto.
  intro h3.
  pose proof (Add_intro2 _ A x) as h4.
  apply h3 in h4.
  contradiction.
Qed.



(*Wyckoff*)
Lemma sub_empty :
  forall {T:Type} (a:T),
    Subtract (Empty_set _) a = Empty_set _.
  intros T a. apply eq_empty.
  red; intros x h1.
  destruct h1; contradiction.
Qed.


(*Wyckoff*)
Definition empty_sig_incl {T:Type} (A:Ensemble T) :=
  exist (fun C => Included C A) (Empty_set _) (incl_empty A).


(*Wyckoff*)
Lemma full_inclusion : forall {T:Type} (S:Ensemble T), 
                         Included S (Full_set _).
intros; red; intros; constructor.
Qed.                                                                


(*Wyckoff*)
Lemma inclusion_reflexive : forall {T:Type} (S:Ensemble T),
                    Included S S.
auto with sets.
Qed.

(*Wyckoff*)
Lemma inclusion_transitive : 
  forall T:Type, 
    Transitive (@Included T).
intro T.
red. intros A B C h1 h2.
red. intros x h3.
apply h1 in h3. apply h2 in h3.
assumption.
Qed.


(*Wyckoff*)
Lemma intersection_full_set : forall {T:Type} (S:Ensemble T), 
  Intersection S (Full_set T) = S.
intros T S.
apply Extensionality_Ensembles.
red. split.
  (* <= *) red. intros ? h. destruct h; assumption.
  (* >= *) red. intros x ?. constructor. assumption. constructor.
Qed.

(*Wyckoff*)
Lemma included2_impl_union_included : forall {T:Type} (A B C:Ensemble T), 
  ((Included A C) /\ (Included B C)) <-> Included (Union A B) C.
intros T A B C.
split.
(* -> *)
unfold Included.
intro h1.
destruct h1 as [h1 h2].
intros x h3.
apply Union_inv in h3.
destruct h3; auto with sets.
(* <- *)
intro h1.
split.
assert (Included A (Union A B)). auto with sets.
auto with sets.
assert (Included B (Union A B)). auto with sets.
auto with sets.
Qed.


(*Wyckoff*)
Lemma feq_im_aux : forall {T U:Type} (A B:Ensemble T) (f:T->U),
                 A = B -> Included (Im A f) (Im B f).
intros T U A B f h1.
red.
intros y h2.
destruct h2; subst.
apply Im_intro with x; auto.
Qed.

(*Wyckoff*)
Lemma feq_im : forall {T U:Type} (A B:Ensemble T) (f:T->U),
                 A = B -> Im A f = Im B f.
intros T U A B f h1.
apply Extensionality_Ensembles; red; split; apply feq_im_aux; auto.
Qed.

(*Wyckoff*)
Lemma im_inj_inj : forall {T U:Type} (A B:Ensemble T) (f:T->U),
                     Injective f ->
                     Im A f = Im B f -> A = B.
intros T U A B f h1 h2.
apply Extensionality_Ensembles.
red. split.
red. intros x h3.
assert (h4:In (Im A f) (f x)). apply Im_intro with x; auto.
rewrite h2 in h4.
inversion h4 as [y h5 ha h6]. subst.
specialize (h1 _ _ h6). subst.
assumption.
red. intros x h3.
assert (h4:In (Im B f) (f x)). apply Im_intro with x; auto.
rewrite <- h2 in h4.
inversion h4 as [y h5 ha h6]. subst.
specialize (h1 _ _ h6). subst.
assumption.
Qed.


                          
(*Wyckoff*)
Lemma im_union : forall {T U:Type} (A B:Ensemble T) (f:T->U),
                      Im (Union A B) f = Union (Im A f) (Im B f).
intros T U A B f.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red.
intros x h1.
destruct h1 as [x h1 y h3].
subst.
destruct h1 as [x h1a | x h1b].
left. apply Im_intro with x; auto.

right. apply Im_intro with x; auto.
(* >= *)
red. 
intros x h1. 
destruct h1 as [x h1a | x h1b].
destruct h1a as [x h1a y h2]. subst.
apply Im_intro with x. left; auto. reflexivity.
destruct h1b as [x h1b y h2]. subst.
apply Im_intro with x; try auto; right; auto.
Qed.                       

(*Wyckoff*)
Lemma im_intersection_incl : 
  forall {T U:Type} (A B:Ensemble T) (f:T->U),
    Included (Im (Intersection A B) f)
             (Intersection (Im A f) (Im B f)).
intros T U A B f.
red. 
intros b h1. 
destruct h1 as [x h1]. subst.
destruct h1 as [x h1l h1r].
constructor.
apply Im_intro with x; auto. apply Im_intro with x; auto.
Qed.


(*Wyckoff*)
Lemma im_intersection_inj : 
  forall {T U:Type} (A B:Ensemble T) (f:T->U),
    injective f ->
    Im (Intersection A B) f = Intersection (Im A f) (Im B f).
intros T U A B f h0.
apply Extensionality_Ensembles.
red. split.
apply im_intersection_incl.
red. intros b h1.
destruct h1 as [b h1 h2].
destruct h1 as [b h1]. subst.
apply Im_intro with b.
inversion h2 as [a h3 c h4]. subst. clear h2.
red in h0. apply h0 in h4. subst.
constructor; auto.
reflexivity.
Qed.


(*Wyckoff*)
Lemma foil_union : 
  forall {T:Type} (A B C D:Ensemble T),
    Intersection (Union A B) (Union C D) = 
    Union (Intersection A C)
          (Union (Intersection A D)
                 (Union (Intersection B C) (Intersection B D))).
intros T A B C D.
rewrite Distributivity.
rewrite Intersection_commutative at 1.
rewrite (Intersection_commutative _ (Union A B) D). 
do 2 rewrite Distributivity.
rewrite (Intersection_commutative _ C A).
rewrite (Intersection_commutative _ C B).
rewrite (Intersection_commutative _ D A).
rewrite (Intersection_commutative _ D B).
rewrite  Union_associative.
f_equal.
rewrite <- Union_associative.
rewrite <- Union_associative.
f_equal.
rewrite Union_commutative; auto.
Qed.

(*Wyckoff*)
Lemma foil_intersection : 
  forall {T:Type} (A B C D:Ensemble T),
    Union (Intersection A B) (Intersection C D) = 
    Intersection (Union A C)
                 (Intersection (Union A D)
                               (Intersection (Union B C) (Union B D))).
intros T A B C D.
rewrite Distributivity'.
rewrite Union_commutative at 1.
rewrite (Union_commutative _ (Intersection A B) D).
do 2 rewrite Distributivity'.
rewrite (Union_commutative _ C A). rewrite (Union_commutative _ C B).
rewrite (Union_commutative _ D A). rewrite (Union_commutative _ D B).
rewrite  Intersection_associative.
f_equal.
rewrite <- Intersection_associative.
rewrite <- Intersection_associative.
f_equal.
rewrite Intersection_commutative; auto.
Qed.



(*Wyckoff*)
Lemma surj_iff : forall {T U:Type} (f:T->U),
                   surjective f <-> Im (Full_set T) f = 
                                    Full_set U.
intros T U f. 
split.
(* -> *)
intro h1.
apply Extensionality_Ensembles.
red. split.
  (* <= *)
  red.
  intros x ?. constructor.
  (* >= *)
  red.
  intros u ?.
  red in h1.
  pose proof (h1 u) as h2.
  destruct h2 as [t].
  apply Im_intro with t.
  constructor.
  symmetry; assumption.
(* <- *)
intro h1.
red.
intro u.
pose proof (Full_intro _ u) as h2.
rewrite <- h1 in h2.
inversion h2 as [t].
exists t. symmetry. assumption.
Qed.

(*Section Abstr. -- toplevel*)
(*Schepler*)
Inductive characteristic_function_abstraction
          {X:Type} (P:X->Prop) (x:X) : Prop :=
  | intro_characteristic_sat: P x ->
    In (characteristic_function_abstraction P) x.

(*Schepler*)
Definition characteristic_function_to_ensemble
           {X:Type} (P:X->Prop) : Ensemble X :=
  characteristic_function_abstraction P.

(*Schepler*)
Notation "[ x : X | P ]" :=
  (characteristic_function_to_ensemble (fun x:X => P))
  (x ident).

Section Abstr.
(*Schepler*)
Lemma characteristic_function_to_ensemble_is_identity:
  forall {X:Type} (P:X->Prop),
    [ x:X | P x ] = P.
Proof.
intros.
apply Extensionality_Ensembles; split; red; intros.
destruct H.
exact H.
constructor.
exact H.
Qed.

(*Wyckoff*)
Lemma sat_in :
  forall {T:Type} (P:T->Prop) (x:T),
    In [t:T | P t] x -> P x.
  intros ? ? ? h1; destruct h1; auto.
Qed.

(*Wyckoff*)
Lemma sat_in' :
  forall {T:Type} (P:T->Prop) (x:T),
    In [t:T | ~ P t] x -> ~ P x.
  intros ? ? ? h1; destruct h1; auto.
Qed.

Lemma sat_in1 :
  forall {T:Type} (P Q:T->Prop) (x:T),
    In [t:T | P t /\ Q t] x -> P x.
  intros ? ? ? ? h1; destruct h1; tauto.
Qed.

Lemma sat_in1' :
  forall {T:Type} (P Q:T->Prop) (x:T),
    In [t:T | ~ P t /\ Q t] x -> ~ P x.
  intros ? ? ? ? h1; destruct h1; tauto.
Qed.

Lemma sat_in2 :
  forall {T:Type} (P Q:T->Prop) (x:T),
    In [t:T | P t /\ Q t] x -> Q x.
  intros ? ? ? ? h1; destruct h1; tauto.
Qed.

Lemma sat_in2' :
  forall {T:Type} (P Q:T->Prop) (x:T),
    In [t:T | P t /\ ~ Q t] x -> ~ Q x.
  intros ? ? ? ? h1; destruct h1; tauto.
Qed.




(*Wyckoff*)
Lemma sat_nin :
  forall {T:Type} (P:T->Prop) (x:T),
    ~ In [t:T | P t] x -> ~P x.
intros T P x h1.
intro h2.
contradict h1. constructor.  assumption.
Qed.

(*Wyckoff*)
Lemma sat_nin' :
  forall {T:Type} (P:T->Prop) (x:T),
    ~ In [t:T | ~ P t] x -> P x.
intros T P x h1.
apply NNPP. intro h2.
contradict h1.
constructor; auto.
Qed.



(*Wyckoff*)
Lemma eq_sat : forall {T:Type} (A:Ensemble T),
                 A = [x:T | In A x].
intros T A.
rewrite characteristic_function_to_ensemble_is_identity.
reflexivity.
Qed.


(*Wyckoff*)
Lemma im_sat :
  forall {T U:Type} (P:T->Prop) (f:T->U),
    Im [x:T | P x] f =
    [y:U | exists x:T, P x /\ y = f x].
  intros T U P f.
  apply Extensionality_Ensembles; red; split; red; intros x h1.
  destruct h1 as [x h1]. subst.
  destruct h1 as [h1]. constructor.
  exists x.
  split; auto.
  destruct h1 as [h1].
  destruct h1 as [y h1].
  destruct h1 as [h1 h2]. subst.
  apply Im_intro with y; auto. constructor; auto. 
Qed.
  

(*Wyckoff*)
Lemma sat_iff : 
  forall {T:Type} (P Q:T->Prop),
    (forall x:T, P x <-> Q x) ->
    [x : T | P x] = [x : T | Q x].
intros T P Q h1.
apply Extensionality_Ensembles.
red. split.
red. intros x h2. destruct h2 as [h2]. rewrite h1 in h2.
constructor; auto.
red. intros x h2. destruct h2 as [h2]. rewrite <- h1 in h2.
constructor; auto.
Qed.

(*Wyckoff*)
Lemma sig_in_sat : 
  forall {T:Type} (P:T->Prop) (x:{t:T | P t}),
    In [t:T | P t] (proj1_sig x).
constructor. apply proj2_sig.
Qed.

(*Wyckoff*)
Lemma sig_in_sat2 : 
  forall {T:Type} (P Q:T->Prop) (x:{t:T | P t}),
    Q (proj1_sig x) ->
    In [t:T | P t /\ Q t] (proj1_sig x).
constructor; intros; split; [apply proj2_sig | auto].
Qed.

(*Wyckoff*)
Lemma sig_in_sat2' : 
  forall {T:Type} (P Q:T->Prop) (x:{t:T | Q t}),
    P (proj1_sig x) ->
    In [t:T | P t /\ Q t] (proj1_sig x).
constructor; intros; split; [auto | apply proj2_sig].
Qed.



End Abstr.
(*End Abstr*)

(*Wyckoff*)
Lemma not_inhabited_empty : forall {T:Type} (S:Ensemble T),
                              ~ Inhabited S -> S = Empty_set _.
intros T S h1. 
apply Extensionality_Ensembles.
red. split.
red.
intros x h2.
pose proof (Inhabited_intro _ _ _ h2). contradiction.
auto with sets.
Qed.


(*Wyckoff*)
Lemma ex_inh : 
  forall {T:Type} (P:T->Prop),
    (exists x:T, P x) ->
    Inhabited [x:T | P x].
intros T P h1.
destruct h1 as [x h1].
apply Inhabited_intro with x; constructor; auto.
Qed.


(*Wyckoff*) 
Lemma empty_alt :
  forall {T:Type} (A:Ensemble T),
    (forall x, ~In A x) -> A = Empty_set _.
intros T A h1. red in h1.
destruct (classic (Inhabited A)) as [h2 | h3]. 
destruct h2 as [x h2].
contradict h1; auto.
intro h1.  apply (h1 _ h2).
apply not_inhabited_empty; auto.
Qed.

Section Complement.

(*Wyckoff*)
Definition excl_midP {T:Type} (P:T->Prop) : Prop :=
  forall x:T, P x \/ ~ P x.

  
  
(*Wyckoff*)
Lemma excl_middle_empty :
  forall {T:Type} (S:Ensemble T),
    Intersection S (Comp S) = Empty_set _.
intros T S.
apply Extensionality_Ensembles. red. split.
(*left*)
red.
intros x h1.
inversion h1; contradiction.
(*right*)
red.
intros; contradiction.
Qed.

(*Wyckoff*)
Lemma excl_middle_full :
  forall {T:Type} (S:Ensemble T),
    Union S (Comp S) = Full_set _.
intros T S.
apply Extensionality_Ensembles. red. split.
(*left*)
red. intros. constructor. auto with sets.
(*right*)
red. intros x h1.
case (classic (In S x)) as [h2 | h3].
  apply Union_introl; trivial.
  apply Union_intror.
  auto with sets.
Qed.

(*Wyckoff*)
Lemma complement_full : forall T:Type, Comp (Full_set T) = Empty_set T.
intro T.
apply Extensionality_Ensembles. red. unfold Included. split.
(* left *)
intros x h2.
unfold Comp in h2. red in h2.
elim h2. apply Full_intro.
(* right *)
intros x h3. contradiction.
Qed.

(*Wyckoff*)
Lemma complement_empty : forall T:Type, Comp (Empty_set T) = Full_set T.
intro T.
pose proof (Complement_Complement _ (Full_set T)) as h1.
rewrite complement_full in h1.
assumption.
Qed.

(*Wyckoff*)
Lemma complement_meets_non_subset : forall {T:Type} (P Q : Ensemble T), 
  (Inhabited P) ->
  (~Included P Q <-> Inhabited (Intersection (Comp Q) P)).
intros T P Q h1.
split.
(*left*)
intro h2. 
unfold Included in h2.
pose proof (not_all_ex_not _ _ h2) as h4.
elim h4. intros x h5.
assert (h6: (In P x) /\ (~In Q x)).
  tauto.
destruct h6 as [h7 h8].
assert (h9: In (Intersection (Comp Q) P) x).
  constructor; auto with sets. 
apply Inhabited_intro with x; trivial.
(*right*)
intro h2.
unfold Included.
inversion h2 as [x h3].
assert (h4: exists x:T, ~(In P x -> In Q x)).
  exists x.
  assert (h5: (In P x) /\ (~ In Q x)).
    split.
    inversion h3 as [h6 h7]. assumption.
    inversion h3 as [y h6 h7].
    compute in h7. auto.
  tauto.
apply ex_not_not_all; trivial.
Qed.


(*Wyckoff*)
Lemma included_empty_complement_int : forall {T:Type} (A B:Ensemble T), 
  Included A B <-> Intersection A (Comp B) = Empty_set _.
intros T A B.
split.
(* -> *)
intros h1.
apply Extensionality_Ensembles. red.
  split.
  (* <= *)
  red in h1.
  red. intros x h2.
  destruct h2 as [x h2l h2r].
  specialize (h1 _ h2l).
  assert (h3:In (Intersection B (Comp B)) x).
    auto with sets.
  rewrite excl_middle_empty in h3. assumption.
  (* >= *) auto with sets.
(* <- *)
intro h1.
red. 
intros x h2.
apply NNPP.
intro h3.
assert (h4:In (Intersection A (Comp B)) x).
  constructor; assumption.
rewrite h1 in h4.
contradiction.
Qed.

(*Wyckoff*)
Lemma elimination_union : 
  forall {T:Type} (A B C:Ensemble T),
    A = Union B C -> C = Intersection A (Union (Comp B) C).
intros T A B C h1.
pose proof (f_equal (fun D => Intersection D (Union (Comp B) C)) h1) as h2.
simpl in h2.
rewrite foil_union in h2.
rewrite excl_middle_empty in h2.
rewrite Intersection_idempotent in h2.
rewrite <- (Union_commutative _ C (Intersection C (Comp B))) in h2.
rewrite abs_sum_psa in h2.
rewrite (Union_commutative _ (Intersection B C) C) in h2.
rewrite (Intersection_commutative _ B C) in h2.
rewrite abs_sum_psa in h2.
rewrite empty_union in h2.
symmetry.
assumption.
Qed.


(*Wyckoff*)
Lemma complement_inv : 
  forall {T:Type} (A:Ensemble T) (x:T), In (Comp A) x -> ~In A x.
auto.
Qed.

(*Wyckoff*)
Lemma complement_inclusion: forall {Y:Type} (S T:Ensemble Y),
  Included S T -> Included (Comp T) (Comp S).
Proof.
intros.
red; intros.
red; intro.
contradiction H0.
auto with sets.
Qed.

(*Wyckoff*)
Lemma setminus_inc : forall {T:Type} (A B:Ensemble T),
                       Included (Setminus A B) A.
intros T A B. red.
intros x h1.
destruct h1; auto with sets.
Qed.

(*Wyckoff*)
Lemma setminus_empty : 
  forall {T:Type} (A:Ensemble T),
    Setminus (Empty_set _) A = 
    Empty_set _.
intros T A.
apply Extensionality_Ensembles; red; split; auto with sets.
red.
intros x h1.
destruct h1.
contradiction.
Qed.

(*Wyckoff*)
Lemma empty_setminus2 : 
  forall {T:Type} (A:Ensemble T),
    Setminus A A = 
    Empty_set _.
intros T A.
apply Extensionality_Ensembles; red; split; auto with sets.
red. intros x h1.
destruct h1; contradiction.
Qed.

(*Wyckoff*)
Lemma setminus_empty' :
  forall {T:Type} (A:Ensemble T),
    Setminus A (Empty_set _) = A.
  intros; apply Extensionality_Ensembles; red; split; red; intros x h2.
  destruct h2; auto.
  constructor; auto.
  intro; contradiction.
Qed.



(*Wyckoff*)
Lemma setminus_int_complement :
  forall {T:Type} (A B:Ensemble T),
    Setminus A B = 
    Intersection A (Comp B).
intros T A B.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red.
intros ? h1.
destruct h1 as [h2 h3].
constructor; auto with sets. 
(* >= *)
red. intros x h1.
destruct h1 as [h2 h3].
constructor; auto with sets.
Qed.

(*Wyckoff*)
Lemma complement_setminus_full : 
  forall {T:Type} (A:Ensemble T), 
    Comp A = Setminus (Full_set T) A.
intros T A.
apply Extensionality_Ensembles.
red. split.
red. intros x h1. constructor. constructor. assumption.
red. intros x h1. red. red.
destruct h1; auto.
Qed.


(*Wyckoff*)
Lemma setminus_sub_sup :
  forall {T:Type} (A B:Ensemble T),
    Included A B -> 
    Setminus A B = Empty_set _.
intros T A B h1.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red. intros x h2.
destruct h2 as [h3 h4].
red in h1.
apply h1 in h3.
contradiction.
(* >= *) auto with sets.
Qed.

(*Wyckoff*)
Lemma int_setminus : forall {T:Type} (A X:Ensemble T),
                       Intersection A (Setminus X A) = Empty_set _.
intros T A X.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red.
intros x h1.
destruct h1 as [x h2 h3].
destruct h3. contradiction.
(* >= *)
auto with sets.
Qed.

(*Wyckoff*)
Lemma setminus_same_int : forall {T:Type} (X Y:Ensemble T),
                            Setminus X (Intersection X Y) =
                            Setminus X Y.
intros T X Y.
apply Extensionality_Ensembles.
red. split.
red.
intros x h1.
destruct h1 as [h1 h2].
constructor. assumption. auto with sets.
red.
intros x h1. constructor.
destruct h1 as [h1 h2].
assumption.
intro h2.
destruct h2; destruct h1; contradiction.
Qed.

(*Wyckoff*)
Lemma setminus_int_eq : 
  forall {T:Type} (A B C:Ensemble T),
    Setminus (Intersection A B) C = 
    Intersection A (Setminus B C).
intros T A B C.
apply Extensionality_Ensembles; red; split; red; intros x h1.
destruct h1 as [h1 h2]. destruct h1 as [x h1a h1b].
constructor; auto. constructor; auto.
destruct h1 as [x h1 h2]. destruct h2 as [h2 h3].
constructor; auto.
constructor; auto.
Qed.



(*Wyckoff*)
Lemma decompose_int_setminus : 
  forall {T:Type} (X Y:Ensemble T),
    X = Union (Intersection X Y) (Setminus X Y).
intros T X Y.
apply Extensionality_Ensembles.
red. split.
red.
intros x h1.
destruct (classic (In Y x)).
left. constructor; auto.
right. constructor; auto.
red.
intros x h1.
destruct h1 as [x h1l | x h1r].
destruct h1l; auto.
destruct h1r; auto.
Qed.



(*Wyckoff*)
Lemma decompose_setminus_inc : 
  forall {T:Type} (X Y:Ensemble T),
    Included Y X ->
    X = Union Y (Setminus X Y).
intros T X Y h1.
rewrite (decompose_int_setminus X Y) at 1.
rewrite inclusion_iff_intersection_eq in h1.
rewrite Intersection_commutative in h1.
rewrite h1.
reflexivity.
Qed.



(*Wyckoff*)
Lemma setminus_add : 
  forall {T:Type} (A B:Ensemble T)
         (a:T),
    Setminus (Add A a) B = 
    Union (Setminus A B) (Setminus (Singleton a) B).
intros T A B a.
apply Extensionality_Ensembles.
red. split.
red. intros x h1.
destruct h1 as [h1 h2].
destruct h1 as [x h1 | x h3].
left. constructor; auto.
destruct h3.
right. constructor; auto. constructor.
red. intros x h1.
destruct h1 as [x h1 | x h2].
destruct h1 as [h1 h2].
constructor. left; auto. assumption.
destruct h2 as [h2 h3].
destruct h2.
constructor. right. constructor.
assumption.
Qed.

(*Wyckoff*)
Lemma setminus_setminus : 
  forall {T:Type} (A B C:Ensemble T),
    Setminus (Setminus A B) C =  
    Setminus A (Union B C).
intros T A B C.  
apply Extensionality_Ensembles.
red; split; red; intros x h1.
destruct h1 as [h1 h2]. 
destruct h1 as [h1 h3].
constructor; auto.
intro h4.
destruct h4; contradiction.
destruct h1 as [h1 h2].
constructor.
constructor; auto.
intro h3.
contradict h2.
left; auto.
intro h3.
contradict h2.
right; auto.
Qed.

(*Wyckoff*)
(*for use in negations*)
Lemma in_setminus_iff :
  forall {T:Type} (A B:Ensemble T) (x:T),
    In (Setminus A B) x <->
    In A x /\ ~In B x.
  intros.
  split; intro h1.
  destruct h1; split; auto.
  destruct h1; constructor; auto.
Qed.

(*Wyckoff*)
(*Uses LEM *)
Lemma setminus_of_same :
  forall {T:Type} (A B:Ensemble T),
    Setminus A (Setminus A B) = Intersection A B.
  intros T A B.
  apply Extensionality_Ensembles; red; split; red; intros x h1.
  destruct h1 as [h1 h2]. 
  rewrite in_setminus_iff in h2.
  apply not_and_or in h2.
  destruct h2 as [h2 | h2]; try contradiction.
  constructor; try apply NNPP; auto.
  destruct h1 as [x h1 h2].
  constructor; auto.
  intro h3.
  destruct h3; contradiction.
Qed.

(*Wyckoff*)
Lemma finite_setminus : 
  forall {T:Type} (A B:Ensemble T),
    Finite A -> Finite (Setminus A B).
intros T A B h1.
induction h1 as [|A h1 h2 a h3].
rewrite setminus_empty.
constructor.
rewrite setminus_add.
apply Union_preserves_Finite; auto.
rewrite setminus_int_complement.
eapply Finite_downward_closed.
apply Singleton_is_finite.
auto with sets.
Qed.

(*Wyckoff*)
Lemma add_downward_closed :
  forall {T:Type} (A:Ensemble T) (x:T),
    Finite (Add A x) -> Finite A.
  intros T A x h1; eapply Finite_downward_closed. apply h1.
  auto with sets.
Qed.


(*Wyckoff*)
Lemma setminus_add1 : forall {T:Type} (E:Ensemble T) (x:T),
                        ~ In E x ->
                        Setminus (Add E x) E = Singleton x.
intros T E x h1.
apply Extensionality_Ensembles.
red. split.
red.
intros y h2.
destruct h2 as [h2 h3].
destruct h2 as [y h2l | y h2r]. contradiction.
assumption.
red. intros y h2. destruct h2; subst. constructor; auto.
right; auto. constructor.
Qed.

(*Wyckoff*)
Lemma setminus_add2 : 
  forall {T:Type} (E:Ensemble T) (e:T),
         Setminus (Add E e) (Singleton e) = 
         Subtract E e.
intros T E e.
apply Extensionality_Ensembles.
red. split.
red. intros x h1.
destruct h1 as [h1 h2].
destruct h1 as [x h1a | x h1b].
constructor; auto.
contradiction.
red.
intros x h1.
red in h1. red in h1. red in h1. destruct h1 as [h1 h2].
constructor. left. assumption.
assumption.
Qed.


   
(*Wyckoff*)
Lemma add_subtract_a : 
  forall {T:Type} (A:Ensemble T) (a:T),
    Subtract (Add A a) a = Subtract A a.
intros T A a.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red.
intros x h1.
destruct h1 as [h1 h2].
destruct h1 as [x h1a | x h1b].
constructor; auto. contradiction.
(* >= *)
red.
intros x h1.
destruct h1 as [h1 h2].
pose proof (Add_intro1 _ A a x h1).
constructor; auto.
Qed.



(*Wyckoff*)
Lemma setminus_couple_eq :
  forall {T:Type} (A:Ensemble T) (a a':T), 
    Setminus A (Couple a a') = Subtract (Subtract A a) a'.
intros T A a a'.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red. intros x h1.
destruct h1 as [h1 h2].
constructor. constructor. assumption.
intro h3. destruct h3; subst.
pose proof (Couple_l _ a a'); contradiction.
intro h3. destruct h3; subst.
pose proof (Couple_r _ a a'); contradiction.
(* >= *)
red. intros x h1.
destruct h1 as [h1 h2].
destruct h1 as [h1a h1b].
constructor. assumption.
intro h3.
destruct h3.
pose proof (In_singleton _ a). contradiction.
pose proof (In_singleton _ a'). contradiction.
Qed.

(*Wyckoff*)
Lemma subtract_couple_l :
  forall {T:Type} (a b:T),
    a <> b ->
    Subtract (Couple a b) a = Singleton b.
  intros T a b h0.
  apply Extensionality_Ensembles; red; split; red; intros x h1.
  destruct h1 as [h1 h2].
  destruct h1. contradict h2. constructor; auto.
  constructor; auto.
  destruct h1.
  constructor. right.
  intro h1.
  destruct h1. contradict h0; auto.
Qed.

(*Wyckoff*)
Lemma subtract_couple_r :
  forall {T:Type} (a b:T),
    a <> b ->
    Subtract (Couple a b) b = Singleton a.
  intros T a b h1.
  apply neq_sym in h1.
  rewrite couple_comm.
  apply subtract_couple_l; auto.
Qed.



(*Wyckoff*)
Lemma subtract2_comm : 
  forall {T:Type} (S:Ensemble T) (a b:T),
    Subtract (Subtract S a) b =
    Subtract (Subtract S b) a.
  intros T S a b.
  apply Extensionality_Ensembles.
  red. split.
  (* <= *)
  red. intros x h1.
  destruct h1 as [h1a h1b].
  destruct h1a as [h1aa h1ab].
  constructor. constructor. 
  assumption. assumption. assumption.
  (* >= *)
  red. intros x h1.
  destruct h1 as [h1a h1b].
  destruct h1a as [h1aa h1ab].
  constructor. constructor.
  assumption. assumption. assumption.
Qed.


(*Wyckoff*)
Lemma union_singleton_eq_setminus :
  forall {T:Type} (A B:Ensemble T) (a b:T), 
    Add A a = Add B b -> 
    Setminus A (Couple a b) = Setminus B (Couple a b).
intros T A B a b h1.
assert (h2:Setminus (Add A a) (Couple a b) = 
           Setminus (Add B b) (Couple a b)).
  f_equal; auto.
repeat rewrite setminus_couple_eq in h2.
rewrite add_subtract_a in h2.
symmetry in h2.
rewrite subtract2_comm in h2.
rewrite add_subtract_a in h2.
repeat rewrite <- setminus_couple_eq in h2.
symmetry.
rewrite couple_comm in h2.
assumption.
Qed.


(*Wyckoff*)
Lemma subtract_preserves_inclusion : 
  forall {T:Type} (A B:Ensemble T) (x:T), 
    Included A B -> Included (Subtract A x) (Subtract B x).
intros T A B x h1.
red. intros y h2.
destruct h2 as [h2 h3].
constructor; auto with sets.
Qed.


(*Wyckoff*)
Lemma subtract_nin : forall {T:Type} (A:Ensemble T) (x:T),
                       ~In A x -> Subtract A x = A.
intros T A x h1.
apply Extensionality_Ensembles.
red. split. red. intros y h2. destruct h2; auto.
red. intros y h2. constructor; auto.
intro h3.
destruct h3; subst.
contradiction.
Qed.


(*Wyckoff*)
Lemma subtract_eq_sing :
  forall {T:Type} (pfdt:type_eq_dec T) (a b:T),
    Subtract (Singleton a) b = Empty_set _ ->
    a = b.
  intros T hdt a b h1.
  destruct (hdt a b) as [|h2]; subst; auto.
  rewrite subtract_nin in h1.
  apply sing_not_empty in h1.
  contradiction.
  rewrite in_sing_iff; auto.
Qed.


(*Wyckoff*)
Lemma incl_add_sub_nin :
  forall {T:Type} (A B:Ensemble T) (b:T),
    ~In B b ->
    Included A (Add B b) ->
    Included (Subtract A b) B.
  intros T A B b h1 h2.
  red; intros x h3. destruct h3 as [h3 h4].
  apply h2 in h3.
  destruct h3 as [|x h3]; auto; try contradiction.
Qed.

(*Wyckoff*)
Lemma incl_union_setminus :
  forall {T:Type} (A B C:Ensemble T),
    Included A (Union B C) ->
    Included (Setminus A C) B.
intros T A B C h1.
red.
intros x h2.
destruct h2 as [h2 h3].
apply h1 in h2.
destruct h2; auto with sets. contradiction.
Qed.

  
(*Wyckoff*)
Lemma union_setminus : forall {T:Type} (A X:Ensemble T),
                         Union A (Setminus X A) = Union A X.
intros T A X.
rewrite setminus_int_complement.
rewrite Distributivity'.
rewrite excl_middle_full.
rewrite intersection_full_set.
reflexivity.
Qed.

(*Wyckoff*)
Lemma union_setminus_incl : forall {T:Type} (A X:Ensemble T),
                          Included A X ->
                          Union A (Setminus X A) = X.
intros T A X h1.
rewrite union_setminus.
rewrite <- inclusion_iff_union.
assumption.
Qed.

(*Wyckoff*)
Lemma setminus_union : 
  forall {T:Type} (A B:Ensemble T),
    Setminus (Union A B) B = Setminus A B.
intros T A B.
apply Extensionality_Ensembles.
red; split; red.
intros x h1.
destruct h1 as [h1 h2].
destruct h1 as [x h1l | x h1r].
constructor; auto. contradiction.
intros x h1.
destruct h1 as [h1 h2].
constructor; auto. left; auto.
Qed.

  
(*Wyckoff*)
Lemma union_eq_disjoint_setminus :
  forall {T:Type} (A B C:Ensemble T),
    disjoint A B ->
    Union A B = C ->
    A = Setminus C B.
  intros T A B C h1 h2; subst.
  rewrite setminus_union.
  symmetry.
  apply disjoint_setminus_compat; auto.
Qed.
 



(*Wyckoff*)
Lemma im_setminus : 
  forall {T U:Type} (A B:Ensemble T) (f:T->U),
    Included (Im (Setminus A B) f)
             (Intersection (Im A f) (Im (Comp B) f)).
intros T U A B f.
red. intros x h1.
destruct h1 as [x h1]. subst. destruct h1 as [h1 h2].
constructor. apply Im_intro with x; auto.
apply Im_intro with x. auto. reflexivity.
Qed.


(*Wyckoff*)
Lemma incl_im_setminus : 
  forall {T U:Type} (A B:Ensemble T),
    Included B A ->
    forall (f:T->U),
    Included (Setminus (Im A f) (Im B f)) (Im (Setminus A B) f).
intros T U A B h1 f.
assert (h2:Im A f = Union (Im B f) (Im (Setminus A B) f)).
  rewrite (decompose_setminus_inc _ _ h1) at 1.
  rewrite im_union. reflexivity.  
apply elimination_union in h2. 
rewrite Distributivity in h2. 
pose proof (setminus_inc A B) as h3.
pose proof h3 as h3'. 
rewrite inclusion_iff_union in h3'. rewrite Union_commutative in h3'.
apply (im_preserves_inclusion _ _ f) in h3.
rewrite inclusion_iff_intersection_eq in h3.
rewrite Intersection_commutative in h3.
rewrite h3 in h2.
symmetry in h2.
rewrite <- inclusion_iff_union in h2.
rewrite <- setminus_int_complement in h2. 
assumption.
Qed.

(*Wyckoff*)
Lemma inj_im_setminus :
  forall {T U:Type} (A B:Ensemble T)
         (f:T->U),
    injective f ->
    Im (Setminus A B) f =
    Setminus (Im A f) (Im B f).
  intros T U A B f h1.
  apply Extensionality_Ensembles; red; split; red; intros x h2.
  destruct h2 as [x h2]; subst.
  destruct h2 as [h2 h3].
  constructor.
  econstructor.
  apply h2.
  reflexivity.
  contradict h3.
  inversion h3 as [? ? ?   h5 h6]; subst.
  apply h1 in h5; subst.
  assumption.
  destruct h2 as [h2 h3].
  destruct h2 as [x h2]; subst.
  econstructor.
  constructor.
  apply h2.
  contradict h3.
  econstructor. apply h3.
  reflexivity.
  reflexivity.
Qed.




(*Wyckoff*)
Lemma im_complement_incl_inj : 
  forall {T U:Type} (A:Ensemble T) (f:T->U),
    injective f ->
    Included (Im (Comp A) f) (Comp (Im A f)).
intros T U A f h0. 
red. intros u h1.
destruct h1 as [u h1]. subst.
red in h1. red in h1. red. red.
intro h2.  inversion h2 as [x h3 h4 h5]. subst. 
apply h0 in h5. subst.
contradiction.
Qed.


(*Wyckoff*)
Lemma im_complement_incl_bij : 
  forall {T U:Type} (A:Ensemble T) (f:T->U),
    bijective f ->
    Included (Im (Comp A) f) 
             (Setminus (Im (Full_set _) f) (Im A f)).
intros T U A f h0.
red in h0. destruct h0 as [h0 h0'].
pose proof (im_complement_incl_inj A f h0) as h1.
rewrite setminus_int_complement.
rewrite Intersection_commutative. 
rewrite surj_iff in h0'. rewrite h0'. 
rewrite intersection_full_set.
assumption.
Qed.


(*Wyckoff*)
Lemma im_complement_bij : 
  forall {T U:Type} (A:Ensemble T) (f:T->U),
    bijective f ->
    Im (Comp A) f = 
    Setminus (Im (Full_set _) f) (Im A f).
intros T U A f h0.
apply Extensionality_Ensembles.
red. split.
apply im_complement_incl_bij. assumption.
red. intros u h1.
red in h0. destruct h0 as [h0 h0']. 
rewrite surj_iff in h0'.
pose proof (Full_intro _ u) as h2.
rewrite <- h0' in h2.
destruct h2 as [u h2]. subst.
apply Im_intro with u.
intro h3. 
inversion h1 as [h4 h5].
contradict h5.
apply Im_intro with u; auto.
reflexivity.
Qed.


(*Wyckoff*)
Lemma im_set_minus_inj : 
  forall {T U:Type} (A B:Ensemble T) (f:T->U),
    injective f ->
    Im (Setminus A B) f = Intersection (Im A f) (Im (Comp B) f).
intros T U A B f h0.
apply Extensionality_Ensembles.
red. split. 
apply im_setminus.
red. intros x h1.
destruct h1 as [u h1 h2].
destruct h1 as [u h1]. subst.
inversion h2 as [y h3 x h4]. subst. 
rewrite h4. 
apply h0 in h4. subst. 
apply Im_intro with y. constructor; auto.
reflexivity.
Qed.



End Complement.


(*Wyckoff*)
Lemma incl_card_fun1_eq :
  forall {T:Type} (A B:Ensemble T),
    Finite B -> Included A B ->
    card_fun1 A = card_fun1 B ->
    A = B.
intros T A B h1. revert A.
induction h1 as [|B h2 h3 x h4].
intros; auto with sets.
intros A h1 h5. 
pose proof (Finite_downward_closed
              _ _
              (Add_preserves_Finite _  _ x h2) _ h1) as hf.
eapply subtract_preserves_inclusion in h1.  
rewrite sub_add_same_nin in h1; auto.
destruct (classic (In A x)) as [h6 | h7].
rewrite card_add_nin' in h5; auto.
rewrite (card_sub_in _ hf x) in h5.
apply S_inj in h5.
specialize (h3 _ h1 h5).
rewrite <- (add_sub_compat_in _ _ h6).
f_equal. assumption. 
assumption.
rewrite subtract_nin in h1; auto.
apply incl_card_fun1 in h1; auto.
rewrite h5 in h1; auto.
rewrite card_add_nin' in h1; auto.
omega.
Qed.


Lemma in2_card :
  forall (A:Ensemble nat) (c d:nat),
    c <> d ->
    Inn A c -> Inn A d ->
    Finite A ->
    1 < card_fun1 A.
  intros A c d h1 h2 h3 h4.
  erewrite card_sub_in.
  erewrite card_sub_in.
  auto with arith.
  apply subtract_preserves_finite.
  assumption.
  constructor.
  apply h3.
  rewrite in_sing_iff.
  apply h1.
  assumption.
  assumption.
Qed.


(*Wyckoff*)
Lemma cardinal_2 : 
  forall {T:Type} (A:Ensemble T),
    cardinal _ A 2 ->
    exists (x y:T), x <> y /\ A = Couple x y.
intros T A h1. 
assert (h2:Inhabited A).
  apply not_empty_Inhabited.
  intro h2.
  subst. inversion h1. subst.
  pose proof (Add_intro2 _ A x) as h3. rewrite H in h3. contradiction.
destruct h2 as [x h2].
pose proof (cardinal_finite _ _ _ h1) as h4.
pose proof (card_sub_in A h4 x h2) as h5.
assert (ha:Finite (Subtract A x)).
  pose proof (incl_subtract A x) as h6.
  apply (Finite_downward_closed _ _ h4 _ h6).
pose proof (card_fun1_compat A) as h6. simpl in h6.
destruct h6 as [h6 h7]. clear h7.
specialize (h6 h4).
assert (h3:Inhabited (Subtract A x)).
  apply not_empty_Inhabited.
  intro h3.
  rewrite h3 in h5.
  rewrite card_fun1_empty in h5.
  rewrite h5 in h6.
  pose proof (cardinal_unicity _  _ _ h1  _ h6). omega.
destruct h3 as [y h3]. 
pose proof (card_sub_in (Subtract A x) ha y h3) as hb.
rewrite hb in h5.
pose proof (card_fun1_compat (Subtract A x)) as h6'. simpl in h6'.
destruct h6' as [h6' h7']. clear h7'.
specialize (h6' ha).

exists x, y. split.
destruct h3.
intro h8. subst.
contradict H0. constructor.
pose proof (cardinal_unicity _ _ _ h1 _ h6) as h7.
rewrite h5 in h7.
assert (h8:card_fun1 (Subtract (Subtract A x) y) = 
           0). omega.
apply card_fun1_O in h8.
destruct h8 as [h8l | h8r].
apply Extensionality_Ensembles.
red. split.
red. intros a h4'.
destruct (eq_dec a x), (eq_dec a y); subst. 
left. left. right.
assert (h5':In (Setminus A (Couple x y)) a).
  constructor; auto.
  intro h5'. destruct h5'; auto. 
rewrite setminus_couple_eq in h5'.
rewrite h8l in h5'. contradiction.
red.
intros a h9.
destruct h9. assumption.
destruct h3; auto.
pose proof (add_preserves_infinite _ y h8r) as h10.
rewrite add_sub_compat_in in h10; auto.
contradiction.
Qed.


(*Wyckoff*)
Lemma card_fun1_couple : 
  forall {T:Type} (x y:T),
    x <> y ->
    card_fun1 (Couple x y) = 2.
intros T x y h1.
pose proof (card_fun1_compat (Couple x y)) as h2.
destruct h2 as [h2l h2r].
specialize (h2l (finite_couple x y)).
pose proof (cardinal_couple _ _ h1) as h3.
eapply cardinal_is_functional; auto.
apply h2l. apply h3.
Qed.



(*Wyckoff*)
Lemma incl_couple_inv : 
  forall {T:Type} (A:Ensemble T) (b c:T),
    Included A (Couple b c) ->
    A = Empty_set _ \/ A = Singleton b \/ A = Singleton c \/
    A = Couple b c.
intros T A b c h1. 
destruct (eq_dec b c) as [he | hne].
subst.
rewrite couple_singleton. rewrite couple_singleton in h1.
apply singleton_inc in h1. 
destruct h1 as [h1l | h1r]. left. assumption.  
right. left. assumption. 
pose proof (finite_couple b c) as h2.
pose proof (card_fun1_couple _ _ hne) as h3.
pose proof (Finite_downward_closed _ _ h2 _ h1) as h4.
pose proof (incl_card_fun1 _ _ h2 h1) as h5.
rewrite h3 in h5. 
destruct (le_lt_eq_dec (card_fun1 A) 2 h5) as [h6 | h7]. 
assert (h7:card_fun1 A <= 1). auto with arith. clear h6.
destruct (le_lt_eq_dec (card_fun1 A) 1 h7) as [h8 | h9].
apply lt1 in h8. clear h7. 
left.
apply card_fun1_O in h8. destruct h8; auto. contradiction.
apply card_fun1_1 in h9.
destruct h9 as [x h9]. subst.
red in h1. specialize (h1 _ (In_singleton _ x)).
destruct h1. right. left. reflexivity.
right. right. left. reflexivity. 
apply incl_card_fun1_eq in h1; auto.
rewrite h7. symmetry. assumption.
Qed.



Section family1.
Variable It T:Type.
Definition Family := Ensemble (Ensemble T).

(*Schepler*)
Inductive FamilyUnion (F:Family): Ensemble T :=
  | family_union_intro: forall (S:Ensemble T) (x:T),
    In F S -> In S x -> In (FamilyUnion F) x.

(*Schepler*)
Inductive FamilyIntersection (F:Family) : Ensemble T :=
  | family_intersection_intro: forall x:T,
    (forall S:Ensemble T, In F S -> In S x) ->
    In (FamilyIntersection F) x.

(*Schepler*)
Definition IndexedFamily := It -> Ensemble T.
Variable FI:IndexedFamily.

(*Schepler*)
Inductive IndexedUnion : Ensemble T :=
  | indexed_union_intro: forall (i:It) (x:T),
    In (FI i) x -> In IndexedUnion x.

(*Wyckoff*)
Inductive seg_union {n}
          (f:seg_fun n (Ensemble T)) : Ensemble T :=
| seg_fun_union_intro :
    forall i (pfi:i < n) (x:T), 
      Inn (f i pfi) x ->
      Inn (seg_union f) x.


(*Schepler*)
Inductive IndexedIntersection : Ensemble T :=
  | indexed_intersection_intro: forall (x:T),
    (forall i:It, In (FI i) x) -> In IndexedIntersection x.

(*Schepler*)
Definition ImageFamily : Family:=
  Im (Full_set _) FI.

(*Schepler*)
Lemma indexed_to_family_union: IndexedUnion = FamilyUnion (ImageFamily).
Proof.
apply Extensionality_Ensembles.
unfold Same_set.
unfold Included.
intuition.
destruct H.
apply family_union_intro with (FI i).
apply Im_intro with i.
constructor.
reflexivity.
assumption.

destruct H.
destruct H.
apply indexed_union_intro with x0.
rewrite <- H1.
assumption.
Qed.                                                        

(*Schepler*)
Lemma indexed_to_family_intersection:
  IndexedIntersection = FamilyIntersection ImageFamily.
Proof.
apply Extensionality_Ensembles.
unfold Same_set.
unfold Included.
intuition.
constructor.
intros.
destruct H.
destruct H0.
rewrite H1.
apply H.

constructor.
intro.
destruct H.
apply H.
apply Im_intro with i.
constructor.
reflexivity.
Qed.
End family1.

(*Wyckoff*)
Definition im_im_proj1_sig {T:Type} {A:Ensemble T} 
           (F:Family (sig_set A)) : Family T :=
  Im F (@im_proj1_sig _ _).


(*Wyckoff*)
Lemma im_proj1_sig_family_union :
  forall {T} (X:Ensemble T) (F:Family (sig_set X)),
    im_proj1_sig (FamilyUnion _ F) =
    FamilyUnion _ [S:Ensemble T |
                   Inn (im_im_proj1_sig F) S].
  intros; apply Extensionality_Ensembles; red; split; red; intros x h1.
  inversion h1; clear h1; subst.
  inversion H; clear H; subst.
  econstructor.
  constructor.
  econstructor.
  apply H0.
  reflexivity.
  econstructor.
  apply H1.
  reflexivity.
  inversion h1; clear h1; subst.
  destruct H as [h1].
  inversion h1; clear h1; subst.
  inversion H0; clear H0; subst.
  econstructor.
  econstructor.
  apply H.
  apply H1.
  reflexivity.
Qed.




(*Wyckoff*)
Lemma family_union_add : forall {T:Type} (F:Family T) (S:Ensemble T),
  FamilyUnion _ (Add F S) = Union S (FamilyUnion _ F).
intros T F S.
apply Extensionality_Ensembles.
  red. split.
  (* <= *)
  red. intros x h4.
  unfold Add in h4.
  inversion h4 as [A y h5 h6].
  pose proof (Union_inv _ _ _ A h5) as h7.
  case h7 as [h8 | h9].
  (*h8*)
  apply Union_intror. 
  apply family_union_intro with A; assumption.
  (*h9*)
  left. 
  pose proof (Singleton_inv _ _ _ h9) as h12.
  rewrite <- h12 in h6; assumption.
  (* >= *)
  red. intros x h2.
  pose proof (Union_inv _ _ _ x h2) as h3.
  case h3 as [h4 | h5].
  (*h4*)
  pose proof (Add_intro2 _ F S) as h5.
  apply family_union_intro with S; assumption.
  (*h5*)
  inversion h5 as [A y h6 h7]. 
  pose proof (Add_intro1 _ F S) as h8.
  pose proof (h8 A h6) as h9.
  apply family_union_intro with A; assumption.
Qed.

(*Wyckoff*)
Lemma familyunion_union_eq : 
  forall {T:Type} (F G:Family T),
    FamilyUnion _ (Union F G) = 
    Union (FamilyUnion _ F) (FamilyUnion _ G).
intros T F G.
apply Extensionality_Ensembles; red; split; red; intros x h1.
destruct h1 as [S x h1 h2].
destruct h1 as [S h1|S h1].
left. apply family_union_intro with S; auto.
right. apply family_union_intro with S; auto.
destruct h1 as [x h1|x h1].
destruct h1 as [S x h1 h2].
apply family_union_intro with S; auto.
left; auto.
destruct h1 as [S x h1 h2].
apply family_union_intro with S; auto.
right; auto.
Qed.


(*Wyckoff*)
Lemma family_union_im : 
  forall {T U:Type} (F:Family T) (f:T->U),
    FamilyUnion _ (Im F (fun A:Ensemble T=>Im A f)) = 
    Im (FamilyUnion _ F) f.
intros T U F f.
apply Extensionality_Ensembles.
red. split.
red.
intros u h1.
destruct h1 as [A u h2 h3].
destruct h2 as [A h2]. subst.
destruct h3 as [u h3]. subst.
apply Im_intro with u. apply family_union_intro with A; auto.
reflexivity.
red.
intros u h1. destruct h1 as [u h1]. subst. 
destruct h1 as [A x h2 h3].
apply family_union_intro with (Im A f).
apply Im_intro with A; auto.
apply Im_intro with x; auto.
Qed.

(*Wyckoff*)
Lemma family_union_im_int :
  forall {T:Type} (F:Family T) (A:Ensemble T), 
    FamilyUnion _ (Im F (fun S : Ensemble T => Intersection A S)) = 
    Intersection (FamilyUnion _ F) A.
  intros T F A.
  apply Extensionality_Ensembles; red; split; red; intros x h1.
  destruct h1 as [S x h1 h2].
  destruct h1 as [S h1]. subst.
  destruct h2 as [x h2 h3].
  constructor; auto.
  apply family_union_intro with S; auto.
  destruct h1 as [x h1 h2].
  destruct h1 as [S x h1 h3].
  eapply family_union_intro.
  eapply Im_intro.
  apply h1.
  apply eq_refl.
  constructor; auto.
Qed.

(*Wyckoff*)
Lemma incl_family_union :
  forall {T:Type} (F:Family T) (E:Ensemble T),
    In F E ->
    Included E (FamilyUnion _ F).
  intros T F E h1 x h2.
  apply family_union_intro with E; auto.
Qed.


(*Wyckoff*)  
Lemma setminus_family_union :
  forall {T:Type} (F:Family T) (A:Ensemble T),
    Setminus (FamilyUnion _ (Add F A)) A = Setminus (FamilyUnion _ F) A.
  intros T F A.
  apply Extensionality_Ensembles; red; split; red; intros x h1.
  destruct h1 as [h1 h2].
  destruct h1 as [S x h3 h4].  
  destruct h3 as [S h5|S h5].
  constructor.
  apply family_union_intro with S; auto.
  assumption.
  destruct h5.
  contradiction.
  rewrite family_union_add.
  rewrite Union_commutative.
  rewrite setminus_union; auto.
Qed.
         
Section family2.

(*Schepler*)
Lemma empty_indexed_intersection: forall {T:Type}
  (F:IndexedFamily False T),
  IndexedIntersection _ _ F = (Full_set _).
Proof.
intros.
apply Extensionality_Ensembles; red; split; red; intros;
  auto with sets.
constructor.
constructor.
destruct i.
Qed.

Variable T:Type.
(*Schepler*)
Lemma empty_family_union: FamilyUnion T (Empty_set _) =
                          (Empty_set _).
Proof.
apply Extensionality_Ensembles.
unfold Same_set.
unfold Included.
intuition.
destruct H.
contradiction H.
contradiction H.
Qed.

(*Wyckoff*)
Lemma family_union_sing : 
  forall (A:Ensemble T),
    FamilyUnion _ (Singleton  A) = A.
intros A.
apply Extensionality_Ensembles.
red. split.
red. intros x h1.
destruct h1 as [B x h1].
destruct h1; subst. assumption.
red. intros x h1.
apply family_union_intro with A; auto with sets.
Qed.

(*Wyckoff*)
Lemma family_union_couple : 
  forall (A B:Ensemble T),
    FamilyUnion _ (Couple A B) = Union A B.
intros A B.
apply Extensionality_Ensembles; red; split; red; intros x h1.
destruct h1 as [S x h1 h2].
destruct h1; [left | right]; auto.
destruct h1 as [x h1 | x h1].
apply family_union_intro with A; auto; left.
apply family_union_intro with B; auto; right.
Qed.


(*Wyckoff*)
Definition chain_set (F:Family T) : Prop := 
  forall (S1 S2:Ensemble T), In F S1 -> 
    In F S2 -> Included S1 S2 \/ Included S2 S1. 



(*Wyckoff*)
Definition pairwise_disjoint (F:Family T) : Prop :=
  forall (A B:Ensemble T),
    A <> B ->
    In F A -> In F B -> disjoint A B.
   
(*Wyckoff*)
Lemma pairwise_disjoint_empty :
  pairwise_disjoint (Empty_set (Ensemble T)).
red.
intros; contradiction.
Qed.


(*Wyckoff*)
Lemma pairwise_disjoint_incl : 
  forall (E F:Family T),
    pairwise_disjoint F -> Included E F -> pairwise_disjoint E.
intros E F h1 h2.
red. red in h1. red in h2.
intros A B h3 h4 h5.
pose proof (h2 _ h4) as h6. pose proof (h2 _ h5) as h7.
apply h1; auto.
Qed.

(*Wyckoff*)
Lemma pairwise_disjoint_add :
  forall (F:Family T),
    pairwise_disjoint F ->
    forall A:Ensemble T,
           (forall S, In F S -> disjoint S A) ->
           pairwise_disjoint (Add F A).
intros F h1 A h2.
red.
intros C D h3 h4 h5.
destruct h4 as [C h4 | C h4].
destruct h5 as [D h5 | D h5].
apply h1; auto.
destruct h5.
apply h2; auto.
destruct h4.
destruct h5 as [D h5 | D h5].
apply sym_disjoint.
apply h2; auto.
destruct h5.
contradict h3.
reflexivity.
Qed.

(*Wyckoff*)
Lemma pairwise_disjoint_add_rev :
  forall (F:Family T) (A:Ensemble T),
    pairwise_disjoint (Add F A) ->
    pairwise_disjoint F.
  intros F A h1.
  red in h1. red.
  intros C D h2 h3 h4.
  pose proof (Add_intro1 _ _ A _ h3) as h5.
  pose proof (Add_intro1 _ _ A _ h4) as h6.
  apply h1; auto.
Qed.

(*Wyckoff*)  
Inductive inhabited_family {T:Type} (F:Family T) : Prop :=
  inhabited_family_intro :
    Inhabited F -> (forall A:Ensemble T, In F A -> Inhabited A) ->
    inhabited_family F.

(*Wyckoff*)
Lemma family_intersection_relativization : 
  forall (F:Family T) (S:Ensemble T),
    In F S ->
    FamilyIntersection _ F =
    FamilyIntersection _ (Im F (fun A => Intersection A S)).
intros F S h1.
apply Extensionality_Ensembles.
red. split.
red. intros x h2.
destruct h2 as [x h2]. 
constructor.
intros S' h3. destruct h3 as [S' h3]. subst.
constructor; apply h2; auto.
red.
intros x h2.
destruct h2 as [x h2]. constructor.
intros S' h3.
specialize (h2 (Intersection S' S)).
assert (h4: In (Im F (fun A : Ensemble T =>
                        Intersection A S)) (Intersection S' S)).
  apply Im_intro with S'.  assumption. 
  reflexivity. 
specialize (h2 h4).
destruct h2; auto.
Qed.

(*Wyckoff*)
Lemma family_incl_intersection_cont : 
  forall (F F':Family T),
    Included F F' -> Included (FamilyIntersection _ F') (FamilyIntersection _ F).
intros F F' h1.
red. intros x h2.
destruct h2 as [x h2]. constructor.
intros S h3. apply h2; auto.
Qed.

(*Wyckoff*)
Definition finc {T:Type} (E:Ensemble T) : Type := 
  {F : Ensemble T | Finite F /\ Included F E}.

(*Wyckoff*)
Lemma family_union_im_full_set_finc_proj1_sig_eq : 
  forall (E:Ensemble T),
    FamilyUnion _ (Im (Full_set (finc E)) (@proj1_sig _ _)) = E.
intro E.
apply Extensionality_Ensembles.
red. split.
red. intros x h1.
destruct h1 as [S  x h1 h2].
destruct h1 as [S h1]. subst. clear h1.
destruct S as [S [h3 h4]]. simpl in h2.
apply h4; auto.
red.
intros x h1.
gterm1.
redterm0 c.
pose proof (family_union_intro _ c0 (Singleton x) x) as h2.
unfold c0 in h2.
apply h2.
pose proof (Singleton_is_finite _ x) as h3.
assert (hinc:Included (Singleton x) E).
  red; intros ? h4; destruct h4; auto.
apply Im_intro with 
(exist (fun F : Ensemble T => Finite F /\ Included F E)
       (Singleton x) (conj h3 hinc)). constructor.
constructor.
constructor.
Qed.

(*Wyckoff*)
Definition finc_union {E:Ensemble T} (A:finc E) (B:finc E) :
  finc E.
  destruct A as [A h1], B as [B h2].
  destruct h1 as [h1a h1b],
                 h2 as [h2a h2b].
refine (exist _ _ _). split.
eapply Union_preserves_Finite.
apply h1a. apply h2a.
rewrite <- included2_impl_union_included.
split; auto.
Defined.

(*Wyckoff*)
Definition finc_intersection {E:Ensemble T} (A:finc E) (B:finc E) :
  finc E.
  destruct A as [A h1], B as [B h2].
  destruct h1 as [h1a h1b], h2 as [h2a h2b].
  refine (exist _ _ _). split.
  eapply intersection_preserves_finite.
  apply h1a. 
  pose proof (Intersection_decreases_l _ A B) as h3.
  pose proof (inclusion_transitive T _ _ _ h3 h1b) as h4.
  apply h4.
Defined.

(*Wyckoff*)
Lemma proj1_sig_finc_union :
  forall {E:Ensemble T} (A:finc E) (B:finc E),
    proj1_sig (finc_union A B) = Union (proj1_sig A) (proj1_sig B).
intros E A B.
destruct A as [A h1], B as [B h2].
destruct h1 as [h1 h1'], h2 as [h2 h2'].
unfold finc_union.
simpl.
reflexivity.
Qed.

(*Wyckoff*)
Lemma proj1_sig_finc_intersection :
  forall {E:Ensemble T} (A:finc E) (B:finc E),
    proj1_sig (finc_intersection A B) =
    Intersection (proj1_sig A) (proj1_sig B).
intros E A B.
destruct A as [A h1], B as [B h2].
destruct h1 as [h1 h1'], h2 as [h2 h2'].
unfold finc_intersection.
simpl.
reflexivity.
Qed.

(*Wyckoff*)
Lemma finc_union_incl_l : 
  forall {E:Ensemble T} (A:finc E) (B:finc E),
    Included (proj1_sig A) (proj1_sig (finc_union A B)).
intros E A B.
rewrite proj1_sig_finc_union.
auto with sets.
Qed.

(*Wyckoff*)
Lemma finc_union_incl_r : 
  forall {E:Ensemble T} (A:finc E) (B:finc E),
    Included (proj1_sig B) (proj1_sig (finc_union A B)).
intros E A B.
rewrite proj1_sig_finc_union.
auto with sets.
Qed.



End family2.

Arguments FamilyUnion [T] _ _.
Arguments FamilyIntersection [T] _ _.
Arguments IndexedUnion _ [T] _ _.
Arguments seg_union [T] [n] _ _.
Arguments IndexedIntersection _ [T] _ _.
Arguments chain_set [T] _.
Arguments disjoint [T] _ _.
Arguments pairwise_disjoint [T] _.

Section Power_set_Algebra_axioms.
Variable Xt:Type.
(*Wyckoff*)
Lemma Union_associative : forall N M P:(Ensemble Xt),
                            (Union N (Union M P)) = 
  (Union (Union N M) P).
intros N M P.
apply Extensionality_Ensembles.  unfold Same_set.
split; unfold Included.
(*left*)
intros x H.
inversion H.
assert (h1 : In (Union N M) x). 
apply Union_introl. assumption.
apply Union_introl. assumption.
inversion H0.
assert (h1 : In (Union N M) x). 
apply Union_intror. assumption.
apply Union_introl. assumption.
apply Union_intror. assumption.
(*right*)
intros x H.
inversion H. inversion H0.
apply (Union_introl _ _ _ _ H2).
assert (h2: In (Union M P) x).
apply (Union_introl _ _ _ _ H2).
apply (Union_intror _ _ _ _ h2).
assert (h2: In (Union M P) x).
apply (Union_intror _ _ _ _ H0).
apply (Union_intror _ _ _ _ h2).
Qed.


(*Wyckoff*)
Lemma Union_commutative : forall N M:(Ensemble Xt),
                            Union N M = Union M N.
intros N M.
apply Extensionality_Ensembles.
unfold Same_set. 
split; unfold Included.
(*left*)
intros x h1.
inversion h1 as [x' h2 h3 | x'' h4 h5].
apply Union_intror; assumption.
apply Union_introl; assumption.
(*right*)
intros x h6.
inversion h6 as [x' h7 h8 | x'' h9 h10].
apply Union_intror; assumption.
apply Union_introl; assumption.
Qed.

(*Wyckoff*)
Lemma Intersection_commutative : 
  forall N M:(Ensemble Xt), Intersection N M = Intersection M N.
intros N M.
apply Extensionality_Ensembles.
unfold Same_set.
split; unfold Included.
(*left*)
intros x h1.
inversion h1 as [x' h2 h3 h4].
apply Intersection_intro; assumption.
(*right*)
intros x h5.
inversion h5 as [x' h6 h7 h8].
apply Intersection_intro; assumption.
Qed.

(*Wyckoff*)
Lemma Distributivity :
  forall N M P:(Ensemble Xt), 
    Intersection P (Union N M) = 
    Union (Intersection P N) (Intersection P M).
intros N M P.
apply Extensionality_Ensembles.
unfold Same_set.
split; unfold Included.
(*left*)
intros x h1.
inversion h1 as [x' h2 h3 h4].
inversion h3 as [x'' h5 h6 | x'' h7 h8].
assert (h9: In (Intersection P N) x).
apply Intersection_intro; assumption.
apply Union_introl; assumption.
assert (h10: In (Intersection P M) x).
apply Intersection_intro; assumption.
apply Union_intror; assumption.
(*right*)
intros x h11.
inversion h11 as [x' h12 h13 | x'' h14 h15].
inversion h12 as [x''' h16 h17 h18].
assert (h19 : In (Union N M) x).
apply Union_introl; assumption.
apply Intersection_intro; assumption.
inversion h14 as [x''' h20 h21 h22].
assert (h23: In (Union N M) x).
apply Union_intror; assumption.
apply Intersection_intro; assumption.
Qed.

(*Wyckoff*)
Lemma dist_prod_psa : 
  forall N M P:(Ensemble Xt), 
    Union P (Intersection N M) = 
    Intersection (Union P N) (Union P M).
intros N M P.
apply Extensionality_Ensembles.
unfold Same_set.
split; unfold Included.
(*left*)
intros x h1.
inversion h1 as [x' h2 h3 | x'' h4 h5].
assert (h6: In (Union P N) x). auto with sets.
assert (h7: In (Union P M) x). auto with sets.
auto with sets.
inversion h4.
assert (h8: In (Union P N) x). auto with sets.
assert (h9: In (Union P N) x). auto with sets.
auto with sets.
(*right*)
intros x h10.
inversion h10 as [x' h11 h12 h13].
inversion h11. 
auto with sets.
inversion h12.
auto with sets.
assert (h14: In (Intersection N M) x). 
auto with sets. auto with sets.
Qed.

(*Wyckoff*)
Lemma comp_sum_psa: 
  forall N:(Ensemble Xt), Union N (Comp N) = Full_set Xt.
intros N.
apply Extensionality_Ensembles.
unfold Same_set. 
split; unfold Included.
(*left*)
intros x h1.
apply Full_intro.
(*right*)
intros x h2.
assert (h3: ~(In N x) \/ (In N x)). 
    tauto.
elim h3; auto with sets.
Qed.

(*Wyckoff*)
Lemma comp_prod_psa:
  forall N:(Ensemble Xt), Intersection N (Comp N) = Empty_set Xt.
intros N.
apply Extensionality_Ensembles.
unfold Same_set.
split; unfold Included.
(*left*)
intros x h1.
contradict h1.
intro h2.
inversion h2 as [x' h3 h4].
unfold In in h3.  unfold In in h4. unfold Comp in h4.
tauto.
(*right*)
intros x h5.
inversion h5.
Qed.
End Power_set_Algebra_axioms.



  
(*Wyckoff*)
Lemma union_eq_empty : forall {T:Type} (A B:Ensemble T),
  Union A B = Empty_set _ -> A = Empty_set _ /\ B = Empty_set _.
intros T A B h1.
split; apply Extensionality_Ensembles; split; auto with sets.
red. intros x h2. pose proof (Union_introl _ A B _ h2) as h3.
rewrite h1 in h3. assumption.
red. intros x h2. pose proof (Union_introl _ B A _ h2) as h3.
rewrite Union_commutative in h3.
rewrite h1 in h3. assumption.
Qed.

(*Wyckoff*)
Lemma intersection_eq_full : forall {T:Type} (A B:Ensemble T),
  Intersection A B = Full_set _ -> A = Full_set _ /\ B = Full_set _. 
intros T A B h1. split; apply Extensionality_Ensembles; split;
try (red; intros; constructor).
red. intros x h2.
rewrite <- h1 in h2. destruct h2; assumption.
red. intros x h2.
rewrite <- h1 in h2. destruct h2; assumption.
Qed.



(*Wyckoff*)
Lemma card_disj_union : 
  forall {T:Type} (A B:Ensemble T) (m n:nat),
    Intersection A B = Empty_set _ ->
         cardinal _ A m ->
         cardinal _ B n ->
         cardinal _ (Union A B) (m+n).
intros T A B m n h1 h2 h3.
revert h1.
induction h2 as [|A m h2 h4 a h5].
intro h4.
assert (h5:0 + n = n). auto with arith.
rewrite h5.
rewrite empty_union.
assumption.
intro h6.
rewrite union_add_comm.
assert (h7:S m + n = S(m+n)). auto with arith.
rewrite h7.
constructor.
assert (h8:Included A (Add A a)). auto with sets.
pose proof (intersection_preserves_inclusion _ _ B h8) as h9.
rewrite (Intersection_commutative _ B (Add A a)) in h9.
rewrite (Intersection_commutative _ B A) in h9.
rewrite h6 in h9.
assert (h10:Intersection A B = Empty_set _). auto with sets.
apply h4; auto.
intro h8.
destruct h8 as [a h8l | a h8r].
contradiction.
pose proof (Add_intro2 _ A a) as h9.
assert (h10:In (Intersection (Add A a) B) a). auto with sets.
rewrite h6 in h10.
contradiction.
Qed.

(*Wyckoff*)
Lemma card_disj_union' : 
  forall {T:Type} (A B:Ensemble T) 
         (pfa:Finite A) (pfb:Finite B),
    Intersection A B = Empty_set _ ->
    (card_fun A pfa) + (card_fun B pfb) = 
    card_fun (Union A B) (Union_preserves_Finite _ _ _ pfa pfb).
intros T A B h1 h2 h3.
pose proof (card_disj_union A B (card_fun A h1) (card_fun B h2) h3) as h4.
pose proof (card_fun_compat A h1) as h5.
pose proof (card_fun_compat B h2) as h6.
specialize (h4 h5 h6).
pose proof (card_fun_compat (Union A B)
                            (Union_preserves_Finite T A B h1 h2)) as h7.
eapply cardinal_is_functional; auto. apply h4. apply h7.
Qed.

(*Wyckoff*)
Lemma card_disj_union'' : 
  forall {T:Type} (A B:Ensemble T),
    Finite A -> Finite B ->
    Intersection A B = Empty_set _ ->
         card_fun1 (Union A B) = (card_fun1 A) + (card_fun1 B).
  intros T A B h1 h2 h3.
  pose proof (card_disj_union' A B h1 h2 h3) as h4.
  do 3 rewrite card_fun_card_fun1_compat in h4.
  rewrite h4; auto.
Qed.      


(*Wyckoff*)
Lemma card_decompose_int_setminus :
  forall {T:Type} (X Y:Ensemble T) (m n p:nat),
    cardinal _ X m -> cardinal _ (Intersection X Y) n ->
    cardinal _ (Setminus X Y) p ->
    m = n + p.
intros T X Y m n p h1 h2 h3.
pose proof (decompose_int_setminus X Y) as h4.
rewrite h4 in h1.
pose proof (setminus_same_int X Y) as h5.
rewrite <- h5 in h1.
assert (h6:Intersection (Intersection X Y)
                         (Setminus X (Intersection X Y)) = Empty_set _).
apply int_setminus.
rewrite <- h5 in h3.
pose proof (card_disj_union _ _ _ _ h6 h2 h3) as h7.
apply (cardinal_unicity _ _ _ h1 _ h7).
Qed.

(*Wyckoff*)
Lemma card_decompose_int_setminus' :
  forall {T:Type} (X Y:Ensemble T) (m n:nat),
    cardinal _ X m -> cardinal _ (Intersection X Y) n ->
    cardinal _ (Setminus X Y) (m-n).
intros T X Y m n h1 h2.
pose proof (card_decompose_int_setminus X Y m n) as h3.
pose proof (cardinal_finite _ _ _ h1) as h4.
pose proof (setminus_inc X Y) as h5.
pose proof (Finite_downward_closed).
pose proof (Finite_downward_closed _ _ h4 _  h5) as h6.
pose proof (finite_cardinal _ _ h6) as h7.
destruct h7 as [r h7].
specialize (h3 _ h1 h2 h7).
subst.
assert (h8:n+r-n = r). auto with arith.
rewrite h8.
assumption.
Qed.


Section Symdiff.

(*Wyckoff*)
Definition Symdiff {T:Type} (A B:Ensemble T) :=
  Union (Setminus A B) (Setminus B A).

(*Wyckoff*)
Lemma symdiff_ref : forall {T:Type} (A:Ensemble T),
                        Symdiff A A = Empty_set _.
intros T A.
unfold Symdiff.
rewrite setminus_sub_sup; auto with sets.
Qed.

(*Wyckoff*)
Lemma symdiff_comm : forall {T:Type} (A B:Ensemble T),
                       Symdiff A B = Symdiff B A.
intros T A B. unfold Symdiff. auto with sets.
Qed.

(*Wyckoff*)
Lemma symdiff_empty_iff_eq : 
  forall {T:Type} (A B:Ensemble T),
    Symdiff A B = Empty_set _ <-> A = B.
intros T A B.
split.
(* -> *)
intro h1.
unfold Symdiff in h1.
apply Extensionality_Ensembles.
red. split.
(* <= *) red.
intros x h2.
apply NNPP.
intro h3.
pose proof (Setminus_intro _ A B x h2 h3) as h4.
pose proof (Union_introl _ (Setminus A B) (Setminus B A) x h4) as h5.
rewrite h1 in h5. contradiction.
(* >= *)
intros x h2.
apply NNPP.
intro h3.
pose proof (Setminus_intro _ B A x h2 h3) as h4.
pose proof (Union_introl _ (Setminus B A) (Setminus A B) x h4) as h5.
rewrite Union_commutative in h5.
rewrite h1 in h5. contradiction.
(* <- *)
intro h1.
rewrite h1.
apply symdiff_ref.
Qed.

(*Wyckoff*)
Definition Setminus_full {T:Type} (F X Y:Ensemble T) := Intersection X (Setminus F Y).

(*Wyckoff*)
Lemma setminus_full_ref : forall {T:Type} (F X:Ensemble T),
                            Setminus_full F X X = Empty_set _.
intros T F X.
apply Extensionality_Ensembles; split; auto with sets.
red.
intros x h1.
destruct h1 as [x h1a h1b].
destruct h1b; contradiction.
Qed.

(*Wyckoff*)
Lemma setminus_full_empty_inc : 
  forall {T:Type} (F X Y:Ensemble T),
    Setminus_full F X Y = Empty_set _ -> Included (Intersection F X) Y.
intros T F X Y h1.
red.
intros x h2.
apply NNPP. intro h3.
unfold Setminus_full in h1.
destruct h2 as [x h2l h2r].
assert (h4:In (Intersection X (Setminus F Y)) x).
  constructor; auto; constructor; auto.
rewrite h1 in h4.
contradiction.
Qed.

(*Wyckoff*)
Lemma setminus_full_empty_inc' : 
  forall {T:Type} (F X Y:Ensemble T),
    Setminus_full F X Y = Empty_set _ -> Included X F -> 
    Included X Y.
intros T F X Y h1 h2.
pose proof (setminus_full_empty_inc F X Y h1) as h3.
rewrite inclusion_iff_intersection_eq in h2.
rewrite Intersection_commutative in h2.
rewrite h2 in h3.
assumption.
Qed.

(*Wyckoff*)
Definition Symdiff_full {T:Type} (F X Y:Ensemble T) := 
  Union (Setminus_full F X Y) (Setminus_full F Y X).

(*Wyckoff*)
Lemma symdiff_full_ref : forall {T:Type} (F A:Ensemble T),
                        Symdiff_full F A A = Empty_set _.
intros T F A. apply Extensionality_Ensembles; split; auto with sets.
red. intros x h1.
destruct h1 as [x h1 |x h2].
rewrite setminus_full_ref in h1. assumption.
rewrite setminus_full_ref in h2. assumption.
Qed.

(*Wyckoff*)
Lemma symdiff_full_comm : forall {T:Type} (F A B:Ensemble T),
                       Symdiff_full F A B = Symdiff_full F B A.
intros T F A B.
unfold Symdiff_full. auto with sets.
Qed.

(*Wyckoff*)
Lemma symdiff_full_empty_iff_eq : 
  forall {T:Type} (F A B:Ensemble T),
    Included A F -> Included B F ->
    (Symdiff_full F A B = Empty_set _ <-> A = B).
intros T F A B h1 h2. split.
(* -> *)
intro h3.
unfold Symdiff_full in h3.
pose proof (union_eq_empty _ _ h3) as h4.
destruct h4 as [h4l h4r].
pose proof (setminus_full_empty_inc' _ _ _ h4l h1) as h5.
pose proof (setminus_full_empty_inc' _ _ _ h4r h2) as h6.
apply Extensionality_Ensembles; red; split; assumption.
(* <- *)
intro h3.
rewrite h3.
apply symdiff_full_ref.
Qed.


End Symdiff.



Section DeMorgan.
Variable T:Type.

(*Wyckoff*)
Lemma comp_union : forall (M N : Ensemble T), Comp (Union M N) =
  Intersection (Comp M) (Comp N).
intros M N.
apply Extensionality_Ensembles. red. split.
(*left*)
red. intros x h1.
assert (h2: Included M (Union M N)).
  auto with sets.
assert (h3: Included N (Union M N)).
  auto with sets.
split.
pose proof (complement_inclusion _ _ h2) as h4.
auto with sets.
pose proof (complement_inclusion _ _ h3) as h5.
auto with sets.
(*right*)
red. intros x h6.
inversion h6 as [y h7 h8].
unfold Comp in h7. red in h7.
unfold Comp in h8. red in h8.
unfold Comp. red.
intro h9.
inversion h9.
contradiction. contradiction.
Qed.

(*Wyckoff*)
Lemma comp_int : forall (M N : Ensemble T), Comp (Intersection M N) =
  Union (Comp M) (Comp N).
intros M N.
pose proof (Complement_Complement _ M) as h1.
pose proof (Complement_Complement _ N) as h2.
rewrite <- h1, <- h2 at 1.
pose proof (comp_union (Comp M) (Comp N)) as h3.
rewrite <- h3.
apply Complement_Complement.
Qed.
End DeMorgan.


(*Wyckoff*)
Lemma incl_union_inv : 
  forall {T:Type} (A B C:Ensemble T),
    Included A (Union B C) -> 
    Included A B \/ Included A C \/ (Intersection A (Symdiff B C)) <> Empty_set _.
intros T A B C h1. 
destruct (classic (Included A (Union B (Comp C)))) as [h2 | h3].
pose proof (conj h1 h2) as h3.
rewrite inclusion_intersection in h3.
rewrite <- dist_prod_psa in h3.
rewrite excl_middle_empty in h3.
rewrite Union_commutative in h3.
rewrite empty_union in h3.
left; auto.
destruct (classic (Included A (Union (Comp B) C))) as [h4 | h5].
rewrite Union_commutative in h1. rewrite Union_commutative in h4.
pose proof (conj h1 h4) as h5.
rewrite inclusion_intersection in h5.
rewrite <- dist_prod_psa in h5.
rewrite excl_middle_empty in h5.
rewrite Union_commutative in h5.
rewrite empty_union in h5.
right. left. auto.
rewrite included_empty_complement_int in h3.
rewrite included_empty_complement_int in h5. 
rewrite comp_union in h5. rewrite comp_union in h3.
rewrite Complement_Complement in h3. rewrite Complement_Complement in h5.
pose proof (union_eq_empty
              (Intersection A (Intersection (Comp B) C))
              (Intersection A (Intersection B (Comp C)))) as h6.
pose proof (contrapos _ _ h6) as h7. clear h6.
assert (h8:  ~
       (Intersection A (Intersection (Comp B) C) =
        Empty_set T /\
        Intersection A (Intersection B (Comp C)) =
        Empty_set T) ).
  apply or_not_and.
  left. auto.
specialize (h7 h8).
rewrite <- Distributivity in h7.
right. right.
unfold Symdiff.
do 2 rewrite setminus_int_complement.
rewrite Union_commutative in h7.
rewrite (Intersection_commutative _ (Comp B) C) in h7.
assumption.
Qed.

(*Wyckoff*)
Lemma symdiff_singleton_nin : 
  forall {T:Type} (A:Ensemble T) (x:T),
    ~In A x ->
    Symdiff A (Singleton x) = Add A x.
intros T A x h1.
apply Extensionality_Ensembles.
red. split.
red. intros a h2.
destruct h2 as [a h2 | a h3].
destruct h2 as [h2 h3]. left; auto.
destruct h3 as [h3 h4]. destruct h3. right. constructor.
red. intros a h2. 
apply Add_inv in h2.
destruct h2 as [h2 | h3].
constructor. constructor; auto.
intro h3. destruct h3; contradiction.
subst.
right.
constructor. constructor. assumption.
Qed.



(*Wyckoff*)
Lemma symdiff_sings : 
  forall {T:Type} (b c:T), b <> c ->
    Symdiff (Singleton b) (Singleton c) = Couple b c.
intros T b c h1.
apply Extensionality_Ensembles.
red. split.
red.
intros x h2. 
destruct h2 as [x h2l | x h2r].
destruct h2l as [h2l]. destruct h2l. left.
destruct h2r as [h2r]. destruct h2r. right.
red. intros x h2.  
destruct h2. left. constructor. 
constructor. intro h2. destruct h2. contradict h1. 
reflexivity.
right. constructor. constructor.
intro h2. destruct h2. contradict h1.
reflexivity.
Qed.




Section Infinite_analogues.
Variable It T:Type.
Variable FI : (IndexedFamily It T).

(*Wyckoff*)
Definition Int_Fam_Ens (P:Ensemble T) (Q:Family T) : Family T := 
  (Im Q (fun (S:Ensemble T) => (Intersection P S))).

(*Wyckoff*)
Definition Union_Fam_Ens (P:Ensemble T) (Q:Family T) : Family T :=
  (Im Q (fun (S:Ensemble T) => (Union P S))).

(*Wyckoff*)
Definition Subtract_Fam_Ens (F:Family T) (x:T) : Family T :=
  (Im F (fun S => Subtract S x)).


(*Wyckoff*)
Definition Int_Ind_Ens (P:Ensemble T) (i:It) : (Ensemble T) :=
  Intersection P (FI i).

(*Wyckoff*)
Definition Union_Ind_Ens (P:Ensemble T) (i:It) : (Ensemble T) :=
  Intersection P (FI i).
  
(*Wyckoff*)
Lemma dist_union_infnt_1 : 
  forall (P:Ensemble T) (Q:Family T) , 
    Intersection P (FamilyUnion Q) = FamilyUnion (Int_Fam_Ens P Q).
intros P Q.
apply Extensionality_Ensembles.
unfold Same_set.
split.
(*left*)
unfold Included.
intros x h1.
inversion h1 as [y h2 h3 h4].
inversion h3 as [S z h5 h6 h7].
assert (h8: In (Intersection P S) x).
  auto with sets.
apply family_union_intro with (S := (Intersection P S)).
unfold Int_Fam_Ens.
apply Im_intro with (x := S).
assumption. reflexivity. assumption.
(*right*)
unfold Included.
intros x h8.
inversion h8 as [S y h9 h10 h11].
unfold Int_Fam_Ens in h9.
apply Intersection_intro.
inversion h9 as [A h12 B h13 h14].
rewrite h13 in h10.
inversion h10.
assumption.
inversion h9 as [A h12 B h13 h14].
rewrite h13 in h10.
assert (h15: In A x).
  inversion h10. assumption.
apply family_union_intro with (S := A).
assumption. assumption.
Qed.

(*Wyckoff*)
Lemma dist_int_infnt_1 : 
  forall (P:Ensemble T) (Q:Family T), 
    Union P (FamilyIntersection Q) =
    FamilyIntersection (Union_Fam_Ens P Q).
intros P Q.
apply Extensionality_Ensembles.
unfold Same_set.
split.
(*left*)
unfold Included.
intros x h1.
inversion h1 as [y h2 | y h3].
apply family_intersection_intro.
intros S h4.
unfold Union_Fam_Ens in h4.
inversion h4 as [S' z h5 h6 h7].
assert (h8: In (Union P S') x). auto with sets.
rewrite h6. assumption.
inversion h3 as [z h9 h10].
apply family_intersection_intro.
intros S h11.
unfold Union_Fam_Ens in h11.
inversion h11 as [A h12 B h13 h14].
assert (h15: In A x).
  apply h9.  assumption.
assert (h16: In (Union P A) x).  auto with sets.
rewrite h13. assumption.
(*right*)
unfold Included.
intros x h17.
inversion h17 as [y h18 h19].
unfold Union_Fam_Ens in h18.
assert (h20: forall (S':Ensemble T), (In Q S') -> In (Union P S') x).
  intros S' h50.
  apply h18.
  apply Im_intro with (x := S'). 
  assumption. reflexivity.
assert (h21 : 
    forall S : Ensemble T, In Q S -> In P x \/ In S x).
  intros S h22.
  apply Union_inv.
  apply h20. assumption.
assert (h22: In P x \/ (forall S : Ensemble T, (In Q S -> In S x))).
  apply prop_dis_forall.
  intro S.
  assert (h23: (In Q S -> In P x \/ In S x)).
    apply h21.
  tauto.
case h22.
  auto with sets.
intro h24.
apply Union_intror.
apply family_intersection_intro.
assumption.
Qed.

(*Wyckoff*)
Lemma dist_union_infnt_1_ind : 
  forall (P:Ensemble T), 
    Intersection P (IndexedUnion _ FI) = IndexedUnion _ (Int_Ind_Ens P).
intro P.
rewrite (indexed_to_family_union _ _ FI).
rewrite indexed_to_family_union.
rewrite (dist_union_infnt_1 P (ImageFamily It T FI)).
apply Extensionality_Ensembles. unfold Same_set.
unfold Int_Fam_Ens.
unfold Int_Ind_Ens.
unfold Included.
split.
(*left*)
intros x h1.
inversion h1.
inversion H.
unfold ImageFamily.
apply family_union_intro with (S := S).
unfold ImageFamily in H2.
rewrite H3.
inversion H2.
apply Im_intro with (X := Full_set It) (x := x2).
assumption.
rewrite H6. reflexivity.
assumption.
(*right*)
intros x h20.
unfold ImageFamily in h20.
inversion h20.
apply family_union_intro with (S := S).
inversion H.
rewrite H3.
apply Im_intro with (x := (FI x1)).
unfold ImageFamily.
apply Im_intro with (x := x1).
assumption.
reflexivity.
reflexivity.
assumption.
Qed.

(*Wyckoff*)
Lemma subtract_family_union :
  forall (x:T) (F:Family T),
    Subtract (FamilyUnion F) x = 
    FamilyUnion (Subtract_Fam_Ens F x).
intros x F.
apply Extensionality_Ensembles.
red. split.
red. intros a h1.
destruct h1 as [h1 h2].
destruct h1 as [A a h1 h3].
apply family_union_intro with (Subtract A x).
apply Im_intro with A; auto. constructor; auto.
red. intros a h1. 
destruct h1 as [A a h1 h2].
destruct h1 as [A h1 ? h4]. subst.
destruct h2 as [h2l h2r].
constructor. apply family_union_intro with A; auto.
assumption.
Qed.


End Infinite_analogues.

(*Wyckoff*)
Lemma pairwise_disjoint_add_family_union : 
  forall {T:Type} (F:Family T) (A:Ensemble T),
    ~Inn F A ->
    pairwise_disjoint (Add F A) ->
    disjoint (FamilyUnion F) A.
intros T F A h0 h1. red.
destruct (eq_dec F (Empty_set _)) as [hem | hnem].
rewrite hem.
rewrite empty_family_union. auto with sets.
pose proof (not_empty_Inhabited _ _ hnem) as hinh.
destruct hinh as [B hin]. 
rewrite Intersection_commutative.
rewrite dist_union_infnt_1.
pose proof (Add_intro2 _ F A) as hadd.
red in h1.
assert (h2:Int_Fam_Ens T A F = Singleton (Empty_set _)).
  apply Extensionality_Ensembles.
  red. split.
  red. intros X h2.
  destruct h2 as [X h2 ? ?]. subst.
  pose proof (Add_intro1 _ _ A _ h2) as h3.
  assert (h4:A <> X).
    intro h4.  subst. contradiction.
  specialize (h1 _ _ h4 hadd h3).
  red in h1. rewrite h1. constructor.
  red. intros X h2.
  destruct h2; subst.
  unfold Int_Fam_Ens.
  apply Im_intro with B. assumption.
  assert (h2:A <> B). intro h3. subst. contradiction.
  pose proof (Add_intro1 _ _ A _ hin) as h3.
  specialize (h1 _ _ h2 hadd h3).
  red in h1. rewrite h1. reflexivity.
rewrite h2.
apply family_union_sing.
Qed.



(*Section Characteristic. Commented to keep the notation top-level*)

(*Wyckoff*)
Definition power_set {T:Type} (A:Ensemble T) :=
  [S:Ensemble T | Included S A].

(*Wyckoff*)
Lemma power_set_empty : 
  forall (T:Type),
    power_set (Empty_set T) = Singleton (Empty_set T).
intro T. unfold power_set.
apply Extensionality_Ensembles.
red. split.
red. intros S h1. destruct h1 as [h1].
assert (h2:Included (Empty_set _) S). auto with sets.
assert (h3:S = Empty_set T). apply Extensionality_Ensembles;
  red; split; auto. subst. constructor.
  red. intros S h1. destruct h1. constructor.
auto with sets.
Qed.

(*Wyckoff*)
Lemma in_power_full : forall {T:Type} (A:Ensemble T),
                        In (power_set (Full_set T)) A.
intros T A.
constructor.
red; intros; constructor.
Qed.

(*Wyckoff*)
Lemma in_power_incl : 
  forall {T:Type} (A B X:Ensemble T),
    In (power_set X) B ->
    Included A B ->
    In (power_set X) A.
intros T A B X h1 h2.
constructor. destruct h1.
auto with sets.
Qed.


(*Wyckoff*)
Lemma in_power_intersection : 
  forall {T:Type} (A B X:Ensemble T),
    In (power_set X) A ->
    In (power_set X) (Intersection A B).
intros T A B X h1.
constructor. 
eapply (in_power_incl _ _ _ h1).
auto with sets.
Qed.


(*Wyckoff*)
Lemma in_power_union : 
  forall {T:Type} (A B X:Ensemble T),
    In (power_set X) A ->
    In (power_set X) B ->
    In (power_set X) (Union A B).
intros T A B X h1 h2.
destruct h1 as [h1]. destruct h2 as [h2].
constructor.
auto with sets.
Qed.

(*Wyckoff*)
Lemma in_power_comp : 
  forall {T:Type} (A X:Ensemble T),
    In (power_set X) A ->
    In (power_set X) (Setminus X A).
intros T A X h1.
constructor.
destruct h1 as [h1].
apply setminus_inc.
Qed.


Definition p_mem {T:Type} (P:T->Prop) (A:Ensemble T) : Prop :=
  forall x:T, In A x -> P x.

Lemma p_mem_downward_closed :
  forall {T:Type} (P:T->Prop) (A A':Ensemble T),
    Included A A' ->
    p_mem P A' ->
    p_mem P A.
  intros T P A A' h1 h2.
  red in h2; red; intros x h3.
  apply h2; auto.
Qed.

Lemma p_mem_setminus :
  forall {T:Type} (P:T->Prop) (A B:Ensemble T),
    p_mem P A ->
    p_mem P (Setminus A B).
  intros T P A B h1.
  red in h1; red.
  intros x h2; apply h1.
  destruct h2; auto.
Qed.


Section Finite_Families.
  
(*Wyckoff*)
Lemma Finite_Finite_Union : forall {T:Type} (F:Family T), 
  (forall (S:(Ensemble T)), In F S -> Finite  S) ->
  Finite F -> Finite (FamilyUnion F).
intros T F h1 h2.
induction h2.
(*Empty*)
pose proof (empty_family_union T) as h3.
rewrite h3.
constructor.
(*Add*)
rename x into S.
rename A into F.
rename H into h3.
rename IHh2 into h4.
pose proof (family_union_add F S) as h5.
rewrite h5.
apply Union_preserves_Finite.
apply h1.
auto with sets.
assert (h6:forall S:Ensemble T, In F S -> Finite S).
  intros S0 h7.
  apply h1.
  auto with sets.
apply (h4 h6).
Qed.

(*Wyckoff*)
Definition finite_mem {T:Type} (F:Family T) : Prop := p_mem (@Finite _) F.

(*Wyckoff*)
Lemma finite_family_union_finite_mem :
  forall {T:Type} (F:Family T),
    Finite (FamilyUnion F) ->
    finite_mem F. 
  intros T F h1. 
  do 2 red; intros.
  eapply Finite_downward_closed.
  apply h1.
  apply incl_family_union; auto.
Qed.

(*Wyckoff*)
Lemma finite_mem_empty :
  forall (T:Type),
    finite_mem (Empty_set (Ensemble T)).
  intro; do 2 red; intros; contradiction.
Qed.

(*Wyckoff*)
Lemma finite_mem_sing_sing :
  forall {T:Type} (a:T),
    finite_mem (Singleton (Singleton a)).
  intros T a. do 2 red; intros S h2.
  destruct h2; subst.
  apply Singleton_is_finite.
Qed.

(*Wyckoff*)
Lemma finite_mem_downward_closed :
  forall {T:Type} (F F':Family T),
    Included F F' ->
    finite_mem F' ->
    finite_mem F.
  intros T F F' h1 h2.
  do 2 red in h2. red.
  intros S h3.
  apply h1 in h3.
  apply h2 in h3; auto.
Qed.

(*Wyckoff*)
Lemma finite_mem_setminus :
  forall {T:Type} (F F':Family T),
    finite_mem F ->
    finite_mem (Setminus F F').
  intros; apply p_mem_setminus; auto.
Qed.

(*Wyckoff*)
Lemma finite_mem_subtract :
  forall {T:Type} (F:Family T) (A:Ensemble T),
    finite_mem F ->
    finite_mem (Subtract F A).
  intros; eapply finite_mem_downward_closed.
  apply incl_subtract.
  assumption.
Qed.



(*Wyckoff*)
(* Singleton Family decomposition of a set *)
Definition SingF {T:Type} (S:Ensemble T) := Im S (@Singleton T).  

(*Wyckoff*)
Lemma union_singF_eq : forall {T:Type} (S:Ensemble T), 
  S = FamilyUnion (SingF S).
intros T S.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red. intros x h1.
assert (h2:In (SingF S) (Singleton x)).
  unfold SingF.
  apply Im_intro with x; trivial.
pose proof (In_singleton _ x) as h3.
apply family_union_intro with (Singleton x); assumption.
(* >= *)
red. intros x h1.

inversion h1 as [S0 ? h2 h3].
unfold SingF in h2.
inversion h2 as [? h4 ? h5].
rewrite h5 in h3.
inversion h3 as [h6].
rewrite <- h6; trivial.
Qed.

(*Wyckoff*)
Lemma fin_fin_singf : forall {T:Type} (S:Ensemble T), Finite S 
  -> Finite (SingF S).
intros.
apply finite_image; assumption.
Qed.


(*Wyckoff*)
Definition ens_S (E:Ensemble nat) : Ensemble nat :=
  Im E S.


(*Wyckoff*)
Definition ens_pred (E:Ensemble nat) : Ensemble nat :=
  Im E pred.


(*Wyckoff*)
Lemma finite_ens_S :
  forall (E:Ensemble nat),
    Finite E ->
    Finite (ens_S E).
  unfold ens_S; intros; apply finite_image; auto.
Qed.

(*Wyckoff*)
Lemma finite_ens_pred :
  forall (E:Ensemble nat),
    Finite E ->
    Finite (ens_pred E).
  unfold ens_pred; intros; apply finite_image; auto.
Qed.

(*Wyckoff*)
Lemma ens_pred_undoes_ens_S :
  forall (E:Ensemble nat),
    ens_pred (ens_S E) = E.
  intro E.
  apply Extensionality_Ensembles; red; split; red; intros x h1.
  destruct h1 as [x h1]. subst.
  destruct h1 as [x h1]. subst.
  simpl.
  assumption.
  unfold ens_S, ens_pred.
  rewrite im_im.
  apply Im_intro with x; auto.
Qed.



(*Wyckoff*)
Lemma incl_pos_ens_S_ens_pred_impl : 
  forall (E:Ensemble nat) (m:nat),
    In (ens_S (ens_pred E)) (S m) ->
    m = 0 \/ In E (S m).
  intros E m h1.
  unfold ens_S, ens_pred in h1.
  rewrite im_im in h1.
  inversion h1 as [m' h2 q hq]. subst. 
  apply S_inj in hq. subst.
  inversion h1 as [m h3 q hq]. subst.
  apply S_inj in hq.  
  destruct (zerop (pred m')) as [h4 | h4].
  left; auto.
  right.
  assert (h5:0 < m'). omega.
  rewrite <- (S_pred _ _ h5).
  apply pred_inj' in hq; auto.
Qed.

(*Wyckoff*)
Lemma ens_S_undoes_pos_ens_pred :
  forall (E:Ensemble nat),
    ~In E 0 ->
    ens_S (ens_pred E) = E.
  intros E h1.
  apply Extensionality_Ensembles; red; split; red; intros x h2.
  destruct h2 as [x h2]. subst.
  destruct h2 as [x h2]. subst.
  assert (h3:0 < x).
    destruct (zerop x) as [h3 | h3].
    subst. contradiction.
    assumption.
  rewrite <- (S_pred _ _ h3).
  assumption.
  unfold ens_S, ens_pred.
  rewrite im_im.
  assert (h3:0 < x).
    destruct (zerop x) as [h3 | h3].
    subst. contradiction.
    assumption. 
  apply Im_intro with x.
  assumption.
  rewrite <- (S_pred _ _ h3); auto.
Qed.


(*Wyckoff*)
Lemma ens_S_undoes_zero_ens_pred1 :
  forall (E:Ensemble nat),
    In E 0 -> ~In E 1 ->
    ens_S (ens_pred E) = Subtract (Add E 1) 0.
  intros E h1 h2.
  unfold ens_S, ens_pred.
  rewrite im_im.
  apply Extensionality_Ensembles; red; split; red; intros n h3.
  destruct h3 as [n h3]. subst.
  constructor.
  destruct n as [|n]. simpl.
  right. constructor.
  simpl.
  left. assumption.
  intro h4.
  inversion h4.
  destruct h3 as [h3 h4].
  destruct h3 as [n h3 | n h3].
  apply Im_intro with n.
  assumption.
  erewrite <- S_pred; auto.
  destruct (zerop n) as [h5 | h5].
  subst.
  contradict h4. constructor.
  apply h5.
  destruct h3. 
  apply Im_intro with 0; auto.
Qed.


(*Wyckoff*)
Lemma ens_pred01_eq :
  forall (E:Ensemble nat),
    In E 0 -> In E 1 ->
    ens_pred E = ens_pred (Subtract E 0).
  intros E h0 h1.  
  unfold ens_pred.
  pose proof (add_sub_compat_in _ _ h0) as h2.
  rewrite <- h2 at 1.
  rewrite Im_add. 
  pose proof (incl_im_subtract E 0 pred) as h3.
  simpl.
  rewrite in_add_eq; auto.
  apply Im_intro with 1. constructor; auto. intro h4. inversion h4.
  simpl. reflexivity.
Qed.

(*Wyckoff*)
Lemma ens_pred_sub_0n1 :
  forall (E:Ensemble nat),
    In E 0 -> ~In E 1 ->
    ens_pred (Subtract E 0) = Subtract (ens_pred E) 0.
  intros E h1 h2.
  apply Extensionality_Ensembles; red; split; red; intros m h3.
  destruct h3 as [m h3]. subst.
  destruct h3 as [h3 h4].
  rewrite in_sing_iff in h4.
  constructor.
  apply Im_intro with m; auto.
  rewrite in_sing_iff.
  destruct m as [|m]. contradict h4; auto.
  simpl.
  intro h5. subst.
  contradiction.
  destruct h3 as [h3 h4].
  rewrite in_sing_iff in h4.
  destruct h3 as [m h3]. subst.
  apply Im_intro with m.
  constructor; auto. rewrite in_sing_iff.
  intro h5. subst. simpl in h4. contradict h4.
  reflexivity.
  reflexivity.
Qed.




(*Wyckoff*)
Lemma in_ens_pred_impl :
  forall (A:Ensemble nat) (n:nat),
    In (ens_pred A) n ->
    {n = 0} + {In A (S n)}.
  intros A n h1.
  destruct n as [|n]. left; auto.
  right.
  inversion h1 as [x h2 q hq]. subst.
  assert (h3:0 < x). omega.
  pose proof (f_equal S hq) as h4.
  rewrite <- (S_pred _ _ h3) in h4.
  subst.
  assumption.
Qed.


(*Wyckoff*)
Lemma ens_S_almost_undoes_ens_pred_zero_01 :
  forall (E:Ensemble nat),
    In E 0 -> In E 1 ->
    ens_S (ens_pred E) = Subtract E 0.
  intros E h0 h1.
  rewrite ens_pred01_eq; auto.
  unfold ens_S, ens_pred.
  rewrite im_im.
  rewrite im_id.
  apply im_ext_in.
  intros x h2.
  destruct h2 as [h2 h3].
  erewrite <- S_pred; auto.
  destruct (zerop x) as [h4 | h4]. subst.
  contradict h3. constructor.
  apply h4.
Qed.


(*Wyckoff*)
Definition ens_S_inj :
  forall (E F:Ensemble nat),
    ens_S E = ens_S F ->
    E = F.
  intros E F h1.
  unfold ens_S in h1.
  apply Extensionality_Ensembles; red; split; red; intros n h2.
  assert (h3:In (Im E S) (S n)).
    apply Im_intro with n; auto.
  rewrite h1 in h3.
  inversion h3 as [m h4 h5 h6]. subst.  
  apply S_inj in h6. subst.
  assumption.
  assert (h3:In (Im F S) (S n)).
    apply Im_intro with n; auto.
  rewrite <- h1 in h3.
  inversion h3 as [m h5 h6 h7]. subst.
  apply S_inj in h7. subst.
  assumption.
Qed.

    
  
(*Wyckoff*)
Lemma intersection_ens_S :
  forall (E F:Ensemble nat),
    Intersection (ens_S E) (ens_S F) =
    ens_S (Intersection E F).
  intros E F.
  apply Extensionality_Ensembles; red; split; red; intros x h1.
  destruct h1 as [x h1 h2].
  destruct h1 as [x h1], h2 as [x' h2 ? h3]. subst.
  apply S_inj in h3. subst.
  apply Im_intro with x'. constructor; auto. reflexivity.
  destruct h1 as [x h1]. subst.
  destruct h1 as [x h1 h2].
  constructor.
  apply Im_intro with x; auto.
  apply Im_intro with x; auto.
Qed.

(*Wyckoff*)
Lemma in_ens_S :
  forall (E:Ensemble nat) (m:nat),
    In E m ->
    In (ens_S E) (S m).
  intros E m h1.
  apply Im_intro with m; auto.
Qed.

(*Wyckoff*)
Lemma in_ens_pred :
  forall (E:Ensemble nat) (m:nat),
    In E (S m) ->
    In (ens_pred E) m.
  intros E m h1.
  apply Im_intro with (S m); auto.
Qed.


(*Wyckoff*)
Lemma in_ens_pred_iff :
  forall (E:Ensemble nat) (m:nat),
    In (ens_pred E) m <->
    (m = 0 /\ (In E 0 \/ In E 1)) \/ In E (S m).
  intros E m; split; intro h1.
  destruct h1 as [m h1]; subst.
  destruct m as [|m].
  left; tauto.
  simpl; right; auto.
  destruct h1 as [h1 | h1].
  destruct h1 as [h1 h2]; subst.
  destruct h2 as [h2 | h2].
  econstructor.
  apply h2.
  reflexivity.
  econstructor.
  apply h2.
  reflexivity.
  econstructor.
  apply h1.
  reflexivity.
Qed.


(*Wyckoff*)
Lemma in_ens_pred0 :
  forall (E:Ensemble nat),
    In E 0 ->
    In (ens_pred E) 0.
  intros E h1.
  apply Im_intro with 0; auto.
Qed.


(*Wyckoff*)
Lemma ens_pred_pos_inj :
  forall (E F:Ensemble nat),
    ens_pred E = ens_pred F ->
    ~In E 0 -> ~In F 0 ->
    E = F.
  assert (h3:forall E F:Ensemble nat,
               ens_pred E = ens_pred F ->
               ~In E 0 -> ~In F 0 ->
               Included E F).
  intros E F h1 h2 h0.
  red; intros x h3.
  assert (h4:x <> 0).
    intro. subst. contradiction.
  assert (h4':0 < x). omega.
  rewrite (S_pred _ _ h4') in h3.
  pose proof (in_ens_pred _ _ h3) as h5.
  rewrite h1 in h5.
  inversion h5 as [x' h6 q hq]. subst. clear h5.
  apply pred_inj0 in hq; auto. subst. assumption.
  destruct x as [|x]. simpl in hq. contradict h4. reflexivity.
  destruct x' as [|x']. simpl in hq. subst.
  simpl in h3.
  contradiction.
  simpl in hq. subst.
  auto with arith.
  intros E F h1 h2 h0.
  apply Extensionality_Ensembles; red; split; red; apply h3; auto.
Qed.


(*Wyckoff*)
Lemma nin_0_ens_S :
  forall (E:Ensemble nat),
    ~In (ens_S E) 0.
  intros E h1.
  inversion h1 as [y h2 q h3 hq]. subst.
  discriminate h3.
Qed.

  
(*Wyckoff*)
Definition fam_S (F:Family nat) :=
  Im F ens_S.

(*Wyckoff*)
Lemma finite_fam_S :
  forall (F:Family nat) (pf:Finite F),
    Finite (fam_S F).
  intros F h1.
  apply finite_image; auto.
Qed.

(*Wyckoff*)
Lemma in_fam_S_pos :
  forall (F:Family nat) (A:Ensemble nat),
    In (fam_S F) A ->
    ~In A 0.
  intros F A h1 h2.
  destruct h1 as [A h1]. subst.
  inversion h2 as [x h3 q hq].
  discriminate hq.
Qed.

(*Wyckoff*)
Lemma impl_in_fam_S :
  forall (F:Family nat) (E:Ensemble nat),
    In F E ->
    In (fam_S F) (ens_S E).
  unfold fam_S.
  intros F E h1.  eapply Im_intro with E; auto.
Qed.  
                        

(*Wyckoff*)
Lemma in_fam_S_impl : 
  forall (F:Family nat) (E:Ensemble nat),
    In (fam_S F) E ->
    In F (ens_pred E).
  intros F E h1.
  destruct h1 as [E h1]. subst.
  rewrite ens_pred_undoes_ens_S.
  assumption.
Qed.

  
(*Wyckoff*)
Lemma in_fam_S_nat_pos_r :
  forall (F:Family nat) (E:Ensemble nat),
    ~In E 0 ->
    In F (ens_pred E) ->
    In (fam_S F) E.
  intros F E h1 h2.
  unfold fam_S.
  apply Im_intro with (ens_pred E); auto.
  rewrite ens_S_undoes_pos_ens_pred; auto.
Qed.

(*Wyckoff*)
Lemma in_fam_S_nat0n1 :
  forall (F:Family nat) (E:Ensemble nat),
    In E 0 -> ~In E 1 ->
    In F (ens_pred E) ->
    In (fam_S F) (Subtract (Add E 1) 0).
  intros F E h1 h2 h3.
  unfold fam_S.
  apply Im_intro with (ens_pred E); auto.
  rewrite ens_S_undoes_zero_ens_pred1; auto.
Qed.


(*Wyckoff*)
Lemma in_fam_ens_nat01 :
  forall (F:Family nat) (E:Ensemble nat),
    In E 0 -> In E 1 ->
    In F (ens_pred E) ->
    In (fam_S F) (Subtract E 0).
  intros F E h1 h2 h3.
  unfold fam_S. 
  apply Im_intro with (ens_pred E); auto.
  rewrite ens_S_almost_undoes_ens_pred_zero_01; auto.
Qed.
  


End Finite_Families.




(*Use these when the underlying type is not really thought of as the 
domain.*)
Section FunctionsOn.


(*Wyckoff*)
Definition fun_dom_ran {T U:Type} (A:Ensemble T) (B:Ensemble U)
           (f:T->U) : Prop :=
  forall x:T, In A x -> In B (f x).

(*Wyckoff*)
Definition inj_on {T U:Type} (A:Ensemble T) (f:T->U) : Prop :=
  forall x y:T, In A x ->
                In A y ->
                f x = f y -> x = y.

(*Wyckoff*)
Definition surj_on {T U:Type} (A:Ensemble T) (B:Ensemble U)
           (f:T->U) : Prop :=
  Included B (Im A f).

(*Wyckoff*)
Definition bij_on {T U:Type} (A:Ensemble T) (B:Ensemble U)
           (f:T->U) : Prop :=
  inj_on A f /\ surj_on A B f.

(*Wyckoff*)
Definition fun_dom_ran_empty1 :
  forall {T U:Type} (B:Ensemble U) (f:T->U),
    fun_dom_ran (Empty_set _) B f.
  intros T U B f. red. intros; contradiction.
Qed.


(*Wyckoff*)
Definition fun_dom_ran_empty2_iff :
  forall {T U:Type} (A:Ensemble T) (f:T->U),
    fun_dom_ran A (Empty_set _) f <->
    A = Empty_set _.
  intros T U A f. split.
  intro h1.
  red in h1.
  apply Extensionality_Ensembles; red; split; red; intros x h2.
  specialize (h1 _ h2).
  contradiction.
  contradiction.
  intros; subst.
  apply fun_dom_ran_empty1.
Qed.


(*Wyckoff*)
Lemma inj_on_empty :
  forall {T U:Type} (f:T->U),
    inj_on (Empty_set _) f.
  intros; red; intros; contradiction.
Qed.

(*Wyckoff*)
Lemma surj_on_empty1 :
  forall {T U:Type} (f:T->U) (B:Ensemble U),
    surj_on (Empty_set _) B f -> B = Empty_set _.
  intros T U f B h1.
  red in h1.
  rewrite image_empty in h1.
  auto with sets.
Qed.

(*Wyckoff*)
Lemma surj_on_im :
  forall {T U:Type} (f:T->U) (A:Ensemble T),
    surj_on A (Im A f) f.
  intros; red; red; intros; auto.
Qed.
    
(*Wyckoff*)
Lemma surj_on_empty2 :
  forall {T U:Type} (f:T->U) (A:Ensemble T),
    surj_on A (Empty_set _) f.
  intros; red; intros; auto with sets.
Qed.

(*Wyckoff*)
Lemma fun_dom_ran_incl1 :
  forall {T U:Type} (A A':Ensemble T) (B:Ensemble U) (f:T->U),
    fun_dom_ran A' B f ->
    Included  A A'->
    fun_dom_ran A B f.
  intros T U A A' B' f h1 h2.
  red in h1. red.
  intros x h3.
  apply (h1 _ (h2 _ h3)).
Qed.

(*Wyckoff*)
Lemma fun_dom_ran_incl2 :
  forall {T U:Type} (A:Ensemble T) (B B':Ensemble U) (f:T->U),
    fun_dom_ran A B f ->
    Included B B' ->
    fun_dom_ran A B' f.
  intros T U A B B' f h1 h2.
  red in h1. red.
  intros x h3.
  specialize (h1 _ h3).
  apply h2; auto.
Qed.

(*Wyckoff*)
Lemma fun_dom_ran_add_inj :
  forall {T U:Type} (A:Ensemble T) (a:T) (B:Ensemble U) (f:T->U),
    ~In A a ->
    fun_dom_ran (Add A a) B f ->
    inj_on (Add A a) f ->
    fun_dom_ran A (Subtract B (f a)) f.
  intros T U A a B f h1 h2 h0.
  red in h2. red.
  intros x h3.
  assert (h4:In (Add A a) x). auto with sets.
  specialize (h2 _ h4).
  constructor; auto.
  intro h5.
  inversion h5 as [h6]. clear h5.
  red in h0.
  pose proof (Add_intro2 _  A a) as h7.
  specialize (h0 _ _ h7 h4 h6).
  subst.
  contradiction.
Qed.



(*Wyckoff*)
Lemma inj_on_add :
  forall {T U:Type} (A:Ensemble T) (a:T) (f:T->U),
    inj_on A f -> ~In (Im A f) (f a) ->
    inj_on (Add A a) f.
  intros T U A a f h1 h2. 
  red in h1. red.
  intros x y h3 h4 h5.
  destruct h3 as [x h3|x h3]. destruct h4 as [y h4 | y h4]; subst.
  apply h1 in h5; auto.
  destruct h4.
  rewrite <- h5 in h2.
  contradict h2. apply Im_intro with x; auto.
  destruct h3.
  destruct h4 as [y h4 | y h4].
  rewrite h5 in h2.
  contradict h2. apply Im_intro with y; auto.
  destruct h4.
  reflexivity.
Qed.

(*Wyckoff*)
Lemma inj_on_incl :
  forall {T U:Type} (A A':Ensemble T) (f:T->U),
    Included A A' ->
    inj_on A' f ->
    inj_on A f.
  intros; red; intros; auto.
Qed.

(*Wyckoff*)
Lemma surj_on_add :
  forall {T U:Type} (A:Ensemble T) (a:T) (B:Ensemble U) (f:T->U),
    ~In A a ->
    surj_on (Add A a) B f -> surj_on A (Subtract B (f a)) f.
  intros T U A a B f h1 h2.
  red in h2. red.
  red in h2. red.
  intros x h3.
  destruct h3 as [h3 h4].
  specialize (h2 _ h3).
  destruct h2 as [x h2]. subst.
  destruct h2 as [x h2 | x h2].
  apply Im_intro with x; auto.
  destruct h2.
  contradict h4. constructor.
Qed.


(*Wyckoff*)
Lemma finite_inj_surj_on_impl :
  forall {T U:Type} (A:Ensemble T) (B:Ensemble U) (f:T->U),
    Finite A -> inj_on A f -> surj_on A B f -> fun_dom_ran A B f ->
    Im A f = B. 
  intros T U A B f h1. revert B.
  induction h1 as [|A h4 h5 c h6].
  intros B h1 h2 h0.
  apply surj_on_empty1 in h2.
  subst.
  rewrite image_empty; auto.
  intros B h1 h2 h0.
  rewrite Im_add.
  assert (h7:Included A (Add A c)). auto with sets.
  pose proof (inj_on_incl A _ f h7 h1) as h8.
  pose proof h2 as h2'.
  apply surj_on_add in h2; auto.
  pose proof h0 as h0'.
  apply fun_dom_ran_add_inj in h0'; auto.
  specialize (h5 _ h8 h2 h0').
  rewrite h5.
  rewrite add_sub_compat_in; auto.
  red in h0.
  apply h0.
  right. constructor.
Qed.

End FunctionsOn.




Section Extends.
(*Wyckoff*)
Definition extends_sig {T U:Type} {A B:Ensemble T}
           (f:sig_set B->U) (g:sig_set A->U) :=
  exists (pf:Included A B),
    forall x:sig_set A,
      g x = f (exist _ _ (pf _ (proj2_sig x))).

(*Wyckoff*)
Lemma refl_extends_sig : 
  forall {T U:Type} {A:Ensemble T} (f:sig_set A->U),
    extends_sig f f.
intros T U A f.
red. exists (inclusion_reflexive _ ).
intro x.
destruct x. simpl. f_equal. apply proj1_sig_injective.
simpl.
reflexivity.
Qed.                               



(*Wyckoff*)
Definition extends_sig1 {T U:Type} {A:Ensemble T} 
           (f:T->U) (g:sig_set A->U)  : Prop :=
  forall x:sig_set A, g x = f (proj1_sig x).

(*Wyckoff*)
(*[prim] fefers to the primitive type of the domain, as opposed
  to assuming a priori that the extended function's domain is 
  a sigma type.*)
Definition extends_prim {T T' U:Type} (f:T->U) (g:T'->U) : Prop :=
  exists A:Ensemble T, 
    T' = sig_set A /\
    forall (x:T) (x':T') (pf:In A x),
      existT id _ x' = existT id _ (exist _ _ pf) ->
      f x = g x'.


(*Wyckoff*)
Lemma extends_sig1_impl_extends_prim : 
  forall {T U:Type} {A:Ensemble T}
         (f:T->U) (g:sig_set A->U),
    extends_sig1 f g -> extends_prim f g.
intros T U A f g h1.
red in h1. red.
exists A. split; auto.
intros x x' h2 h3.
specialize (h1 (exist _ _ h2)).
simpl in h1.
rewrite <- h1.
f_equal.
apply inj_pair2 in h3. symmetry.
assumption.
Qed.

(*See TypeUtilities for the converse of the above*)

(*Wyckoff*)
Definition extends_sig_prop {T U:Type} {P P':T->Prop}
           (f:{x:T | P' x} ->U) (g:{x:T | P x}->U) :=
  exists (pf:forall x:T, P x -> P' x),
    forall x:{t:T | P t},
      g x = f (exist _ _ (pf _ (proj2_sig x))).




(*Search past in_im_proj1_sig_iff to get another "extends"
  function, [extends_sig_sig].*)

(*Wyckoff*)
Definition common_extension2 {T1 T2 T U:Type}
           (f:T->U) (g1:T1->U) (g2:T2->U) : Prop :=
  extends_prim f g1 /\ extends_prim f g2.

(*Wyckoff*)
Definition fam_fun_one_ran {U:Type} :=
  Ensemble {T:Type & T->U}.

(*Wyckoff*)
Definition common_extension_fam {T U:Type} (f:T->U) 
           (F:Ensemble {V:Type & V->U}) :=
  forall g, In F g -> extends_prim f (projT2 g).         

(*Wyckoff*)
Lemma common_extension2_fam_compat_iff : 
  forall {T1 T2 T U:Type}
         (f:T->U) (g1:T1->U) (g2:T2->U),
    common_extension2 f g1 g2 <->
    common_extension_fam f (Couple (existT _ _ g1) (existT _ _ g2)).
intros T1 T2 T U f g1 g2.
split.
intro h1.
red in h1. red. intros g h2. 
destruct h1 as [h1a h1b]. destruct h2; auto.
intro h2.
red in h2. red.  
pose proof (h2 (existT _ _ g1) (Couple_l _ _ _)) as h3. 
pose proof (h2 (existT _ _ g2) (Couple_r _ _ _)) as h4.
simpl in h3. simpl in h4.
split; auto.
Qed.

(*Wyckoff*)
Definition directed_fun_fam {U:Type} 
           (F:Ensemble {V:Type & V->U}) : Prop :=
  forall f g, In F f -> In F g -> 
              exists k, In F k /\
                        common_extension2 (projT2 k) (projT2 f) 
                                          (projT2 g).


(*Wyckoff*)
Definition agree_on {T U:Type} (f g:T->U)
           (A:Ensemble T) : Prop :=
  forall x:T, In A x -> f x = g x.


(*Wyckoff*)
Definition restriction_fun {T U:Type} (f:T->U) (A:Ensemble T) :
  sig_set A -> U :=
  fun x => f (proj1_sig x).


(*Wyckoff*)
Definition restriction_fun2 {T U:Type} (f:T->T->U) (A:Ensemble T) :
  sig_set A -> sig_set A -> U :=
  fun x y => f (proj1_sig x) (proj1_sig y).


(*Wyckoff*)
Definition restriction_sig 
           {T U:Type} {A:Ensemble T}
           (f:sig_set A->U) (B:Ensemble T)
           (pf:Included B A) : sig_set B->U :=
  fun x=>f (exist _ _ (pf _ (proj2_sig x))).

(*Wyckoff*)
Definition restriction_sig2 
           {T U:Type} {A:Ensemble T}
           (f:sig_set A->sig_set A->U) (B:Ensemble T)
           (pf:Included B A) : sig_set B->sig_set B->U :=
  fun x y=>f (exist _ _ (pf _ (proj2_sig x))) (exist _ _ (pf _ (proj2_sig y))).

(*Wyckoff*)
Definition closed_fun {T:Type} {A B:Ensemble T} 
           (f:sig_set A->sig_set A) (pf:Included B A) : Prop :=
           forall (x:T) (pfi:In B x), 
                     In B (proj1_sig (f (exist _ _ (pf _ pfi)))).

(*Wyckoff*)
Definition closed_fun2 {T:Type} {A B:Ensemble T} 
           (f:sig_set A->sig_set A->sig_set A) (pf:Included B A) : Prop :=
           forall (x y:T) (pfx:In B x) (pfy:In B y), 
                     In B (proj1_sig (f (exist _ _ (pf _ pfx))
                                        (exist _ _ (pf _ pfy)))).

(*Wyckoff*)
Definition restriction_sig' 
           {T:Type} {A:Ensemble T}
           (f:sig_set A->sig_set A) (B:Ensemble T)
           (pf:Included B A) 
           (pfcl:closed_fun f pf)
: sig_set B->sig_set B :=
  fun x=>exist _ _ (pfcl _ (proj2_sig x)).

(*Wyckoff*)
Definition restriction_sig2' 
           {T:Type} {A:Ensemble T}
           (f:sig_set A->sig_set A->sig_set A) (B:Ensemble T)
           (pf:Included B A) 
           (pfcl:closed_fun2 f pf)
: sig_set B->sig_set B->sig_set B :=
  fun x y=>exist _ _ (pfcl _ _  (proj2_sig x) (proj2_sig y)).

         
Definition same_im_subsets_sig 
           {T U:Type} {A:Ensemble T}
           (f:sig_set A->U) :=
  [B:Ensemble T | exists pf:Included B A, 
                            Im (full_sig A) f =
                            Im (full_sig B) (restriction_sig f B pf)].

Lemma extends_sig_iff :
  forall {T U:Type} {A B:Ensemble T}
         (f:sig_set B->U) (g:sig_set A->U),
    extends_sig f g <->
    exists (pf:Included A B),
      restriction_sig f _ pf  = g.
  intros T U A B f g.
  split.
  intro h1.
  red in h1. destruct h1 as [h1 h2].
  exists h1.
  apply functional_extensionality.
  intro x.
  specialize (h2 x).
  rewrite h2.
  unfold restriction_sig.
  reflexivity.
  intro h1.
  destruct h1 as [h1 h2].
  red. exists h1.
  intro x.
  subst.
  unfold restriction_sig.
  reflexivity.
Qed.


(*Wyckoff*)
Lemma im_full_sig_restriction :
  forall {T U:Type} (E:Ensemble T) (p:T->U),
    Im (full_sig E) (restriction_fun p E) = Im E p.
  intros T U E p.
  apply Extensionality_Ensembles; red; split; red; intros x h1.
  destruct h1 as [x h1]; subst.
  apply Im_intro with (proj1_sig x).
  apply proj2_sig.
  unfold restriction_fun.
  reflexivity.
  destruct h1 as [x h1]; subst.
  apply Im_intro with (exist _ _ h1).
  constructor.
  unfold restriction_fun; auto.
Qed.


(*Wyckoff*)
Lemma im_full_sig_union_eq : 
  forall {T U:Type} {A B:Ensemble T}
         (f:sig_set (Union A B)->U),
    Im (full_sig (Union A B)) f =
    Union (Im (full_sig A) (restriction_sig f _ (Union_increases_l _ A B)))
          (Im (full_sig B) (restriction_sig f _ (Union_increases_r _ A B))).
intros T U A B f. unfold restriction_sig.
apply Extensionality_Ensembles; red; split; red; intros x h1.
destruct h1 as [x h1]. subst.
clear h1.
destruct x as [x h1].
destruct h1 as [x h1 | x h1].
left.
apply Im_intro with (exist _ _ h1). constructor.
f_equal.
apply proj1_sig_injective; auto.
right.
apply Im_intro with (exist _ _ h1). constructor.
f_equal.
apply proj1_sig_injective; auto.
destruct h1 as [x h1 | x h1].
destruct h1 as [x h1]. subst.
clear h1.
destruct x as [x h1].
assert (h2:In (Union A B) x). left; auto with sets.
apply Im_intro with (exist _ _ h2). constructor.
f_equal. apply proj1_sig_injective; auto.
destruct h1 as [x h1]. subst.
clear h1.
destruct x as [x h1].
assert (h2:In (Union A B) x). right; auto with sets.
apply Im_intro with (exist _ _ h2). constructor.
f_equal.
apply proj1_sig_injective; auto.
Qed.

(*Wyckoff*)
Lemma im_full_sig_union_eq' : 
  forall {T U:Type} {A B C:Ensemble T}
         (f:sig_set A->U)
         (pfb:Included B A) (pfc:Included C A),
    A = Union B C  ->
    Im (full_sig A) f =
    Union (Im (full_sig B) (restriction_sig f _ pfb))
          (Im (full_sig C) (restriction_sig f _ pfc)).
intros T U A B C f h1 h2 h3.
subst.
rewrite im_full_sig_union_eq.
unfold restriction_sig.  
f_equal. 
apply im_ext_in; intros; simpl; f_equal. apply proj1_sig_injective; auto.
apply im_ext_in; intros; simpl; f_equal. apply proj1_sig_injective; auto.
Qed.


(*Wyckoff*)
Definition restriction_sig_prop {T U:Type} (P P':T->Prop)
           (pf:forall x:T, P x -> P' x)
           (f:{t:T | P' t} -> U) : {t:T | P t}->U.
intros x. destruct x as [x h1]. apply pf in h1.
refine (f (exist _ _ h1)).
Defined.

(*Wyckoff*)
Lemma extends_restriction_sig_prop :
  forall {T U:Type} (P P':T->Prop)
         (pf:forall x:T, P x -> P' x)
         (f:{t:T | P' t} -> U),
    extends_sig_prop f (restriction_sig_prop P P' pf f).
intros T U P P' h1 f.
red.
exists h1.
intro x.
unfold restriction_sig_prop.
destruct x as [x h2]. simpl.
reflexivity.
Qed.

(*Wyckoff*)
Lemma im_full_sig_add : 
  forall {T U:Type} {A:Ensemble T} {a:T}
         (f:sig_set (Add A a)->U),
         Im (full_sig (Add A a)) f =
         Add (Im (full_sig A) (restriction_sig f _ (incl_add A a)))
             (f (exist _ _ (Add_intro2 _ A a))).
intros T U A a f.
apply Extensionality_Ensembles.
red. split.
red. intros y h1.
destruct h1 as [y h1]. subst. clear h1.
destruct y as [y h1]. destruct h1 as [y h1 | y h2]. left.
apply Im_intro with (exist _ _ h1). constructor.
unfold restriction_sig. f_equal. apply proj1_sig_injective.
simpl. reflexivity.
destruct h2. right.
assert (h1: f (exist (In (Add A a)) a (Add_intro2 T A a)) = 
            f
               (exist (fun x : T => In (Add A a) x) a
                      (Union_intror T A (Singleton a)
                                    a (In_singleton T a)))).
  f_equal. apply proj1_sig_injective. simpl. reflexivity.
rewrite h1. constructor.
red. intros y h1. 
destruct h1 as [y h1 | y h2].
destruct h1 as [x h1]. subst. clear h1. destruct x as [x h1].
apply Im_intro with (exist _ _ (Add_intro1 _ A a x h1)). constructor.
unfold restriction_sig. f_equal. apply proj1_sig_injective.
simpl. reflexivity.
destruct h2.
apply Im_intro with (exist (In (Add A a)) a (Add_intro2 T A a)); auto.
constructor.
Qed.


(*Wyckoff*)
Lemma setminus_im_full_sig_incl_setminus : 
  forall {T U:Type} (A B:Ensemble T) (pf:Included B A)
         (f:sig_set A->U),
    Included (Setminus (Im (full_sig A) f) 
                       (Im (full_sig B) (restriction_sig f _ pf)))
             (Im (full_sig (Setminus A B)) (restriction_sig
                                              f _
                                              (setminus_inc A B))).
intros T U A B h1 f. red. intros y h2.
destruct h2 as [h2 h3].
destruct h2 as [y h2 u]. subst. clear h2.
destruct y as [y h4].
assert (h5:~In B y).
  intro h5.
  contradict h3.
  apply Im_intro with (exist _ _ h5). constructor.
  unfold restriction_sig. simpl.
  f_equal. apply proj1_sig_injective. simpl. reflexivity.
assert (h6:In (Setminus A B) y).
  constructor; auto.
apply Im_intro with (exist _ _ h6). constructor.
unfold restriction_sig. simpl.
f_equal.
apply proj1_sig_injective. simpl. reflexivity.
Qed.


End Extends.


(*This Section defines the Von Neumann notion of a natural number as the
set of its predecessors ([seg]) and one that counts from one ([seg_one])
 (the latter so as to give more flexibility to those utilizing a back-and
-forth argument, where you don't have complete control of the terms.*)
Section SegSet.

(*Wyckoff*)
Definition seg (n:nat) := [m:nat | m < n].

(*Wyckoff*)
Definition segT (n:nat) := {m:nat | m < n}.

(*Wyckoff*)
Inductive seg_ens (n:nat) : Ensemble nat -> Prop :=
| seg_ens_intro : forall (E:Ensemble nat),
                    Included E (seg n) ->
                    seg_ens n E.
 

End SegSet.

Notation "n @" := (segT n) (at level 30, right associativity).

Section SegSet'.
(*Wyckoff*)
Definition seg_weak (n:nat) := [m:nat | m <= n].

(*Wyckoff*)
Definition seg_one (n:nat) := [m:nat | 1 <= m /\ m <= n].

(*Wyckoff*)
Definition seg_one_t (n:nat) := {m:nat | 1 <= m /\ m <= n}.
 
(*Wyckoff*)
Lemma segT_eq_dec :
  forall (n:nat),
    type_eq_dec (n @).
  intros n. red. intros a b.
  destruct a as [a ha], b as [b h].
  destruct (nat_eq_dec a b) as [h1 | h1]; subst.
  left. f_equal. apply proof_irrelevance.
  right. intro h2.
  apply exist_injective in h2.
  contradiction.
Qed.


(*Wyckoff*)
Lemma in_st :
  forall {n:nat} (m:n @),
    In (seg n) (proj1_sig m).
  intros n m.
  constructor.
  apply proj2_sig.
Qed.

(*Wyckoff*)
Lemma in_seg :
  forall (m n:nat),
    m < n ->
    In (seg n) m.
  intros; constructor; auto.
Qed.

(*Wyckoff*)
Lemma in_pred_seg :
  forall (n:nat),
    0 < n ->
    Inn (seg n) (pred n).
  intro; constructor; auto with arith.
Qed.


(*Wyckoff*)
Lemma seg_in :
  forall (m n:nat),
    In (seg n) m ->
    m < n.
  intros ? ? h1; destruct h1; auto.
Qed.



(*Wyckoff*)
Lemma in_sot :
  forall {n:nat} (m:seg_one_t n),
    In (seg_one n) (proj1_sig m).
  intros n m.
  constructor.
  apply proj2_sig.
Qed.


(*Wyckoff*)
Lemma seg_one_t_compat : 
  forall (n:nat),
    im_proj1_sig (Full_set (seg_one_t n)) = seg_one n.
intro n.
apply Extensionality_Ensembles; red; split; red; intros m h1.
destruct h1 as [m h1]. subst.
constructor. apply proj2_sig.
destruct h1 as [h1].
unfold im_proj1_sig.
apply Im_intro with (exist (fun r => 1 <= r <= n) _ h1).
constructor.
auto.
Qed.


(*Wyckoff*)
Lemma segT_compat : 
  forall (n:nat),
    im_proj1_sig (Full_set (n @)) = seg n.
intro n.
apply Extensionality_Ensembles; red; split; red; intros m h1.
destruct h1 as [m h1]. subst.
constructor. apply proj2_sig.
destruct h1 as [h1]. 
unfold im_proj1_sig.
apply Im_intro with (exist (fun r => r < n) _ h1).
constructor.
auto.
Qed.


(*Wyckoff*)
Lemma seg_one_t_lt :
  forall {n:nat} (m:seg_one_t n),
    pred (proj1_sig m) < n.
  intros n m; destruct m; simpl; omega.
Qed.

(*Wyckoff*)
Lemma seg_one_t_le_le :
  forall {n:nat} (m:seg_one_t n),
    1 <= S (pred (proj1_sig m)) <= n.
  intros n m; destruct m; simpl; omega.
Qed.

(*Wyckoff*)
Lemma segT_le_le :
  forall {n:nat} (m:n @),
    1 <= S (proj1_sig m) <= n.
  intros ? m; destruct m; simpl; omega.
Qed.


(*Wyckoff*)
Definition sot_to_st {n:nat} (m:seg_one_t n) : n @.
  refine (exist (fun c => c < n) _ _).
  apply (seg_one_t_lt m).
Defined.


(*Wyckoff*)
Lemma sot_to_st_compat :
  forall {n:nat} (m:seg_one_t n),
    proj1_sig (sot_to_st m) = pred (proj1_sig m).
  intros n m.
  destruct m as [m h1].
  unfold sot_to_st.
  destruct h1 as [h1 h2]; simpl; auto.
Qed.

(*Wyckoff*)
Definition st_to_sot {n:nat} (m:n @) : seg_one_t n.
  destruct m as [m h1]. red.
  assert (h2:1 <= (S m) <= n). auto with arith.
  refine (exist (fun a => 1 <= a <= n) _ h2).
Defined.

(*Wyckoff*)
Lemma st_to_sot_compat :
  forall {n:nat} (m:n @),
    proj1_sig (st_to_sot m) = S (proj1_sig m).
  intros n m.
  destruct m as [m h1]; simpl; auto.
Qed.

(*Wyckoff*)
Definition st_to_sot_eq :
  forall {n:nat} (m:n @),
    let pf := iff1 (lt_sot_iff _ _) (seg_one_t_lt (st_to_sot m)) in
    st_to_sot m = exist (fun c => 1 <= c <= n) _ pf.
  intros n m; simpl.
  apply proj1_sig_injective; simpl.
  rewrite st_to_sot_compat; simpl; auto.
Qed.


(*Wyckoff*)
Definition sot_to_st_eq :
  forall {n:nat} (m:seg_one_t n),
    let pf := seg_one_t_lt m in
    sot_to_st m = exist (fun c => c < n) _ pf.
  intros n m; simpl.
  apply proj1_sig_injective; simpl; auto.
Qed.

(*Wyckoff*)
Definition st_to_sot_to_st :
  forall {n:nat} (m:seg_one_t n),
    st_to_sot (sot_to_st m) = m.
  unfold st_to_sot, sot_to_st.
  intros n m.
  destruct m as [m h2]; apply proj1_sig_injective; simpl; omega.
Qed.

(*Wyckoff*)
Definition sot_to_st_to_sot :
  forall {n:nat} (m:n @),
    sot_to_st (st_to_sot m) = m.
  unfold st_to_sot, sot_to_st.
  intros n m.
  destruct m as [m h2]; apply proj1_sig_injective; simpl; omega.
Qed.


(*Wyckoff*)
Lemma full_set_sot :
  forall (n:nat), 
    Full_set (seg_one_t n) =
    Im (Full_set (n @)) (st_to_sot).
  intro n.
  apply Extensionality_Ensembles; auto with sets.
  red; split; red; intros x h1.
  apply Im_intro with (sot_to_st x).
  constructor.
  rewrite st_to_sot_to_st; auto.
  constructor.
Qed.

(*Wyckoff*)
Lemma full_set_st :
  forall (n:nat), 
    Full_set (n @) =
    Im (Full_set (seg_one_t n)) (sot_to_st).
intro n.
  apply Extensionality_Ensembles; auto with sets.
  red; split; red; intros x h1.
  apply Im_intro with (st_to_sot x).
  constructor.
  rewrite sot_to_st_to_sot; auto.
  constructor.
Qed.



(*Wyckoff*)
Definition st_conv {n:nat} (m:n @) :
  sig_set (seg n) := exist _ _ (in_st m).

(*Wyckoff*)
Definition st_conv_rev {n:nat} (m:sig_set (seg n)) : n @ :=
  exist _ _ (match (proj2_sig m) with
               | intro_characteristic_sat P => P
             end).


(*A conversion function for functions on [seg_sset_t] (st)*)
(*Wyckoff*)
Definition fun_st_conv {U:Type} {n:nat} (f:n @ -> U) :
  sig_set (seg n) -> U.
  intro x.
  refine (f _).
  refine (exist _ _ (match (proj2_sig x) with | intro_characteristic_sat P => P end)).
Defined.

(*Wyckoff*)
Definition fun_st_conv_rev {U:Type} {n:nat}  (f:sig_set (seg n) -> U) : n @ -> U.
  intro m.
  refine (f _).
  red.
  refine (exist _ _ (in_st m)).
Defined.



(*A conversion function for functions on [seg_sset_one_t] (sot)*)
(*Wyckoff*)
Definition fun_sot_conv {U:Type} {n:nat} (f:seg_one_t n -> U) :
  sig_set (seg_one n) -> U.
  intro x.
  refine (f _).
  refine (exist _ _ (match (proj2_sig x) with | intro_characteristic_sat P => P end)).
Defined.

(*Wyckoff*)
Definition fun_sot_conv_rev {U:Type} {n:nat}
           (f:sig_set (seg_one n) -> U) : seg_one_t n -> U.
  intro m.
  refine (f _).
  red.
  refine (exist _ _ (in_sot m)).
Defined.



(*Wyckoff*)
Definition ens_ss_conv {n:nat} (A:Ensemble (sig_set (seg n))) :
  Ensemble (n @) :=
  Im A
     (fun x =>
        exist _ _
              (match (proj2_sig x) with
                 | P =>
                   match P with
                     | intro_characteristic_sat P' => P'
                   end
               end)).



(*Wyckoff*)
Definition ens_sso_conv {n:nat} (A:Ensemble (sig_set (seg_one n))) :
  Ensemble (seg_one_t n) :=
  Im A (fun x => exist _ _
                       (match (proj2_sig x) with
                          | P =>
                            match P with
                              | intro_characteristic_sat P' => P'
                            end
                        end)).

(*Wyckoff*)
Definition ens_st_to_sot {n:nat} (A:Ensemble (n @)) :
  Ensemble (seg_one_t n).
  refine (Im A _).  
  intro m.
  refine (exist _ (S (proj1_sig m))  _).
  apply segT_le_le.
Defined.


(*Wyckoff*)
Definition ens_sot_to_st {n:nat} (A:Ensemble (seg_one_t n)) :
  Ensemble (n @).
  refine (Im A _).
  intro m.
  refine (exist _ (pred (proj1_sig m)) _).
  apply seg_one_t_lt.
Defined.

(*Wyckoff*)
Lemma ens_st_to_sot_to_st :
  forall {n:nat} (A:Ensemble (seg_one_t n)),
    ens_st_to_sot (ens_sot_to_st A) = A.
  intros n A.
  unfold ens_st_to_sot, ens_sot_to_st.
  rewrite im_im.
  rewrite im_id.
  apply im_ext_in.
  intros m h1; simpl.
  apply proj1_sig_injective; simpl; auto.
  destruct m as [m hm]; simpl.
  omega.
Qed.

(*Wyckoff*)
Lemma ens_sot_to_st_to_sot :
  forall {n:nat} (A:Ensemble (n @)),
    ens_sot_to_st (ens_st_to_sot A) = A.
  intros n A.
  unfold ens_st_to_sot, ens_sot_to_st.
  rewrite im_im.
  rewrite im_id.
  apply im_ext_in.
  intros m h1; simpl.
  apply proj1_sig_injective; simpl; auto.
Qed.

(*Wyckoff*)
Lemma ens_sot_to_st_im :
  forall {n:nat} (A:Ensemble (seg_one_t n)),
    A = Im (ens_sot_to_st A) st_to_sot.
  intros n A.
  unfold ens_sot_to_st, st_to_sot.
  rewrite im_im.
  rewrite im_id at 1.
  apply im_ext_in.
  intros m h1.
  apply proj1_sig_injective; simpl; auto.
  destruct m; simpl; omega.
Qed.

(*Wyckoff*)
Lemma ens_st_to_sot_im :
  forall {n:nat} (A:Ensemble (n @)),
    A = Im (ens_st_to_sot A) sot_to_st.
  intros n A.
  unfold ens_st_to_sot, sot_to_st.
  rewrite im_im.
  rewrite im_id at 1.
  apply im_ext_in.
  intros m h1.
  apply proj1_sig_injective; simpl; auto.
Qed.

(*Wyckoff*)
Lemma im_ens_sot :
  forall {T:Type} {n:nat} (A:Ensemble (seg_one_t n))
         (f:seg_one_t n-> T),
    Im A f =
    Im (ens_sot_to_st A) (compose f st_to_sot).
  intros T n A f.
  unfold ens_sot_to_st.
  rewrite im_im.
  apply im_ext_in.
  intros m h1.
  unfold compose, st_to_sot.
  f_equal.
  apply proj1_sig_injective; simpl.
  destruct m; simpl; omega.
Qed.

(*Wyckoff*)
Lemma im_ens_st :
  forall {T:Type} {n:nat} (A:Ensemble (n @)) (f:n @-> T),
    Im A f =
    Im (ens_st_to_sot A) (compose f sot_to_st).
  intros T n A f.
  unfold ens_st_to_sot.
  rewrite im_im.
  apply im_ext_in.
  intros m h1.
  unfold compose, sot_to_st.
  f_equal.
  apply proj1_sig_injective; simpl; auto.
Qed.


(*Wyckoff*)
Lemma ens_st_to_sot_full_set :
  forall (n:nat),
    ens_st_to_sot (Full_set (n @)) = Full_set (seg_one_t n).
  intro n.
  unfold ens_st_to_sot.
  apply Extensionality_Ensembles; auto with sets.
  red; split; red; intros m h1.
  constructor.
  apply Im_intro with (sot_to_st m).
  constructor.
  apply proj1_sig_injective; simpl.
  destruct m; simpl; omega.
Qed.

(*Wyckoff*)
Lemma ens_sot_to_st_full_set :
  forall (n:nat),
    ens_sot_to_st (Full_set (seg_one_t n)) = Full_set (n @).
  intro n.
  unfold ens_sot_to_st.
  apply Extensionality_Ensembles; auto with sets.
  red; split; red; intros m h1. 
  constructor.
  apply Im_intro with (st_to_sot m).
  constructor.
  apply proj1_sig_injective; simpl; auto.
  destruct m; simpl; omega.
Qed.



(*Wyckoff*)
Definition segT_fun_to_nat_fun {T:Type} {n:nat} (f:n @->T) (def:T) :=
  (fun m:nat => match (lt_le_dec m n) with
                  | left P =>
                    f (exist (fun a => a < n) _ P)
                  | _ => def
                end).

(*Wyckoff*)
Lemma segT_fun_to_nat_fun_compat :
  forall {T:Type} {n:nat} (f:n @->T) (def:T) (m:nat)
         (pf:m < n),
    segT_fun_to_nat_fun f def m = 
    f (exist (fun a => a < n) _ pf).
  intros T n f def m h1.
  unfold segT_fun_to_nat_fun.
  destruct lt_le_dec as [h2 | h2].
  f_equal. f_equal. apply proof_irrelevance.
  omega.
Qed.

(*Wyckoff*)
Definition segT_fun_to_seg_one {T:Type} {n:nat} (f:n @->T) :
  seg_one_t n -> T.
  intro m.
  destruct m as [m h1].
  destruct m as [|m].
  destruct h1 as [h1].
  apply le0 in h1.
  discriminate.
  rewrite <- lt_sot_iff in h1.
  refine (f (exist (fun c => c < n) _ h1)).
Defined.

(*Wyckoff*)
Lemma segT_fun_to_seg_one_compat :
  forall {T:Type} {n:nat} (f:n @->T) (m:nat) (pf':1 <= S m <= n),
    
    let pf := iff2 (lt_sot_iff _ _) pf' in
    let f' := segT_fun_to_seg_one f in
    f' (exist (fun c => 1 <= c <= n) _ pf') = 
    f (exist (fun c => c < n) _ pf).
  simpl; intros; repeat f_equal; apply proof_irrelevance.
Qed.

(*Wyckoff*)
Definition seg_one_t_fun_to_seg {T:Type} {n:nat}
           (f:seg_one_t n -> T) : n @->T.
  intro m. 
  destruct m as [m h1].
  rewrite lt_sot_iff in h1.
  refine (f (exist (fun c => 1 <= c <= n) _ h1)).
Defined.

(*Wyckoff*)
Lemma seg_one_t_fun_to_seg_compat :
  forall {T:Type} {n:nat} (f:seg_one_t n->T) (m:nat) (pf:m < n),
    let pf' := iff1 (lt_sot_iff _ _) pf in
    let f' := seg_one_t_fun_to_seg f in
    f' (exist (fun c => c < n) _ pf) = 
    f (exist (fun c => 1 <= c <= n) _ pf').
  simpl; intros; repeat f_equal; apply proof_irrelevance.
Qed.

(*Wyckoff*)
Lemma inj_segT_fun_to_seg_one_iff :
  forall {T:Type} {n:nat} (f:n @->T),
    injective (segT_fun_to_seg_one f) <->
    injective f.
  intros T n f; split; intro h1; red in h1; red; intros x y h2.
  destruct x as [x hx], y as [y hy].
  pose proof hx as h3.
  pose proof hy as h4.
  rewrite lt_sot_iff in h3, h4.
  specialize (h1 (exist _ _ h3) (exist _ _ h4)).
  hfirst h1.
  do 2 rewrite segT_fun_to_seg_one_compat.
  assert (hx = iff2 (lt_sot_iff _ _) h3).
    apply proof_irrelevance.
  assert (hy = iff2 (lt_sot_iff _ _) h4).
    apply proof_irrelevance.
  subst; auto.
  specialize (h1 hf).
  apply proj1_sig_injective; auto; simpl.
  apply exist_injective in h1; auto.
  destruct x as [x hx], y as [y hy].
  destruct x as [|x], y as [|y]; try omega.
  do 2 rewrite segT_fun_to_seg_one_compat in h2.
  apply h1 in h2.
  apply exist_injective in h2.
  subst; f_equal; apply proof_irrelevance.
Qed.

  
(*Wyckoff*)
Lemma inj_seg_one_t_fun_to_seg_iff :
  forall {T:Type} {n:nat} (f:seg_one_t n->T),
    injective (seg_one_t_fun_to_seg f) <->
    injective f.
  intros T n f; split; intro h1; red in h1; red; intros x y h2.
  destruct x as [x hx], y as [y hy].
  pose proof hx as h3.
  pose proof hy as h4.
  destruct x as [|x], y as [|y]; try omega.
  rewrite <- lt_sot_iff in h3, h4.
  specialize (h1 (exist (fun c => c < n) _ h3)
                 (exist (fun c => c < n) _ h4)).
  do 2 rewrite seg_one_t_fun_to_seg_compat in h1.
  hfirst h1.
  assert (h5:hx = (iff1 (lt_sot_iff _ _) h3)).  apply proof_irrelevance.
  assert (h6:hy = (iff1 (lt_sot_iff _ _) h4)).  apply proof_irrelevance.
  subst; auto.
  specialize (h1 hf).
  apply exist_injective in h1; subst; f_equal; apply proof_irrelevance.
  destruct x as [x hx], y as [y hy].
  pose proof hx as h3. pose proof hy as h4.
  rewrite lt_sot_iff in h3, h4.
  specialize (h1 (exist _ _ h3) (exist _ _ h4)).
  do 2 rewrite seg_one_t_fun_to_seg_compat in h2.
  assert (h5:h3 = iff1 (lt_sot_iff _ _) hx). apply proof_irrelevance.
  assert (h6:h4 = iff1 (lt_sot_iff _ _) hy). apply proof_irrelevance.
  subst.
  specialize (h1 h2).
  apply exist_injective in h1; apply S_inj in h1; subst; f_equal;
  apply proof_irrelevance.
Qed.

  
(*Wyckoff*)
Lemma surj_segT_fun_to_seg_one_iff :
  forall {T:Type} {n:nat} (f:n @->T),
    surjective (segT_fun_to_seg_one f) <->
    surjective f.
  intros T n f; split; intro h1; red; red in h1; intros y;
  specialize (h1 y); destruct h1 as [x h2]; subst.
  exists (sot_to_st x).
  destruct x as [x h1]; subst.
  destruct x as [|x]; try omega.
  rewrite segT_fun_to_seg_one_compat.
  f_equal.
  apply proj1_sig_injective; simpl.
  destruct h1; simpl; auto.
  exists (st_to_sot x).
  rewrite st_to_sot_eq.
  rewrite segT_fun_to_seg_one_compat.
  f_equal; apply proj1_sig_injective; simpl; auto.
  rewrite st_to_sot_compat; simpl; auto.
Qed.

  
(*Wyckoff*)
Lemma surj_seg_one_t_fun_to_seg_iff : 
  forall {T:Type} {n:nat} (f:seg_one_t n->T),
    surjective (seg_one_t_fun_to_seg f) <->
    surjective f.
  intros T n f; split; intro h1; red; red in h1; intros y;
  specialize (h1 y); destruct h1 as [x h2]; subst. 
  exists (st_to_sot x).
  destruct x as [x hx].
  rewrite seg_one_t_fun_to_seg_compat.
  rewrite st_to_sot_eq.
  repeat f_equal; apply proof_irrelevance.

  exists (sot_to_st x).
  rewrite sot_to_st_eq. 
  rewrite seg_one_t_fun_to_seg_compat.
  f_equal; apply proj1_sig_injective; simpl; auto.
  pose proof (seg_one_t_lt x) as h1.
  destruct x as [x [h2 h3]].
  simpl in h1; simpl.
  omega.
Qed.


  
(*Wyckoff*)
Lemma bij_segT_fun_to_seg_one_iff :
  forall {T:Type} {n:nat} (f:n @->T),
    bijective (segT_fun_to_seg_one f) <->
    bijective f.
  intros T n f; split; intro h1; red; red in h1;
  destruct h1 as [h1 h2]; split.
  rewrite inj_segT_fun_to_seg_one_iff in h1; auto.
  rewrite surj_segT_fun_to_seg_one_iff in h2; auto.
  rewrite inj_segT_fun_to_seg_one_iff; auto.
  rewrite surj_segT_fun_to_seg_one_iff; auto.
Qed.

  
(*Wyckoff*)
Lemma bij_seg_one_t_fun_to_seg_iff :
  forall {T:Type} {n:nat} (f:seg_one_t n->T),
    bijective (seg_one_t_fun_to_seg f) <->
    bijective f.
  intros T n f; split; intro h1; red; red in h1;
  destruct h1 as [h1 h2]; split.
  rewrite inj_seg_one_t_fun_to_seg_iff in h1; auto.
  rewrite surj_seg_one_t_fun_to_seg_iff in h2; auto.
  rewrite inj_seg_one_t_fun_to_seg_iff; auto.
  rewrite surj_seg_one_t_fun_to_seg_iff; auto.
Qed.




(*Wyckoff*)
Lemma inj_st_conv_iff :
  forall {U:Type} {n:nat} (f:n @->U),
    injective f <-> injective (fun_st_conv f).
  intros U n f.
  split.
  intro h1.
  red in h1. red.
  intros x y h2.
  destruct x as [x h3], y as [y h4].
  unfold fun_st_conv in h2.
  specialize (h1 _ _ h2).
  apply exist_injective in h1. simpl in h1. subst.
  f_equal. apply proof_irrelevance.
  intro h1.
  red in h1. red.
  intros x y h2.
  unfold fun_st_conv in h1.
  destruct x as [x h3], y as [y h4].
  assert (h5:In (seg n) x). constructor; auto.
  assert (h6:In (seg n) y). constructor; auto.
  specialize (h1 (exist _  _ h5) (exist _ _ h6)).
  apply proj1_sig_injective. simpl.
  simpl in h1.
  let P := type of h1 in match P with ?h1 -> _ => assert (h7:h1) end.
  destruct h5 as [h5], h6 as [h6].
  pose proof (proof_irrelevance _ h3 h5).
  pose proof (proof_irrelevance _ h4 h6).
  subst; auto.
  specialize (h1 h7).
  apply exist_injective in h1; auto.
Qed.

(*Wyckoff*)
Lemma inj_sot_conv_iff :
  forall {U:Type} {n:nat} (f:seg_one_t n->U),
    injective f <-> injective (fun_sot_conv f).
  intros U n f.
  split.
  intro h1.
  red in h1. red.
  intros x y h2.
  destruct x as [x h3], y as [y h4].
  unfold fun_st_conv in h2.
  specialize (h1 _ _ h2).
  apply exist_injective in h1. simpl in h1. subst.
  f_equal. apply proof_irrelevance.
  intro h1.
  red in h1. red.
  intros x y h2.
  unfold fun_st_conv in h1.
  destruct x as [x h3], y as [y h4].
  assert (h5:In (seg_one n) x). constructor; auto.
  assert (h6:In (seg_one n) y). constructor; auto.
  specialize (h1 (exist _  _ h5) (exist _ _ h6)).
  apply proj1_sig_injective. simpl.
  simpl in h1.
  let P := type of h1 in match P with ?h1 -> _ => assert (h7:h1) end.
  destruct h5 as [h5], h6 as [h6].
  pose proof (proof_irrelevance _ h3 h5).
  pose proof (proof_irrelevance _ h4 h6).
  subst; auto.
  specialize (h1 h7).
  apply exist_injective in h1.
  assumption.
Qed.

(*Wyckoff*)
Lemma surj_st_conv_iff :
  forall {U:Type} {n:nat} (f:n @->U),
    surjective f <-> surjective (fun_st_conv f).
  intros U n f.
  split.
  intro h1.
  red in h1. red.
  intro y. specialize (h1 y). destruct h1 as [x h1]. subst.
  assert (h1:In (seg n) (proj1_sig x)). constructor; apply proj2_sig.
  exists (exist _ _ h1).
  unfold fun_st_conv.
  f_equal. apply proj1_sig_injective. simpl. reflexivity.
  intro h1.
  red in h1. red.
  intro y. specialize (h1 y).
  destruct h1 as [x h1]; subst.
  destruct x as [x h1]. destruct h1 as [h1].
  exists (exist (fun a => a < n) _ h1).
  unfold fun_st_conv.
  f_equal.
Qed.

(*Wyckoff*)
Lemma surj_sot_conv_iff :
  forall {U:Type} {n:nat} (f:seg_one_t n->U),
    surjective f <-> surjective (fun_sot_conv f).
  intros U n f.
  split.
  intro h1. red in h1.
  intro y. specialize (h1 y). destruct h1 as [x h1]. subst.
  assert (h1:In (seg_one n) (proj1_sig x)). constructor; apply proj2_sig.
  exists (exist _ _ h1).
  unfold fun_sot_conv.
  f_equal. apply proj1_sig_injective. simpl. reflexivity.
  intro h1.
  red in h1. red.
  intro y. specialize (h1 y).
  destruct h1 as [x h1]; subst.
  destruct x as [x h1]. destruct h1 as [h1].
  exists (exist (fun a => 1 <= a <= n) _ h1).
  unfold fun_sot_conv.
  f_equal.
Qed.

(*Wyckoff*)
Lemma bij_st_conv_iff :
  forall {U:Type} {n:nat} (f:n @->U),
    bijective f <-> bijective (fun_st_conv f).
  intros U n f.
  split.
  intro h1. red in h1. red.
  destruct h1 as [h1 h2].
  rewrite inj_st_conv_iff in h1.
  rewrite surj_st_conv_iff in h2.
  split; auto.
  intro h1. red in h1. red.
  destruct h1 as [h1 h2].
  rewrite <- inj_st_conv_iff in h1.
  rewrite <- surj_st_conv_iff in h2.
  split; auto.
Qed.

(*Wyckoff*)
Lemma bij_sot_conv_iff :
  forall {U:Type} {n:nat} (f:seg_one_t n->U),
    bijective f <-> bijective (fun_sot_conv f).
  intros U n f.
  split.
  intro h1.
  red in h1. red.
  destruct h1 as [h1 h2].
  rewrite inj_sot_conv_iff in h1.
  rewrite surj_sot_conv_iff in h2.
  split; auto.
  intro h1. red in h1. red.
  destruct h1 as [h1 h2].
  rewrite <- inj_sot_conv_iff in h1.
  rewrite <- surj_sot_conv_iff in h2.
  split; auto.
Qed.

(*Wyckoff*)
Definition st_nest {n:nat} {m:n @} (p:(proj1_sig m) @) : n @.
  destruct m as [m h1]. simpl in p.
  destruct p as [p h2]. red.
  refine (exist (fun a => a < n)_ (lt_trans _ _ _ h2 h1)).
Defined.

(*Wyckoff*)
Lemma proj1_sig_st_nest_eq :
  forall {n:nat} {m:n @} (p:(proj1_sig m) @),
    proj1_sig (st_nest p) = proj1_sig p.
  intros n m p.
  destruct p as [p h1].
  simpl.
  unfold st_nest.
  destruct m as [m h2]. simpl.
  reflexivity.
Qed.

  
(*Wyckoff*)
Definition sot_nest {n:nat} {m:seg_one_t n}
           (p:seg_one_t (proj1_sig m)) : seg_one_t n.
  destruct m as [m h1].  simpl in p.
  destruct p as [p h2]. red.
  destruct h1 as [h1a h1b], h2 as [h2a h2b].
  refine (exist _ _ (conj h2a (le_trans _ _ _ h2b h1b))).
Defined.


(*Wyckoff*)
Lemma segT_0_false : 
  forall n:0 @,
    False.
  intro n.
  destruct n; omega.
Qed.

  
(*Wyckoff*)
Definition segT_S {n:nat} (m:n @) : (S n) @.
  destruct m as [m h1].
  red.
  refine (exist (fun a => a < S n) _ (lt_n_S _ _ h1)).
Defined.

(*Wyckoff*)
Definition segT_S_compat :
  forall {n:nat} (m:n @),
    proj1_sig (segT_S m) = S (proj1_sig m).
  intros ? m; destruct m; auto.
Qed.

(*Wyckoff*)
Definition segT_S_pred {n} (m:pred n @) : n @ :=
  exist (fun c => c < n) _ (lt_pred' _ _ (proj2_sig m)).

(*Wyckoff*)
Lemma segT_S_pred_compat :
  forall {n} (m:pred n @),
    proj1_sig (segT_S_pred m) =
    proj1_sig m.
  auto.
Qed.


(*Wyckoff*)
Definition segT_S_pred_to {n} (m:S (pred n) @) (pf:0 < n) : n @ :=
  exist (fun c => c < n)
        (proj1_sig m)
        (lt_eq_trans _ _ _ (proj2_sig m)
                     (eq_sym (S_pred _ _ pf))).

(*Wyckoff*)
Lemma segT_S_pred_to_compat :
  forall {n} (m:S (pred n) @) (pf:0 < n),
    proj1_sig (segT_S_pred_to m pf) = 
    proj1_sig m.
  auto.
Qed.


(*Wyckoff*)
(*This decrements the value of m*)
Definition segT_pred {n:nat} (m:n @)
           (pf:0 < proj1_sig m) : (pred n) @.
  destruct m as [m h1].
  simpl in pf.
  pose proof (S_pred _ _ pf) as h2.
  rewrite h2 in h1.
  pose proof (lt_pred _ _ h1) as h3.
  refine (exist (fun c => c < pred n) _ h3).
Defined.


(*Wyckoff*)
Definition segT_pred_compat :
  forall {n:nat} (m:n @) (pf:0 < proj1_sig m),
    proj1_sig (segT_pred m pf) =
    pred (proj1_sig m).
  intros ? m; destruct m; auto.
Qed.

(*Wyckoff*)
Definition segT_incr {n:nat} (m:n @) : (S n) @ :=
  exist (fun c => c < S n) (S (proj1_sig m)) (lt_n_S _ _ (proj2_sig m)).

(*Wyckoff*)
Definition segT_incr_compat :
  forall {n:nat} (m:n @),
    proj1_sig (segT_incr m) =
    S (proj1_sig m).
  auto.
Qed.


(*This only narrows the upper bound n (keeping the value of m the same)*)
(*Wyckoff*) 
Definition segT_narrow {n:nat} (m:n @) (pfn:proj1_sig m <> pred n) :
  (pred n) @ :=
  let pflt := lt_neq_impl _ _  (proj2_sig m) pfn in
  exist (fun c => c < pred n) (proj1_sig m) pflt.


(*Wyckoff*)
Lemma segT_narrow_compat :
  forall {n:nat} (m:n @) (pf:proj1_sig m <> pred n),
    proj1_sig (segT_narrow m pf) =
    proj1_sig m.
  auto.
Qed.

(*Wyckoff*)
Definition segT_widen {n:nat} (m:n @) : S n @ :=
  let pflt := lt_S _ _ (proj2_sig m) in
  exist (fun c => c < S n) (proj1_sig m) pflt.

(*Wyckoff*)
Lemma segT_widen_compat :
  forall {n:nat} (m:n @),
    proj1_sig (segT_widen m) = proj1_sig m.
  auto.
Qed.
  


Definition segT_fun n U : Type :=
  n @ -> U.

Definition segT_fun' n m : Type :=
  n @ -> m @.

Definition segged_fun T n : Type :=
  T -> n @.

Definition segged_fun_le {n m T} (f:segged_fun T n) (pfle:m >= n) :
  segged_fun T m :=
  fun x =>
    let (fx, pffx) := f x in
    let hfx' := lt_le_trans _ _ _ pffx pfle in
    exist (fun c => c < m) fx hfx'.


Definition segT_fun_app {n U} (f:segT_fun (S n) U) : nat -> U :=
  fun i =>
    f (exist (fun c => c < S n) (nat_bound (S n) i)
             (lt_bound _ _)).

Definition segT_fun_app' {n m} (f:segT_fun' (S n) m) : nat -> m @ :=
  segT_fun_app f.


End SegSet'.


Notation "n --->  U" := (segT_fun n U) (at level 40, no associativity).
Notation "n ----> m" := (segT_fun' n m) (at level 40, no associativity).
Notation "T _--> n" := (segged_fun T n) (at level 40, no associativity).
Notation "f |--->  i" := (segT_fun_app f i)
                           (at level 40, no associativity).
Notation "f |----> i" := (segT_fun_app' f i)
                           (at level 40, no associativity).



Section SegSet''.

Definition seg_fun_to_segT_fun {U n}
           (f:seg_fun n U) :
  n ---> U :=
  fun i => f (proj1_sig i) (proj2_sig i).


  

Lemma lt_segT_fun_app :
  forall {n U i}
         (f:segT_fun (S n) U)
         (pfi:i < S n),
    f |---> i =
    f (exist (fun c => c < S n) i pfi).
  intros n U i f hi.
  unfold segT_fun_app.
  f_equal; apply proj1_sig_injective; simpl.
  rewrite nat_bound_eq_same_iff; auto.
Qed.

Lemma ge_segT_fun_app :
  forall {n U i}
         (f:segT_fun (S n) U)
         (pfi:i >= S n),
    f |---> i =
    f (exist (fun c => c < S n) n (lt_n_Sn _)).
  intros n U i f hi.
  unfold segT_fun_app.
  f_equal; apply proj1_sig_injective; simpl.
  rewrite nat_bound_eq_iff.
  right; simpl; auto.
Qed.



(*Wyckoff*)
Lemma segT_functional :
  forall (m n:nat),
    m = n ->
    m @ = n @.
  intros; subst; auto.
Qed.

(*Wyckoff*)
Lemma segT_functional_eq_refl :
  forall (n:nat),
    segT_functional n n eq_refl = eq_refl.
    intro; apply proof_irrelevance.
Qed.

(*Wyckoff*)
(*This preserves values.*)
(*Analogous to [seg_from_S].*)
Definition segged_from_Sn {T n} (f: T _--> S n) :
  {x:T | proj1_sig (f x) <> n} _--> n :=
  fun x' =>
    let (x, pfx) := x' in
    let pffx := lt_S_neq _ _ (proj2_sig (f x)) pfx in
    exist (fun c => c < n) (proj1_sig (f x)) pffx.

(*Wyckoff*)
(*This decrements values.*)
(*Analogous to [seg_to_S]*)
Definition segged_from_S0 {T n} (f: T _--> S n) :
  {x:T | 0 < proj1_sig (f x)} _--> n :=
  fun x' =>
    let (x, pfx) := x' in
    let pffx := lt_S_pred' _ _ pfx (proj2_sig (f x)) in
    exist (fun c => c < n) (pred (proj1_sig (f x))) pffx.


  
(*Wyckoff*)
Lemma segged_from_Sn_compat' :
  forall {T n} (f: T _--> S n) (x:{t:T | proj1_sig (f t) <> n}),
    proj1_sig (segged_from_Sn f x) =
    proj1_sig (f (proj1_sig x)).
  unfold segged_from_Sn; simpl; intros ? ? ? x; destruct x; auto.
Qed.

(*Wyckoff*)
Lemma segged_from_Sn_compat :
  forall {T n} (f: T _--> S n) (x:{t:T | proj1_sig (f t) <> n}),
    let pflt := eq_lt_trans _ _ _
                            (eq_sym (segged_from_Sn_compat' f x))
                            (proj2_sig (segged_from_Sn f x)) in
    segged_from_Sn f x = exist (fun c => c < n)
                               (proj1_sig (f (proj1_sig x)))
                               pflt.
  unfold segged_from_Sn; intros ? ? ? x; destruct x; f_equal;
  apply proof_irrelevance.
Qed.

(*Wyckoff*)
Definition segged_from_S0_compat' :
  forall {T n} (f: T _--> S n) (x:{t:T | 0 < proj1_sig (f t)}),
    proj1_sig (segged_from_S0 f x) =
    pred (proj1_sig (f (proj1_sig x))).
  unfold segged_from_S0; simpl; intros ? ? ? x; destruct x; auto.
Qed.
  
(*Wyckoff*)
Lemma segged_from_S0_compat :
  forall {T n} (f: T _--> S n) (x:{t:T | 0 < proj1_sig (f t)}),
    let pflt := eq_lt_trans _ _ _
                            (eq_sym (segged_from_S0_compat' f x))
                            (proj2_sig (segged_from_S0 f x)) in
    segged_from_S0 f x = exist (fun c => c < n)
                               (pred (proj1_sig (f (proj1_sig x))))
                               pflt.
  unfold segged_from_S0; intros ? ? ? x; destruct x; f_equal;
  apply proof_irrelevance.
Qed.


Definition segged_cond_from_S0 {T n i}
           {f:T _--> S n}
           (pfp:forall x, proj1_sig (f x) <> S i) :
  forall x, (proj1_sig (segged_from_S0 f x)) <> i.
  intro x; destruct x as [x hx]; rewrite segged_from_S0_compat;
  simpl.
  apply S_inj_neq.
  rewrite <- (S_pred _ _ hx); auto.
Defined.


(*Wyckoff*)
(*This preserves values before [i], decrements after*)
(*Analogous to [remove_seg_fun] in [FunctionProperties2].*)
Fixpoint segged_from_i {T n} :
  forall (f:T _--> n) i (pfi:i < n)
         (pfp:forall x, proj1_sig (f x) <> i),
    T _--> pred n :=
  match n return (  forall (f:T _--> n) i (pfi:i < n)
         (pfp:forall x, proj1_sig (f x) <> i),
    T _--> pred n)
  with
    | 0 => fun _ _ pfi x => False_rect _ (lt_n_0 _ pfi)
    | S n' => 
      fun f =>
        let f' := segged_from_S0 f  in
        let f_ := segged_from_Sn f  in
        fun i =>
          match i with
            | 0 =>
              fun _ pfp x =>
                f' (exist (fun c => 0 < proj1_sig (f c))
                          x
                          (neq_lt0 _ (neq_sym _ _ (pfp x)))) 
            | S i' =>
              fun pfi pfp =>
                let pfi' := lt_S_pred _ _ pfi in
                let pf0 := Olt _ _ pfi' in
                let pf0' := lt_neq _ _ pf0 in
                let pfp' := segged_cond_from_S0 pfp in
                let f0 := segged_from_i f' i' pfi' pfp' (n := n') in
                fun x =>
                  match (zerop (proj1_sig (f x))) with
                    | left pfe =>
                      let pf0'' := eq_neq_trans _ _ _ pfe pf0' in
                      f_ (exist (fun c => proj1_sig (f c) <> n')
                                          x
                                          pf0'')
                    | right pflt =>
                      segT_S_pred_to (segT_S (f0 (exist (fun c => 0 < proj1_sig (f c)) x pflt))) pf0
                  end
          end
  end.


Lemma segged_from_i_lt :
  forall {T n} (f:T _--> n)
         i j
         (pfi: i < n)
         (pfij:j < i)
         (pfp:forall x, proj1_sig (f x) <> i)
         x,
    proj1_sig (f x) = j ->
    let pfj := lt_trans _ _ _ pfij pfi in
    proj1_sig (segged_from_i f i pfi pfp x) =
    proj1_sig (f x).
  intros T n. revert T.
  induction n as [|n ih]; simpl.
  intros; omega.
  intros T f i j hi hji h1 x h2.
  destruct i as [|i]; simpl; subst.
  omega.
  lr_match_goal'; simpl; auto. 
  specialize (ih _ (segged_from_S0 f) i (pred (proj1_sig (f x)))
                 (lt_S_n _ _ hi)).
  hfirst ih.
  apply lt_S_dec in hji.
  destruct hji; subst; auto.
  omega.
  omega. 
  specialize (ih hf).
  hfirst ih.
  intro y; destruct y as [y hy]. 
  rewrite segged_from_S0_compat; simpl.
  specialize (h1 y).
  omega.
  rewrite segT_S_compat.
  specialize (ih hf0 (exist (fun c => 0 < proj1_sig (f c)) _ hr)).
  hfirst ih. 
  rewrite segged_from_S0_compat; simpl; auto.
  specialize (ih hf1).
  simpl in ih. 
  erewrite f_equal_dep. 
  rewrite ih at 1.
  rewrite <- (S_pred _ _ hr); auto.
  apply proof_irrelevance.
Qed.


Lemma segged_from_i_ge :
  forall {T n} (f:T _--> n)
         i j
         (pfi: i < n)
         (pfij:i <= j)
         (pfp:forall x, proj1_sig (f x) <> i)
         x,
    proj1_sig (f x) = j ->
    proj1_sig (segged_from_i f i pfi pfp x) =
    pred (proj1_sig (f x)).
  intros T n; revert T.
  induction n as [|n ih]; simpl.
  intros; omega. 
  intros T f i.
  destruct i as [|i]; simpl.
  intros; auto.
  intro j.
  destruct j as [|j].
  intros hi h1 h2 x h4.  
  lr_match_goal'; simpl.
  rewrite h4; auto.
  rewrite segT_S_compat.
  omega.
  intros hi hij hp x h1.
  lr_match_goal'; simpl.
  rewrite hl in h1; discriminate.
  rewrite segT_S_compat.
  specialize (ih _ (segged_from_S0 f) i j (lt_S_n _ _ hi)
                 (ge_S_n _ _ hij)
                 (segged_cond_from_S0 hp)
                 (exist (fun c => 0 < proj1_sig (f c)) x hr)).
  hfirst ih.
  rewrite segged_from_S0_compat; simpl; omega.
  specialize (ih hf).
  erewrite f_equal_dep.
  rewrite ih at 1.
  rewrite segged_from_S0_compat; simpl.
  rewrite h1; simpl.
  destruct (zerop j) as [|h2]; subst. 
  rewrite segged_from_S0_compat in e, hij; simpl in e, hij.
  rewrite e in hij; simpl in hij.
  apply le_S_n in hij.
  apply le0 in hij; subst.
  pose proof (hp x). omega.
  rewrite <- (S_pred _ _ h2); auto.
  apply proof_irrelevance.
Qed.


(*Wyckoff*)
Lemma inj_segged_from_S0 : 
  forall {T n} (f: T _--> S n),
    Injective f ->
    Injective (segged_from_S0 f).
  unfold segged_from_S0; intros T n f h1; red in h1; red; intros x y h2;
  destruct x as [x hx], y as [y hy].
  apply exist_injective in h2.
  apply pred_inj0 in h2; auto.
  apply proj1_sig_injective in h2.
  apply h1 in h2; subst; f_equal; apply proof_irrelevance.
Qed.

(*Wyckoff*)
Lemma inj_segged_from_Sn : 
  forall {T n} (f: T _--> S n),
    Injective f ->
    Injective (segged_from_Sn f).
  unfold segged_from_S0; intros T n f h1; red in h1; red; intros x y h2;
  destruct x as [x hx], y as [y hy].
  apply exist_injective in h2.
  apply proj1_sig_injective in h2.
  apply h1 in h2; subst; f_equal; apply proof_irrelevance.
Qed.


(*Wyckoff*)
Inductive sparse_segged_fun {T n} (f:T _--> n) : Prop :=
| sparse_segged_fun_intro :
    Injective f ->
    forall i,
      i < n ->
      (forall x, proj1_sig (f x) <> i) ->
      sparse_segged_fun f.

Lemma sparse_segged_from_i : 
  forall {T n} (f:T _--> n) i (pfi:i < n)
         (pfn:forall x : T, proj1_sig (f x) <> i),
    sparse_segged_fun (segged_from_i f i pfi pfn) ->
    sparse_segged_fun f.
  intros T n; revert T.
  induction n as [|n ih].
  intros; omega. 
  intros T f i hi hp h1.

  inversion h1 as [h2 k h3 h4]; simpl in h3.
  econstructor.
  red; intros x y h5.
  specialize (h2 x y). 
  destruct (lt_le_dec (proj1_sig (f x)) i) as [h6|h6].
  hfirst h2.
  apply proj1_sig_injective.
  rewrite (segged_from_i_lt _ _ _ _ h6).
  pose proof h6 as h6'.
  rewrite (segged_from_i_lt _ _ _ _ h6').
  rewrite h5; auto.
  rewrite h5; auto.
  rewrite h5; auto.
  specialize (h2 hf); auto.
  hfirst h2.
  apply proj1_sig_injective.
  rewrite (segged_from_i_ge _ _ _ _ h6).
  pose proof h6 as h6'.
  rewrite (segged_from_i_ge _ _ _ _ h6').
  rewrite h5; auto.
  rewrite h5; auto.
  rewrite h5; auto.
  specialize (h2 hf); auto.
  apply hi.
  assumption.
Qed.


(*Wyckoff*)
Definition seg_fun_to_segged {U n} (f:seg_fun n U) :
  n ---> U :=
  fun i => f (proj1_sig i) (proj2_sig _).

(*Wyckoff*)
Definition seg_fun_type_to_ind {n} :
  seg_fun n Type -> 
  n ---> Type :=
fun f => seg_fun_to_segged f.


(*Wyckoff*)
Lemma seg_one_t_functional :
  forall (m n:nat),
    m = n ->
    seg_one_t m = seg_one_t n.
  intros; subst; auto.
Defined.

(*Wyckoff*)
Lemma seg_one_t_functional_eq_refl :
  forall (n:nat),
    seg_one_t_functional n n eq_refl = eq_refl.
    intro; apply proof_irrelevance.
Qed.



(*Wyckoff*)
Lemma nin_seg : 
  forall n:nat,
    ~In (seg n) n.
intros n h1.
destruct h1.
omega.
Qed.

(*Wyckoff*)
Lemma nin_seg_0 : 
  forall n:nat, 
    ~In (seg 0) n.
intros n h1.
destruct h1.
omega.
Qed.

(*Wyckoff*)
Lemma in_seg_Sn :
  forall n:nat,
    In (seg (S n)) 0.
  intro; constructor; auto with arith.
Qed.

(*Wyckoff*)
Lemma in_seg_n_S :
  forall m n:nat,
    In (seg n) m ->
    In (seg (S n)) (S m).
  intros m n h1.
  destruct h1 as [h1]. constructor.
  apply lt_n_S; auto.
Qed.



(*Wyckoff*)
Lemma seg_0_eq :
  seg 0 = Empty_set nat.
  apply Extensionality_Ensembles; red; split; auto with sets;
  red; intros n h1; apply nin_seg_0 in h1; contradiction.
Qed.


(*Wyckoff*)
Lemma seg_1_eq :
  seg 1 = Singleton 0.
  apply Extensionality_Ensembles; red; split; red; intros x h1;
  destruct h1 as [h1].
  assert (x = 0). omega. subst. constructor.
  constructor. omega.
Qed.

(*Wyckoff*)
Lemma seg_eq_empty :
  forall n,
    seg n = Empty_set _ ->
    n = 0.
  intros n h1; destruct n; auto.
  pose proof (lt_0_Sn n) as h2.
  apply in_seg in h2.
  rewrite h1 in h2; inversion h2.
Qed.


(*Wyckoff*)
Lemma seg_eq_singleton :
  forall (a n:nat),
    seg n = Singleton a ->
    n = 1 /\ a = 0.
  intros a n h1. 
  pose proof (In_singleton _ a) as h3.
  rewrite <- h1 in h3.
  destruct h3 as [h3].
  rewrite (S_pred _ _ h3) in h1.
  pose proof (lt_0_Sn  (pred n)) as h4.
  apply in_seg in h4.
  rewrite h1 in h4. 
  destruct h4; subst.
  pose proof (lt_n_Sn (pred n)) as h4.
  apply in_seg in h4.
  rewrite h1 in h4.
  inversion h4; subst.
  rewrite <- (S_pred _ _ h3); auto.
Qed.



(*Wyckoff*)
Lemma seg_one_0 : 
  seg_one 0 = Empty_set _.
apply Extensionality_Ensembles; red; split; red; intros x h1;
try contradiction; destruct h1; omega.
Qed.

(*Wyckoff*)
Lemma seg_one_t_0_false : 
  forall x:seg_one_t 0,
    False.
intro x.
destruct x; omega.
Qed.

(*Wyckoff*)
Lemma full_set_segT0 :
  Full_set (0 @) = Empty_set _.
  apply Extensionality_Ensembles; red; split; auto with sets.
  red; intros m h1.
  pose proof (segT_0_false m).
  contradiction.
Qed.



(*Wyckoff*)
Definition sot_1 : seg_one_t 1.
  assert (h1:1 <= 1 <= 1). auto with arith.
  refine (exist (fun m => 1 <= m <= 1) _ h1).
Defined.


(*Wyckoff*)
Lemma seg_incl_seg_weak : 
  forall (n:nat),
    Included (seg n) (seg_weak n).
intros n. red. intros m h1.
constructor. destruct h1. omega.
Qed.

(*Wyckoff*)
Lemma seg_one_incl_seg_weak : 
  forall n:nat,
    Included (seg_one n) (seg_weak n).
intros; red; intros m h1; constructor; destruct h1; omega.
Qed.

(*Wyckoff*)
Lemma seg_weak_O : seg_weak 0 = Singleton 0.
apply Extensionality_Ensembles.
red. split.
red. intros m h1.
destruct h1 as [h1]. assert (m = 0). omega. subst.
constructor.
red. intros m h1.
destruct h1; subst.
constructor. auto with arith.
Qed.

(*Wyckoff*)
Lemma seg_weak_Sn : 
  forall n:nat, 
    seg_weak (S n) = Add (seg_weak n) (S n).
intro n.
apply Extensionality_Ensembles.
red. split.
red. intros m h1.
destruct h1 as [h1].
destruct (le_lt_eq_dec _ _ h1) as [h2 | h3].
left. constructor. omega.
subst. right. constructor.
red. intros m h1.
destruct h1 as [m h1l | m h1r].
destruct h1l as [h1l]. constructor. omega.
destruct h1r; subst.
constructor. auto with arith.
Qed.

(*Wyckoff*)
Lemma seg_one_Sn : 
  forall n:nat, 
    seg_one (S n) = Add (seg_one n) (S n).
intro n.
apply Extensionality_Ensembles.
red. split.
red. intros m h1.
destruct h1 as [h1].
destruct h1 as [h1 h1'].
destruct (le_lt_eq_dec _ _ h1') as [h2 | h3].
left. constructor. omega.
subst. right. constructor.
red. intros m h1.
destruct h1 as [m h1l | m h1r].
destruct h1l as [h1l]. constructor. omega.
destruct h1r; subst.
constructor. auto with arith.
Qed.

(*Wyckoff*)
Lemma seg_O : seg O = Empty_set _.
apply Extensionality_Ensembles.
red; split; auto with sets.
red. intros m h1. destruct h1; omega.
Qed.

(*Wyckoff*)
Lemma seg_Sn : 
  forall n:nat, 
    seg (S n) = Add (seg n) n.
intro n.
apply Extensionality_Ensembles.
red. split.
red. intros m h1.
destruct h1 as [h1].
assert (h2: m <= n). omega.
destruct (le_lt_eq_dec _ _ h2) as [h3 | h4].
left. constructor. omega. 
subst. right. constructor.
red. intros m h1.
destruct h1 as [m h1l | m h1r].
destruct h1l as [h1l]. constructor. omega.
destruct h1r; subst.
constructor. auto with arith.
Qed.


(*Wyckoff*)
Lemma seg_eq_couple :
  forall (a b n:nat),
    seg n = Couple a b->
    n = 1 /\ a = b /\ a = 0 \/
                      (n = 2 /\ ((a = 0 /\ b = 1) \/
                                 (a = 1 /\ b = 0))).
  intros a b n h1.
  destruct (nat_eq_dec a b) as [|h2]; subst.
  left.
  rewrite couple_singleton in h1.
  apply seg_eq_singleton in h1.
  destruct h1; subst; auto.
  right.
  destruct (nat_eq_dec n 2) as [h3 | h3]; subst.
  pose proof (in_seg _ _ (lt_0_Sn 1)) as h3.
  rewrite h1 in h3. 
  inversion h3; subst.
  split; auto. 
  left. 
  pose proof (lt_n_Sn  1) as h4.
  apply in_seg in h4.
  rewrite  h1 in h4.
  inversion h4; subst; auto.
  pose proof (lt_n_Sn  1) as h4.
  apply in_seg in h4.
  rewrite h1 in h4.
  inversion h4; subst; auto.
  assert (h5:n <> 1).
    intro; subst. 
  rewrite seg_1_eq in h1.
  symmetry in h1.  
  apply couple_eq_sing in h1.
  destruct h1; subst.
  contradict h2; auto. 
  assert (h6: n <> 0).
    intro; subst.
  rewrite seg_0_eq in h1.
  symmetry in h1.  
  apply couple_not_empty in h1; contradiction.
  assert (h7:0 < n). omega.
  pose proof h7 as h8.
  apply in_seg in h7.
  rewrite h1 in h7.  
  inversion h7; subst.
  pose proof (lt_n_Sn (pred n)) as h9.
  rewrite <- (S_pred _ _ h8) in h9.
  apply in_seg in h9.
  rewrite  h1 in h9.
  inversion h9. 
  omega.
  subst. 
  rewrite (S_pred _ _ h8) in h1 at 1.
  rewrite  seg_Sn in h1.
  rewrite couple_add_sing in h1.
  apply add_nin_inj_eq in h1.
  apply seg_eq_singleton in h1.
  destruct h1 as [h1 h10].
  omega.
  apply nin_seg.
  rewrite in_sing_iff; auto.

  pose proof (lt_n_Sn (pred n)) as h9.
  rewrite <- (S_pred _ _ h8) in h9.
  apply in_seg in h9.
  rewrite  h1 in h9.
  inversion h9; subst.
  rewrite (S_pred _ _ h8) in h1 at 1.
  rewrite seg_Sn in h1.
  rewrite couple_comm in h1.
  rewrite couple_add_sing in h1.
  apply add_nin_inj_eq in h1.
  apply seg_eq_singleton in h1.
  destruct h1 as [h1 h10].
  omega.
  apply nin_seg.
  rewrite in_sing_iff; auto.
  omega.
Qed.


(*Wyckoff*)
Lemma finite_seg_weak : 
  forall n:nat, Finite (seg_weak n).
intro n.
induction n as [|n h1].
rewrite seg_weak_O.
apply Singleton_is_finite.
rewrite seg_weak_Sn.
constructor; auto.
intro h2.
destruct h2 as [h2]. omega.
Qed.

(*Wyckoff*)
Lemma finite_seg : 
  forall n:nat, Finite (seg n).
intros n.
pose proof (seg_incl_seg_weak n) as h1.
pose proof (finite_seg_weak n) as h2.
apply (Finite_downward_closed _ _ h2 _ h1).
Qed.

(*Wyckoff*)
Lemma finite_full_set_segT : 
  forall n:nat, Finite (Full_set (n @)).
  intro n.
  pose proof (finite_seg n) as h1.
  rewrite <- segT_compat in h1.
  rewrite finite_im_proj1_sig_iff in h1.
  assumption.
Qed.


(*Wyckoff*)
Lemma finite_seg_one : 
  forall n:nat, Finite (seg_one n).
intro n.
pose proof (seg_one_incl_seg_weak n) as h1.
pose proof (finite_seg_weak n) as h2.
apply (Finite_downward_closed _ _ h2 _ h1).
Qed.




(*Wyckoff*)
Lemma finite_full_set_seg_one_t :
  forall n:nat, Finite (Full_set (seg_one_t n)).
intro n.
pose proof (finite_seg_one n) as h1.
rewrite <- seg_one_t_compat in h1.
rewrite finite_im_proj1_sig_iff in h1.
assumption.
Qed.

(*Wyckoff*)
Lemma full_set_segT_S_eq :
  forall m:nat,
    Full_set ((S m) @) =
    Add (Im (Full_set (m @))
            (fun c =>
               (exist (fun c => c < S m) _
                      (lt_trans _ _ _ (proj2_sig c) (lt_n_Sn _)))))
        (exist (fun c => c < S m) _ (lt_n_Sn _)).
  intros m.
  apply Extensionality_Ensembles; red; split; red; intros c h1.
  clear h1.
  destruct c as [c h1].
  destruct (lt_S_dec _ _ h1) as [h2 | h2].
  left.
  apply Im_intro with (exist (fun c => c < m) _ h2). constructor.
  apply proj1_sig_injective; auto.
  subst.
  right.
  pose proof (proof_irrelevance _ h1 (lt_n_Sn m)); subst.
  constructor; auto.
  constructor.
Qed.


(*Wyckoff*)
Lemma inh_seg : forall (n:nat),
                      0 < n ->
                      Inhabited (seg n).
  intros n h1.
  apply Inhabited_intro with 0.
  constructor.
  assumption.
Qed.

(*Wyckoff*)
Lemma segT_pos : 
  forall (n:nat)  (m:seg_one_t n),
      0 <= proj1_sig m.
  intros ? m.
  destruct m; auto with arith.
Qed.

  
(*Wyckoff*)
Lemma inh_seg_weak :
  forall n:nat,
    Inhabited (seg_weak n).
  intro n.
  apply Inhabited_intro with 0.
  constructor; auto with arith.
Qed.


(*Wyckoff*)
Lemma inh_seg_one :
  forall (n:nat),
    0 < n ->
    Inhabited (seg_one n).
  intros n h1.
  apply Inhabited_intro with 1.
  constructor; auto with arith.
Qed.




(*Wyckoff*)
Lemma seg_one_t_pos : 
  forall (n:nat)  (m:seg_one_t n),
      1 <= proj1_sig m.
intros ? m.
destruct m as [m h1].
destruct h1; auto.
Qed.

(*Wyckoff*)
Lemma seg_one_decompose : 
  forall m n:nat,
    m < n ->
    seg_one n = Union (seg_one m) [r:nat | m < r <= n].
intros m n h1.
apply Extensionality_Ensembles; red; split; red; intros x h2.
destruct h2 as [h2]. destruct h2 as [h2 h3].
destruct (lt_eq_lt_dec x m) as [h4 | h4].
destruct h4 as [h4 | h4].
left. constructor. omega.
subst.
left. constructor. omega.
right. constructor. omega.

destruct h2 as [x h2 | x h2].
destruct h2 as [h2]. constructor.
omega.
destruct h2 as [h2]. constructor.
omega.
Qed.

(*Wyckoff*)
Lemma incl_lt_seg :
  forall m n:nat,
    m < n ->
    Included (seg m) (seg n).
  intros m n h1.
  red; intros x h2.
  destruct h2 as [h2].
  constructor.
  eapply lt_trans. apply h2. assumption.
Qed.


(*Wyckoff*)
Lemma incl_le_seg :
  forall m n:nat,
    m <= n ->
    Included (seg m) (seg n).
  intros m n h1.
  red; intros x h2.
  destruct h2 as [h2].
  constructor.
  eapply lt_le_trans. apply h2. assumption.
Qed.
 


(*Wyckoff*)
Lemma incl_le_seg_one :
  forall m n:nat,
    m <= n ->
    Included (seg_one m) (seg_one n).
intros m n h1.
red; intros x h2.
destruct h2 as [h2]. constructor. omega.
Qed.

(*Wyckoff*)
Lemma incl_lt_seg_one :
  forall m n:nat,
    m < n ->
    Included (seg_one m) (seg_one n).
intros m n h1.
red; intros x h2.
destruct h2 as [h2]. constructor. omega.
Qed.

(*Wyckoff*)
Lemma sot_pred_lt_length :
  forall {n:nat} (m:seg_one_t n),
      pred (proj1_sig m) < n.
  intros n m.
  destruct m as [m h1].
  destruct h1 as [h1 h2].
  simpl.
  assert (h3:0 < m). auto with arith.
  apply lt_S_n.
  rewrite <- (S_pred _ _ h3).
  auto with arith.
Qed.


(*Wyckoff*)
Definition st01 : 1 @ :=
  exist (fun a => a < 1) _ st01_pf.


(*Wyckoff*)
Definition st02 : 2 @ :=
  exist (fun a => a < 2) _ st02_pf.

(*Wyckoff*)
Definition st12 : 2 @ :=
  exist (fun a => a < 2) _ st12_pf.



(*Wyckoff*)
Definition sot11 : seg_one_t 1 :=
  exist (fun a => 1 <= a <= 1) _ sot11_pf.


(*Wyckoff*)
Definition sot12 : seg_one_t 2 :=
  exist (fun a => 1 <= a <= 2) _ sot12_pf.

(*Wyckoff*)
Definition sot22 : seg_one_t 2 :=
  exist (fun a => 1 <= a <= 2) _ sot22_pf.




(*Wyckoff*)
Lemma full_set_segT1 :
  Full_set (1 @) = Singleton st01.
  apply Extensionality_Ensembles; red; split; red; intros x h1;
  destruct x as [x h2]; subst; simpl; unfold st01.
  assert (x = 0). omega. subst.
  pose proof (proof_irrelevance _ h2 st01_pf); subst; constructor.
  constructor.
Qed.


(*Wyckoff*)
Lemma full_set_seg_one_t_1 :
  Full_set (seg_one_t 1) = Singleton sot11.
  apply Extensionality_Ensembles; red; split; red; intros x h1.
  destruct x as [x h2].
  assert (h4:x = 1). destruct h2; auto with arith.
  subst.
  unfold sot11.
  pose proof (proof_irrelevance _ sot11_pf h2).
  subst.
  constructor.
  constructor.
Qed.


(*Wyckoff*)
Lemma full_set_seg_one_t_2 :
  Full_set (seg_one_t 2) = Couple sot12 sot22.
  apply Extensionality_Ensembles; red; split; red; intros x h1.
  destruct x as [x h2].
  destruct h2 as [h2 h3].
  pose proof h2 as h2'.
  apply le_lt_eq_dec in h2'.
  destruct h2' as [h4 | h4].
  assert (h5:x = 2). auto with arith.
  subst.
  unfold sot22.
  pose proof (proof_irrelevance _ (sot22_pf) (conj h2 h3)) as h5.
  rewrite <- h5.
  right.
  subst.
  unfold sot12.
  pose proof (proof_irrelevance _ (sot12_pf) (conj h2 h3)) as h6.
  rewrite <- h6.
  left.
  constructor.
Qed.


(*Wyckoff*)
Lemma seg_one_t_2_dec :
  forall (n:seg_one_t 2),
    {proj1_sig n = 1} + {proj1_sig n = 2}.
  intro n.
  destruct n as [n h1]. destruct h1 as [h1 h2]; simpl.
  apply le_lt_eq_dec in h1.
  destruct h1. right. omega.
  subst. left; auto.
Qed.



(*Wyckoff*)
Lemma im_seg_one_t_decompose : 
  forall {T:Type} (m n:nat)
         (f:seg_one_t n -> T)
         (pf:m < n),
    let f' := 
        restriction_sig_prop _ _ (lt_prop_seg_one_t m n pf) f in
    let f'' :=
        restriction_sig_prop _ _ (lt_le_prop_seg_one_t m n pf) f in
    Im (Full_set (seg_one_t n)) f = 
    Union (Im (Full_set (seg_one_t m)) f')
          (Im (Full_set {r:nat | m < r <= n}) f'').
intros T m n f h1 f' f''.
apply Extensionality_Ensembles; red; split; red; intros x h2.
destruct h2 as [x h2]. subst. clear h2.
destruct x as [x h2].
destruct h2 as [h2 h2'].
destruct (le_lt_dec  x m) as [h3 | h3].
left.
apply Im_intro with (exist (fun s => 1 <= s <= m) _ (conj h2 h3)).
constructor.
simpl.
f_equal. apply proj1_sig_injective; auto.
right.
apply Im_intro with (exist (fun s => m < s  <= n) _ (conj h3 h2')).
constructor.
simpl.
f_equal.
apply proj1_sig_injective; auto.
destruct h2 as [x h2 | x h2].
destruct h2 as [x h2]. subst. clear h2.
destruct x as [x h2].
assert (h3: 1 <= x <= n). omega.
apply Im_intro with (exist (fun s=>1 <= s <= n)  _ h3).
constructor.
simpl. f_equal. apply proj1_sig_injective; auto.
destruct h2 as [x h2]. subst. clear h2.
destruct x as [x h2].
assert (h3:1 <= x <= n). omega.
apply Im_intro with (exist (fun s=> 1 <= s <= n) _ h3).
constructor.
simpl. f_equal. apply proj1_sig_injective; auto.
Qed.


(*Wyckoff*)
Lemma seg_one_t_S_dec :
  forall {T:Type} {l:list T} (x:seg_one_t (S (length l))),
    {proj1_sig x = 1} + {1 <= pred (proj1_sig x) <= length l}.
  intros T l x. 
  destruct x as [x h1]; simpl.
  destruct h1 as [h1 h2].
  apply le_lt_eq_dec in h1.
  destruct h1 as [h1 | h1].
  assert (h3:2 <= x <= S (length l)). omega.
  right. omega.
  subst.
  left.
  reflexivity.
Qed.

(*Wyckoff*)
Lemma card_fun1_seg_eq :
  forall (n:nat),
    card_fun1 (seg n) = n.
  intro n.
  induction n as [|n h1].
  rewrite seg_0_eq.
  apply card_fun1_empty.
  rewrite seg_Sn.
  rewrite card_add_nin'.
  rewrite h1; auto.
  apply finite_seg.
  intro h2.
  destruct h2 as [h2].
  apply lt_irrefl in h2.
  contradiction.
Qed.


(*Wyckoff*)
Lemma in_ens_S_seg :
  forall (m n:nat),
    m < n ->
    In (ens_S (seg n)) (S m).
  intros; apply in_ens_S; constructor; auto.
Qed.

(*Wyckoff*)
Lemma inj_on_seg1 :
  forall {T:Type} (f:nat->T),
    inj_on (seg 1) f.
  intros T f. red. intros x y h1 h2 h3.
  destruct h1, h2; omega.
Qed.

(*Wyckoff*)
Definition fun_to_sig_seg
           {U:Type}
           (f:nat->U) (n:nat)
: seg_fun n (sig_set (Im (seg n) f)) :=
  fun i (pf:i < n) =>
    exist _ _
          (Im_intro _ _ (seg n) f i (in_seg _ _ pf) _ eq_refl).


(*Wyckoff*)
Lemma fun_to_sig_seg_compat : 
  forall {U:Type}
         (f:nat->U) (n:nat)
         (i:nat) (pf:i < n),
    fun_to_sig_seg f n i pf =
    exist _ (f i) (Im_intro _ _ (seg n) f i (in_seg _ _ pf) _ eq_refl).
  unfold fun_to_sig_seg; intros; auto.
Qed.


(*Wyckoff*)
Definition seg_fun_to_ens
           {U:Type} {n:nat}
           (f:seg_fun n U) : fun_in_ens (seg n) U :=
  fun i (pf:Inn (seg n) i) =>
    f i (seg_in _ _ pf).



(*Wyckoff*)
Lemma seg_fun_to_ens_compat : 
  forall {U:Type} {n:nat}
         (f:seg_fun n U) (i:nat)
         (pf:Inn (seg n) i),
    seg_fun_to_ens f i pf = 
    f i (seg_in _ _ pf).
  unfold seg_fun_to_ens; intros; auto.
Qed.


End SegSet''.


Section CartProd.
(*Wyckoff*)
Definition cart_prod {T U:Type} (A:Ensemble T) (B:Ensemble U) :
  Ensemble (T*U) := [pr:T*U | 
                      In A (fst pr) /\ In B (snd pr)].

(*Wyckoff*)
Lemma in_cart_prod : forall {T U:Type} 
                            (A:Ensemble T) (B:Ensemble U)
                            (x:T) (y:U),
                       In A x -> In B y ->
                       In (cart_prod A B) (x, y).
intros A B x y h1 h2.
constructor.
simpl.
split; assumption.
Qed.

(*Wyckoff*)
Lemma in_cart_prod_comm : 
  forall {T U:Type} (A:Ensemble T) (B:Ensemble U) (pr:T*U),
    In (cart_prod A B) pr -> In (cart_prod B A) (snd pr, fst pr).
intros T U A B pr h1.
constructor.
simpl.
destruct h1 as [h1]. 
destruct h1; split; assumption.
Qed.
                     
(*Wyckoff*)
Lemma cart_prod_sing : 
  forall {T U:Type} (A:Ensemble T) (y:U), 
    cart_prod A (Singleton y) = 
    Im A (fun x:T => (x, y)).
  intros T U A y.
  apply Extensionality_Ensembles.
  red. split.
  (* <= *)
  red.
  intros pr h1.
  unfold cart_prod in h1.
  destruct h1 as [h1].
  destruct h1 as [h1 h2].
  apply  Im_intro with (fst pr). assumption.
  pose proof (Singleton_inv _ _ _ h2) as h3.
  subst.
  apply surjective_pairing.
  (* >= *)
  red.
  intros pr h1.
  destruct h1 as [x h2 pr].
  subst.
  unfold cart_prod.
  constructor. split. simpl. assumption.
  simpl. constructor.
Qed.

  
(*Wyckoff*)
Lemma cart_prod_sing' : 
  forall {T U:Type} (B:Ensemble U) (x:T), 
    cart_prod (Singleton x) B = 
    Im B (fun y:U => (x, y)).
  intros T U B x.
  apply Extensionality_Ensembles.
  red. split.
  (* <= *)
  red.
  intros pr h1.
  unfold cart_prod in h1.
  destruct h1 as [h1].
  destruct h1 as [h1 h2].
  apply  Im_intro with (snd pr). assumption.
  pose proof (Singleton_inv _ _ _ h1) as h3.
  subst.
  apply surjective_pairing.
  (* >= *)
  red.
  intros pr h1.
  destruct h1 as [z h2 pr].
  subst.
  unfold cart_prod.
  constructor. split. simpl. constructor.
  simpl. assumption.
Qed.

(*Wyckoff*)
Lemma cart_prod_sing_rev : 
  forall {T U:Type} (A:Ensemble T) (y:U), 
    A = Im (cart_prod A (Singleton y)) (fun pr:T*U => fst pr).
intros T U A y.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red.
intros x h1.
apply Im_intro with (x, y).
unfold cart_prod. constructor. simpl.
split; auto with sets. reflexivity.
(* >= *)
red. 
intros x h1.
destruct h1 as [pr h1]; subst.
destruct h1 as [h1].
destruct h1; assumption.
Qed.

(*Wyckoff*)
Lemma cart_prod_sing_rev' : 
  forall {T U:Type} (B:Ensemble U) (x:T), 
    B = Im (cart_prod (Singleton x) B) (fun pr:T*U => snd pr).
intros T U B x.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red.
intros y h1.
apply Im_intro with (x, y).
unfold cart_prod. constructor. simpl.
split; auto with sets. reflexivity.
(* >= *)
red. 
intros y h1.
destruct h1 as [pr h1]; subst.
destruct h1 as [h1].
destruct h1; assumption.
Qed.

(*Wyckoff*)
Lemma card_cart_prod_sing : 
  forall {T U:Type} (A:Ensemble T) (n:nat) (b:U),
    cardinal _ A n -> cardinal _ (cart_prod A (Singleton b)) n.
intros T U A n b h1.
rewrite cart_prod_sing.
pose proof (injective_preserves_cardinal).
pose proof (cardinal_finite _ _ _ h1) as h2.
pose proof (finite_image _ _ _ (fun x:T=>(x, b)) h2) as h3.
pose proof (finite_cardinal _ _ h3) as h4.
destruct h4 as [n' h4].
assert (h5:Image.injective _ _ (fun x:T=>(x, b))).
  red. intros x' y' h5. injection h5. auto.
pose proof (injective_preserves_cardinal _ _ _ _ _ h5 h1 _ h4) as h6. 
subst.
assumption.
Qed.





(*Wyckoff*)
Lemma cart_prod_sing_fin : forall {T U:Type} (A:Ensemble T) (y:U),
                             Finite A ->
                             Finite (cart_prod A (Singleton y)).
intros T U A y h1.
pose proof (cart_prod_sing A y) as h2.
rewrite h2.
apply finite_image.
assumption.
Qed.

(*Wyckoff*)
Lemma cart_prod_sing_fin' : forall {T U:Type} (B:Ensemble U) (x:T),
                             Finite B ->
                             Finite (cart_prod (Singleton x) B).
intros T U B x h1.
pose proof (cart_prod_sing' B x) as h2.
rewrite h2.
apply finite_image.
assumption.
Qed.

(*Wyckoff*)
Lemma cart_prod_sing_fin_rev : forall {T U:Type} (A:Ensemble T) (y:U),
                                 Finite (cart_prod A (Singleton y)) ->                            
                                 Finite A.
                            
intros T U A y h1.
pose proof (cart_prod_sing_rev A y) as h2.
rewrite h2.
apply finite_image.
assumption.
Qed.

(*Wyckoff*)
Lemma cart_prod_sing_fin_rev' : forall {T U:Type} (B:Ensemble U) (x:T),
                                 Finite (cart_prod (Singleton x) B) ->                            
                                 Finite B.
                            
intros T U B x h1.
pose proof (cart_prod_sing_rev' B x) as h2.
rewrite h2.
apply finite_image.
assumption.
Qed.


(*Wyckoff*)
Lemma card_cart_prod_sing' : 
  forall {T U:Type} (A:Ensemble T),
    Finite A -> forall (b:U),
    card_fun1 (cart_prod A (Singleton b)) = card_fun1 A.
intros T U A h1 b.
pose proof (finite_cardinal _ _ h1) as h2.
destruct h2 as [n h2].
pose proof (card_cart_prod_sing A n b h2) as h3.
pose proof (card_fun1_compat (cart_prod A (Singleton b))) as h4.
pose proof (card_fun1_compat A) as h5.
destruct h4 as [h4l h4r].
destruct h5 as [h5l h5r].
specialize (h5l h1).
pose proof (cart_prod_sing_fin _ b h1) as h6.
specialize (h4l h6).
pose proof (cardinal_is_functional _ _ _ h2 _ _ h5l (eq_refl)). subst.
pose proof (cardinal_is_functional _ _ _ h3 _ _ h4l (eq_refl)). symmetry.
assumption.
Qed.


(*Wyckoff*)
Lemma cart_prod_inc_add1 : 
  forall {T U:Type} (A:Ensemble T) (B:Ensemble U) (x:T),
    Included (cart_prod A B) (cart_prod (Add A x) B).
intros T U A B x.
red. intros pr h1.
destruct h1 as [h1].
destruct h1 as [h1l h1r].
constructor.
split; try assumption.
left. assumption.
Qed.

(*Wyckoff*)
Lemma cart_prod_inc_add2 : 
  forall {T U:Type} (A:Ensemble T) (B:Ensemble U) (y:U),
    Included (cart_prod A B) (cart_prod A (Add B y)).
intros T U A B y.
red. intros pr h1.
destruct h1 as [h1].
destruct h1 as [h1l h1r].
constructor.
split; try assumption.
left. assumption.
Qed.

(*Wyckoff*)
Lemma cart_prod_eq : 
  forall {T U:Type} (A:Ensemble T) (B:Ensemble U), 
    cart_prod A B =
    FamilyUnion 
      [S:Ensemble (T*U) | 
       exists y:U,
                In B y /\ S = cart_prod A (Singleton y)].
intros T U A B.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red.
intros pr h1.
destruct h1 as [h1l h1r].
destruct h1l as [h1ll h1lr].
assert (h2:In (cart_prod A (Singleton (snd pr))) pr).
  rewrite cart_prod_sing.
  apply Im_intro with (fst pr). assumption.
  apply surjective_pairing.
  apply family_union_intro with 
  (cart_prod A (Singleton (snd pr))).
  constructor.
  exists (snd pr). 
  split; [assumption | reflexivity]. assumption.
(* >= *)
red. intros pr h1.
destruct h1 as [S pr h1 h2].
destruct h1 as [h1].
destruct h1 as [y h1].
destruct h1 as [h1l h1r].
subst.
rewrite cart_prod_sing in h2.
destruct h2 as [x].
subst.
constructor; split; assumption.
Qed.

(*Wyckoff*)
Lemma cart_prod_eq' : 
  forall {T U:Type} (A:Ensemble T) (B:Ensemble U), 
    cart_prod A B =
    FamilyUnion 
      [S:Ensemble (T*U) | 
       exists x:T,
                In A x /\ S = cart_prod (Singleton x) B].

intros T U A B.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red.
intros pr h1.
destruct h1 as [h1l h1r].
destruct h1l as [h1ll h1lr].
assert (h2:In (cart_prod (Singleton (fst pr)) B) pr).
  rewrite cart_prod_sing'.
  apply Im_intro with (snd pr). assumption.
  apply surjective_pairing.
  apply family_union_intro with 
  (cart_prod (Singleton (fst pr)) B).
  constructor.
  exists (fst pr). 
  split; [assumption | reflexivity]. assumption.
(* >= *)
red. intros pr h1.
destruct h1 as [S pr h1 h2].
destruct h1 as [h1].
destruct h1 as [y h1].
destruct h1 as [h1l h1r].
subst.
rewrite cart_prod_sing' in h2.
destruct h2 as [x].
subst.
constructor; split; assumption.
Qed.

(*Wyckoff*)
Lemma cart_prod_empty : forall {T U:Type} (B:Ensemble U),
                          cart_prod (Empty_set T) B =
                          Empty_set _.
intros T U B.
apply Extensionality_Ensembles.
red. split.
red.
intros pr h1.
unfold cart_prod in h1.
destruct h1 as [h1].
destruct h1; contradiction.
auto with sets.
Qed.

(*Wyckoff*)
Lemma cart_prod_empty_empty : forall {T U:Type} (B:Ensemble U),
                          cart_prod (Empty_set T) B =
                          cart_prod (Empty_set _) (Empty_set _).
intros T U B.
rewrite cart_prod_empty.
rewrite <- (cart_prod_empty (Empty_set _)).
reflexivity.
Qed.

(*Wyckoff*)
Lemma cart_prod_empty' : forall {T U:Type} (A:Ensemble T),
                          cart_prod A (Empty_set U)=
                          Empty_set _.
intros T U A.
apply Extensionality_Ensembles.
red. split.
red.
intros pr h1.
unfold cart_prod in h1.
destruct h1 as [h1].
destruct h1; contradiction.
auto with sets.
Qed.

(*Wyckoff*)
Lemma cart_prod_empty_rev : 
  forall {T U:Type} (A:Ensemble T) (B:Ensemble U),
    cart_prod A B = Empty_set _ -> 
    (A = Empty_set _ \/ B = Empty_set _).
intros T U A B h1.
apply NNPP.
intro h2.
pose proof (not_or_and _ _ h2) as h3.
destruct h3 as [h3 h4].
pose proof (not_empty_Inhabited _ _ h3) as h5.
pose proof (not_empty_Inhabited _ _ h4) as h6.
destruct h5 as [a h5].
destruct h6 as [b h6].
assert (h7:In (cart_prod A B) (a, b)).
  constructor. simpl. split; assumption.
rewrite h1 in h7.
contradiction.
Qed.

(*Wyckoff*)
Lemma cart_prod_empty_empty' : forall {T U:Type} (A:Ensemble U),
                          cart_prod A (Empty_set U) =
                          cart_prod (Empty_set _) (Empty_set _).
intros T U B.
rewrite cart_prod_empty'.
rewrite <- (cart_prod_empty' (Empty_set _)).
reflexivity.
Qed.

(*Wyckoff*)
Lemma add_ex_family : 
  forall {T U:Type} (A:Ensemble U) (a:U)
         (f:U -> Ensemble T),
    [S:Ensemble T | exists u:U, Inn (Add A a) u /\ S = f u] =
    Add [S:Ensemble T | exists u:U, Inn A u /\ S = f u]
        (f a).
  intros T U A a f.
    apply Extensionality_Ensembles.
    red. split.
    (* <= *)
    red.
    intros S h6.
    destruct h6 as [h6].
    destruct h6 as [y h6].
    destruct h6 as [h6l h6r].
    destruct h6l as [y h6la | y h6lb].
    left.
    constructor. exists y. split; assumption.
    destruct h6lb.
    right. rewrite h6r.
    constructor.
    (* >= *)
    red.
    intros S h6.
    destruct h6 as [S h6 | S h7'].
    destruct h6 as [h6].
    destruct h6 as [y h6].
    destruct h6 as [h6a h6b].
    constructor.
    exists y.
    split. left. assumption. assumption.
    destruct h7'.
    constructor.
    exists a. split.
    right. constructor. reflexivity.
Qed.

(*Wyckoff*)
Lemma cart_prod_fin : 
  forall {T U:Type} (A:Ensemble T) (B:Ensemble U), 
    Finite A -> Finite B -> Finite (cart_prod A B).
  intros T U A B h1.
  induction h1.
    intros h2. rewrite cart_prod_empty.
    constructor.
  intro h2.
  rewrite cart_prod_eq.
  apply Finite_Finite_Union.
  intros S h3.
  destruct h3 as [h3].
  destruct h3 as [y h3].
  destruct h3 as [h3l h3r].
  rewrite h3r.
  apply cart_prod_sing_fin.
  apply Add_preserves_Finite. assumption.
  induction h2 as [|B h3 h4 z h5].
  gterm0.
  rename c into C.
  fold C.
  assert (h3:C = Empty_set _).
    unfold C.
    apply Extensionality_Ensembles.
    red. split.
    red.
    intros S h2. destruct h2 as [h2].
    destruct h2 as [y h2].
    destruct h2; contradiction.
    auto with sets.
    rewrite h3.
    constructor.
    pose proof (Add_preserves_Finite _ B z h3) as h0.
    specialize (IHh1 h0). 
    rewrite cart_prod_eq in IHh1.
    assert (hs : Included (FamilyUnion [S:Ensemble (T*U) |
                            exists y:U, In B y /\ 
                                        S = cart_prod A (Singleton y)]) (FamilyUnion
              [S : Ensemble (T * U)
              | exists y : U, In (Add B z) y /\ S = cart_prod A (Singleton y)])).
      red. intros pr h6.
      destruct h6 as [S pr h7].
      apply family_union_intro with S.
      destruct h7 as [h7].
      destruct h7 as [y h7].
      constructor. exists y.
      split.
      left. tauto. tauto. assumption.
  pose proof (Finite_downward_closed _ _ IHh1 _ hs) as h7.
  rewrite <- cart_prod_eq in h7.
  hfirst h4.
  tauto.
  specialize (h4 hf).
  pose proof (add_ex_family B z (fun u:U => cart_prod (Add A x)
                                                      (Singleton u))) as h6.
  simpl in h6.
  rewrite h6.
  constructor. assumption.
  intro h9.
  destruct h9 as [h9].
  destruct h9 as [a h9].
  destruct h9 as [h9a h9b].
  assert (h10:In (cart_prod (Add A x) (Singleton z)) (x, z)).
    constructor.
    simpl. split.
    right. constructor.
    constructor.
  rewrite h9b in h10.
  destruct h10 as [h10].
  destruct h10 as [h10a h10b].
  simpl in h10b.
  destruct h10b.
  contradiction.
Qed.

(*Wyckoff*)
Lemma cart_prod_fin_rev1 : 
  forall {T U:Type} (A:Ensemble T) (B:Ensemble U), 
    Finite (cart_prod A B) -> Inhabited B ->
    Finite A.
intros T U A B h1 h2.

dependent induction h1.
destruct h2 as [y h2].

assert (h3:A = Empty_set _).
  apply Extensionality_Ensembles.
  red. split.
  (* <= *)
  red.
  intros z h1.
  assert (h3:In (cart_prod A B) (z, y)). 
    constructor. simpl. split; assumption.
  rewrite <- x in h3.
  contradiction.
  (* >= *)
  auto with sets.
rewrite h3. constructor.
pose proof (Add_preserves_Finite _ _ x0 h1) as h3.
rewrite x in h3.
rewrite cart_prod_eq in h3.
pose proof (Finite_downward_closed
              _ _ h3
              (cart_prod A (Singleton (snd x0)))) as hf.
hfirst hf.
red.
  intros pr h4.
  apply family_union_intro with (cart_prod A (Singleton (snd x0))).
  constructor.
  exists (snd x0).
  split.
  pose proof (Add_intro2 _ A0 x0) as h5.
  rewrite x in h5.
  destruct h5 as [h5]. destruct h5; assumption.
  reflexivity.
  assumption.
specialize (hf hf0).
apply cart_prod_sing_fin_rev with (snd x0); auto.
Qed.

(*Wyckoff*)
Lemma cart_prod_fin_rev2 : 
  forall {T U:Type} (A:Ensemble T) (B:Ensemble U), 
    Finite (cart_prod A B) -> Inhabited A ->
    Finite B.
intros T U A B h1 h2.

dependent induction h1.
destruct h2 as [a h2].

assert (h3:B = Empty_set _).
  apply Extensionality_Ensembles.
  red. split.
  (* <= *)
  red.
  intros b h1.
  assert (h3:In (cart_prod A B) (a, b)). 
    constructor. simpl. split; assumption.
  rewrite <- x in h3.
  contradiction.
  (* >= *)
  auto with sets.
rewrite h3. constructor.
pose proof (Add_preserves_Finite _ _ x0 h1) as h3.
rewrite x in h3.
rewrite cart_prod_eq' in h3.
assert (h4:Included (cart_prod (Singleton (fst x0)) B)
(FamilyUnion
            [S : Ensemble (T * U)
            | exists x1 : T, In A x1 /\ S = cart_prod (Singleton x1) B])).
  red.
  intros pr h4.
  apply family_union_intro with (cart_prod (Singleton (fst x0)) B).
  constructor.
  exists (fst x0).
  split.
  pose proof (Add_intro2 _ A0 x0) as h5.
  rewrite x in h5.
  destruct h5 as [h5]. destruct h5; assumption.
  reflexivity.
  assumption.
pose proof (Finite_downward_closed _ _ h3 _ h4) as h5.
apply cart_prod_sing_fin_rev' with (fst x0).
assumption.
Qed.

(*Wyckoff*)
Lemma cart_prod_fin_rev : 
  forall {T U:Type} (A:Ensemble T) (B:Ensemble U), 
    Finite (cart_prod A B) -> Inhabited (cart_prod A B) ->
    Finite A /\ Finite B.
intros T U A B h1 h2.
destruct h2 as [pr h2].
destruct h2 as [h2].
destruct h2 as [h2l h2r].
pose proof (Inhabited_intro _ _ _ h2l) as h3.
pose proof (Inhabited_intro _ _ _ h2r) as h4.
split.
apply cart_prod_fin_rev1 with B; assumption.
apply cart_prod_fin_rev2 with A; assumption.
Qed.

(*Wyckoff*)
Lemma cart_prod_fin_rev_or : 
  forall {T U:Type} (A:Ensemble T) (B:Ensemble U),
    Finite (cart_prod A B) ->
    Finite A \/ Finite B.
intros T U A B h1.
destruct (classic (Inhabited A)) as [h2 | h3].
right.
apply cart_prod_fin_rev2 with A; assumption.
left.
assert (h4:A = Empty_set _).
  apply Extensionality_Ensembles;split. 
  red. intros x h4. 
  pose proof (Inhabited_intro _ A _ h4).
  contradiction. 
  auto with sets. 
rewrite h4.
constructor.
Qed.


(*Wyckoff*) 
Lemma cart_prod_proj1_surj : 
  forall {T U:Type} (A:Ensemble T) (B:Ensemble U),
    let f := (fun pr:T*U => fst pr) in
    Inhabited B -> Im (cart_prod A B) f = A.
intros T U A B f h1.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red.
intros y h2.
destruct h2 as [x h2].
subst.
destruct h2 as [h2].
destruct h2 as [h2l h2r].
unfold f. simpl.
assumption.
(* >= *)
red.
intros x h2.
unfold f.
destruct h1 as [y h1].
apply Im_intro with (x, y).
split; simpl; split; assumption.
simpl. reflexivity.
Qed.

(*Wyckoff*)
Lemma cart_prod_proj2_surj : 
  forall {T U:Type} (A:Ensemble T) (B:Ensemble U),
    let f := (fun pr:T*U => snd pr) in
    Inhabited A -> Im (cart_prod A B) f = B.
intros T U A B f h1.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red.
intros y h2.
destruct h2 as [x h2].
subst.
destruct h2 as [h2].
destruct h2 as [h2l h2r].
unfold f. simpl.
assumption.
(* >= *)
red.
intros y h2.
unfold f.
destruct h1 as [x h1].
apply Im_intro with (x, y).
split; simpl; split; assumption.
simpl. reflexivity.
Qed.


(*Wyckoff*)
Lemma cart_prod_im : 
  forall {T U V W:Type} 
         (A:Ensemble T) (B:Ensemble U) (f:T->V) (g:U->W),
         cart_prod (Im A f) (Im B g) = 
         Im (cart_prod A B) (fun pr => (f (fst pr), g (snd pr))).
intros T U V W A B f g.
apply Extensionality_Ensembles.

red. split.
(* <= *)
red.
intros pr h1.
inversion h1 as [h2]. clear h1.
destruct h2 as [h2l h2r].
inversion h2l as [x h3 v h4 h5]. clear h2l.
inversion h2r as [y h6 w h7 h8]. clear h2r.
rewrite (surjective_pairing pr).
subst. rewrite h4. rewrite h7.
apply Im_intro with (x, y). constructor; auto.
simpl. reflexivity.
(* >= *)
red. intros pr h1.
inversion h1 as [prx  h3 pry  h4 h5]. clear h1. 
subst.
rewrite (surjective_pairing prx) in h3.
destruct h3 as [h3].
simpl in h3.
destruct h3 as [h3l h3r].
constructor. simpl. split.
apply Im_intro with (fst prx).  assumption. reflexivity.
apply Im_intro with (snd prx).  assumption. reflexivity.
Qed.


(*Wyckoff*)
Lemma full_set_prod : 
  forall {T U:Type}, Full_set (T*U) = cart_prod (Full_set T) (Full_set U).
intros T U.
apply Extensionality_Ensembles.
red. split.
red. intros pr h1.
constructor. split; constructor.
red. intros pr h1. 
constructor.
Qed.


End CartProd.
 
Section Relations.
 
(*Wyckoff*)
Definition dom_rel {T U:Type} (S:Ensemble (T*U)) : Ensemble T :=
  [x:T | exists y:U, In S (x, y)].

(*Wyckoff*)
Definition ran_rel {T U:Type} (S:Ensemble (T*U)) : Ensemble U :=
  [y:U | exists x:T, In S (x, y)].

(*Wyckoff*)
Lemma dom_rel_eq : forall {T U:Type} (S:Ensemble (T*U)),
                     dom_rel S = Im S (@fst _ _).  
intros T U S.
apply Extensionality_Ensembles.
red. split. 
red. intros x h1.
destruct h1 as [h1].
destruct h1 as [y h1].
apply Im_intro with (x, y); auto. 
red. intros x h1.
destruct h1 as [pr h1 x]. subst.
constructor.
exists (snd pr). rewrite (surjective_pairing pr) in h1.
assumption.
Qed.

(*Wyckoff*)
Lemma ran_rel_eq : forall {T U:Type} (S:Ensemble (T*U)),
                     ran_rel S = Im S (@snd _ _).  
intros T U S.
apply Extensionality_Ensembles.
red. split. 
red. intros y h1.
destruct h1 as [h1].
destruct h1 as [x h1].
apply Im_intro with (x, y); auto. 
red. intros x h1.
destruct h1 as [pr h1 y]. subst.
constructor.
exists (fst pr). rewrite (surjective_pairing pr) in h1.
assumption.
Qed.

(*Wyckoff*)
Lemma dom_rel_empty : forall (T U:Type),
                    dom_rel (Empty_set (T*U)) = Empty_set _.
intros T U.
rewrite dom_rel_eq.
apply image_empty.
Qed.

(*Wyckoff*)
Lemma ran_rel_empty : forall (T U:Type),
                    ran_rel (Empty_set (T*U)) = Empty_set _.
intros T U.
rewrite ran_rel_eq.
apply image_empty.
Qed.

(*Wyckoff*)
Lemma dom_rel_singleton : 
  forall {T U:Type} (pr:T*U),
    dom_rel (Singleton pr) = Singleton (fst pr).
intros T U pr.
apply Extensionality_Ensembles.
red; split.
red; intros x h1.
destruct h1 as [h1].
destruct h1 as [y h1].
inversion h1. subst. clear h1.
simpl.
constructor.
red. intros x h1.
destruct h1.
constructor.
exists (snd pr).
rewrite <- surjective_pairing.
constructor.
Qed.

(*Wyckoff*)
Lemma ran_rel_singleton : 
  forall {T U:Type} (pr:T*U),
    ran_rel (Singleton pr) = Singleton (snd pr).
intros T U pr.
apply Extensionality_Ensembles.
red; split.
red; intros x h1.
destruct h1 as [h1].
destruct h1 as [y h1].
inversion h1. subst. clear h1.
simpl.
constructor.
red. intros x h1.
destruct h1.
constructor.
exists (fst pr).
rewrite <- surjective_pairing.
constructor.
Qed.

(*Wyckoff*)
Lemma dom_rel_couple : 
  forall {T U:Type} (pr pr':T*U),
    dom_rel (Couple pr pr') = Couple (fst pr) (fst pr').
intros T U pr pr'.
apply Extensionality_Ensembles.
red; split.
red; intros x h1.
destruct h1 as [h1].
destruct h1 as [y h1].
inversion h1. 
subst. clear h1.
simpl.
left.
subst. simpl.
right. 
red. intros x h1.
destruct h1.
constructor.
exists (snd pr).
rewrite <- surjective_pairing.
left.
constructor.
exists (snd pr').
rewrite <- surjective_pairing.
right.
Qed.


(*Wyckoff*)
Lemma ran_rel_couple : 
  forall {T U:Type} (pr pr':T*U),
    ran_rel (Couple pr pr') = Couple (snd pr) (snd pr').
intros T U pr pr'.
apply Extensionality_Ensembles.
red; split.
red; intros x h1.
destruct h1 as [h1].
destruct h1 as [y h1].
inversion h1. 
subst. clear h1.
simpl.
left.
subst. simpl.
right. 
red. intros x h1.
destruct h1.
constructor.
exists (fst pr).
rewrite <- surjective_pairing.
left.
constructor.
exists (fst pr').
rewrite <- surjective_pairing.
right.
Qed.


(*Wyckoff*)
Lemma dom_rel_finite : 
  forall {T U:Type} (S:Ensemble (T*U)),
    Finite S -> Finite (dom_rel S).
intros T U S h1.
rewrite dom_rel_eq.
apply finite_image; auto.
Qed.

(*Wyckoff*)
Lemma ran_rel_finite : 
  forall {T U:Type} (S:Ensemble (T*U)),
    Finite S -> Finite (ran_rel S).
intros T U S h1.
rewrite ran_rel_eq.
apply finite_image; auto.
Qed.

(*Wyckoff*)
Lemma dom_rel_add : 
  forall {T U:Type} (S:Ensemble (T*U)) (pr:T*U),
    dom_rel (Add S pr) = Add (dom_rel S) (fst pr).
intros T U S pr. unfold dom_rel.
apply Extensionality_Ensembles.
red. split.
red. intros x h1. 
destruct h1 as [h1].
destruct h1 as [y h1]. 
inversion h1 as [pr' h1l | pr' h1r]. subst.
left. constructor. exists y. assumption.
inversion h1r. subst. simpl.
right. constructor.
red. intros x h1.
destruct h1 as [x h1l | x h1r].
constructor.
destruct h1l as [h1l].
destruct h1l as [y h1l].
exists y. left. assumption.
inversion h1r. subst. constructor.
exists (snd pr). right. rewrite surjective_pairing at 1.
constructor.
Qed.

(*Wyckoff*)
Lemma ran_rel_add : 
  forall {T U:Type} (S:Ensemble (T*U)) (pr:T*U),
    ran_rel (Add S pr) = Add (ran_rel S) (snd pr).
intros T U S pr. unfold dom_rel.
apply Extensionality_Ensembles.
red. split.
red. intros y h1. 
destruct h1 as [h1].
destruct h1 as [x h1]. 
inversion h1 as [pr' h1l | pr' h1r]. subst.
left. constructor. exists x. assumption.
inversion h1r. subst. simpl.
right. constructor.
red. intros y h1.
destruct h1 as [y h1l | y h1r].
constructor.
destruct h1l as [h1l].
destruct h1l as [x h1l].
exists x. left. assumption.
inversion h1r. subst. constructor.
exists (fst pr). right. rewrite surjective_pairing at 1.
constructor.
Qed.


(*Wyckoff*)
Lemma dom_rel_incl : 
  forall {T U:Type} (S S':Ensemble (T*U)),
    Included S S' -> Included (dom_rel S) (dom_rel S').
intros T U S S' h1.
red.
intros x h2.
destruct h2 as [h2]. destruct h2 as [y h2].
constructor. exists y. auto with sets.
Qed.

(*Wyckoff*)
Lemma ran_rel_incl : 
  forall {T U:Type} (S S':Ensemble (T*U)),
    Included S S' -> Included (ran_rel S) (ran_rel S').
intros T U S S' h1.
red.
intros y h2.
destruct h2 as [h2]. destruct h2 as [x h2].
constructor. exists x. auto with sets.
Qed.

(*Wyckoff*)
Definition strict_included_exist_rel {T:Type} (A:Ensemble T) :
  {C:Ensemble T | Included C A} -> {C:Ensemble T | Included C A} -> Prop :=
  exist_rel (@Strict_Included T) (fun C => Included C A).


(*Wyckoff*)
Definition decreasing_or_zero {T:Type} (R:T->T->Prop) (f:T->T) : Prop :=
  exists z:T,
    f z = z /\
    (forall x:T, R (f x) x \/ x = z).

(*Wyckoff*)
Definition decreasing_or_zero' {T U:Type} (R:T->T->Prop) (f:T->T) (z:T) : Prop :=
  f z = z /\
  (forall x:T, R (f x) x \/ x = z).

(*Wyckoff*)
Definition decreasing_or_zero2 {T U:Type} (R:U->U->Prop) (f:T*U->T*U) (z:U) : Prop :=
  (forall x:T, snd (f (x, z)) = z )/\
  (forall pr:T*U, R (snd (f pr)) (snd pr) \/ snd (f pr) = z).


(*Wyckoff*)
(*Placed here for dependency issues.*)
Lemma iter_uo_decreasing_or_zero :
      forall {T U:Type} (R:U->U->Prop) (f:T*U->T*U) (z:U),
        decreasing_or_zero2 R f z ->
        forall (pr:T*U) (n:nat),
          let x' := snd (iter_uo f pr (S n)) in
          let x := snd (iter_uo f pr n) in
          R x' x \/
          x' = z.
  intros T U R f z h1 pr n. simpl.
  red in h1.
  destruct h1 as [h1 h2].
  apply h2; auto.
Qed.


(*Wyckoff*)
Definition rel_und_set {T:Type} (R:Relation T) :=
  [y:T | exists x:T, R x y].


(*Wyckoff*)
Definition foreset {T:Type} (R:Relation T) (y:T) :=
  [x:T | R x y].

(*Wyckoff*)
Definition afterset {T:Type} (R:Relation T) (x:T) :=
  [y:T | R x y].


(*Wyckoff*)
Definition foreset' {T:Type} (R:Relation T) (S:Ensemble T) (y:T) :=
  [x:T | R x y /\ In S y].

(*Wyckoff*)
Definition afterset' {T:Type} (R:Relation T) (S:Ensemble T) (x:T) :=
  [y:T | R x y /\ In S x].



(*Wyckoff*)
Definition ex_max_set {T:Type} (R:Relation T) (S:Ensemble T) : Prop :=
  exists x:T, 
    In S x /\ forall w:T, In S w -> R w x \/ w = x.

(*Wyckoff*)
Definition ex_unq_max_set {T:Type} (R:Relation T) (S:Ensemble T) : Prop :=
  exists! x:T, 
    In S x /\ forall w:T, In S w -> R w x \/ w = x.
           




(*Wyckoff*)
Lemma min_set_sortable_ex :
  forall {T:Type} (S:Ensemble T),
    Finite S -> Inhabited S ->
    forall (R:T->T->Prop),
      sortable R ->
      exists! m:T,
        In S m /\
        (forall x:T,
           In (Subtract S m) x ->
           R m x).
  intros T S h1.
  induction h1 as [|S h2 h3 a h4].
  intro h1.
  destruct h1; contradiction.
  intros h1 R h60.
  inversion h60 as [h6 h6' h6''].
  destruct (eq_dec S (Empty_set _)) as [h5 | h5]; subst.
  exists a.
  red. split. split. right; auto. constructor.
  intros x h7.
  rewrite add_empty_sing in h7.
  rewrite subtract_sing in h7. contradiction.
  intros x h7. destruct h7 as [h7 h8].
  rewrite add_empty_sing in h7.
  destruct h7; auto.
  apply not_empty_Inhabited in h5.
  specialize (h3 h5 R h60).
  destruct h3 as [m h3]. red in h3.
  destruct h3 as [h3a h3b].
  red in h6. 
  specialize (h6 a m).
  destruct h6 as [h6 | h6].
  exists a.
  red; split. split. right; auto. constructor.
  intros x h7.
  rewrite sub_add_same_nin in h7; auto.
  destruct h3a as [h3a h3c].
  destruct (eq_dec x m) as [h8 | h8]. subst; auto.
  assert (h9:In (Subtract S m) x).
    constructor; auto. intro h9. destruct h9. contradict h8; auto. 
  specialize (h3c _ h9).
  eapply h6'.  apply h6. assumption.
  intros x h7. 
  destruct h7 as [h7 h8].  
  destruct h7 as [x h7 | x h9]. 
  assert (h9: In S x /\ (forall x0 : T, In (Subtract S x) x0 -> R x x0)).
    split; auto.
    intros x' h9.
    assert (h10:In (Subtract (Add S a) x) x').
      destruct h9 as [h9 h10].
      constructor. left; auto. auto.
    apply h8; auto.
  specialize (h3b _ h9). subst.
  specialize (h8 a).
  assert (h10:  In (Subtract (Add S a) x) a).
    constructor. right. constructor. intro h10. destruct h10.
    apply h6'' in h6.
    contradiction.
  specialize (h8 h10).
  red in h6'. specialize (h6' _ _ _ h8 h6).
  apply h6'' in h6'.
  contradiction.
  destruct h9; auto.
  destruct h3a as [h3a h3c].
  exists m.
  red; split. split. left; auto.
  intros x h9.
  destruct h9 as [h9 h10].
  destruct h9 as [x h9 | x h11].
  apply h3c. constructor; auto.
  destruct h11.
  destruct h6 as [h6 | h6]. subst.
  contradict h10; constructor; auto.
  assumption.
  intros x' h9.
  destruct h9 as [h9 h10].
  destruct h9 as [x' h9 | x' h9].
  assert (h11:In S x' /\ (forall x : T, In (Subtract S x') x -> R x' x)).
    split; auto.
    intros x h11.
    destruct h11 as [h11 h12].
    apply h10. constructor. left; auto. assumption. 
  apply h3b; auto.
  destruct h9.
  destruct h6 as [h6 | h11]; subst; auto.
  assert (h12:In (Subtract (Add S a) a) m).
    rewrite sub_add_same_nin; auto.
  specialize (h10 _ h12).
  specialize (h6' _ _ _ h10 h11).
  apply h6'' in h6'.
  contradiction.
Qed.

(*Wyckoff*)
Definition min_set_sortable
           {T:Type} (S:Ensemble T) (pff:Finite S) (pfi: Inhabited S)
           (R:T->T->Prop) (pfso:sortable R) := 
  proj1_sig (constructive_definite_description _ (min_set_sortable_ex S pff pfi R pfso)).

(*Wyckoff*)
Lemma min_set_sortable_compat :
  forall {T:Type} (S:Ensemble T) (pff:Finite S) (pfi: Inhabited S)
         (R:T->T->Prop) (pfso: sortable R),
    let m := min_set_sortable S pff pfi R pfso in
    In S m /\
    (forall x:T,
       In (Subtract S m) x ->
       R m x).
  intros.
  unfold m, min_set_sortable.
  destruct constructive_definite_description; simpl; auto.
Qed.

(*Wyckoff*)
Lemma min_set_sortable_functional :
  forall {T:Type} (S S':Ensemble T) (pff:Finite S) (pff':Finite S')
         (pfi: Inhabited S) (pfi':Inhabited S')
         (R:T->T->Prop) (pfso pfso':sortable R),
    S = S' ->
    min_set_sortable S pff pfi R pfso = 
    min_set_sortable S' pff' pfi' R pfso'.
  intros T S S' h1 h2 h3 h4 R h5 h5' h6.
  subst.
  pose proof (proof_irrelevance _ h1 h2).
  pose proof (proof_irrelevance _ h3 h4).
  subst.
  apply f_equal_prop_dep; auto.
Qed.

Lemma min_set_sortable_sing :
  forall {T:Type} (a:T) (pff:Finite (Singleton a))
         (pfi:Inhabited (Singleton a)) (R:T->T->Prop)
         (pfso:sortable R),
    min_set_sortable (Singleton a) pff pfi R pfso = a.
  intros T a h1 h2 R h3.
  pose proof (min_set_sortable_compat _ h1 h2 R h3) as h6.
  simpl in h6.
  destruct h6 as [h6].
  inversion h6; auto.
Qed.


Lemma min_set_sortable_add :
  forall {T:Type} (A:Ensemble T) (a:T) (pff:Finite (Add A a))
         (pfi:Inhabited (Add A a)) (R:T->T->Prop)
         (pfso:sortable R),
    let (pftso, _, _) := pfso in
    let pftso' := tso_dec_compat _  pftso in
    (forall (pfi':Inhabited A),
       let pff' :=
           Finite_downward_closed _ _ pff _  (incl_add A a) in
       min_set_sortable (Add A a) pff pfi R pfso = 
       minR R pftso' (min_set_sortable A pff' pfi' R pfso)
               a) /\
    (A = Empty_set T ->
     min_set_sortable (Add A a) pff pfi R pfso  =
     a).
  intros T A a h1 h2 R h3.
  destruct h3 as [h3 h4 h5].
  split.
  intro h6. simpl.
  pose proof (min_set_sortable_compat  A (Finite_downward_closed T (Add A a) h1 A (incl_add A a)) h6 R (sort_intro _ h3 h4 h5)) as h7. simpl in h7.
  destruct h7 as [h7 h8]. 
  unfold min_set_sortable.
  destruct constructive_definite_description as [m h10]. simpl.
  destruct constructive_definite_description as [m' h11]. simpl.
  destruct h10 as [h10 h12]. destruct h11 as [h11 h13]. 
  destruct h10 as [m h10|m h10]; subst.
  unfold minR.    
  destruct (tso_dec_compat R h3 m' a) as [h14 | h14]. 
  red in h3.
  pose proof (h3 m m') as h3'.
  destruct h3' as [h3' | h3'].
  apply NNPP.
  intro h15. 
  specialize (h13 m).
  assert (h16:In (Subtract A m') m).
    constructor; auto.
    intro h16. destruct h16.
    red in h5. pose proof (h5 m'). contradiction.
  specialize (h13 h16).
  red in h4. pose proof (h4 m m' m h3' h13) as h4'.
  red in h5. pose proof (h5 m) as h5'. contradiction.
  destruct h3' as [|h3']; auto.
  specialize (h12 m').
  assert (h16:In (Subtract (Add A a) m) m').
    constructor. left; auto.
    intro h16. destruct h16.
    red in h5. pose proof (h5 m). contradiction.
    specialize (h12 h16).
  red in h4.
  pose proof (h4 _ _ _ h12 h3').
  red in h5. pose proof (h5 m). contradiction.
  red in h3.
  pose proof (h3 m a) as h3'.
  destruct h3' as [h3' | h3'].
  pose proof (h13 m) as h13'.
  red in h4. 
  pose proof (h4 _ _ _ h3' h14) as h15.
  hfirst h13'.
  constructor; auto.
  rewrite in_sing_iff.
  intro; subst.
  apply h5 in h15; auto.
  specialize (h13' hf).
  pose proof (h4 _ _ _ h13' h15) as h16.
  apply h5 in h16; contradiction.
  destruct h3' as [|h3']; auto.
  specialize (h12 a).
  hfirst h12.
  constructor. right; auto. constructor; auto.
  rewrite in_sing_iff.
  intro; subst.
  apply h5 in h3'; auto.
  specialize (h12 hf).
  pose proof (h4 _ _ _ h12 h3') as h15.
  apply h5 in h15; contradiction.
  destruct h10.
  unfold minR. 
  destruct (tso_dec_compat R h3 m' a) as [h14 | h14].
  destruct h14 as [h14|]; subst.
  specialize (h12 m').
  assert (h15: In (Subtract (Add A a) a) m').
    constructor. left; auto.
    intro h16. destruct h16.
    pose proof (h5 a). contradiction.
  specialize (h12 h15).
  pose proof (h4 _ _ _ h12 h14).
  pose proof (h5 a). contradiction.
  reflexivity.
  reflexivity.
  intro. subst.
  pose proof (add_empty_sing a) as h6.
  erewrite min_set_sortable_functional.
  apply min_set_sortable_sing.
  assumption.
  Grab Existential Variables.
  eapply sort_intro; auto.
  apply inh_sing.
  apply Singleton_is_finite.
Qed.

(*Wyckoff*)
Lemma min_set_sortable_couple :
  forall {T:Type} (a b:T)
         (pff:Finite (Couple a b)) (pfi:Inhabited (Couple a b))
         (R:T->T->Prop) (pfso:sortable R),
    R a b ->
    min_set_sortable (Couple a b) pff pfi R pfso = a.
  intros T a b hf hi R h1 h4.
  destruct h1 as [h1 h2 h3].
  pose proof (couple_add_sing a b) as h5.
  assert (h6:Finite (Add (Singleton a) b)).
    rewrite <- h5. apply finite_couple.
    pose proof (min_set_sortable_functional
                  _ _ hf h6 hi (inh_add _ _) R
                  (sort_intro _ h1 h2 h3)
                  (sort_intro _ h1 h2 h3) h5) as h7.
  rewrite h7.
  pose proof (min_set_sortable_add
                (Singleton a) b h6
                (inh_add (Singleton a) b) R (sort_intro _ h1 h2 h3))
    as h8.
  destruct h8 as [h8 h9].
  specialize (h8 (inh_sing _)). simpl in h8.
  rewrite h8.
  unfold minR.
  destruct (tso_dec_compat R h1  (min_set_sortable (Singleton a)
            (Finite_downward_closed T (Add (Singleton a) b) h6 
               (Singleton a) (incl_add (Singleton a) b)) 
            (inh_sing a) R (sort_intro _ h1 h2 h3))) as [h10 | h10].
  rewrite min_set_sortable_sing.
  reflexivity.
  rewrite min_set_sortable_sing in h10.
  pose proof (h2 _ _ _ h4 h10) as h11.
  apply h3 in h11.
  contradiction.
Qed.


(*Wyckoff*)
Lemma min_set_sortable_couple' :
  forall {T:Type} (a b:T)
         (pff:Finite (Couple a b)) (pfi:Inhabited (Couple a b))
         (R:T->T->Prop) (pfso:sortable R),
    let (pftso, _, _) := pfso in
    let pftso' := tso_dec_compat _ pftso in
    min_set_sortable (Couple a b) pff pfi R pfso = 
    minR R pftso' a b.
  intros T a b h1 h2 R h3.
  destruct h3 as [h3 h4 h5].
  unfold minR.
  destruct (tso_dec_compat R h3 a b) as [h6 | h6].
  destruct h6 as [h6|]; subst.
  apply min_set_sortable_couple; auto.
  pose proof (couple_singleton b) as h7.
  pose proof (min_set_sortable_functional
                _ _ h1
                (Singleton_is_finite _ _)
                h2 (inh_sing _) R
                (sort_intro _ h3 h4 h5) (sort_intro _ h3 h4 h5) h7) as h8.
  rewrite h8.
  rewrite min_set_sortable_sing; auto.
  pose proof (couple_comm a b) as h7.
  pose proof (min_set_sortable_functional
                _ _ h1 (finite_couple _ _)
                h2 (inh_couple _ _) R
                (sort_intro _ h3 h4 h5) (sort_intro _ h3 h4 h5) h7) as h8.
  rewrite h8.
  apply min_set_sortable_couple; auto.
Qed.

(*Wyckoff*)
Lemma max_set_sortable_ex :
  forall {T:Type} (S:Ensemble T),
    Finite S -> Inhabited S ->
    forall (R:T->T->Prop),
      sortable R ->
      exists! m:T,
        In S m /\
        (forall x:T,
           In (Subtract S m) x ->
           R x m).
  intros T S h1.
  induction h1 as [|S h2 h3 a h4].
  intro h1.
  destruct h1; contradiction.
  intros h1 R h60.
  inversion h60 as [h6 h6' h6''].
  destruct (eq_dec S (Empty_set _)) as [h5 | h5]; subst.
  exists a.
  red. split. split. right; auto. constructor.
  intros x h7.
  rewrite add_empty_sing in h7.
  rewrite subtract_sing in h7. contradiction.
  intros x h7. destruct h7 as [h7 h8].
  rewrite add_empty_sing in h7.
  destruct h7; auto.
  apply not_empty_Inhabited in h5.
  specialize (h3 h5 R h60).
  destruct h3 as [m h3]. red in h3.
  destruct h3 as [h3a h3b].
  red in h6.
  specialize (h6 m a).
  destruct h6 as [h6 | h6].
  exists a.
  red; split. split. right; auto. constructor.
  intros x h7.
  rewrite sub_add_same_nin in h7; auto.
  destruct h3a as [h3a h3c].
  destruct (eq_dec x m) as [h8 | h8]. subst; auto.
  assert (h9:In (Subtract S m) x).
    constructor; auto. intro h9. destruct h9. contradict h8; auto. 
  specialize (h3c _ h9).
  eapply h6'. apply h3c. assumption.  
  intros x h7.
  destruct h7 as [h7 h8].
  destruct h7 as [x h7 | x h9].
  specialize (h3b x).
  hfirst h3b.
    split; auto.
    intros x' h9.
    specialize (h8 x').
    hfirst h8.
    destruct h9 as [h9 h10].
   constructor. left; auto. auto.
    apply h8; auto.
  specialize (h3b hf). subst.
  specialize (h8 a).
  hfirst h8.
    constructor. right. constructor. intro h10. destruct h10.
    apply h6'' in h6.
    contradiction.
  specialize (h8 hf0).
  red in h6'. specialize (h6' _ _ _ h8 h6).
  apply h6'' in h6'.
  contradiction.
  destruct h9; auto.
  destruct h3a as [h3a h3c].
  exists m.
  red; split. split. left; auto.
  intros x h9.
  destruct h9 as [h9 h10].
  destruct h9 as [x h9 | x h11].
  apply h3c. constructor; auto.
  destruct h11.
  destruct h6 as [h6 | h6]. subst.
  contradict h10; constructor; auto.
  assumption.
  intros x' h9.
  destruct h9 as [h9 h10].
  destruct h9 as [x' h9 | x' h9].
  specialize (h3b x').
  hfirst h3b.
    split; auto.
    intros x h11.
    destruct h11 as [h11 h12].
    apply h10. constructor. left; auto. assumption.
  apply h3b; auto.
  destruct h9.
  destruct h6 as [h6 | h11]; subst; auto.
  specialize (h10 m).
  hfirst h10.
    rewrite sub_add_same_nin; auto.
  specialize (h10 hf).
  specialize (h6' _ _ _ h10 h11).
  apply h6'' in h6'.
  contradiction.
Qed.

(*Wyckoff*)
Definition max_set_sortable
           {T:Type} (S:Ensemble T) (pff:Finite S) (pfi: Inhabited S)
           (R:T->T->Prop) (pfso:sortable R) :=
  proj1_sig (constructive_definite_description
               _ (max_set_sortable_ex S pff pfi R pfso)).

(*Wyckoff*)
Lemma max_set_sortable_compat :
  forall {T:Type} (S:Ensemble T) (pff:Finite S) (pfi: Inhabited S)
         (R:T->T->Prop) (pfso: sortable R),
    let m := max_set_sortable S pff pfi R pfso in
    In S m /\
    (forall x:T,
       In (Subtract S m) x ->
       R x m).
  intros.
  unfold m, max_set_sortable.
  destruct constructive_definite_description; simpl; auto.
Qed.

(*Wyckoff*)
Lemma max_set_sortable_functional :
  forall {T:Type} (S S':Ensemble T) (pff:Finite S) (pff':Finite S')
         (pfi: Inhabited S) (pfi':Inhabited S')
         (R:T->T->Prop) (pfso pfso':sortable R),
    S = S' ->
    max_set_sortable S pff pfi R pfso =
    max_set_sortable S' pff' pfi' R pfso'.
  intros T S S' h1 h2 h3 h4 R h5 h5' h6.
  subst.
  pose proof (proof_irrelevance _ h1 h2).
  pose proof (proof_irrelevance _ h3 h4).
  subst.
  apply f_equal_prop_dep; auto.
Qed.

(*Wyckoff*)
Lemma max_set_sortable_sing :
  forall {T:Type} (a:T) (pff:Finite (Singleton a))
         (pfi:Inhabited (Singleton a)) (R:T->T->Prop)
         (pfso:sortable R),
    max_set_sortable (Singleton a) pff pfi R pfso = a.
  intros T a h1 h2 R h3.
  pose proof (max_set_sortable_compat _ h1 h2 R h3) as h6.
  simpl in h6.
  destruct h6 as [h6].
  inversion h6; auto.
Qed.

(*Wyckoff*)
Lemma max_set_sortable_add :
  forall {T:Type} (A:Ensemble T) (a:T) (pff:Finite (Add A a))
         (pfi:Inhabited (Add A a)) (R:T->T->Prop)
         (pfso:sortable R),
    let (pftso, _, _) := pfso in
    let pftso' := tso_dec_compat _  pftso in
    (forall (pfi':Inhabited A),
       let pff' :=
           Finite_downward_closed _ _ pff _  (incl_add A a) in
       max_set_sortable (Add A a) pff pfi R pfso =
       maxR R pftso' (max_set_sortable A pff' pfi' R pfso)
               a) /\
    (A = Empty_set T ->
     max_set_sortable (Add A a) pff pfi R pfso  =
     a).
  intros T A a h1 h2 R h3.
  simpl.
  destruct h3 as [h3 h4 h5].
  split.
  intro h6.
  unfold max_set_sortable, maxR.
  destruct constructive_definite_description as [m h7]; simpl.
  destruct constructive_definite_description as [m' h8]; simpl.
  destruct h7 as [h7 h9], h8 as [h8 h10].
  lr_if_goal; fold hlr; destruct hlr as [hl | hr].
  destruct hl as [hl | hl].
  destruct h7 as [m h7 | m h7]; subst. 
  pose proof (h10 m) as h10'.
  hfirst h10'.
  constructor; auto.
  intro h11; destruct h11.
  pose proof (h9 a) as h9'.
  hfirst h9'.
  constructor. right; constructor; auto.
  intro h11. destruct h11.
  apply h5 in hl; contradiction.
  specialize (h9' hf).
  pose proof (h4 _ _ _ hl h9') as h11.
  apply h5 in h11; contradiction.
  specialize (h10' hf).
  pose proof (h4 _ _ _ h10' hl) as h11.
  pose proof (h9 a) as h9'.
  hfirst h9'.
  constructor. right; constructor; auto.
  intro h12; destruct h12.
  apply h5 in h11; contradiction.
  specialize (h9' hf0).
  pose proof (h4 _ _ _ h11 h9') as h12.
  apply h5 in h12; contradiction.
  destruct h7; auto.
  subst.
  destruct h7 as [m h7 | m h7].
  pose proof (h3 m a) as h3'.
  destruct h3' as [h3' | h3']; auto.
  specialize (h9 a). 
  hfirst h9. 
  constructor. right; constructor; auto.
  intro h11; destruct h11.
  apply h5 in h3'; contradiction.
  specialize (h9 hf).
  pose proof (h4 _ _ _ h9 h3') as h11.
  apply h5 in h11; contradiction.
  destruct h3' as [|h3']; auto.
  specialize (h10 m).
  hfirst h10.
  constructor; auto.
  intro h11; destruct h11.
  apply h5 in h3'; contradiction.
  specialize (h10 hf).
  pose proof (h4 _ _ _ h10 h3') as h11. apply h5 in h11; contradiction.
  destruct h7; auto.
  destruct h7 as [m h7 | m h7]; subst.
  pose proof (h3 m m') as h11.
  destruct h11 as [h11 | h11].
  specialize (h9 m').
  hfirst h9.
  constructor. left; auto.
  intro h12; destruct h12.
  apply h5 in h11; auto.
  specialize (h9 hf).
  pose proof (h4 _ _ _ h9 h11) as h12.
  apply h5 in h12; contradiction.
  destruct h11 as [|h11]; auto.
  specialize (h10 m).
  hfirst h10.
  constructor; auto.
  intro h12. destruct h12.
  apply h5 in h11; contradiction.
  specialize (h10 hf).
  pose proof (h4 _ _ _ h10 h11) as h12.
  apply h5  in h12; contradiction.
  destruct h7.
  specialize (h9 m').
  hfirst h9.
  constructor.
  left; auto.
  intro h11; destruct h11.
  apply h5 in hr; contradiction.
  specialize (h9 hf).
  pose proof (h4 _ _ _ h9 hr) as h11.
  apply h5 in h11; contradiction.
  intro; subst.
  generalize h1, h2.
  rewrite add_empty_sing.
  intros h1' h2'.
  rewrite max_set_sortable_sing; auto.
Qed.


(*Wyckoff*)
Lemma max_set_sortable_couple :
  forall {T:Type} (a b:T)
         (pff:Finite (Couple a b)) (pfi:Inhabited (Couple a b))
         (R:T->T->Prop) (pfso:sortable R),
    R a b ->
    max_set_sortable (Couple a b) pff pfi R pfso = b.
  intros T a b hf hi R h1 h4.
  generalize hf hi.
  rewrite couple_add_sing.
  intros hf' hi'.
  pose proof (max_set_sortable_add (Singleton a) b hf' hi' R h1) as h8.
  simpl in h8.
  destruct h1 as [h1 h2 h3].
  destruct h8 as [h8 h9].
  specialize (h8 (inh_sing _)). simpl in h8.
  rewrite h8 at 1.
  unfold maxR. unfold maxR in h8.
  lr_if_goal'; auto.
  rewrite max_set_sortable_sing.
  rewrite max_set_sortable_sing in hr.
  pose proof (h2 _ _ _ h4 hr) as h11.
  apply h3 in h11; contradiction.
Qed.


(*Wyckoff*)
Lemma max_set_sortable_couple' :
  forall {T:Type} (a b:T)
         (pff:Finite (Couple a b)) (pfi:Inhabited (Couple a b))
         (R:T->T->Prop) (pfso:sortable R),
    let (pftso, _, _) := pfso in
    let pftso' := tso_dec_compat _ pftso in
    max_set_sortable (Couple a b) pff pfi R pfso =
    maxR R pftso' a b.
  intros T a b h1 h2 R h3.
  destruct h3 as [h3 h4 h5].
  unfold maxR.
  destruct (tso_dec_compat R h3 a b) as [h6 | h6].
  destruct h6 as [h6|]; subst.
  apply max_set_sortable_couple; auto.
  generalize h1 h2.
  rewrite couple_singleton.
  intros h1' h2'.
  rewrite max_set_sortable_sing; auto.
  generalize h1 h2.
  rewrite couple_comm.
  intros.
  rewrite max_set_sortable_couple; auto.
Qed.


Section FunctionallyPaired.

Inductive functionally_paired {T U:Type} (A:Ensemble T)
          (B:Ensemble U) (S:Ensemble (T*U)) : Prop :=
  functional_pairs_intro : 
    (forall (x:T), 
      In A x -> (exists! y:U, 
                  In B y /\ In S (x, y))) ->
    (forall (pr:T*U), In S pr -> In A (fst pr) /\
    In B (snd pr)) ->
                  functionally_paired A B S.



Lemma fp_in_dom : 
  forall {T U:Type} (A:Ensemble T) 
         (B:Ensemble U) (S:Ensemble (T*U)),
    functionally_paired A B S -> 
    forall pr:T*U, In S pr -> In A (fst pr).
  intros T U A B S h1 pr h2.
  destruct h1 as [? h1].
  apply h1; assumption.
Qed.

Lemma fp_in_ran :
  forall {T U:Type} (A:Ensemble T) 
         (B:Ensemble U) (S:Ensemble (T*U)),
    functionally_paired A B S -> 
    forall pr:T*U, In S pr -> In B (snd pr).
  intros T U A B S h1 pr h2.
  destruct h1 as [? h1].
  apply h1; assumption.
Qed.

Lemma fp_functional :
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U}
         {S:Ensemble (T*U)},
    functionally_paired A B S ->
    forall (x:T) (y1 y2:U), In S (x, y1) ->
                            In S (x, y2) ->
                            y1 = y2.
intros T U A B S h1 x y1 y2 h3 h4.
inversion h1 as [h5 h6].
pose proof (fp_in_dom _ _ _ h1 _ h3) as h7.
simpl in h7.
pose proof (h5 x h7) as h8.
destruct h8 as [y h9].
red in h9.
destruct h9 as [h9a h9b].
pose proof (fp_in_ran _ _ _ h1 _ h3) as h10.
pose proof (fp_in_ran _ _ _ h1 _ h4) as h11.
simpl in h10, h11.
pose proof (h9b y1) as h12.
pose proof (h9b y2) as h13.
cut (y = y1).
cut (y = y2).
congruence. tauto. tauto.
Qed.


Definition fun_well_defined {T U:Type} (S:Ensemble (T*U)) :=
  functionally_paired (dom_rel S) (ran_rel S) S.

Lemma fun_well_defined_empty : forall (T U:Type),
                        fun_well_defined (Empty_set (T*U)).
intros T U. red.
rewrite dom_rel_empty. rewrite ran_rel_empty.
constructor; intros; contradiction.
Qed.

        
Lemma fun_well_defined_im: forall {T U:Type} (A:Ensemble T) (f:T->U),
                         fun_well_defined (Im A (fun x=>(x, f x))).
intros T U A f.
red. constructor.
intros x h1. 
rewrite dom_rel_eq in h1.
rewrite im_im in h1.
simpl in h1.
destruct h1. subst.
exists (f x). red.
split. split.
constructor. exists x. apply Im_intro with x. assumption.
reflexivity.
apply Im_intro with x. assumption. reflexivity. 
intros y h1. 
destruct h1 as [h1l h1r]. 
rewrite ran_rel_eq in h1l. 
rewrite im_im in h1l. simpl in h1l. destruct h1l. subst.
inversion h1r. inversion H2. reflexivity.
intros pr h1. split.
rewrite dom_rel_eq. rewrite im_im. simpl.
destruct h1. subst.
simpl. apply Im_intro with x; auto.
rewrite ran_rel_eq. rewrite im_im. simpl.
destruct h1. subst. simpl. apply Im_intro with x; auto.
Qed.

Lemma fun_well_defined_incl : 
  forall {T U:Type} (R S:Ensemble (T*U)),
    Included R S -> fun_well_defined S -> fun_well_defined R.
intros T U R S h1 h2. 
constructor. 
intros x h3. 
destruct h3 as [h3].
destruct h3 as [y h3].
exists y. red. split. split. constructor. exists x. assumption.
assumption.
intros y' h4.
inversion h2 as [h2l h2r].
assert (h5:In S (x, y)). auto with sets.
specialize (h2r _ h5).
simpl in h2r. destruct h2r as [h2a h2b].
pose proof (h2l _ h2a) as h2'.
destruct h2' as [y'' h2']. red in h2'. destruct h2' as [h6 h7].
specialize (h7 _ (conj h2b h5)). subst. 
destruct h4 as [h4l h4r]. destruct h6 as [h6l h6r].
assert (h7: In S (x, y')). auto with sets.
red in h2.
pose proof (fp_functional h2 _ _ _ h6r h7). assumption.
intros pr h3. split.
constructor. exists (snd pr).
rewrite (surjective_pairing pr) in h3. assumption.
constructor. exists (fst pr).
rewrite (surjective_pairing pr) in h3. assumption.
Qed.

    


Lemma fp_sub_add : 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U}
         {S:Ensemble (T*U)}
         (x:T) (y y':U),
    y <> y' ->
    functionally_paired A B (Add S (x, y)) -> 
    Subtract (Add S (x, y)) (x, y') = Add S (x, y).
intros T U A B S  x y y' h0 h1. 
apply Extensionality_Ensembles. red. split.
apply incl_subtract. 
red. 
intros pr h2. 
constructor. assumption.
intro h3. inversion h3. subst. clear h3.
pose proof (Add_intro2 _ S (x, y)) as h3.
pose proof (fp_functional h1 _ _ _ h2 h3). subst.
contradict h0. reflexivity.
Qed.


Lemma fp_dom_rel :
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U} 
         {S:Ensemble (T*U)},
    functionally_paired A B S ->
    dom_rel S = A.
intros T U A B S h1.
apply Extensionality_Ensembles.
red. split.
red. intros x h2.
destruct h2 as [h2]. destruct h2 as [y h2].
destruct h1 as [h1l h1r].
specialize (h1r _ h2). simpl in h1r. destruct h1r; auto.
red. intros x h2.
constructor.
destruct h1 as [h1l h1r].
specialize (h1l _ h2).
destruct h1l as [y h1l].
red in h1l. destruct h1l as [[h1l ?] ?].
exists y; auto.
Qed.



Lemma fp_add : 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U}
         {S:Ensemble (T*U)}
         (pf:functionally_paired A B S)
         (x:T) (y:U),
    ~ In A x -> In B y -> ~ In S (x, y) ->
    functionally_paired (Add A x) B (Add S (x, y)).
intros T U A B S h1 x y h2 h3 h4.
constructor.
intros x' h5.
destruct h5 as [x' h5 | x' h6].
destruct h1 as [h1l h1r].
specialize (h1l _ h5).
destruct h1l as [y' h1l].
red in h1l.
exists y'.
red.
destruct h1l as [h1a h1b]. split.
destruct h1a as [h1aa h1ab].
split; auto. left. assumption.
intros y'' h6.
destruct h6 as [h6l h6r].
inversion h6r. subst.
specialize (h1b _ (conj h6l H)). assumption.
subst.
inversion H. subst.
contradiction.
inversion h6. subst.
exists y. red.
split. split; try right; auto. constructor.
intros y' h7. destruct h7 as [h7l h7r].
inversion h7r. subst. 
destruct h1 as [h1l h1r].
specialize (h1r _ H).
simpl in h1r. destruct h1r; contradiction. subst.
inversion H. reflexivity.
intros pr h5. destruct h5 as [pr h5l |pr  h5r].
destruct h1 as [h1l h1r].
specialize (h1r _ h5l).
destruct h1r as [h6  h7].
split; try left; auto.
inversion h5r. subst. simpl. split; try right; auto.
constructor.
Qed.


Lemma fun_well_defined_add : 
  forall {T U:Type} {S:Ensemble (T*U)},
    fun_well_defined S ->
    forall (x:T) (y:U), 
      ~In (dom_rel S) x ->
      fun_well_defined (Add S (x, y)).
intros T U S h1 x y h2.
red. constructor.
intros a h3.
rewrite dom_rel_add in h3. simpl in h3.
destruct h3 as [a h3l | a h3r].
destruct h3l as [h3l]. destruct h3l as [b h3l].
exists b. red. split.
split.
rewrite ran_rel_add. simpl. left. constructor. exists a.
assumption.
left. assumption.
intros y' h4.
destruct h4 as [h4l h4r]. 
inversion h4r as [? h5 | ? h6]. 
subst. 
apply (fp_functional h1 _ _ _ h3l h5). subst.
inversion h6. subst. 
pose proof (fp_in_dom _ _ _ h1 _ h3l) as h7.
simpl in h7. contradiction.
inversion h3r. subst. clear h3r.
exists y. red.
split. split. rewrite ran_rel_add. simpl. right. constructor.
right. constructor.
intros y' h3. destruct h3 as [h3l h3r].
inversion h3r as [? h3a | ? h3b]. subst.
pose proof (fp_in_dom _ _ _ h1 _ h3a) as h4. simpl in h4. contradiction.
subst.
inversion h3b. reflexivity.
intros pr h3. split. constructor. exists (snd pr).
rewrite <- surjective_pairing. assumption.
constructor. exists (fst pr). rewrite <- surjective_pairing. assumption.
Qed.

Lemma fun_well_defined_subtract : 
  forall {T U:Type} {S:Ensemble (T*U)},
    fun_well_defined S ->
    forall pr:T*U,
      fun_well_defined (Subtract S pr).
intros T U S h1 pr.
red. constructor.
intros x h2.
destruct h2 as [h2]. destruct h2 as [y h2].
inversion h2 as [h3 h4].
red in h1.
exists y. red. split. split.
constructor. exists x. assumption. assumption.
intros y' h5.
destruct h5 as [h5l h5r].
inversion h5r as [h6 h7].
apply (fp_functional h1 _ _ _ h3 h6).
intros pr' h2. split.
constructor. exists (snd pr'). rewrite <- surjective_pairing.
assumption.
constructor. exists (fst pr'). rewrite <- surjective_pairing.
assumption.
Qed.



Lemma fp_cart_prod_sing : 
  forall {T U:Type} (A:Ensemble T) (b:U),
    functionally_paired A (Singleton b) 
                        (cart_prod A (Singleton b)).
intros T U A b.
constructor.
intros x h1.
exists b.
red. split. split. constructor. constructor.
split. simpl.  assumption. simpl. constructor.
intros u h2.
destruct h2 as [h2l h2r].
destruct h2l; subst. reflexivity.
intros pr h1.
destruct h1 as [h1].
assumption.
Qed.




Lemma fps_inc_cart_prod :
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U}
         {S:Ensemble (T*U)},
  functionally_paired A B S -> Included S (cart_prod A B).
intros T U A B S h1.
red.
intros pr h2.
constructor.
destruct h1 as [h3 h4].
apply h4. assumption.
Qed.



Lemma functionally_paired_restriction : 
  forall {T U:Type} (A A':Ensemble T) (B:Ensemble U)
         (S':Ensemble (T*U)),
    Included A A' ->
    let S := [pr : T * U | In S' pr /\ In A (fst pr)] :
               Ensemble (T * U) in
    functionally_paired A' B S' ->
    functionally_paired A B S.
intros T U A A' B S' h1 S h2.
destruct h2 as [h2l h2r].
constructor. 
intros x h3.
assert (h4:In A' x). auto with sets.
specialize (h2l _ h4). destruct h2l as [y h2l].
red in h2l. destruct h2l as [h2a h2b].
destruct h2a as [h2a h2a'].
exists y. red. split. unfold S. split; auto.
constructor. split; auto.
intros u h5.
destruct h5 as [h5l h5r].
specialize (h2b u).
destruct h5r as [h5r]. simpl in h5r.
destruct h5r as [h5ra h5rb].
specialize (h2b (conj h5l h5ra)). assumption.
intros pr h3.
destruct h3 as [h3].
destruct h3 as [h3l h3r].
specialize (h2r _ h3l).
destruct h2r.
split; auto.
Qed.



Lemma fp_empty1 : forall (T U:Type) (B:Ensemble U),
                        functionally_paired (Empty_set T) 
                                            B (Empty_set (T*U)).
intros; constructor; intros; contradiction.
Qed.

 
Lemma fp_cart_empty11 : forall {T U V:Type} (A:Ensemble U) (B:Ensemble V),
                          functionally_paired (cart_prod (Empty_set T) A)
                                              B (Empty_set (T*U*V)).
intros T U V A B.
constructor. intros ? h1. destruct h1 as [h1].
destruct h1; contradiction.
intros; contradiction.
Qed.

Lemma fp_cart_empty21 : forall {T U V:Type} (A:Ensemble T) (B:Ensemble V),
                        functionally_paired (cart_prod A (Empty_set U))
                                            B (Empty_set (T*U*V)).
intros T U V A B.
constructor. intros ? h1. destruct h1 as [h1].
destruct h1; contradiction.
intros; contradiction.
Qed.


Lemma fp_empty1_s : forall {T U:Type} (B:Ensemble U) (S:Ensemble (T*U)),
                      functionally_paired (Empty_set T) B S ->
                      S = Empty_set (T*U).
intros T U B S h1.
apply Extensionality_Ensembles; red; split; auto with sets.
red.
intros pr h2.
destruct h1 as [h3 h4].
specialize (h4 _ h2).
destruct h4; contradiction.
Qed.


Lemma no_fp_empty2 : forall {T U:Type} (A:Ensemble T) 
                            (S:Ensemble (T*U)),
                        functionally_paired A (Empty_set U) S ->
                        A = Empty_set _.
intros T U A S h1.
apply Extensionality_Ensembles.
red; split; auto with sets; red.
intros x h2.
destruct h1 as [h3 h4].
specialize (h3 _ h2).
destruct h3 as [? h3].
red in h3.
destruct h3 as [h3].
destruct h3; contradiction.
Qed.


Lemma no_fp_empty2' : forall {T U:Type} (A:Ensemble T) 
                            (S:Ensemble (T*U)),
                        functionally_paired A (Empty_set U) S ->
                        S = Empty_set _.
intros T U A S h1.
apply Extensionality_Ensembles.
red; split; auto with sets; red.
intros pr h2.
destruct h1 as [h3 h4].
specialize (h4 _ h2).
destruct h4; contradiction.
Qed.



End FunctionallyPaired.

Section RelExt.
  
  Definition graph {T U:Type} (f:T->U) : Ensemble (T*U) :=
    Im (Full_set T) (fun c => (c, f c)).

  Definition graph_seg {T U:Type} {A:Ensemble T} (f:sig_set A->U) :
    Ensemble (T * U) :=
    Im (full_sig A) (fun c => (proj1_sig c, f c)).
    
  Definition rel_ext {T U:Type} (R:Ensemble (T*U)) (f:T->U) : Prop :=
    Included (graph f) R.

  Definition rel_ext_seg {T U:Type} {A:Ensemble T}
             (R:Ensemble (T*U)) (f:sig_set A->U) : Prop :=
    Included (graph_seg f) R.

  
End RelExt.

Section PopEns.
  
(*Wyckoff*)
Definition pop_ens {T:Type} (S:Ensemble T) (pff:Finite S)
           (R:T->T->Prop) (pfso:sortable R) (def:T) : T*Ensemble T.
  pose proof pff as h1.
  apply finite_inh_or_empty in h1.
  apply or_to_sumbool in h1.
  destruct h1 as [h1 | h1].
  pose (min_set_sortable S pff h1 R pfso) as m.
  refine (m, Subtract S m).
  refine (def, Empty_set _).
Defined.




(*Wyckoff*)
Lemma incl_pop_ens :
  forall {T:Type} (S:Ensemble T) (pff:Finite S)
         (R:T->T->Prop) (pfso:sortable R) (def:T)
         (C:{A:Ensemble T | Included A S}),
    let pfc := Finite_downward_closed _ _ pff _ (proj2_sig C) in
    Included (snd (pop_ens (proj1_sig C) pfc R pfso def)) S.
  intros T S h1 R h2 def C h0.
  destruct h2 as [h2 h3 h4].
  red.
  intros x h5.
  unfold pop_ens in h5.
  destruct or_to_sumbool as [h6 | h6]. simpl in h5. 
  destruct h5 as [h5 h7].
  apply (proj2_sig C); auto.
  simpl in h5.
  contradiction.
Qed.

(*Wyckoff*)
Lemma pop_ens_functional :
  forall {T:Type} (S S':Ensemble T) (pff:Finite S) (pff':Finite S')
         (R:T->T->Prop) (pfso:sortable R) (def:T),
    S = S' ->
    pop_ens S  pff  R pfso def =
    pop_ens S' pff' R pfso def.
  intros T S S' h1 h2 R h3 def h6.
  destruct h3 as [h3 h4 h5].
  subst.
  pose proof (proof_irrelevance _ h1 h2); subst; auto.
Qed.


(*Wyckoff*)
Lemma pop_ens_empty :
  forall {T:Type} (pff:Finite (Empty_set _))
         (R:T->T->Prop) (pfso:sortable R)
         (def:T),
    pop_ens (Empty_set _) pff R pfso def = (def, Empty_set _).
  intros T h1 R h2 def.
  destruct h2 as [h2 h3 h4].
  unfold pop_ens.
  destruct or_to_sumbool as [h5 | h5].
  destruct h5 as [c h5]. destruct h5.
  reflexivity.
Qed.

  
(*Wyckoff*)
Lemma pop_ens_sing :
  forall {T:Type} (a:T) (pff:Finite (Singleton a))
         (R:T->T->Prop) (pfso:sortable R) (def:T),
    pop_ens (Singleton a) pff R pfso def = 
    (a, Empty_set _).
  intros T a h1 R h3 def.
  destruct h3 as [h2 h3 h4].
  unfold pop_ens.
  destruct or_to_sumbool as [h5 | h5].
  rewrite min_set_sortable_sing.
  f_equal.
  rewrite subtract_sing; auto.
  pose proof (In_singleton _ a) as h6.
  rewrite h5 in h6.
  contradiction.
Qed.

(*Wyckoff*)
Lemma pop_ens_couple :
  forall {T:Type} (a b:T) (pff:Finite (Couple a b))
         (R:T->T->Prop) (pfso:sortable R) (def:T),
    let (pftso, _, _) := pfso in
    let pftso' := tso_dec_compat _ pftso in
    pop_ens (Couple a b) pff R pfso def = 
    (minR R pftso' a b, Subtract (Couple a b) (minR R pftso' a b)).
  intros T a b h1 R h2 def.
  destruct h2 as [h2 h3 h4].
  unfold pop_ens.
  destruct or_to_sumbool as [h5 | h5].
  rewrite (min_set_sortable_couple'
             _ _ _ _ _
             (sort_intro R h2 h3 h4)); auto.
  pose proof (Couple_l _ a b) as h6.
  rewrite h5 in h6. contradiction.
Qed.


(*Wyckoff*)

(*This looks much more complicated than it is and basically will be used
in an application of [iter_uo] in the next function to do the iteration.*)
Definition pop_ens_iter_aux {T:Type} (S:Ensemble T) (pff:Finite S)
           (pfi:Inhabited S) (R:T->T->Prop) (pfso: sortable R) (def:T)  :
  T * {A:Ensemble T | Included A S} ->
  T * {A:Ensemble T | Included A S} :=
  let m := min_set_sortable S pff pfi R pfso in
  (fun C':T*{A:Ensemble T | Included A S} =>
     let (_, C) := C' in
     let (x, _) := pop_ens (proj1_sig C)
                           (Finite_downward_closed
                              _ _ pff _ (proj2_sig C)) R pfso def in
     let pf := incl_pop_ens S pff R pfso  def C in
     (x, exist (fun D => Included D S) _ pf)).


(*Wyckoff*)  
Lemma pop_ens_iter_aux_sing :
  forall {T:Type} (a:T) (pff:Finite (Singleton a))
         (pfi:Inhabited (Singleton a))
         (R:T->T->Prop) (pfso:sortable R)
         (def:T) ,
      pop_ens_iter_aux (Singleton a) pff pfi R pfso def =
      (fun pr:(T * {A : Ensemble T | Included A (Singleton a)}) =>
         if (or_to_sumbool
               _ _
               (singleton_inc _ _ (proj2_sig (snd pr)))) then
           (def, (exist (fun A => Included A (Singleton a))
                        (Empty_set _) (incl_empty _)))
         else
           (a,
            (exist
               (fun A => Included A (Singleton a))
               (Empty_set _) (incl_empty _)))).
  intros T a h1 h2 R h3 def.
  destruct h3 as [h3 h4 h5].
  apply functional_extensionality.
  intro C.
  destruct C as [c C]. destruct C as [C h6].
  simpl.
  destruct or_to_sumbool as [h7 | h7].
  subst.
  rewrite pop_ens_empty at 1.
  f_equal. apply proj1_sig_injective; simpl.
  rewrite pop_ens_empty at 1. simpl.
  reflexivity.
  subst.
  rewrite pop_ens_sing at 1.
  f_equal. apply proj1_sig_injective; simpl.
  rewrite pop_ens_sing.
  simpl.
  reflexivity.
Qed.

(*Wyckoff*)
Lemma pop_ens_iter_aux_couple :
  forall {T:Type} (a b:T) (pff:Finite (Couple a b))
         (pfi:Inhabited (Couple a b))
         (R:T->T->Prop) (pfso:sortable R) (def:T),
    let (pftso, _, _) := pfso in
    let pftso' := tso_dec_compat _ pftso in
    pop_ens_iter_aux (Couple a b) pff pfi R pfso def =
    (fun pr:(T * {A : Ensemble T | Included A (Couple a b)}) =>
       match (or_to_sumbool
                _ _
                (incl_couple_inv _ _ _ (proj2_sig (snd pr)))) with
         | left _ =>
           (def, (exist (fun A => Included A (Couple a b))
                        (Empty_set _) (incl_empty _)))
         | right P =>
           match (or_to_sumbool _ _ P) with
             | left Q => 
               (a, (exist (fun A => Included A (Couple a b))
                          (Empty_set _) (incl_empty _)))
               | right F =>
               match (or_to_sumbool _ _ F) with
                 | left W =>
                   (b, (exist (fun A => Included A (Couple a b))
                              (Empty_set _) (incl_empty _)))
                 | right _ =>
                         ((minR R pftso' a b),
                          (exist (fun A => Included A (Couple a b))
                                 (Subtract (Couple a b)
                                           (minR R pftso' a b))
                                 (incl_subtract _ _)))
               end
           end
       end).
  intros T a b h1 h2 R h3 def.
  destruct h3 as [h3 h4 h5].
  apply functional_extensionality. intro pr.
  destruct pr as [x A]. destruct A as [A h6].
  simpl.
  destruct or_to_sumbool as [h7 | h7]. subst.
  rewrite pop_ens_empty at 1.
  f_equal. apply proj1_sig_injective; simpl.
  rewrite pop_ens_empty. simpl. reflexivity.
  destruct or_to_sumbool as [h8 | h8]. subst.
  rewrite pop_ens_sing at 1.
  f_equal. apply proj1_sig_injective; simpl.
  rewrite pop_ens_sing. simpl; auto.
  destruct or_to_sumbool as [h9 | h9]. subst.
  rewrite pop_ens_sing at 1.
  f_equal. apply proj1_sig_injective; simpl.
  rewrite pop_ens_sing at 1. simpl; auto.
  subst.
  pose proof (pop_ens_couple a b  (Finite_downward_closed T (Couple a b) h1 (Couple a b) h6) R (sort_intro R h3 h4 h5) def) as h9.
  simpl in h9.
  rewrite h9 at 1.
  f_equal. apply proj1_sig_injective; simpl.
  rewrite h9; simpl; auto.
Qed.


Definition pop_ens_iter {T:Type} (S:Ensemble T) (pff:Finite S)
           (pfi:Inhabited S) (R:T->T->Prop) (pfso:sortable R) (def:T) :
  nat -> T :=
  let m := min_set_sortable S pff pfi R pfso  in
  (fun n:nat =>
     fst (iter_uo
            (pop_ens_iter_aux S pff pfi R pfso def)
            (m, (exist (fun C => Included C S) _ (incl_subtract S m)))
            n)).


(*Wyckoff*)
Lemma pop_ens_iter_functional :
  forall {T:Type} (S S':Ensemble T) (pff:Finite S) (pff':Finite S')
         (pfi:Inhabited S) (pfi': Inhabited S')
         (R:T->T->Prop) (pfso:sortable R) (def:T),
    S = S' ->
    pop_ens_iter S pff pfi R pfso def =
    pop_ens_iter S' pff' pfi' R pfso def.
  intros T S S' h1 h2 h3 h4 R h5 def h8.
  subst.
  pose proof (proof_irrelevance _ h1 h2).
  pose proof (proof_irrelevance _ h3 h4).
  subst.
  reflexivity.
Qed.


(*Wyckoff*)
Lemma pop_ens_iter0 :
  forall {T:Type} (S:Ensemble T) (pff:Finite S)
         (pfi:Inhabited S) (R:T->T->Prop) (pfso:sortable R) (def:T),
    pop_ens_iter S pff pfi R pfso def 0 =
    min_set_sortable S pff pfi R pfso.
  intros; unfold pop_ens_iter; auto.
Qed.

(*Wyckoff*)
Lemma pop_ens_iter_add0 :
  forall {T:Type} (A:Ensemble T) (a:T) (pff:Finite (Add A a))
         (pfi:Inhabited (Add A a))
         (R:T->T->Prop) (pfso:sortable R) (def:T),
    let (pftso, _, _) := pfso in
    let pftso' := tso_dec_compat _ pftso in
    (forall pf':Inhabited A,
       let pff' := Finite_downward_closed _ _ pff _ (incl_add A a) in
    pop_ens_iter (Add A a) pff pfi R pfso def 0 =
    minR R pftso' (min_set_sortable A pff' pf' R pfso) a) /\
    (A = Empty_set _ ->
     pop_ens_iter (Add A a) pff pfi R pfso def 0 = a).
  intros T A a h1 h2 R h3 def.
  destruct h3 as [h3 h4 h5].
  simpl.
  pose proof (min_set_sortable_add A a h1 h2 R
                                   (sort_intro _ h3 h4 h5)) as h7.
  simpl in h7. destruct h7 as [h7 h8].
  rewrite pop_ens_iter0.
  split.
  intro h6. simpl.
  specialize (h7 h6); auto.
  assumption.
Qed.


(*Wyckoff*)
Lemma pop_ens_iter_couple1 :
  forall {T:Type} (a b:T) (R:T->T->Prop) (pfso:sortable R)
         (def:T),
    R a b ->
    let (pftso, _, _) := pfso in
    pop_ens_iter (Couple a b) (finite_couple a b)
                 (inh_couple a b) R pfso def 1 = b.
  intros T a b R h1 def h4.
  destruct h1 as [h1 h2 h3].
  unfold pop_ens_iter.
  rewrite min_set_sortable_couple; auto.
  simpl.
  assert (h5:a <> b).
    intro. subst. pose proof (h3 b). contradiction.
  pose proof (subtract_couple_l _ _ h5) as h6.
  pose proof (pop_ens_functional
                _ _
                (Finite_downward_closed
                   T (Couple a b) 
                   (finite_couple a b) (Subtract (Couple a b) a)
               (incl_subtract (Couple a b) a))
                (Singleton_is_finite _ b) R
                (sort_intro _ h1 h2 h3) def h6) as h7.
  rewrite h7 at 1.
  rewrite pop_ens_sing.
  simpl.
  reflexivity.
Qed.
End PopEns.






End Relations.



Lemma inhabited_family_cond :
  forall {T:Type} (F:Family T),
         inhabited_family F ->
  forall A:sig_set F, exists y:T, In (proj1_sig A) y.
intros T F h1 A.
destruct h1 as [h1 h2].
destruct A as [A h3].
simpl.
apply h2 in h3.
destruct h3 as [y h3].
exists y; auto.
Qed.

Section InverseImage.

(*Schepler*)
Definition inverse_image {X Y:Type} (f:X->Y) (T:Ensemble Y)
  : Ensemble X :=
  [ x:X | In T (f x) ].
Hint Unfold inverse_image : sets.

(*Schepler*)
Lemma inverse_image_increasing: forall {X Y:Type} (f:X->Y)
  (T1 T2:Ensemble Y), Included T1 T2 ->
  Included (inverse_image f T1) (inverse_image f T2).
Proof.
intros.
red; intros.
destruct H0.
constructor.
auto.
Qed.

(*Schepler*)
Lemma inverse_image_empty: forall {X Y:Type} (f:X->Y),
  inverse_image f (Empty_set _) = Empty_set _.
Proof.
intros.
apply Extensionality_Ensembles; split; red; intros.
destruct H as [[]].
destruct H.
Qed.

(*Schepler*)
Lemma inverse_image_full: forall {X Y:Type} (f:X->Y),
  inverse_image f (Full_set _) = Full_set _.
Proof.
intros.
apply Extensionality_Ensembles; split; red; intros;
  constructor; constructor.
Qed.

(*Schepler*)
Lemma inverse_image_intersection: forall {X Y:Type} (f:X->Y)
  (T1 T2:Ensemble Y), inverse_image f (Intersection T1 T2) =
  Intersection (inverse_image f T1) (inverse_image f T2).
Proof.
intros.
apply Extensionality_Ensembles; split; red; intros.
destruct H.
inversion H.
constructor; constructor; trivial.

destruct H as [? [] []].
constructor; constructor; trivial.
Qed.

(*Schepler*)
Lemma inverse_image_union: forall {X Y:Type} (f:X->Y)
  (T1 T2:Ensemble Y), inverse_image f (Union T1 T2) =
  Union (inverse_image f T1) (inverse_image f T2).
Proof.
intros.
apply Extensionality_Ensembles; split; red; intros.
destruct H.
inversion H.
left; constructor; trivial.
right; constructor; trivial.

constructor.
destruct H as [? []|? []].
left; trivial.
right; trivial.
Qed.

(*Schepler*)
Lemma inverse_image_complement: forall {X Y:Type} (f:X->Y)
  (T:Ensemble Y), inverse_image f (Comp T) =
  Comp (inverse_image f T).
Proof.
intros.
apply Extensionality_Ensembles; split; red; intros.
red; intro.
destruct H.
destruct H0.
contradiction H.

constructor.
intro.
contradiction H.
constructor; trivial.
Qed.

(*Schepler*)
Lemma inverse_image_composition: forall {X Y Z:Type} (f:Y->Z) (g:X->Y)
  (U:Ensemble Z), inverse_image (fun x:X => f (g x)) U =
  inverse_image g (inverse_image f U).
Proof.
intros.
apply Extensionality_Ensembles; split; red; intros.
constructor; constructor.
destruct H.
assumption.

destruct H; inversion H.
constructor; trivial.
Qed.

Hint Resolve @inverse_image_increasing : sets.
Hint Rewrite @inverse_image_empty @inverse_image_full
  @inverse_image_intersection @inverse_image_union
  @inverse_image_complement @inverse_image_composition : sets.

End InverseImage.


Section RandomFacts.


(*Wyckoff*)
Definition char_fun {T:Type} (A:Ensemble T)
           : T->bool := 
  fun x:T => if (classic_dec (In A x)) then true else false.


(*Wyckoff*)
Lemma char_fun_int : 
  forall {T:Type} (A B:Ensemble T) (x:T),
    char_fun (Intersection A B) x = char_fun A x && char_fun B x.
intros T A B x.  
pose proof (in_intersection_iff A B x) as h1.
unfold char_fun.
lr_if_goal'.
rewrite h1 in hl.
destruct hl.
lr_if_goal'; simpl; auto.
lr_if_goal'; auto.
contradiction.
contradiction.
rewrite h1 in hr.
apply not_and_or in hr.
destruct hr as [h2 | h2].
lr_if_goal'; simpl; auto. contradiction.
lr_if_goal'; simpl; auto.
lr_if_goal'; simpl; auto.
Qed.


(*Wyckoff*)
Lemma char_fun_union : 
  forall {T:Type} (A B:Ensemble T) (x:T),
    char_fun (Union A B) x = char_fun A x || char_fun B x.
intros T A B x.  
unfold char_fun.
lr_if_goal'.
lr_if_goal'; simpl; auto.
lr_if_goal'; auto.
destruct hl; contradiction.
lr_if_goal'; simpl.
contradict hr; left; auto.
rewrite in_union_iff in hr.
lr_if_goal'; auto.
Qed.

(*Wyckoff*)
Lemma char_fun_comp : 
  forall {T:Type} (A:Ensemble T) (x:T),
    char_fun (Comp A) x = negb (char_fun A x).
intros T A x.
unfold char_fun.
lr_if_goal'; simpl; auto.
lr_if_goal'; simpl; auto.
destruct hl; auto.
lr_if_goal'; simpl; auto.
Qed.


(*Wyckoff*)
Lemma injective_preserves_card_fun1 : 
  forall {T U:Type} (f:T->U),
    injective f ->
    forall (S:Ensemble T),
      Finite S ->
    card_fun1 S = card_fun1 (Im S f).
intros T U f h1 S h2.
pose proof (card_fun1_compat S) as h3. simpl in h3.
destruct h3 as [h3 h4].
specialize (h3 h2).  
rewrite inj_image_fun_prop_compat in h1.
symmetry.
apply (injective_preserves_cardinal _ _ _  _ _ h1  h3).
apply card_fun1_compat.
apply finite_image; auto.
Qed.

(*Wyckoff*)
Lemma le_im_card :
  forall {T U:Type} (A:Ensemble T) (f:T->U),
    Finite A -> 
    card_fun1 (Im A f) <= card_fun1 A.
  intros T U A f h1.
  pose proof (card_fun1_compat A) as h3.
  pose proof (card_fun1_compat (Im A f)) as h4.
  simpl in h3, h4.
  destruct h3 as [h3 h5]. destruct h4 as [h4 h6]. clear h5 h6.
  hfirst h4.
  apply finite_image; auto.
  specialize (h4 hf).
  specialize (h3 h1).
  eapply cardinal_decreases.
  apply h3.
  apply h4.
Qed.


(*Wyckoff*)
Lemma excl_middle_sat :
  forall {T:Type} (P:T->Prop),
    Full_set T = Union [x:T | P x] [x:T | ~ (P x)].
intros T P. apply Extensionality_Ensembles.
red. split.
red.
intros x h1.
destruct (classic (P x)) as [h2 | h3].
left. constructor. assumption.
right. constructor. assumption.
red. intros; constructor.
Qed.

(*Wyckoff*)
Lemma excl_middle_sat' : forall {T:Type} (S:Ensemble T)
                         (P:T->Prop),
                           S = Union [x:T | In S x /\ P x]
                                     [x:T | In S x /\ ~ (P x)].
intros T S P.
apply Extensionality_Ensembles.
red. split.
red.
intros x h1.
destruct (classic (P x)) as [h2 | h3].
left. constructor. split; auto.
right. constructor. split; auto.
red. intros x h1.
destruct h1 as [x h1l | x h1r].
destruct h1l as [h1l]. destruct h1l; assumption.
destruct h1r as [h1r]. destruct h1r; assumption.
Qed.



(*Wyckoff*)
Lemma surj_iff' :
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U}
         (f:sig_set A -> sig_set B),
    surjective f <-> im_proj1_sig (Im (full_sig A) f) = B.
intros T U A B f.
split.
intro h1.
apply surj_iff in h1.
rewrite h1 at 1.
rewrite im_proj1_sig_full_set_sig.
reflexivity.
intro h1.
red.
intro y.
destruct y as [y h2].
pose proof h2 as h2'.
rewrite <- h1 in h2'.
destruct h2' as [y h3].
clear h1.
subst.
destruct h3 as [x h3]. subst.
clear h3.
exists x.
apply proj1_sig_injective; auto.
Qed.

(*Wyckoff*)
Definition extends_sig_sig 
           {T U:Type} {A B:Ensemble T} {E:Ensemble (sig_set A)}
           (f:sig_set B->U) (g:sig_set E->U) :=
  exists pf:Included (im_proj1_sig E) B,
    forall x:sig_set (im_proj1_sig E),
      g (exist _ _  (sig_set_im_proj1_sig_in x)) =
      f (exist _ _ (pf _ (proj2_sig x))).

(*Wyckoff*)
Definition extends_sig_sig'
           {T U:Type} {A B:Ensemble T} {E:Ensemble (sig_set A)}
           (f:sig_set B->U) (g:sig_set E->U) :=
  forall (pf:Included (im_proj1_sig E) B)
         (x:sig_set (im_proj1_sig E)),
      g (exist _ _  (sig_set_im_proj1_sig_in x)) =
      f (exist _ _ (pf _ (proj2_sig x))).

(*Wyckoff*)
Lemma extends_sig_sig_impl_prime :
  forall {T U:Type} {A B:Ensemble T} {E:Ensemble (sig_set A)}
         (f:sig_set B->U) (g:sig_set E->U),
    extends_sig_sig f g -> extends_sig_sig' f g.
  intros T A B E f g h1 h2.
  red in h2. red.
  intro h3. destruct h2 as [h2 h4].
  pose proof (proof_irrelevance _ h2 h3); subst.
  assumption.
Qed.

(*Wyckoff*)
Lemma extends_sig_sig_extends_sig1_iff :
  forall {T U:Type} {A:Ensemble T} (E:Ensemble (sig_set A))
         (f:sig_set A->U) (g:sig_set E->U),
    extends_sig_sig f g <-> extends_sig1 f g.
  intros T U A E f g.
  split.
  intro h1. red in h1. red.
  destruct h1 as [h1 h2].
  intro x. destruct x as [x h3].
  pose proof h3 as h3'.
  rewrite <- in_im_proj1_sig_iff in h3'.
  specialize (h2 (exist _ _ h3')).
  simpl. simpl in h2.
  let P := type of h2 in match P with g ?c' = _ => pose c' as c end.
  let P := type of h2 in match P with _ = f ?d' => pose d' as d end.
  match goal with |- g ?e' = _ => pose e' as e end.
  assert (h4:x = d). unfold d. apply proj1_sig_injective; auto.
  assert (h5:e = c). unfold e, c. apply proj1_sig_injective; simpl.
    apply proj1_sig_injective; auto.
  fold e. fold c in h2.
  rewrite h5.  rewrite h2. fold d. rewrite h4.
  reflexivity.

  intro h1. red in h1. red.
  assert (h2:Included (im_proj1_sig E) A).
    red; intros x h3. destruct h3 as [x h3]; subst. apply proj2_sig.
  exists h2.
  intro x.
  destruct x as [x h3]. simpl. destruct h3 as [x h3]. subst.
  specialize (h1 (exist _ _ h3)).
  simpl in h1.
  match goal with |- g ?c' = _ => pose c' as c end.
  match goal with |- _ = f ?d' => pose d' as d end.
  fold c d.
  assert (h4:c = exist _ _ h3). unfold c. apply proj1_sig_injective; simpl.
    apply proj1_sig_injective; auto.
  rewrite h4.
  assert (h5:x = d).
    unfold d. apply proj1_sig_injective; auto.
  rewrite <- h5; auto.
Qed.




(*Wyckoff*)
Lemma card_fun1_im_proj1_sig : 
  forall {T:Type} (A:Ensemble T)
         (S:Ensemble (sig_set A)),
    Finite S ->
    card_fun1 S = card_fun1 (im_proj1_sig S).
intros; apply injective_preserves_card_fun1; auto.
red; intros; apply proj1_sig_injective; auto.
Qed.


(*Wyckoff*)
Lemma card_fun1_im_proj1_sig' : 
  forall {T:Type} (P:T->Prop),
    let U:= {x:T | P x} in
    forall S:Ensemble U,
      Finite S ->
      card_fun1 S = card_fun1 (im_proj1_sig' S).
intros; apply injective_preserves_card_fun1; auto.
red; intros; apply proj1_sig_injective; auto.
Qed.  

(*See [SetUtilities2], for [card_fun1_im_proj2_sig].*)
Lemma im_proj1_sig_full_sig :
  forall {T:Type} (S:Ensemble T),
    S = im_proj1_sig (full_sig S).
  intros T S.
  apply Extensionality_Ensembles; red; split; red; intros x h1.
  apply Im_intro with (exist _ _ h1).
  constructor. auto.
  destruct h1 as [x h1]; subst.
  clear h1. apply proj2_sig.
Qed.




(*Wyckoff*)
Lemma card_union :
  forall {T:Type} (A B:Ensemble T) (m n p:nat),
         cardinal _ A m ->
         cardinal _ B n ->
         cardinal _ (Intersection A B) p ->
         cardinal _ (Union A B) (m+n-p).
intros T A B m n p h1 h2 h3.
pose proof (decompose_int_setminus (Union A B) B) as h4. 
assert (h5:Included B (Union A B)). auto with sets.
rewrite Intersection_commutative in h4.
rewrite inclusion_iff_intersection_eq in h5.
rewrite h5 in h4.
rewrite setminus_union in h4.
pose proof (int_setminus B (Union A B)) as h6.
rewrite setminus_union in h6.
pose proof (card_decompose_int_setminus' _ _ _ _ h1 h3) as h7.
pose proof (card_disj_union _ _ _ _ h6 h2 h7) as h8.
rewrite h4.
assert (h9:n + (m-p) = m + n - p).
  rewrite (plus_comm m n).    
  apply plus_minus_assoc.
  assert (h9:Included (Intersection A B) A). auto with sets.
  apply (incl_card_le _ _ _ _ _ h3 h1 h9).
rewrite <- h9.
assumption.
Qed.

(*Wyckoff*)
Lemma card_gt_O_inh : 
  forall {T:Type} (A:Ensemble T) (n:nat),
    n > O -> cardinal _ A n -> Inhabited A.
intros T A n; destruct n as [|n]; simpl.
intros h1 h2. omega.
intros ? h1. inversion h1. apply Inhabited_intro with x. right.
constructor.
Qed.



(*Wyckoff*)
Lemma sub_empty_sing_or_empty :
  forall {T:Type} (A:Ensemble T) (x:T),
    Subtract A x = Empty_set _ ->
    A = Singleton x \/ A = Empty_set _.
intros T A x h1.
destruct (classic (In A x)) as [h2 | h3].
left.
apply Extensionality_Ensembles; red; split.
red. intros a h3. 
apply NNPP.
intro h4.
assert (h5:In (Subtract A x) a). constructor; auto.
rewrite h1 in h5. contradiction.
red. intros a h3. destruct h3; subst. assumption.
right.
rewrite subtract_nin in h1; auto.
Qed.





 

(*Wyckoff*)
Lemma atom_sing : 
  forall {T:Type} (E A:Ensemble T), 
    Finite E -> Included A E -> A <> Empty_set _ ->
    ((forall F:Ensemble T, Included F E ->
                          Intersection F A = A \/
                          Intersection F A = Empty_set _) <->
    (exists (a:T), In E a /\ A = Singleton a)).
intros T E A h1. revert A.
induction h1 as [|E h2 h3 x h4].
intros A h1 h2.
assert (h3:A = Empty_set _). auto with sets.
contradiction.
intros A h1 h5. 
split. 
(* -> *)
intro h6.
pose proof (subtract_preserves_inclusion A (Add E x) x h1) as h7.
rewrite sub_add_same_nin in h7; auto.
destruct (classic (In A x)) as [h8 | h9].
exists x. split. right. constructor. 
pose proof (Inhabited_intro _ _ _ h8) as h9.

assert (h10:Included (Singleton x) (Add E x)).
  red. intros x' h11. destruct h11. right. constructor.
specialize (h6 (Singleton x) h10).
destruct h6 as [h6l | h6r].
assert (h11:Intersection (Singleton x) A = Singleton x).
  apply Extensionality_Ensembles.
  red. split. auto with sets.
  red. intros x' h11. destruct h11. constructor; auto.
  constructor.
rewrite h11 in  h6l.
rewrite h6l. reflexivity.
assert (h11:In (Intersection (Singleton x) A) x).
  constructor; auto. constructor.
rewrite h6r in h11. contradiction.
pose proof (subtract_preserves_inclusion A (Add E x) x h1) as h10.
rewrite subtract_nin in h10; auto.
rewrite sub_add_same_nin in h10; auto.
specialize (h3 _ h10 h5).
assert (h11:(forall F : Ensemble T,
        Included F E ->
        Intersection F A = A \/ Intersection F A = Empty_set T)).
  intros F h12. 
  assert (h13:Included F (Add E x)). red. intros; auto with sets.
  apply h6; auto.
rewrite h3 in h11.
destruct h11 as [a h11].
destruct h11 as [h11l h11r].
exists a. split; auto. left. assumption.
intros h6.
destruct h6 as [a h6].
destruct h6 as [h6l h6r].
intros F h7.
subst.
destruct (classic (In F a)) as [h8 | h9].
left. apply Extensionality_Ensembles. red. split.  auto with sets.
red. intros a' h9. destruct h9; subst. constructor; auto.
constructor.
right.
apply Extensionality_Ensembles.
red. split; auto with sets.
red. intros a' h10. destruct h10 as [a' h10l h10r].
destruct h10r; subst.
contradiction.
Qed.

(*Wyckoff*)
Inductive incl_sig {T:Type} (A B:Ensemble T)
          (pfi:Included A B) : Ensemble {x:T|In B x} :=
  incl_sig_intro : forall (x:T) (pfa:In A x),
                    In (incl_sig A B pfi) 
                       (exist _ _ (pfi _ pfa)).

(*Wyckoff*)
Lemma incl_sig_iff :   
  forall {T:Type} (A B:Ensemble T) (pf:Included A B)
         x, In (incl_sig A B pf) x <->
            In A (proj1_sig x).
intros T A B h0 x.
split. intros h1.
destruct h1. simpl. assumption.
intro h1. 
pose proof (incl_sig_intro _ _ h0 _ h1) as h2.
assert (h3:x = (exist (In B) (proj1_sig x) (h0 (proj1_sig x) h1))). 
  rewrite unfold_sig at 1. apply proj1_sig_injective. simpl.
  reflexivity.
rewrite <- h3 in h2.
assumption.
Qed.


(*Wyckoff*)
Definition subset_sig {T:Type} (A B:Ensemble T)
           (pf:Included A B) : Ensemble (sig_set B) :=
  (Im (full_sig A) (fun x => exist
                               _ _
                               (pf (proj1_sig x) (proj2_sig x)))). 


(*Wyckoff*)
Lemma subset_sig_compat : 
  forall {T:Type} (A B:Ensemble T) (pf:Included A B)
         x, In (subset_sig A B pf) x <->
                In A (proj1_sig x).
intros T A B h0 x.
split.
intro h1.
destruct h1 as [x h2 y h4]. subst. simpl.
inversion h2. subst. apply proj2_sig.
intro h1.
unfold subset_sig.
apply Im_intro with  (exist _ _ h1).
constructor. simpl.
apply proj1_sig_injective. simpl.
reflexivity.
Qed.


(*Wyckoff*)
Lemma incl_sig_eq_subset_sig : forall {T:Type} (A B:Ensemble T)
  (pf:Included A B), incl_sig A B pf = subset_sig A B pf.
intros T A B pf.
apply Extensionality_Ensembles.
red. split.
red.
intros x h1.
destruct h1 as [x h1].
unfold subset_sig.
apply Im_intro with (exist _ _ h1).
constructor. apply proj1_sig_injective. simpl. reflexivity.
red. 
intros x h1.
destruct h1 as [x h1]. subst.
constructor.
Qed.




(*Wyckoff*)
Lemma full_sig_family_union_eq :
  forall {T:Type} (F:Family T),
    full_sig (FamilyUnion F) =
    FamilyUnion
      [S:Ensemble (sig_set (FamilyUnion F)) |
       In F (Im S (@proj1_sig _ _))].
intros T F. 
apply Extensionality_Ensembles. red. split.
Focus 2.
red. intros x h3. constructor.
red. intros x h3.
destruct x as [x h4].
destruct h4 as [S x h4 h5].
assert (h7:forall s:T, In S s -> In S s). auto.
apply family_union_intro with
(im_in _ _  (fun x':{s:T|In S s} =>
               (exist
                  _ _
                  (family_union_intro
                     _ _ _
                     (proj1_sig x') h4 (proj2_sig x')))) h7).
constructor.
assert (h8:S =
           (Im
              (im_in
                 S
                 (In S)
                 (fun x' : {s : T | In S s} =>
                    exist (In (FamilyUnion F)) (proj1_sig x')
                          (family_union_intro T F S (proj1_sig x')
                                              h4 (proj2_sig x')))
                 h7)
              (proj1_sig (P:=fun x0 : T => In (FamilyUnion F) x0)))).
apply Extensionality_Ensembles.
red. split.
red. intros s h9.
apply Im_intro with (exist _ _ (family_union_intro _ _ _ _ h4 h9)).
unfold im_in. simpl.
apply Im_intro with (exist _ _ h9).
constructor.
apply proj1_sig_injective.
simpl. reflexivity. simpl. reflexivity.
red. intros s h9.
destruct h9 as [s h9]. subst.
destruct h9 as [s h9]. subst.
simpl.
destruct s. simpl. assumption.
pose proof h4 as h4'. rewrite h8 in h4'.
assumption.
unfold im_in. simpl.
apply Im_intro with (exist _ _ h5).
constructor.
apply proj1_sig_injective. simpl.
reflexivity.
Qed.

(*Wyckoff*)
Lemma full_sig_decompose_setminus_incl : 
  forall {T:Type} (A B:Ensemble T) (pf:Included B A),
    full_sig A =
    Union
      (Im (full_sig (Setminus A B)) 
          (fun x => exist _ _
                          ((setminus_inc A B)
                             _ (proj2_sig x))))
      (Im (full_sig B)
          (fun x => exist _ _ (pf _ (proj2_sig x)))).
intros T A B h1.
apply Extensionality_Ensembles.
red. split.
red.
intros x h2. clear h2.
destruct x as [x h2].
pose proof (decompose_setminus_inc _ _ h1) as h3.
pose proof h2 as h2'. rewrite h3 in h2'.
destruct h2' as [x h4 | x h5].
right.
apply Im_intro with (exist _ _ h4). constructor.
apply proj1_sig_injective.
simpl. reflexivity.
left. apply Im_intro with (exist _ _ h5).
constructor.
apply proj1_sig_injective. simpl. reflexivity.
red. intros x h2. constructor.
Qed.


(*Wyckoff*)
Lemma full_false_empty : 
  forall (T:Type), 
    Full_set {_ : T | False} = Empty_set _.
intro T.
apply Extensionality_Ensembles.
red. split; auto with sets.
red.
intros x. destruct x. contradiction.
Qed.


(*Wyckoff*)
Lemma full_empty_sig_empty :
  forall (T:Type),
    full_sig (Empty_set T) = Empty_set _.
intros T.
apply Extensionality_Ensembles.
red. split.
red.
intros x h1.
pose proof (proj2_sig x) as h2. simpl in h2.
contradiction.
auto with sets.
Qed.




(*Wyckoff*)
Lemma finite_image_rev_inj : 
  forall {T U:Type} (A:Ensemble T) (f:T->U), 
    injective f -> Finite (Im A f) ->
    Finite A.
intros T U A f h1 h2. 
apply NNPP.
intro h3. 
contradict h2.
intro h4.
pose proof (Pigeonhole_bis _ _ _ _ h3 h4).
contradiction.
Qed.

(*Wyckoff*)
Definition inv_im {T U:Type} (B:Ensemble U) (f:T->U) :=
  [x : T | In B (f x)].

(*Wyckoff*)
Definition inv_im1 {T U:Type} (f:T->U) (y:U)  :=
  [x : T | f x = y].

(*Wyckoff*)
Definition inv_im_seg {U:Type} {n:nat} (f:seg_fun n U) (y:U)  :=
  [i : nat | exists pf:i < n, f i pf = y].


(*Wyckoff*)
Lemma inv_im1_compat :
  forall {T U:Type} (f:T->U) (x:T) (y:U),
    Inn (inv_im1 f y) x ->
    f x = y.
  intros; destruct H; auto.
Qed.

(*Wyckoff*)
Lemma inv_im1_compat' :
  forall {T U:Type} (f:T->U) (y:U),
    inv_im1 f y = inv_im (Singleton y) f.
  intros T U f y.
  apply Extensionality_Ensembles; red; split; red; intros x h1;
  destruct h1; subst; econstructor.
  constructor.
  destruct H; auto.
Qed.

(*Wyckoff*)
Lemma inv_im_seg_compat :
  forall {U:Type} {n:nat} (f:seg_fun n U) (y:U) (i:nat),
    Inn (inv_im_seg f y) i ->
    exists pf:i < n,
      f i pf = y.
  intros; destruct H; auto.
Qed.

(*Wyckoff*)
Lemma lt_inv_im_seg : 
  forall {U:Type} {n:nat} (f:seg_fun n U) (y:U) (i:nat),
    Inn (inv_im_seg f y) i ->
    i < n.
  intros.
  destruct H.
  destruct H; auto.
Qed.

(*Wyckoff*)
(*The set of all pre-images of points in [U]*)
Definition inv_im1s {T U} (f:T->U) : Family T :=
  Im (Full_set U) (inv_im1 f).




(*Wyckoff*)
Lemma inv_im_empty : forall {T U:Type} (f:T->U),
                       inv_im (Empty_set U) f = Empty_set T.
intros T U f.
apply Extensionality_Ensembles; red; split; auto with sets.
red.
intros x h1.
destruct h1. contradiction.
Qed.


(*Wyckoff*)
Lemma im_inv_im : 
  forall {T U:Type} (f:T->U) (B:Ensemble U),
    Im (inv_im B f) f = Intersection B (Im (Full_set T) f).
intros T U  f B.
apply Extensionality_Ensembles.
unfold inv_im.
red. split.
(* <= *)
red.
intros b h1. 
inversion h1 as [a h2 b' h3]. subst.
constructor. destruct h2. assumption.
apply Im_intro with a. constructor. reflexivity.
red.
intros u h1.
inversion h1 as [u' h2 h3]. subst.
inversion h3 as [x h4 y h5 h6]. subst.
apply Im_intro with x.
constructor. assumption.
reflexivity.
Qed.

(*Wyckoff*)
Lemma im_inv_im1_in : 
  forall {T U:Type} (f:T->U) (y:U),
    In (Im (Full_set _) f) y ->
    Im (inv_im1 f y) f = Singleton y.
  intros T U f y h1.
  destruct h1; subst. 
  apply Extensionality_Ensembles; red; split; red; intros y h1;
  destruct h1; subst.
  destruct H0.
  rewrite H0.
  constructor.
  econstructor.
  econstructor.
  reflexivity.
  reflexivity.
Qed.


(*Wyckoff*)
Lemma im_inv_im_seg_in : 
  forall {U:Type} {n:nat} (f:seg_fun n U) (y:U),
    Inn (im_in_ens f) y ->
    im_in_ens (fun i (pf:Inn (inv_im_seg f y) i) =>
                 f i (lt_inv_im_seg _ _ _ pf)) = Singleton y.
  intros U n f y h1.
  destruct h1; subst.
  red in pf.
  apply Extensionality_Ensembles; red; split; red;
  intros y h1; destruct h1; subst.
  destruct pf0.
  destruct e as [h1 h2].
  rewrite in_sing_iff.
  rewrite <- h2.
  f_equal.
  apply proof_irrelevance.
  econstructor.
  f_equal.
  apply proof_irrelevance.

  Grab Existential Variables.
  econstructor.
  exists pf; auto.
Qed.
  


(*Wyckoff*)
Lemma in_im_inv_im1 : 
  forall {T U:Type} (f:T->U) (x:T),
    Inn (inv_im1 f (f x)) x.
  intros; econstructor; auto.
Qed.




(*Wyckoff*)
Lemma in_inv_im_seg :
  forall {U:Type} {n:nat} (f:seg_fun n U) (i:nat)
         (pf:i < n),
    Inn (inv_im_seg f (f i pf)) i.
  intros; econstructor; auto.
  exists pf; auto.
Qed.



(*Wyckoff*)
Lemma inv_im_int : 
  forall {T U:Type} (A B:Ensemble U) (f:T->U),
    inv_im (Intersection A B) f =
    Intersection (inv_im A f) (inv_im B f).
intros T U A B f.
apply Extensionality_Ensembles.
red. split.
red.
intros x h1.
destruct h1 as [h1].
inversion h1 as [? h2 h3]. subst.
constructor; constructor; auto.
red.
intros x h1.
destruct h1 as [x h1 h2].
constructor.
destruct h1 as [h1]. destruct h2 as [h2].
constructor; auto.
Qed.

(*Wyckoff*)
Lemma inv_im_union : 
  forall {T U:Type} (A B:Ensemble U) (f:T->U),
    inv_im (Union A B) f = Union (inv_im A f) (inv_im B f).
intros T U A B f.
apply Extensionality_Ensembles.
red. split.
red. intros x h1.
destruct h1 as [h1].
inversion h1 as [? h2 |? h3]. subst.
left. constructor; auto.
subst. right. constructor; auto.
red. intros x h1.
destruct h1 as [x h1 | x h2].
destruct h1 as [h1].
constructor. left. auto.
destruct h2 as [h2].
constructor. right.
assumption.
Qed.

(*Wyckoff*)
Lemma inv_im_comp : 
  forall {T U:Type} (A:Ensemble U) (f:T->U),
    inv_im (Comp A) f = Comp (inv_im A f).
intros T U A f.
apply Extensionality_Ensembles.
red. split.
red. intros x h1.
destruct h1 as [h1]. 
intro h2.
destruct h2 as [h2].
auto with sets.
red. intros x h1. 
constructor.
intro h2. 
unfold inv_im in h1.
apply complement_inv in h1. 
apply sat_nin in h1.
contradiction.
Qed.

(*Wyckoff*)
Lemma inh_surj :
  forall {T U:Type} (f:T->U) (y:U),
    surjective f ->
    Inhabited (inv_im1 f y).
  intros T U f y h1.
  red in h1.
  specialize (h1 y).
  destruct h1; subst.
  econstructor.
  econstructor.
  reflexivity.
Qed.

(*Wyckoff*)
Lemma inh_surj_seg :
  forall {U:Type} {n:nat} (f:seg_fun n U) (y:U),
    surj_seg f ->
    Inhabited (inv_im_seg f y).
  intros U n f y h1.
  specialize (h1 y).
  destruct h1 as [i h2].
  destruct h2; subst.
  econstructor.
  econstructor.
  exists x; auto.
Qed.

(*Wyckoff*)
Lemma inh_inv_im_seg :
  forall {U:Type} {n:nat} (f:seg_fun n U) (y:sig_set (im_in_ens f)),
    Inhabited (inv_im_seg f (proj1_sig y)).
  intros U n f y.
  destruct y as [y hy]; subst.
  destruct hy as [hy]; subst.
  simpl.
  econstructor.
  econstructor.
  exists pf; auto.
Qed.

(*Wyckoff*)
Lemma inh_inv_im_seg' :
  forall {U:Type} {n:nat} (f:nat->U)
         (y:sig_set (Im (seg n) f)),
    Inhabited (inv_im_seg (fun_to_seg f n) (proj1_sig y)).
  intros U n f y.
  destruct y as [y hy]; subst.
  destruct hy as [y hy]; subst.
  simpl.
  econstructor.
  econstructor.
  destruct hy as [hy].
  exists hy; auto.
Qed.

(*Wyckoff*)
Lemma inh_inv_im_sig_seg :
  forall {U:Type} (f:nat->U) (n:nat) (y:sig_set (Im (seg n) f)),
    Inhabited (inv_im_seg (fun_to_sig_seg f n) y).
  intros U f n y.
  destruct y as [y hy]; subst.
  destruct hy as [i hy]; subst.
  destruct hy as [hy].
  econstructor.
  econstructor.
  exists hy.
  unfold fun_to_sig_seg.
  apply proj1_sig_injective; auto.
Qed.


(*Wyckoff*)
Lemma finite_inv_im_seg :
  forall {U:Type} {n:nat} (f:seg_fun n U) (y:U),
    Finite (inv_im_seg f y).
  intros U n f y.
  unfold inv_im_seg.
  eapply Finite_downward_closed.
  apply finite_seg.
  red.
  intros i h1.
  destruct h1 as [h1].
  destruct h1; subst.
  constructor.
  apply x.
Qed.


(*Wyckoff*)
Lemma inv_im_surj_inj : 
  forall {T U:Type} (f:T->U),
    surjective f ->
    forall (A B:Ensemble U),
      inv_im A f = inv_im B f->
      A = B.
intros T U f h1 B C h2.
red in h1.
apply Extensionality_Ensembles.
red. split.
red. intros y h3.
specialize (h1 y).
destruct h1 as [x h1r].
subst.
assert (h4:In (inv_im B f) x).
  constructor; auto.
rewrite h2 in h4.
destruct h4.
assumption.
red. intros y h3.
specialize (h1 y).
destruct h1 as [x h1r]. subst.
assert (h4:In (inv_im C f) x).
  constructor; auto.
rewrite <- h2 in h4.
destruct h4.
assumption.
Qed.

(*Wyckoff*)
Lemma inv_im1_surj_inj : 
  forall {T U:Type} (f:T->U)
         (y z:U),
    surjective f ->
    inv_im1 f y = inv_im1 f z ->
    y = z.
  intros T U f y z h1 h2.
  pose proof (inh_surj f y h1) as h3.
  destruct h3 as [x h3].
  destruct h3 as [h3]; subst.
  pose proof (in_im_inv_im1 f x) as h3.
  rewrite h2 in h3.
  destruct h3; auto.
Qed.

(*Wyckoff*)
Lemma inv_im_seg_surj_inj : 
  forall {U:Type} (n:nat)
         (f:seg_fun n U)
         (y z:U),
    surj_seg f ->
    inv_im_seg f y = inv_im_seg f z ->
    y = z.
  intros U n f y z h1 h2.
  pose proof (inh_surj_seg _ y h1) as h3.
  destruct h3 as [m h4].
  rewrite h2 in h4.
  destruct h4 as [h4].
  destruct h4 as [h4 h5]; subst.
  pose proof (in_inv_im_seg f m h4) as h5.
  rewrite <- h2 in h5.
  destruct h5.
  destruct H; subst.
  f_equal.
  apply proof_irrelevance.
Qed.


(*Wyckoff*)
Lemma inv_im_inj_sing : 
  forall {T U:Type} (f:T->U),
    injective f -> 
    forall y:U,
      {exists! x:T, f x = y} +
      {inv_im (Singleton y) f = Empty_set _}.
intros T U f h1 y.
destruct (classic_dec (exists! x:T, f x = y)) as [h2 | h3].
left; auto. right.
apply Extensionality_Ensembles; red; split; auto with sets.
red. intros x h4. destruct h4 as [h4]. inversion h4; clear h4.
subst.
contradict h3.
exists x.
red; split; auto.
Qed.

(*Wyckoff*)
Lemma inv_im1_inj_sing : 
  forall {T U:Type} (f:T->U),
    injective f -> 
    forall y:U,
      {exists! x:T, f x = y} +
      {inv_im1 f y = Empty_set _}.
  intros; rewrite inv_im1_compat'; apply inv_im_inj_sing; auto.
Qed.


(*Wyckoff*)
Lemma inv_im_inj_sing' : 
  forall {T U:Type} (f:T->U),
    injective f -> 
    forall y:U,
      card_fun1 (inv_im (Singleton y) f) <= 1.
intros T U f h1 y.
pose proof (inv_im_inj_sing f h1 y) as h2.
destruct h2 as [h2a | h2b].
destruct h2a as [x h3]. red in h3. destruct h3 as [? h3]. subst.
assert (h4:inv_im (Singleton (f x)) f = Singleton x).
  apply Extensionality_Ensembles.
  red. split. red. intros x' h4.
  destruct h4 as [h4].
  inversion h4 as [h5]. clear h4.
  symmetry in h5. apply h3 in h5. subst. constructor.
  red. intros x' h4. destruct h4. constructor. 
  constructor. 
rewrite h4.
rewrite card_fun1_sing. auto with arith.
rewrite h2b.
rewrite card_fun1_empty. auto with sets.
Qed.



(*Wyckoff*)
Lemma inv_im1_inj_sing' : 
  forall {T U:Type} (f:T->U),
    injective f -> 
    forall y:U,
      card_fun1 (inv_im1 f y) <= 1.
  intros; rewrite inv_im1_compat'; apply inv_im_inj_sing'; auto.
Qed.

(*Wyckoff*)
Lemma left_inverse_ex :
  forall {T U:Type} (f:T->U),
    injective f ->
    forall (A:Ensemble T),
      exists! f':sig_set (Im A f)->T,
        forall (x:T) (pf:In A x),
          f' (exist _ _ (Im_intro _ _ A f _ pf _ (eq_refl _))) = x.
intros T U f h1 A.
assert (h2:forall y:sig_set (Im A f), exists! x:T, f x = proj1_sig y).
  intro y.
  pose proof (inv_im_inj_sing _  h1 (proj1_sig y)) as h2.
  destruct h2 as [h2l | h2r]; auto.
  destruct y as [y h3]. destruct h3 as [x h3]. subst.
  simpl in h2r. simpl.
  contradict h2r.
  apply Inhabited_not_empty.
  apply Inhabited_intro with x. constructor. constructor.
  exists (fun x =>
            (proj1_sig
               (constructive_definite_description _ (h2 x)))). red. split.
intros x h3. destruct constructive_definite_description as [x' h5].
simpl in h5.
simpl.
apply h1.
assumption.
intros f' h3.
apply functional_extensionality.
intro x.
destruct constructive_definite_description as [x' h4].
simpl.
destruct x as [x h5]. simpl in h4.
destruct h5 as [x h5]. subst.
apply h1 in h4.
subst.
symmetry.
apply h3.
Qed.

(*Wyckoff*)
Definition left_inverse {T U:Type} (f:T->U)
           (pf:injective f) (A:Ensemble T) : sig_set (Im A f)-> T :=
  proj1_sig (constructive_definite_description
               _
               (left_inverse_ex f pf A)).


(*Wyckoff*)
Lemma left_inverse_compat : 
  forall {T U:Type} (f:T->U) (pf:injective f)
         (A:Ensemble T),
      let f' := left_inverse f pf A in
      forall (x:T) (pf:In A x),
        f' (exist _ _ (Im_intro _ _ A f _ pf _ (eq_refl _))) = x.
intros T U f h1 S f'.
unfold f'. unfold left_inverse.
destruct constructive_definite_description as [f'' h2].
simpl.
apply h2.
Qed.


(*Wyckoff*)
Definition left_inverse_full 
           {T U:Type} (f:T->U) (pf:injective f) :
  sig_set (Im (Full_set T) f) -> T :=
  left_inverse f pf (Full_set T).

(*Wyckoff*)
Lemma left_inverse_full_compat : 
  forall {T U:Type} (f:T->U) (pf:injective f) x,
    left_inverse_full
      f pf
      (exist (fun c => Inn (Im (Full_set _) f) c)  _ (in_im x f)) = x.
  intros T U f h1 x.
  rewrite <- (left_inverse_compat f h1 (Full_set _) x (Full_intro _ x)).
  unfold left_inverse_full; f_equal; apply proj1_sig_injective; simpl.
  f_equal.
  rewrite left_inverse_compat; auto.
Qed.



(*Wyckoff*)
Lemma left_inverse_inj : 
  forall {T U:Type} (f:T->U) (pf:injective f)
         (A:Ensemble T),
    injective (left_inverse f pf A).
intros T U f h1 A.
red.
intros y y' h0.
destruct y as [y h2], y' as [y' h3].
pose proof (left_inverse_compat f h1 A) as h4.
simpl in h4.
destruct h2 as [x h2], h3 as [x' h3]. subst.
do 2 rewrite h4 in h0.
subst.
apply proj1_sig_injective. simpl.
reflexivity.
Qed.


(*Wyckoff*)
Lemma incl_im_full_sig_im_left_inverse :
  forall {T U:Type} (f:T->U) (pf:injective f) (A:Ensemble T),
    Included (Im (full_sig (Im A f)) (left_inverse f pf A)) A.
intros T U f h1 A.
red. intros x h2.
destruct h2 as [x h2]. subst. clear h2.
destruct x as [x h2].
destruct h2 as [x h2]. subst.
rewrite left_inverse_compat.
assumption.
Qed.




(*Wyckoff*)
Lemma inv_im_inj_im_compat : 
  forall {T U:Type} (f:T->U)
         (pf:injective f) (B:Ensemble U),
    inv_im B f = Im [u:sig_set (Im (Full_set T) f) |
                     In B (proj1_sig u)] (left_inverse_full f pf).
intros T U f h1 B.
apply Extensionality_Ensembles.
red. split.
red. intros x h2. 
destruct h2 as [h2].
apply Im_intro with
(exist _ _ (Im_intro _ _ (Full_set T)
                     f _ (Full_intro _ x) _ (eq_refl _))).
constructor; simpl; auto. 
unfold left_inverse_full.
pose proof (left_inverse_compat f h1 (Full_set T)) as h3.
simpl in h3.
rewrite  h3.
reflexivity.
red.
intros x h2.
destruct h2 as [y h2]. subst.
destruct y as [y h3]. destruct h3 as [x h3]. subst. 
inversion h2 as [h4]. clear h2. simpl in h4.
constructor.
pose proof (left_inverse_compat f h1 (Full_set T)) as h3'.
simpl in h3'.
unfold left_inverse_full. rewrite  h3'.
assumption.
Qed.

(*Wyckoff*)
Lemma finite_im_full_set_f_in_b : 
  forall {T U:Type} (B:Ensemble U),
    Finite B ->
    forall (f:T->U)
           (pf:injective f),
      Finite  (Im [u : sig_set (Im (Full_set T) f) | In B (proj1_sig u)]
        (left_inverse f pf (Full_set T))).
intros T U B h1.
induction h1 as [|B h2 h3 b h4].
intros.
gterm0.
rename c into E.
assert (h1:E = Empty_set _).
  apply Extensionality_Ensembles; red; split; auto with sets.
  red. intros x h2. destruct h2 as [x h2]. destruct h2 as [h2].
  contradiction.
subst.
rewrite h1.
constructor.

intros f h1. 
gterm0.
redterm1 c.
clear c.
rename c0 into E.
assert (h7: E = 
            Union  [u : sig_set (Im (Full_set T) f) |
                    In B (proj1_sig u)]
                   [u : sig_set (Im (Full_set T) f) | proj1_sig u = b]).
  apply Extensionality_Ensembles.
  red. split.
  red. intros x h7. destruct h7 as [h7]. inversion h7 as [b' h8 | b' h9]; clear h7.
  subst. left. constructor. assumption.
  inversion h9. subst.
  right. constructor. reflexivity.
  red.
  intros x h7.
  destruct h7 as [x h7l | x h7r]. destruct h7l as [h7l].
  constructor. left. assumption.
  destruct h7r as [h7r]. unfold E; clear E; subst.
  constructor. right. constructor.
unfold E in h7.
clear E.
rewrite h7.
rewrite im_union.
apply Union_preserves_Finite.
apply h3.
pose proof (inv_im_inj_im_compat _ h1 (Singleton b)) as h8.
assert (h9:Im [u : sig_set (Im (Full_set T) f) |
               In (Singleton b) (proj1_sig u)]
              (left_inverse_full f h1) =
           Im [u : sig_set (Im (Full_set T) f) | proj1_sig u = b]
              (left_inverse f h1 (Full_set T))).
f_equal. apply Extensionality_Ensembles. red. split.
  red. intros x h9. destruct h9 as [h9]. inversion h9. subst.
  constructor. reflexivity.
  red. intros x h9. destruct h9 as [h9]. subst. constructor. constructor.
rewrite <- h9.
rewrite <- h8.
pose proof (inv_im_inj_sing' _ h1 b) as h10.
apply le_lt_eq_dec in h10.
destruct h10 as [h10l | h10r].
assert (h11:card_fun1 (inv_im (Singleton b) f) = 0). omega.
apply card_fun1_O in h11.
destruct h11 as [h11l | h11r].
rewrite h11l. constructor. 
pose proof (approximant_can_be_any_size _ (inv_im (Singleton b) f)(inv_im (Singleton b) f)   h11r 2) as h12.
destruct h12 as [X h13].
destruct h13 as [h13l h13r].
destruct h13r as [h14 h15].
apply cardinal_2 in h13l.
destruct h13l as [x [y [h13a h13b]]]. subst.
pose proof (h15 _ (Couple_l _ x y)) as h16.
pose proof (h15 _ (Couple_r _ x y)) as h17.
destruct h16 as [h16], h17 as [h17]. 
inversion h16; subst; clear h16.
inversion h17; subst; clear h17.
apply h1 in H0.
contradiction.
apply card_fun1_1 in h10r.
destruct h10r as [x' h10r].
rewrite h10r.
apply Singleton_is_finite.
Qed.

Inductive inv_im_eq {T U V W} (f:T->U) (g:V->W) : Prop :=
  inv_im_eq_intro : forall (h:fun_in_ens' (inv_im1s f) (inv_im1s g)),
                      invertible_in_ens h ->
                      (forall S (pfs:Inn (inv_im1s f) S),
                       exists (k:fun_in_ens' S (proj1_sig (h S pfs))),
                         invertible_in_ens k) ->
                      inv_im_eq f g.



(*Wyckoff*)
Definition full_wrap_fun1 {T U:Type} (f:T -> U) :
  sig_set (Full_set T) -> U :=
  (fun x  => f (proj1_sig x)).

(*Wyckoff*)
Definition full_wrap_fun2 {T U:Type} (f:T -> U) :
  T ->
  sig_set (Full_set U) := (fun x => (exist _ _ (Full_intro _ (f x)))).

(*Wyckoff*)
Lemma inj_full_wrap_fun1_iff :
  forall {T U:Type} (f:T->U),
    injective f <-> injective (full_wrap_fun1 f).
  intros T U f.
  split.
  intro h1.
  red in h1. red.
  intros x y h4. destruct x as [x h2], y as [y h3].
  unfold full_wrap_fun1  in h4.
  simpl in h4.
  specialize (h1 x y h4).
  subst.
  apply proj1_sig_injective; auto.
  intro h1.
  red in h1. red.
  intros x y h4.
  specialize (h1 (exist _ _ (Full_intro _ x))
                 (exist _ _ (Full_intro _ y))).
  unfold full_wrap_fun1 in h1. simpl in h1.
  specialize (h1 h4). apply exist_injective in h1. subst.
  reflexivity.
Qed.

(*Wyckoff*)
Lemma surj_full_wrap_fun1_iff :
  forall {T U:Type} (f:T->U),
    surjective f <-> surjective (full_wrap_fun1 f).
  intros T U f.
  split.
  intro h1. red in h1. red.
  intro y. specialize (h1 y).
  destruct h1 as [x h2]. subst.
  exists (exist _ _ (Full_intro _ x)).
  unfold full_wrap_fun1. simpl. reflexivity.
  intro h1. red in h1. red.
  intro y. specialize (h1 y).
  destruct h1 as [x h2]. destruct x as [x h3].
  unfold full_wrap_fun1 in h2.
  simpl in h2. subst.
  exists x; auto.
Qed.

(*Wyckoff*)
Lemma bij_full_wrap_fun1_iff :
  forall {T U:Type} (f:T->U),
    bijective f <-> bijective (full_wrap_fun1 f).
  intros T U f.
  unfold bijective.
  rewrite inj_full_wrap_fun1_iff, surj_full_wrap_fun1_iff.
  tauto.
Qed.

(*Wyckoff*)
Lemma inj_full_wrap_fun2_iff :
  forall {T U:Type} (f:T->U),
    injective f <-> injective (full_wrap_fun2 f).
  intros T U f.
  split.
  intro h1.
  red in h1. red.
  intros x y h4. 
  unfold full_wrap_fun2  in h4.
  apply exist_injective in h4.
  specialize (h1 x y h4).
  assumption.
  intro h1.
  red in h1. red.
  intros x y h4.
  specialize (h1 x y).
  unfold full_wrap_fun2 in h1. simpl in h1.
  pose proof (f_equal (fun x => exist _ _ (Full_intro _ x)) h4) as h5.
  simpl in h5.
  specialize (h1 h5).
  assumption.
Qed.

(*Wyckoff*)
Lemma surj_full_wrap_fun2_iff :
  forall {T U:Type} (f:T->U),
    surjective f <-> surjective (full_wrap_fun2 f).
  intros T U.
  split.
  intro h1. red in h1. red.
  intro y. destruct y as [y h2].
  specialize (h1 y). destruct h1 as [x h3].
  subst.
  exists x.
  unfold full_wrap_fun2. apply proj1_sig_injective; auto.
  intro h1. red in h1. red.
  intro y. specialize (h1 (exist _ _ (Full_intro _ y))).
  destruct h1 as [x h2].
  unfold full_wrap_fun2 in h2.
  apply exist_injective in h2.
  subst.
  exists x; auto.
Qed.

(*Wyckoff*)
Lemma bij_full_wrap_fun2_iff :
  forall {T U:Type} (f:T->U),
    bijective f <-> bijective (full_wrap_fun2 f).
  intros T U f.
  unfold bijective.
  rewrite inj_full_wrap_fun2_iff, surj_full_wrap_fun2_iff.
  tauto.
Qed.


(*Wyckoff*)
Definition sig_fun_app {T U:Type} {A:Ensemble T}
           (f:sig_set A->U) (def:U) : T->U.
intro x.
destruct (classic_dec (In A x)) as [h1 | h2].
refine (f (exist _ _ h1)).
refine def.
Defined.



(*Wyckoff*)
(*Uses LEM*)
Lemma im_sig_eq :
  forall {T U:Type} {P:T->Prop} (S:Ensemble {x | P x})
         (f:{x | P x}-> U) (def:U),
    Im S f = Im [x:T | In (im_proj1_sig S) x /\ P x] (sig_fun_app f def).
  intros T U P S f def.
  unfold sig_fun_app.
  apply Extensionality_Ensembles; red; split; red; intros x h1.
  destruct h1 as [x h1]. subst.
  apply Im_intro with (proj1_sig x).
  constructor. split. rewrite in_im_proj1_sig_iff; auto. apply proj2_sig; auto.
  destruct classic_dec as [h2 | h2].
  f_equal. apply proj1_sig_injective; auto.
  contradict h2. apply proj2_sig. 
  destruct h1 as [x h1]. subst.
  destruct h1 as [h1]. destruct h1 as [h1 h2].
  destruct h1 as [x h1]. subst.
  apply Im_intro with x; auto.
  lr_match_goal'.
  f_equal. apply proj1_sig_injective; auto.
  contradiction.
Qed.  

(*Wyckoff*)
Lemma im_full_eq :
  forall {T U:Type} {P:T->Prop} (f:{x | P x}-> U) (def:U),
    Im (Full_set {x | P x}) f = Im [x:T |  P x] (sig_fun_app f def).
  intros T U P f def.
  unfold sig_fun_app.
  apply Extensionality_Ensembles; red; split; red; intros x h1.
  destruct h1 as [x h1]. subst.
  clear h1.
  apply Im_intro with (proj1_sig x).
  constructor. apply proj2_sig.
  lr_match_goal'.
  f_equal. apply proj1_sig_injective; auto.
  contradict hr. apply proj2_sig.
  destruct h1 as [x h1]. subst. destruct h1 as [h1].
  apply Im_intro with (exist _ _ h1). constructor.
  lr_match_goal'.
  f_equal. apply proj1_sig_injective; auto.
  contradiction.
Qed.




(*Wyckoff*)
Definition sig_fun_injectors {T U:Type} {A:Ensemble T} 
           (f:sig_set A->U) : Family T :=
  [B:Ensemble T | exists pf:Included B A,
                  injective (restriction_sig f B pf) /\
                  Inn (same_im_subsets_sig f) B].

Lemma incl_sig_fun_injectors :
  forall {T U:Type} {A B:Ensemble T} (f:sig_set A->U),
    In (sig_fun_injectors f) B ->
    Included B A.
intros T U A B f h1.
destruct h1 as [h1]. destruct h1; auto.
Qed.


(*Wyckoff*)
Definition min_sig_fun_injectors
  {T U:Type} {A:Ensemble T}
  (f:sig_set A->U) : Ensemble T :=
  FamilyIntersection (sig_fun_injectors f).


(*Wyckoff*)
Lemma incl_min_sig_fun_injectors : 
  forall  {T U:Type} {A:Ensemble T}
  (f:sig_set A->U),
    Inhabited (sig_fun_injectors f) ->
    Included (min_sig_fun_injectors f) A.
intros T U A f h0.
red. 
intros x h1.
destruct h1 as [x h1].
destruct h0 as [B h0].  
specialize (h1 _ h0).
destruct h0 as [h0].
destruct h0 as [h0 [h2 h3]].
apply h0; auto.
Qed.


(*Wyckoff*)
Lemma incl_min_sig_fun_injectors' : 
  forall  {T U:Type} {A:Ensemble T}
  (f:sig_set A->U)
  (B:Ensemble T),
    In (sig_fun_injectors f) B ->
    Included (min_sig_fun_injectors f) B.
intros T U A f p h1.
red. intros x h2.
destruct h2 as [x h2].
apply h2; auto.
Qed.

(*Wyckoff*)
Lemma in_sig_fun_injectors_bij : 
  forall {T U:Type} {A:Ensemble T}
         (f:sig_set A->U)
         (B:Ensemble T)
         (pf:In (sig_fun_injectors f) B),
    let f' := (fun x:sig_set A => (exist _ _ (Im_intro _ _ (Full_set _) f _ (Full_intro _ x) _ (eq_refl _)))) in
    bijective (restriction_sig f' _  (incl_sig_fun_injectors f pf )).
  intros T U A f B h1 f'.
  destruct h1 as [h1].
  destruct h1 as [h1 h2]. destruct h2 as [h2 h3].
  red. split.
  red; intros x y h4.
  unfold f', restriction_sig in h4. simpl in h4.
  apply exist_injective in h4.
  red in h2.
  unfold restriction_sig in h2.
  pose proof (h2 x y) as h2'.
  terml h4.
  termr h4.
  redterm0 x0.
  redterm0 c.
  redterm0 y0.
  redterm0 c1.
  clear c c1 x0 y0.
  pose proof (proof_irrelevance _ (h1 _ (proj2_sig x)) c0) as h5.
  pose proof (proof_irrelevance _ (h1 _ (proj2_sig y)) c2) as h6.
  unfold c0 in h5; unfold c2 in h6.
  rewrite h5, h6 in h2'.
  specialize (h2' h4). assumption.
  rewrite surj_iff.
  destruct h3 as [h3]. destruct h3 as [h3 h4].
  apply im_proj1_sig_injective.
  rewrite im_proj1_sig_full_set_sig.
  unfold full_sig in h4.
  rewrite h4 at 4.
  unfold im_proj1_sig.
  rewrite im_im.
  apply im_ext_in.
  intros x ?.
  unfold restriction_sig, f'. simpl.
  f_equal. apply proj1_sig_injective; auto.
Qed.

(*Wyckoff*)
Lemma inj_restriction_sig :
  forall {T U:Type} (A B C:Ensemble T) (f:sig_set A->U)
         (pfbc:Included B C) (pfca:Included C A),
      injective (restriction_sig f C pfca) ->
      injective (restriction_sig f B (inclusion_transitive _ _ _ _ pfbc pfca)).
  intros T U A B C f h1 h2 h3.
  red in h3. red.
  intros x y h4.
  destruct x as [x h5], y as [y h6].
  pose proof (h1 _ h5) as h5'. pose proof (h1 _ h6) as h6'.
  specialize (h3 (exist _ _ h5') (exist _ _ h6')).
  unfold restriction_sig in h4. simpl in h4.
  unfold restriction_sig in h3. simpl in h3.
  pose proof (proof_irrelevance
                _ 
                (h2 x h5')
                (inclusion_transitive T B C A h1 h2 x h5)) as h7.
  pose proof (proof_irrelevance
                _ (h2 y h6')
                (inclusion_transitive T B C A h1 h2 y h6)) as h8.
  rewrite <- h7, <- h8 in h4.
  specialize (h3 h4).
  apply proj1_sig_injective. simpl.
  apply exist_injective in h3.
  assumption.
Qed.

(*Wyckoff*)
Lemma im_full_sig_sub_decompose_inj :
  forall {T U:Type} (pfdt:type_eq_dec T)
         (E:Ensemble T) (x:T) (f:sig_set E ->U)
         (pfe:Inn E x),
     let f' := restriction_sig f _ (incl_subtract _ _) in
     Im (full_sig E) f = Add (Im (full_sig (Subtract E x)) f')
                             (f (exist _ _ pfe)).
  intros T U h0 E x f h1. simpl.
  apply Extensionality_Ensembles; red; split; red; intros c h3.
  destruct h3 as [c h3]; subst.
  clear h3. destruct c as [c h3].
  destruct (h0 x c) as [h4 | h4]; subst.
  right.
  pose proof (proof_irrelevance _ h3 h1); subst.
  constructor.
  left.
  rewrite <- in_sing_iff in h4.
  apply Im_intro with
  (exist (fun c => In (Subtract E x) c) c (conj h3 h4)).
  constructor.
  unfold restriction_sig. simpl.
  f_equal. f_equal. apply proof_irrelevance.

  destruct h3 as [c h3 | c h3]; subst.
  destruct h3 as [c h3]; subst.
  clear h3. destruct c as [c h3]. destruct h3 as [h3 h4].
  apply Im_intro with (exist _ _ h3).
  constructor.
  unfold restriction_sig.
  f_equal. f_equal. apply proof_irrelevance.
  destruct h3.
  apply Im_intro with (exist _ _ h1). constructor.
  reflexivity.
Qed.


Section EquivalenceClass.

(*Wyckoff*)
Definition rel_classes 
           {T:Type} (A:Ensemble T) (R:T->T->Prop) :=
  Im A (fun x:T => [y:T | In A y /\ R x y]).

(*Wyckoff*)
Lemma rel_classes_empty :
  forall {T:Type} (R:T->T->Prop),
    rel_classes (Empty_set _) R = Empty_set (Ensemble T).
  intros T R.
  unfold rel_classes.
  rewrite image_empty.
  reflexivity.
Qed.

(*Wyckoff*)
Definition rep_to_eq_class
           {T:Type} (A:Ensemble T) (R:T->T->Prop) (a:T) (pf:In A a) :=
           [y:T | R a y].

(*Wyckoff*)
Lemma refl_rel_classes_inhabited : 
  forall {T:Type} (A:Ensemble T),
    Inhabited A ->
  forall (R:T->T->Prop),
    Reflexive R ->
    inhabited_family (rel_classes A R).
intros T A h1 R h2.
constructor.
destruct h1 as [a h1].
specialize (h2 a).
unfold rel_classes.
apply Inhabited_intro with [y : T | In A y /\ R a y].
apply Im_intro with a; auto.
intros A' h3.
destruct h3 as [a' h3]. subst.
apply Inhabited_intro with a'.
constructor.
specialize (h2 a').
split; auto.
Qed.

(*wyckoff*)
Lemma equiv_classes_inhabited : 
  forall {T:Type} (A:Ensemble T),
    Inhabited A ->
  forall (R:T->T->Prop),
    Equivalence R ->
    inhabited_family (rel_classes A R).
intros T A h1 R h2.
destruct h2 as [h2].
apply refl_rel_classes_inhabited; auto.
Qed.



(*Wyckoff*)
Definition partition_fam {T:Type} (F:Family T) (A:Ensemble T) : Prop 
  :=
    pairwise_disjoint F /\
    FamilyUnion F = A.

(*Wyckoff*)
Definition partition_fam_impl_pairwise_disjoint :
  forall {T:Type} {F:Family T},
    pairwise_disjoint F -> partition_fam F (FamilyUnion F).
  unfold partition_fam; intros; auto.
Qed.


(*Wyckoff*)
Lemma partition_fam_empty :
  forall {T:Type} (A:Ensemble T),
    partition_fam (Empty_set (Ensemble T)) A ->
    A = Empty_set _.
  intros T A h1.
  red in h1.
  destruct h1 as [h1 h2].
  rewrite empty_family_union in h2; auto.
Qed.

(*Wyckoff*)
Lemma partition_fam_add :
  forall {T:Type} (F:Family T) (A B:Ensemble T),
    ~In F A ->
    partition_fam (Add F A) B ->
    partition_fam F (Setminus B A).
  intros T F A B h0 h1.
  red in h1. red.
  destruct h1 as [h1 h2].
  pose proof (pairwise_disjoint_add_family_union F A h0 h1) as h3.
  apply pairwise_disjoint_add_rev in h1.
  split; auto.
  subst.
  rewrite setminus_family_union.
  rewrite disjoint_setminus_compat; auto.
Qed.

(*Wyckoff*)
Lemma partition_fam_eq :
  forall {T:Type} (F:Family T) (A B C:Ensemble T) (a:T),
    partition_fam F A ->
    In F B -> In F C ->
    In B a -> In C a ->
    B = C.
  intros T F A B C a h1 h2 h3 h4 h5.
  red in h1.
  destruct h1 as [h1 h6].
  subst.
  red in h1.
  apply NNPP.
  intro h6.
  specialize (h1 _ _ h6 h2 h3).
  red in h1.
  contradict h1.
  intro h7.
  pose proof (Intersection_intro _ _ _ _ h4 h5) as h8.
  rewrite h7 in h8.
  contradiction.
Qed.

(*Wyckoff*)
Lemma partition_fam_dec :
  forall {T:Type} (F:Family T) (A B U:Ensemble T),
    partition_fam F U ->
    In F A -> In F B ->
    {A = B} + {disjoint A B}.
  intros T F A B U h1 h2 h3.
  red in h1. destruct h1 as [h1 h4].
  subst.
  red in h1.
  destruct (eq_dec A B) as [|h4]. left; auto.
  right. apply h1; auto.
Qed.


(*Wyckoff*)
Lemma rep_to_partition_class_ex :
  forall {T:Type} (F:Family T) (pf:Finite F) (A:Ensemble T),
    partition_fam F A ->
    forall a:T,
      Inn A a ->
  exists! S:Ensemble T,
    In F S /\ In S a.
  intros T F h1.
  induction h1 as [|F h2 h3 A h4]. 
  intros A h1 a h2.
  apply partition_fam_empty in h1. subst.
  contradiction.
  intros B h5 a h6.
  pose proof (partition_fam_add _ _ _ h4 h5) as h5'.  
  red in h5.
  destruct h5 as [h5 h9]. subst.
  rewrite family_union_add in h6.
  destruct h6 as [a h6 | a h6].
  exists A.
  red. split. split; auto. right. constructor.
  intros S h9.
  destruct h9 as [h9 h10].
  red in h5.
  destruct h9 as [S h9 | S h9].
  specialize (h5 S A).
  assert (h11:S <>  A). 
    intro h11. subst. contradiction.
  pose proof (Add_intro1 _ _ A _ h9) as h12.
  pose proof (Add_intro2 _  F A) as h13.
  specialize (h5 h11 h12 h13).
  red in h5.
  pose proof (Intersection_intro _ _ _ _ h10 h6) as h14.
  rewrite h5 in h14.
  contradiction.
  destruct h9; auto.
  rewrite setminus_family_union in h5'.
  red in h5'.
  destruct h5' as [h7 h8].
  destruct h6 as [S a h6 h0].
  exists S.
  red. split. split.
  left; auto. assumption.
  intros S' h9.
  destruct h9 as [h9 h10].
  destruct h9 as [S' h9 | S' h9].
  apply NNPP.
  intro h11.
  red in h7.
  specialize (h7 _ _ h11 h6 h9).
  red in h7.
  pose proof (Intersection_intro _ _ _ _ h0 h10) as h12.
  rewrite h7 in h12.
  contradiction.
  destruct h9.
  pose proof (family_union_intro _ F _ _ h6 h0) as h11.
  rewrite h8 in h11.
  destruct h11; contradiction.
Qed.

(*Wyckoff*)
Definition rep_to_partition_class
           {T:Type} (F:Family T) (pff:Finite F) (A:Ensemble T)
           (pfp:partition_fam F A) (a:T)
           (pfa:In A a) : Ensemble T :=
  proj1_sig (constructive_definite_description
               _ (rep_to_partition_class_ex F pff A pfp a pfa)).



(*Wyckoff*)
Lemma rep_to_partition_class_compat :
  forall {T:Type} (F:Family T) (pff:Finite F) (A:Ensemble T)
         (pfp:partition_fam F A) (a:T)
         (pfa:In A a),
    let S := rep_to_partition_class F pff A pfp a pfa in
    In F S /\ In S a.
  unfold rep_to_partition_class; intros;
  destruct constructive_definite_description; auto.
Qed.


(*Wyckoff*)
Lemma finite_rep_to_partition_class :
  forall {T:Type} (F:Family T) (pff:Finite F) (A:Ensemble T)
         (pfp:partition_fam F A) (a:T)
         (pfa:In A a),
    Finite A ->
    Finite (rep_to_partition_class F pff A pfp a pfa).
  intros T F h1 A h2 a h4 h5.
  unfold rep_to_partition_class.
  destruct constructive_definite_description as [B h6]. simpl.
  red in h2. destruct h2 as [h2 h7].
  subst.
  eapply finite_family_union_finite_mem.
  apply h5.
  destruct h6; auto.
Qed.
  


(*Wyckoff*)
Lemma in_rep_to_partition_class :
  forall {T:Type} (F:Family T) (pff:Finite F) (A:Ensemble T)
         (pfp:partition_fam F A) (a b:T)
         (pfa:Inn A a),
    let S := rep_to_partition_class F pff A pfp a pfa in
    In S b -> In A b.
  intros T F h1 A h2 a b h3 S h4.
  red in h2. destruct h2 as [h2]. unfold S in h4.  clear S. subst.
  pose proof (rep_to_partition_class_compat
                F h1
                (FamilyUnion F) (conj h2 eq_refl) a h3) as h5.
  simpl in h5.
  destruct h5 as [h5 h6].
  eapply family_union_intro.
  apply h5. assumption.
Qed.

(*Wyckoff*)
Lemma rep_to_partition_class_functional :
  forall {T:Type} (F:Family T) (pff pff':Finite F) (A:Ensemble T)
         (pfp pfp':partition_fam F A) (a a':T)
         (pfa:Inn A a) (pfa':Inn A a'),
    a = a' ->
    rep_to_partition_class F pff  A pfp  a  pfa =
    rep_to_partition_class F pff' A pfp' a' pfa'.
  intros T F h1 h2 A h3 h4 a a' h5 h6 h7.
  subst.
  pose proof (proof_irrelevance _ h1 h2).
  pose proof (proof_irrelevance _ h5 h6).
  pose proof (proof_irrelevance _ h3 h4).
  subst; auto.
Qed.

(*Wyckoff*)
Lemma in_rep_to_partition_class_eq_iff :
  forall {T:Type} (F:Family T) (pff pff':Finite F) (A:Ensemble T)
         (pfp pfp':partition_fam F A) (a a':T)
         (pfa:In A a) (pfa':In A a'),
    Inn (rep_to_partition_class F pff A pfp a pfa) a' <->
    rep_to_partition_class F pff  A  pfp  a  pfa =
    rep_to_partition_class F pff' A pfp' a' pfa'.
  intros T F h1 h2 A h3 h4 a a' h5 h6.
  unfold rep_to_partition_class.
  destruct constructive_definite_description as [C h9].
  destruct constructive_definite_description as [D h10]. 
  simpl. 
  destruct h9 as [h9 h11], h10 as [h10 h12].
  split.
  intro h8.
  eapply partition_fam_eq. apply h3. assumption. assumption.
  apply h8. apply h12.
  intro. subst.
  assumption.
Qed.


(*Wyckoff*)
Lemma equiv_classes_partition_fam : 
  forall {T:Type} (A:Ensemble T) (R:T->T->Prop),
    Equivalence R -> partition_fam (rel_classes A R) A.
intros T A R h1.
destruct h1 as [h1a h1b h1c]. 
red. split. 
red. intros S S' h4 h2 h3.
destruct h2 as [a h2 S]. subst. destruct h3 as [a' h3 S']. subst.
red.
assert (h0:~R a a').
  intro h5.
  contradict h4.
  apply Extensionality_Ensembles.
  red. split.
  red. intros x h6. destruct h6 as [h6]. destruct h6 as [h6a h6b].
  apply h1b in h6b. specialize (h1c _ _ _ h6b h5).
  constructor; auto.
  red. intros x h6. destruct h6 as [h6]. destruct h6 as [h6a h6b].
  apply h1b in h6b. apply h1b in h5. specialize (h1c _ _ _ h6b h5).
  constructor; auto.
apply Extensionality_Ensembles; red; split; auto with sets. 
red. intros x h5.
destruct h5 as [x h5 h6]. destruct h5 as [h5]. destruct h6 as [h6].
destruct h5 as [h5a h5b]. destruct h6 as [h6a h6b].
apply h1b in h6b.
pose proof (h1c _ _ _ h5b h6b) as h7.
contradiction.
apply Extensionality_Ensembles.
red. split.
red. intros x h2.
destruct h2 as [S x h2 h3]. destruct h2 as [y h2 S]. subst.
destruct h3 as [h3]. destruct h3; auto.
red.
intros x h2.  
apply family_union_intro with [y:T |In A y /\ R x y].
unfold rel_classes.
apply Im_intro with x; auto.
specialize (h1a x).
constructor; auto.
Qed.


(*Wyckoff*)
Lemma partition_fam_S :
  forall (F:Family nat) (A:Ensemble nat),
    Finite F -> partition_fam F A->
    partition_fam (fam_S F) (ens_S A).
  intros F A h1 h2.
  red in h2. red. 
  destruct h2 as [h2 h3].
  subst.
  red in h2.
  split. red.
  intros A B h3 h4 h5.
  destruct h4 as [A h4]. destruct h5 as [B h5]. subst.
  assert (h6:A <> B).
    intro h6. subst.
    contradict h3; auto.
  specialize (h2 _ _ h6 h4 h5).
  red in h2. red.
  rewrite intersection_ens_S.
  rewrite h2.
  apply image_empty.
  unfold ens_S, fam_S.
  rewrite <- family_union_im.
  unfold ens_S.
  reflexivity.
Qed.

(*Wyckoff*)
Lemma rep_to_partition_class_fam_S_eq :
  forall (F:Family nat) (pff:Finite F) (A:Ensemble nat)
         (pfp:partition_fam F A) (n:nat)
         (pfn:Inn A n),
    rep_to_partition_class (fam_S F)
                           (finite_fam_S  _ pff)
                                    (ens_S A)
                                    (partition_fam_S _ _ pff pfp)
                                    (S n) (in_ens_S _ _ pfn) =
    ens_S (rep_to_partition_class F pff A pfp n pfn).
  intros F h1 A h2 n h3.
  unfold rep_to_partition_class, ens_S.
  destruct constructive_definite_description as [B h4]. 
  destruct constructive_definite_description as [C h5]. simpl.
  destruct h4 as [h4 h6], h5 as [h5 h7].
  destruct h4 as [B h4]. subst.
  inversion h6 as [m h11 p h12]. subst.
  apply S_inj in h12. subst.
  clear h6.
  pose proof (rep_to_partition_class_ex F h1 A h2 m h3) as h12.
  destruct h12 as [S h12].
  red in h12. destruct h12 as [h12 h13].
  pose proof (h13 _ (conj h5 h7)). subst.
  pose proof (h13 _ (conj h4 h11)). subst.
  reflexivity.
Qed.


(*Wyckoff*)
Lemma rep_to_partition_class_fam_S_eq' :
  forall (F:Family nat) (pff:Finite F) (A:Ensemble nat)
         (pfp:partition_fam F A) (n:nat)
         (pfn:Inn A n)
         (pfSn:Inn (ens_S A) (S n)),
    rep_to_partition_class (fam_S F)
                           (finite_fam_S  _ pff)
                                    (ens_S A)
                                    (partition_fam_S _ _ pff pfp)
                                    (S n) pfSn =
    ens_S (rep_to_partition_class F pff A pfp n pfn).
  intros F h1 A h2 n h3 h4.
  pose proof (rep_to_partition_class_fam_S_eq F h1 A h2 n h3) as h5.
  pose proof (proof_irrelevance _ h4 (in_ens_S _ _ h3)); subst.
  assumption.
Qed.


(*Wyckoff*)
Lemma in_rep_to_partition_class_fam_S :
  forall (F:Family nat) (pff:Finite F) (A:Ensemble nat)
         (pfp:partition_fam F A) (m n:nat)
         (pfn:Inn A n),
    In (rep_to_partition_class F pff A pfp n pfn) m ->
    In (rep_to_partition_class (fam_S F)
                               (finite_fam_S  _ pff)
                               (ens_S A)
                               (partition_fam_S _ _ pff pfp)
                               (S n) (in_ens_S _ _ pfn))
       (S m).
  intros F h1 A h2 m n h3 h4.
  pose proof (rep_to_partition_class_compat  (fam_S F)
            (finite_fam_S F h1) (ens_S A)
            (partition_fam_S F A h1 h2) (S n)
            (in_ens_S A n h3)) as h5.
  simpl in h5.
   destruct h5 as [h5 h6]. 
  inversion h5 as [B h8 ? h9]. subst.
  rewrite h9.
  rewrite rep_to_partition_class_fam_S_eq in h9.
  apply ens_S_inj in h9.
  rewrite h9 in h4.
  econstructor.
  apply h4.
  reflexivity.
Qed.


(*Converse*)
(*Wyckoff*)
Lemma in_rep_to_partition_class_fam_S_impl :
  forall (F:Family nat) (pff:Finite F) (A:Ensemble nat)
         (pfp:partition_fam F A) (m n:nat)
         (pfn:Inn A n),
         In (rep_to_partition_class (fam_S F)
                                    (finite_fam_S  _ pff)
                                    (ens_S A)
                                    (partition_fam_S _ _ pff pfp)
                                    (S n) (in_ens_S _ _ pfn))
            (S m) ->
         In (rep_to_partition_class F pff A pfp n pfn) m. 
  intros F h1 A h2 m n h3 h4.
  pose proof (rep_to_partition_class_compat  (fam_S F)
            (finite_fam_S F h1) (ens_S A)
            (partition_fam_S F A h1 h2) (S n)
            (in_ens_S A n h3)) as h5. 
  simpl in h5. destruct h5 as [h5 h6]. 
  inversion h5 as [B h8 ? h9]. subst.
  rewrite h9 in h4.
  inversion h4 as [m' h10 n' h11 h12].
  apply S_inj in h11.  subst. clear h4.
  rewrite rep_to_partition_class_fam_S_eq in h9.
  apply ens_S_inj in h9.
  subst.
  assumption.
Qed.




(*Wyckoff*)
Definition im_rel {T U:Type} (f:T->U) := (fun x y:T=>f x = f y).

(*Wyckoff*)
Lemma equiv_im_rel :
  forall {T U:Type} (f:T->U),
    Equivalence (im_rel f).
intros T U f. constructor.
red. intro x. red. reflexivity.
red. intros x y h1.
red in h1. red. rewrite h1. reflexivity.
red. intros x y z h1 h2. red in h1, h2. red. rewrite h1, h2.
reflexivity.
Qed.

(*Wyckoff*)
Definition rel_classes_im_rel 
           {T U:Type} (f:T->U) : Ensemble (Ensemble T) := 
  rel_classes (Full_set T) (im_rel f).

(*Wyckoff*)
Definition rel_classes_im_rel_set
           {T U:Type} (A:Ensemble T) (f:T->U) : 
  Ensemble (Ensemble T) := 
  rel_classes A (im_rel f).

(*Wyckoff*)
Lemma partition_fam_rel_classes_im_rel_set : 
  forall {T U:Type} (A:Ensemble T) (f:T->U),
    partition_fam (rel_classes_im_rel_set A f) A.
intros T U A f. 
apply equiv_classes_partition_fam.
apply equiv_im_rel.
Qed.

(*Wyckoff*)
Definition rel_classes_im_rel_sig_fun 
           {T U:Type} {A:Ensemble T} (f:sig_set A->U) :
  Ensemble (Ensemble T) := 
  Im (rel_classes (full_sig A) (im_rel f))
     (fun A=>Im A (@proj1_sig _ _)).

(*Wyckoff*)
Lemma rel_classes_im_rel_eq_inv_im_singletons : 
  forall {T U:Type} {A:Ensemble T} (f:sig_set A->U),
    rel_classes_im_rel_sig_fun f = 
    [S:Ensemble T | exists u:U, In (Im (full_sig A) f) u /\
                                S = (Im (inv_im (Singleton u) f)
                                        (@proj1_sig _ _))].
intros T U A f.
apply Extensionality_Ensembles.
red. split.
red. intros B h1.
destruct h1 as [B h1]. subst.
destruct h1 as [x h1]. subst. clear h1.
constructor.
exists (f x). split.
apply Im_intro with x. constructor. reflexivity.
f_equal. unfold inv_im. unfold im_rel.
apply Extensionality_Ensembles.
red. split.
red.
intros x' h1. destruct h1 as [h1]. destruct h1 as [h1l h1r].
clear h1l. constructor. rewrite h1r. constructor.
red.
intros x' h1. destruct h1 as [h1']. inversion h1'.
constructor. rewrite H0.
split; auto. constructor.
red.
intros B h1.
destruct h1 as [h1].
destruct h1 as [u h1].
destruct h1 as [h1l h1r]. subst. 
destruct h1l as [x h1]. subst. clear h1.
unfold rel_classes_im_rel_sig_fun.
unfold rel_classes. rewrite im_im.
apply Im_intro with x. constructor.
f_equal.
unfold inv_im.
apply Extensionality_Ensembles.
red. split.
red. intros x' h1.
destruct h1 as [h1]. inversion h1 as [h2].
constructor. split; try constructor.
red. assumption.
red. intros x' h1.
destruct h1 as [h1].
destruct h1 as [h0 h1]. clear h0.
constructor. red in h1. rewrite h1.
constructor.
Qed.

(*Wyckoff*)
Lemma rel_classes_im_rel_eq_inv_im_singletons' : 
  forall {T U:Type} {A:Ensemble T} (f:sig_set A->U),
    rel_classes_im_rel_sig_fun f = 
    Im (Im (full_sig A) f)
       (fun u:U =>
          Im (inv_im (Singleton u) f) (@proj1_sig _ _)).
intros T U A f.
rewrite rel_classes_im_rel_eq_inv_im_singletons.
rewrite im_im.
apply Extensionality_Ensembles.
red. split.
red.
intros B h1.
destruct h1 as [h1]. destruct h1 as [u h1]. destruct h1 as [h1l h1r].
subst.
destruct h1l as [u h1]. subst.
apply Im_intro with u. assumption.
reflexivity.
red.
intros B h1.
destruct h1 as [x h1]. subst. clear h1.
constructor.
exists (f x). split. apply Im_intro with x. constructor.
reflexivity.
reflexivity.
Qed.

(*Wyckoff*)
Lemma empty_rel_classes_im_rel_eq_inv_im_singletons : 
  forall {T U:Type} {A:Ensemble T} (f:sig_set A->U),
    rel_classes_im_rel_sig_fun f =
    Empty_set _ -> A = Empty_set _.
intros T U A f h1.
apply Extensionality_Ensembles; red; split; auto with sets.
red.
intros x h2.
unfold rel_classes_im_rel_sig_fun in h1.
unfold rel_classes in h1.
rewrite im_im in h1.
assert (h3:In (Im (full_sig A)
                  (fun x : sig_set A =>
                     Im [y : sig_set A |
                         In (full_sig A) y /\ im_rel f x y]
                        (proj1_sig (P:=fun x0 : T => In A x0))))
               ((fun x : sig_set A =>
          Im [y : sig_set A | In (full_sig A) y /\ im_rel f x y]
            (proj1_sig (P:=fun x0 : T => In A x0))) (exist _ _ h2))).
  apply Im_intro with (exist _ _ h2).
  constructor.
  reflexivity.
rewrite h1 in h3.
contradiction.
Qed.

(*Wyckoff*)
Lemma inh_rel_classes_im_rel_sig_fun : 
  forall {T U:Type} {A:Ensemble T} (f:sig_set A->U),
    forall (S:Ensemble T),
      Inn (rel_classes_im_rel_sig_fun f) S ->
      Inhabited S.
intros T U A f S h2.
destruct h2 as [S h2]. subst.
destruct h2 as [x h2]. subst. clear h2.
apply Inhabited_intro with (proj1_sig x).
apply Im_intro with x.
constructor. split. constructor. red. reflexivity.
reflexivity.
Qed.

(*Wyckoff*)
Lemma inh_rel_classes_im_rel_set : 
  forall {T U:Type} {A:Ensemble T} (f:T->U),
    forall (S:Ensemble T),
      Inn (rel_classes_im_rel_set A f) S ->
      Inhabited S.
intros T U A f S h1.
destruct h1 as [a h1]. subst.
apply Inhabited_intro with a. constructor.
split; auto. red. auto.
Qed.




(*Wyckoff*)
Lemma rel_classes_im_rel_sig_fun_consistent_membership : 
  forall {T U:Type} {A:Ensemble T} (f:sig_set A->U)
         (B:Ensemble T),
    In (rel_classes_im_rel_sig_fun f) B ->
    forall (x y:sig_set A),
      (f x = f y) -> In B (proj1_sig x) ->
      In B (proj1_sig y).
intros T U A f B h1 x y h2 h3.
destruct h1 as [B h1]. subst.
destruct h1 as [x' h1]. subst.
apply Im_intro with y.
constructor. split. constructor.
inversion h3 as [y' h5]. subst.
destruct h5 as [h5]. destruct h5 as [h5l h5r]. clear h5l.
red in h5r. red.
rewrite <- h2. rewrite h5r.
f_equal.
destruct y', x. simpl in H. subst.
apply proj1_sig_injective. simpl. reflexivity.
reflexivity.
Qed.

(*Wyckoff*)
Lemma partition_fam_rel_classes_im_rel_sig_fun : 
  forall {T U:Type} {A:Ensemble T} (f:sig_set A->U),
    partition_fam (rel_classes_im_rel_sig_fun f) A.
intros T U A f.
pose proof (equiv_classes_partition_fam (full_sig A)
                                        _ (equiv_im_rel f)) as h1.
red in h1. destruct h1 as [h1l h1r].
red.
split.
intros S S' h4 h2 h3.
destruct h2 as [S h2]. destruct h3 as [S' h3]. subst.
red in h1l.
assert (h5:S <> S'). intro h5. subst. contradict h4. reflexivity.
specialize (h1l _ _ h5 h2 h3). 
red in h1l. red.
apply NNPP.
intro h6.
apply not_empty_Inhabited in h6.
destruct h6 as [x h6].
destruct h6 as [x h6 h7]. 
destruct h6 as [x h6]. destruct x as [x h8].
subst. simpl in h7. destruct h7 as [x h7]. destruct x as [x h9]. subst.
simpl in h6, h8.
pose proof (proof_irrelevance _ h8 h9); subst.
assert (h10:In (Intersection S S') (exist _ _ h9)). auto with sets.
rewrite h1l in h10. contradiction.
unfold rel_classes_im_rel_sig_fun.
rewrite family_union_im.
rewrite h1r.
apply Extensionality_Ensembles.
red. split.
red. intros x h2.
destruct h2 as [x h2]. subst.
apply proj2_sig.
red.
intros x h2. apply Im_intro with (exist _ _ h2); auto.
constructor.
Qed.


(*Wyckoff*)
Lemma im_elt_of_rel_classes_im_rel_sig_fun :
  forall {T U:Type} {A:Ensemble T} (f:sig_set A->U) (B:Ensemble T)
         (pfinc:Included B A),
    In (rel_classes_im_rel_sig_fun f) B ->
    forall (b:T) (pfb:In B b),
        Im (full_sig B) (fun x => f (exist _ _ (pfinc _ (proj2_sig x)))) = 
      Singleton (f (exist _ _ (pfinc _ pfb))).
intros T U A f B h1 h2 b h3.
apply Extensionality_Ensembles.
red. split.
red.
intros u h4.
destruct h4 as [u h4]. subst. clear h4. destruct u as [u h4].
destruct h2 as [B h2]. subst. simpl. 
destruct h2 as [x h2]. subst. clear h2.
destruct h3 as [b h3]. subst. destruct h3 as [h3].
destruct h3 as [h3a h3b]. 
destruct h4 as [u h4]. subst. destruct h4 as [h4].
destruct h4 as [h4a h4b].
red in h3b. red in h4b.
destruct b as [b h5]. destruct u as [u h6].
simpl.
pose proof h4b as h4b'. rewrite h3b in h4b'.
gterm0.
redterm0 c.
gterm1.
redterm0 c1.
redterm0 c2.
assert (h7: c3 = exist _ _ h5).
  apply proj1_sig_injective; simpl; auto.
fold c3.
assert (h8:c0 = exist _ _ h6).
  unfold c0. apply proj1_sig_injective; simpl; auto.
fold c0.
rewrite h7, h8 at 1.
rewrite h4b' at 1.
constructor.
red.
intros x h4.
destruct h4.
apply Im_intro with (exist _ _ h3). constructor.
simpl.
reflexivity.
Qed.



(*Wyckoff*)
Lemma rel_classes_im_rel_sig_fun_setminus : 
  forall {T U:Type} (A B:Ensemble T) (f:sig_set A -> U),
    In (rel_classes_im_rel_sig_fun f) B ->
    rel_classes_im_rel_sig_fun
      (restriction_sig f (Setminus A B) (setminus_inc A B)) = 
    Subtract (rel_classes_im_rel_sig_fun f) B.
intros T U A B f h1.
unfold restriction_sig.
do 2 rewrite  rel_classes_im_rel_eq_inv_im_singletons.
apply Extensionality_Ensembles.
red. split. 
red. intros S h2.
destruct h2 as [h2]. destruct h2 as [u h2]. destruct h2 as [h2 h3].
subst.
destruct h2 as [u h2]. subst. clear h2.
destruct u as [u h2]. destruct h2 as [h2 h3].
constructor.
constructor. 
exists (f (exist _ _ h2)).
split. apply Im_intro with (exist _ _ h2). constructor. reflexivity.
unfold restriction_sig. simpl.
pose proof (proof_irrelevance
              _ h2
              (setminus_inc A B u (conj h2 h3))) as h4.
rewrite <- h4.
apply Extensionality_Ensembles.
red. split.
red.
intros x h5. destruct  h5 as [x h5].  clear h4. subst.
destruct x as [x h6]. destruct h6 as [h6 h7]. simpl. 
apply Im_intro with (exist _ _ h6).
inversion h5 as [h8]. clear h5. simpl in h8.
constructor.
inversion h8 as [h9]. clear h8.
pose proof (proof_irrelevance _ h6
                              (setminus_inc A B x (conj h6 h7))) as h10. 
rewrite h9.
rewrite <- h10.
constructor. simpl. reflexivity.
red.
intros x h5.
destruct h5 as [x h5]. clear h4. subst.
destruct h5 as [h5]. inversion h5 as [h6]. clear h5.
assert (h7:~In B (proj1_sig x)).
  intro h8.
  symmetry in h6.
  pose proof (rel_classes_im_rel_sig_fun_consistent_membership
                f B h1 _ _ h6 h8) as h9.
  simpl in h9. contradiction.
destruct x as [x h8]. simpl in h7. simpl.
assert (h9:In (Setminus A B) x). auto with sets.
apply Im_intro with (exist _ _ h9).
constructor.
simpl.
rewrite h6.
pose proof (proof_irrelevance _ h8 (setminus_inc A B x h9)) as h10.
subst.
constructor. simpl. reflexivity.
intro h4.
inversion h4 as [h5]. clear h4. simpl in h5.
termr h5.
assert (h6:Inn y u).
  unfold y.
  assert (h6:In (Setminus A B) u). auto with sets.
  apply Im_intro with (exist _ _ h6).
  constructor. simpl.
  pose proof (proof_irrelevance _ h6 (conj h2 h3)) as h7.
  rewrite h7.
  constructor. simpl. reflexivity.
unfold y in h6.
rewrite <- h5 in h6.
contradiction.
red.
intros S h2.
destruct h2 as [h2 h3]. assert (h4:S <> B). intro h4. subst.
contradict h3. constructor. clear h3.
destruct h2 as [h2].
destruct h2 as [u h2].
destruct h2 as [h2 h3]. subst.
destruct h2 as [x h2]. subst. clear h2. 
destruct x as [x h2].
constructor. 
exists (f (exist _ _ h2)). split. 
assert (h3:~In B x).
nterml h4.
assert (h5:Inn x0 x).
  unfold x0.
  apply Im_intro with (exist _ _ h2).
  constructor. constructor. simpl. reflexivity.
  intro h6.
  assert (h7:In (rel_classes_im_rel_sig_fun f) x0).
    unfold x0.
    rewrite rel_classes_im_rel_eq_inv_im_singletons.
    constructor.
    exists (f (exist _ _ h2)).
    split. apply Im_intro with (exist _ _ h2).
    constructor. reflexivity.
    f_equal.
  pose proof (partition_fam_rel_classes_im_rel_sig_fun f) as h8.
  red in h8.
  destruct h8 as [h8 h9].
  red in h8.
  specialize (h8 _ _ h4 h7 h1).
  red in h8.
  terml h8.
  assert (h10:Inn x1 x).
    unfold x1.
    auto with sets.
  unfold x1 in h10.
  rewrite h8 in h10.
  contradiction.
assert (h5:In (Setminus A B) x). auto with sets.
apply Im_intro with (exist _ _ h5).
constructor. simpl.
f_equal. apply proj1_sig_injective. reflexivity.
apply Extensionality_Ensembles.
red. split.
red. intros x' h5.
destruct h5 as [x' h5]. subst. destruct x' as [x' h6].
inversion h5 as [h7]. clear h5. inversion h7 as [h8]. clear h7.
simpl.
nterml h4.
  assert (h7:In (rel_classes_im_rel_sig_fun f) x0).
    unfold x0.
    rewrite rel_classes_im_rel_eq_inv_im_singletons.
    constructor.
    exists (f (exist _ _ h2)).
    split. apply Im_intro with (exist _ _ h2).
    constructor. reflexivity.
    f_equal.
unfold x0 in h7.
pose proof (partition_fam_rel_classes_im_rel_sig_fun f) as h9.
red in h9. destruct h9 as [h9 h10].
specialize (h9 _ _ h4 h7 h1). red in h9.
assert (h11:Inn x0 x).
  unfold x0.
  apply Im_intro with (exist _ _ h2).
  constructor. constructor. reflexivity.
  pose proof (rel_classes_im_rel_sig_fun_consistent_membership
                _ _ h7 _ _ h8 h11) as h12.
simpl in h12.
assert (h13:~In B x').
  intro h14.
  terml h9.
  assert (h15:Inn x1 x').
    unfold x1.
    auto with sets.
  unfold x1 in h15.
  rewrite h9 in h15. contradiction.
  assert (h14:In (Setminus A B) x'). auto with sets.
  apply Im_intro with (exist _ _ h14).
  constructor.
  simpl.
  rewrite h8 at 1.
  pose proof (proof_irrelevance
                _ h6 
                (setminus_inc A B x' h14)) as h15.
  rewrite h15. 
  constructor.
  reflexivity.
red.
intros x' h5.
destruct h5 as [x' h5]. subst.
destruct h5 as [h5]. destruct x' as [x' h6]. simpl in h5.
inversion h5 as [h7]. clear h5. simpl.
destruct h6 as [h6 h8].
apply Im_intro with (exist _ _ h6).
constructor. rewrite h7 at 1.
pose proof (proof_irrelevance
              _ h6
              (setminus_inc A B x' (conj h6 h8))) as h9. 
rewrite h9 at 2.
constructor.
reflexivity.
Qed.


(*Wyckoff*)
Lemma in_rep_to_partition_class_fam_S_seg :
  forall (F:Family nat) (pff:Finite F) (m n p:nat) 
         (pfp:partition_fam F (seg n))
         (pflt:m < n),
    In (rep_to_partition_class F pff (seg n) pfp m (in_seg _ _ pflt)) p -> 
    In (rep_to_partition_class (fam_S F)
                               (finite_fam_S  _ pff)
                               (ens_S (seg n))
                               (partition_fam_S _ _ pff pfp)
                               (S m) (in_ens_S_seg _ _ pflt))
       (S p).
  intros F h1 m n p h2 h3 h4.
  apply in_rep_to_partition_class_fam_S in h4.
  pose proof (proof_irrelevance
                _
                (in_ens_S (seg n) m (in_seg m n h3))
                (in_ens_S_seg m n h3)) as h5.
  rewrite h5 in h4; auto.
Qed.


(*Wyckoff*)
Lemma in_rep_to_partition_class_fam_S_seg_impl :
  forall (F:Family nat) (pff:Finite F) (m n p:nat) 
         (pfp:partition_fam F (seg n))
         (pflt:m < n),
    In (rep_to_partition_class (fam_S F)
                               (finite_fam_S  _ pff)
                               (ens_S (seg n))
                               (partition_fam_S _ _ pff pfp)
                               (S m) (in_ens_S_seg _ _ pflt))
       (S p) ->
    In (rep_to_partition_class F pff (seg n) pfp m (in_seg _ _ pflt)) p. 
  intros F h1 m n p h2 h3 h4.
  pose proof (proof_irrelevance
                _ 
                (in_ens_S_seg _ _ h3)
                (in_ens_S _ _ (in_seg _ _ h3))) as h5.
  rewrite h5 in h4.
  apply in_rep_to_partition_class_fam_S_impl in h4; auto.
Qed.


End EquivalenceClass.


Section EvenOddPos.
(*Wyckoff*)
Definition pos_evens :=
  [n:nat | 0 < n /\ even n].

(*Wyckoff*)
Definition pos_odds :=
  [n:nat | odd n].

(*Wyckoff*)
Definition pos_evens_t :=
  {n:nat | 0 < n /\ even n}.

(*Wyckoff*)
Definition pos_odds_t :=
  {n:nat | odd n}.

(*Wyckoff*)
Definition pet_lt :=
  exist_rel lt (fun n:nat => 0 < n /\ even n).

(*Wyckoff*)
Definition pot_lt :=
  exist_rel lt (fun n:nat => odd n).

(*Wyckoff*)
Definition pet_le :=
  exist_rel le (fun n:nat => 0 < n /\ even n).

(*Wyckoff*)
Definition pot_le :=
  exist_rel le (fun n:nat => odd n).

(*Wyckoff*)
Lemma pet_le_refl : 
  forall x, pet_le x x.
intro; red; red; auto with arith.
Qed.

(*Wyckoff*)
Lemma pot_le_refl : 
  forall x, pot_le x x.
intro; red; red; auto with arith.
Qed.

(*Wyckoff*)
Lemma pet_ge_2 : 
  forall m:pos_evens_t,
    2 <= (proj1_sig m).
intro m. destruct m as [m h1]. simpl.
destruct h1 as [h1 h2].
apply even_decompose in h2.
destruct h2 as [k h2].
omega.
Qed.

(*Wyckoff*)
Lemma pet_ge_1 :
  forall m:pos_evens_t,
    1 <= (proj1_sig m).
  intro m. pose proof (pet_ge_2 m).
  auto with arith.
Qed.

(*Wyckoff*)
Definition o1 : pos_odds_t := exist _ _ odd_1.

(*Wyckoff*)
Definition e2 : pos_evens_t := exist _ _ (conj lt_0_2 even_2).


(*Wyckoff*)
Lemma pos_evens_eq : 
  pos_evens = Im (Full_set nat) (fun n => 2*n + 2).
apply Extensionality_Ensembles; red; split; red; intros x h1.
destruct h1 as [h1]. destruct h1 as [h1 h2].
apply even_decompose in h2.
destruct h2 as [k h2].
assert (h3:k > 0). omega.
apply S_pred in h3. 
apply Im_intro with (pred k). constructor.
omega.
destruct h1 as [x h1]. subst.
constructor. split; try omega.
apply even_even_plus.
apply even_mult_l. repeat constructor. repeat constructor.
Qed.

(*Wyckoff*)
Lemma pos_odds_eq : 
  pos_odds = Im (Full_set nat) (fun n => 2*n + 1).
apply Extensionality_Ensembles; red; split; red; intros x h1.
destruct h1 as [h1].
apply odd_decompose in h1.
destruct h1 as [k h1]. subst.
apply Im_intro with k;  constructor.
destruct h1 as [x h1]. subst.
constructor.
apply odd_plus_r.
apply even_mult_l. repeat constructor.
repeat constructor.
Qed.

(*Wyckoff*)
Definition pet_pred (n:pos_evens_t) : pos_odds_t.
pose proof (pet_ge_2 n) as h0.
destruct n as [n h1]. 
destruct h1 as [h1 h2]. simpl in h0.  
rewrite <- (odd_pred_iff n h1) in h2.
refine (exist _ _ h2).
Defined.


(*Wyckoff*)
(*Note: This returns 2 if n = 1, kind of like pred 0 = 0*)
Definition pot_pred (n:pos_odds_t) : pos_evens_t.
destruct (le_lt_dec 3 (proj1_sig n)) as [h1 | h2].
destruct n as [n h3]. simpl in h1.
assert (h4:0 < n). destruct (zerop n) as [h4 | h4]. subst.
  inversion h1. assumption.
rewrite <- (even_pred_iff n h4) in h3. 
assert (h5:0 < pred n). red.
  apply le_S_n.
  simpl.
  rewrite <- (S_pred _ _ h4).
  auto with arith.
refine (exist _ _ (conj h5 h3)).
assert (h6:0 < 2). red. auto with arith.
refine (exist _ _ (conj h6 even_2)).
Defined.


(*Wyckoff*)
Lemma pot_pred_one :
  forall (pf:odd 1),
    pot_pred (exist _ _ pf) = exist _ _ (conj lt_0_2 even_2).
  intro h1.
  unfold pot_pred.
  simpl.
  apply proj1_sig_injective; auto.
Qed.


(*Wyckoff*)
Lemma le_proj1_sig_pot_pred :  
  forall (m:nat) (n:pos_odds_t),
    even m -> m <= proj1_sig n ->
    m <= proj1_sig (pot_pred n).
intros m n h1 h2.
unfold pot_pred.
destruct  le_lt_dec as [h3 | h3].
destruct n as [n h4]. simpl. simpl in h2, h3. 
apply le_even_odd; auto. 
simpl.
omega.
Qed.

(*Wyckoff*)
Lemma le_proj1_sig_pet_pred : 
  forall (m:nat) (n:pos_evens_t),
    odd m -> m <= proj1_sig n ->
    m <= proj1_sig (pet_pred n).
intros m n h1 h2.
unfold pet_pred. 
destruct n as [n h4].  destruct h4 as [h4 h5].
simpl. simpl in h2.
apply le_odd_even; auto.
Qed.

(*Wyckoff*)
Lemma pet_le_pot_pred :
  forall (m:pos_evens_t) (n:pos_odds_t),
    1 < proj1_sig n ->
    pet_le m (pot_pred n) ->
    proj1_sig m <= proj1_sig n.
  intros m n h0 h1.
  red in h1. unfold exist_rel in h1.
  unfold pot_pred in h1.
  destruct le_lt_dec as [h2 | h2].
  simpl in h1.
  destruct n as [n h3].
  simpl in h1.
  simpl.
  omega.
  simpl in h1.
  omega.
Qed.

(*Wyckoff*)
Lemma le_pot_le_pet_pred :
  forall (n:pos_evens_t) (o:pos_odds_t),
    proj1_sig o <= proj1_sig n ->
    pot_le o (pet_pred n).
  intros n o h1.
  red. unfold exist_rel.
  unfold pet_pred.
  destruct n as [n h2], o as [o h3].
  simpl.
  destruct h2 as [h2 h4].
  simpl.
  simpl in h1. 
  destruct n as [|n].
  assert (h5:o = 0). auto with arith.
  subst.
  inversion h3.
  simpl.
  destruct (le_lt_eq_dec _ _ h1); try omega.
  subst.
  pose proof (not_even_and_odd _ h4 h3).
  contradiction.
Qed.



(*Wyckoff*)
Definition pos_evens_fun_to_sig_fun 
           {T:Type} (f:pos_evens_t -> T) : sig_set (pos_evens)->T.
intro n.
destruct n as [n h1]. destruct h1 as [h1].
refine (f (exist _ _ h1)).
Defined.


(*Wyckoff*)
Definition pos_odds_fun_to_sig_fun 
           {T:Type} (f:pos_odds_t -> T) : sig_set (pos_odds)->T.
intro n.
destruct n as [n h1]. destruct h1 as [h1].
refine (f (exist _ _ h1)).
Defined.


(*Wyckoff*)
Lemma inj_pos_evens_fun_to_sig_fun_iff : 
  forall {T:Type} (f:pos_evens_t -> T),
    injective f <-> injective (pos_evens_fun_to_sig_fun f).
intros T f.
split.
intro h1. red in h1. red.
intros x y h2.
unfold pos_evens_fun_to_sig_fun in h2.
destruct x as [x h3], y as [y h4].
destruct h3 as [h3], h4 as [h4].
apply h1 in h2.
apply proj1_sig_injective; simpl.
apply exist_injective in h2. assumption.

intro h1. red in h1. red.
intros x y h2.
destruct x as [x h3], y as [y h4].
assert (h5:In pos_evens x).
  constructor; auto.
assert (h6:In pos_evens y).
  constructor; auto.
specialize (h1 (exist _ _ h5) (exist _ _ h6)).
unfold pos_evens_fun_to_sig_fun in h1.
destruct h5 as [h5], h6 as [h6].
pose proof (proof_irrelevance _ h3 h5); subst.
pose proof (proof_irrelevance _ h4 h6); subst.
specialize (h1 h2).
apply exist_injective in h1. subst.
apply proj1_sig_injective; auto.
Qed.


(*Wyckoff*)
Lemma inj_pos_odds_fun_to_sig_fun_iff : 
  forall {T:Type} (f:pos_odds_t -> T),
    injective f <-> injective (pos_odds_fun_to_sig_fun f).
intros T f.
split.
intro h1. red in h1. red.
intros x y h2.
unfold pos_odds_fun_to_sig_fun in h2.
destruct x as [x h3], y as [y h4].
destruct h3 as [h3], h4 as [h4].
apply h1 in h2.
apply proj1_sig_injective; simpl.
apply exist_injective in h2. assumption.

intro h1. red in h1. red.
intros x y h2.
destruct x as [x h3], y as [y h4].
assert (h5:In pos_odds x).
  constructor; auto.
assert (h6:In pos_odds y).
  constructor; auto.
specialize (h1 (exist _ _ h5) (exist _ _ h6)).
unfold pos_odds_fun_to_sig_fun in h1.
destruct h5 as [h5], h6 as [h6].
pose proof (proof_irrelevance _ h3 h5).
pose proof (proof_irrelevance _ h4 h6); subst.
specialize (h1 h2).
apply exist_injective in h1. subst.
apply proj1_sig_injective; auto.
Qed.


(*Wyckoff*)
Lemma surj_pos_evens_fun_to_sig_fun_iff : 
  forall {T:Type} (f:pos_evens_t -> T),
    surjective f <-> surjective (pos_evens_fun_to_sig_fun f).
intros T f.
split.
intro h1. red in h1. red.
intro y. specialize (h1 y).
destruct h1 as [x h1]. subst.
destruct x as [x h1].
assert (h2:In pos_evens x). constructor; auto.
exists (exist _ _ h2).
unfold pos_evens_fun_to_sig_fun.
destruct h2.
f_equal. apply proj1_sig_injective; auto.

intro h1. red in h1. red.
intro y. specialize (h1 y).
destruct h1 as [x h1].
destruct x as [x h2]. destruct h2 as [h2].
exists (exist (fun s=>0 < s /\ even s)  _ h2).
subst.
unfold pos_evens_fun_to_sig_fun.
reflexivity.
Qed.


(*Wyckoff*)
Lemma surj_pos_odds_fun_to_sig_fun_iff : 
  forall {T:Type} (f:pos_odds_t -> T),
    surjective f <-> surjective (pos_odds_fun_to_sig_fun f).
intros T f.
split.
intro h1. red in h1. red.
intro y. specialize (h1 y).
destruct h1 as [x h1]. subst.
destruct x as [x h1].
assert (h2:In pos_odds x). constructor; auto.
exists (exist _ _ h2).
unfold pos_odds_fun_to_sig_fun.
destruct h2.
f_equal. apply proj1_sig_injective; auto.

intro h1. red in h1. red.
intro y. specialize (h1 y).
destruct h1 as [x h1].
destruct x as [x h2]. destruct h2 as [h2].
exists (exist (fun s=>odd s)  _ h2).
subst.
unfold pos_odds_fun_to_sig_fun.
reflexivity.
Qed.


(*Wyckoff*)
Lemma bij_pos_evens_fun_to_sig_fun_iff : 
  forall {T:Type} (f:pos_evens_t -> T),
    bijective f <-> bijective (pos_evens_fun_to_sig_fun f).
intros T f. 
unfold bijective.
rewrite inj_pos_evens_fun_to_sig_fun_iff.
rewrite surj_pos_evens_fun_to_sig_fun_iff.
tauto.
Qed.

(*Wyckoff*)
Lemma bij_pos_odds_fun_to_sig_fun_iff : 
  forall {T:Type} (f:pos_odds_t -> T),
    bijective f <-> bijective (pos_odds_fun_to_sig_fun f).
intros T f. 
unfold bijective.
rewrite inj_pos_odds_fun_to_sig_fun_iff.
rewrite surj_pos_odds_fun_to_sig_fun_iff.
tauto.
Qed.



End EvenOddPos.

Section Unite.
  (*These "unite functions" take an infinite function on all 
[odds/evens] and an [[even/odd]/[odd/even]] number and a finite 
function on some [evens/odds] and unites it all into a single 
finite function on 1 <= m <= n (seg_one_t n).*)

(*Wyckoff*)
Definition pos_odds_fun_unite_finite_evens_aux 
           {T:Type} (p:pos_odds_t->T) (n:pos_evens_t)
           (p':{m:pos_evens_t | pet_le m n}->T) : seg_one_t (proj1_sig n) -> T.
intro m.
destruct m as [m h1].
destruct (even_odd_dec m) as [h2 | h2].
assert (h3:0 < m). destruct h1. auto with arith.
pose (exist (fun a => 0 < a /\ even a) _ (conj h3 h2)) as m'.
assert (h4:pet_le m' n).
  red. red. unfold m'. simpl. destruct h1; auto with arith.
refine (p' (exist (fun a => pet_le a n) _ h4)).
refine (p (exist _ _ h2)).
Defined.

(*Wyckoff*)
Lemma pos_odds_fun_unite_finite_evens_ex :
  forall {T:Type} (p:pos_odds_t->T) (n:pos_evens_t)
         (p':{m:pos_evens_t | pet_le m n}->T),
  exists !f : seg_one_t (proj1_sig n) -> T,
    (forall (o:pos_odds_t) (pf:proj1_sig o <= proj1_sig n),
       p o = f (exist (fun m => 1 <= m <= (proj1_sig n)) _ (conj (odd_ge_1 _ (proj2_sig o)) pf))) /\
    (forall (e:pos_evens_t) (pf:pet_le e n),
       p' (exist (fun m => pet_le m n)  _ pf) =
       f (exist (fun m => 1 <= m <= proj1_sig n) _ (conj (pet_ge_1 e) pf))).
  intros T p n p'.
  exists (pos_odds_fun_unite_finite_evens_aux p n p').
  red. split.
  split.
  intros o h1.
  unfold pos_odds_fun_unite_finite_evens_aux.
  destruct even_odd_dec as [h2 | h2].
  destruct o as [o h3]. simpl in h2.
  pose proof (not_even_and_odd _ h2 h3).
  contradiction.
  destruct o as [o h3].
  f_equal. apply proj1_sig_injective; auto.
  intros e h4.
  destruct e as [e h3]. 
  unfold pos_odds_fun_unite_finite_evens_aux.
  destruct even_odd_dec as [h5 | h5].
  f_equal. apply proj1_sig_injective; simpl.
  apply proj1_sig_injective; auto.
  simpl in h5.
  destruct n as [n h6]. destruct h3 as [h3 h7].
  pose proof (not_even_and_odd _ h7 h5).
  contradiction.
  intros f h1.
  apply functional_extensionality.
  intros x.
  destruct x as [m h2].
  destruct h1 as [h1a h1b].
  unfold pos_odds_fun_unite_finite_evens_aux.
  destruct even_odd_dec as [h3 | h3].
  assert (h4: 0 < m). omega.  
  specialize (h1b (exist _ _ (conj h4 h3))
                  (match h2 with
                     | conj _ H0 => H0
                   end)).
  terml h1b.
  gterml.
  fold x0.
  fold x in h1b.
  assert (h6:x = x0).
    unfold x, x0.
    f_equal. apply proj1_sig_injective; auto. simpl.
    apply proj1_sig_injective. simpl.
    reflexivity.
  rewrite <- h6 at 1.
  rewrite h1b.
  f_equal.
  apply proj1_sig_injective; auto.

  destruct h2 as [h2 h4].
  specialize (h1a (exist _ _ h3) h4). simpl in h1a.
  rewrite h1a.
  f_equal.
  apply proj1_sig_injective; auto.
Qed.


(*Wyckoff*)
Definition pos_odds_fun_unite_finite_evens 
  {T:Type} (p:pos_odds_t->T) (n:pos_evens_t)
  (p':{m:pos_evens_t | pet_le m n}->T) :
  seg_one_t (proj1_sig n) -> T :=
  proj1_sig (constructive_definite_description
               _ (pos_odds_fun_unite_finite_evens_ex p n p')).


(*Wyckoff*)
Lemma pos_odds_fun_unite_finite_evens_compat :
  forall {T:Type} (p:pos_odds_t->T) (n:pos_evens_t)
         (p':{m:pos_evens_t | pet_le m n}->T),
    let f := pos_odds_fun_unite_finite_evens p n p' in
    (forall (o:pos_odds_t) (pf:proj1_sig o <= proj1_sig n),
       p o =
       f (exist
            (fun m =>
               1 <= m <= (proj1_sig n)) _
            (conj (odd_ge_1 _ (proj2_sig o)) pf))) /\
    (forall (e:pos_evens_t) (pf:pet_le e n),
       p' (exist (fun m => pet_le m n)  _ pf) =
       f (exist (fun m => 1 <= m <= proj1_sig n) _
                (conj (pet_ge_1 e) pf))).
  unfold pos_odds_fun_unite_finite_evens; intros;
  destruct constructive_definite_description; auto.
Qed.


(*Wyckoff*)
Definition pos_odds_fun_unite_finite_evens_aux' 
           {T:Type} (p:pos_odds_t->T) (n:pos_odds_t)
           (p':{m:pos_evens_t | pet_le m (pot_pred n)}->T) :
  seg_one_t (proj1_sig n) -> T.
intro m. 
destruct m as [m h1]. 
destruct (even_odd_dec m) as [h2 | h2].
assert (h3:0 < m).
  destruct h1 as [h1 h3].
  red.
  assumption.
pose (exist (fun a => 0 < a /\ even a) _ (conj h3 h2)) as m'.
assert (h4:pet_le m' (pot_pred n)).
  red. red. apply le_proj1_sig_pot_pred; auto.
  unfold m'. simpl. destruct h1; auto.
refine (p' (exist (fun m:pos_evens_t => pet_le m (pot_pred n)) _ h4)).
refine (p (exist _ _ h2)).
Defined.


(*Wyckoff*)
Lemma pos_odds_fun_unite_finite_evens_ex' :
  forall {T:Type} (p:pos_odds_t->T) (n:pos_odds_t)
         (p':{m:pos_evens_t | pet_le m (pot_pred n)}->T),
  exists! f:seg_one_t (proj1_sig n) -> T,
    (forall (o:pos_odds_t) (pf:proj1_sig o <= proj1_sig n),
       p o = f (exist
                  (fun m =>
                     1 <= m <=
                     (proj1_sig n)) _
                  (conj (odd_ge_1 _ (proj2_sig o)) pf))) /\
    (forall (e:pos_evens_t) (pf:pet_le e (pot_pred n))
            (pflt: 1 < proj1_sig n),
       p' (exist (fun m => pet_le m (pot_pred n))  _ pf) =
       f (exist (fun m => 1 <= m <= proj1_sig n) _
                (conj (pet_ge_1 e) (pet_le_pot_pred _ _ pflt pf)))).
  intros T p n p'.
  exists (pos_odds_fun_unite_finite_evens_aux' p n p'). split.
  split.
  intros o h1.
  unfold pos_odds_fun_unite_finite_evens_aux'.
  destruct even_odd_dec as [h2 | h2].
  destruct o as [o h3]. simpl in h2.
  pose proof (not_even_and_odd _ h2 h3).
  contradiction.
  destruct o as [o h3].
  f_equal. apply proj1_sig_injective; auto.
  intros e h4 hlt.
  destruct e as [e h3]. 
  unfold pos_odds_fun_unite_finite_evens_aux'.
  destruct even_odd_dec as [h5 | h5].
  f_equal. apply proj1_sig_injective; simpl.
  apply proj1_sig_injective; auto.
  simpl in h5.
  destruct n as [n h6]. destruct h3 as [h3 h7].
  pose proof (not_even_and_odd _ h7 h5).
  contradiction.
  intros f h1.
  destruct h1 as [h1 h2].
  unfold pos_odds_fun_unite_finite_evens_aux'.
  apply functional_extensionality.
  intro m.
  simpl.
  destruct m as [m h3]. 
  destruct even_odd_dec as [h4 | h4]. 
    assert (h5:0 < m). omega.
  specialize (h2 (exist _ _ (conj h5 h4))).
  hfirst h2.  
    red. red. simpl. 
    unfold pot_pred.
    destruct (le_lt_dec 3 (proj1_sig n)).
    simpl.
    destruct n as [n h6].
    simpl.
    simpl in h3, l.
    destruct n as [|n]. omega.
    pose proof h6 as h6'. rewrite odd_Sn_iff in h6'.
    simpl.
    apply even_decompose in h6'.
    destruct h6' as [k h6'].
    pose proof h4 as h4'.
    apply even_decompose in h4'.
    destruct h4' as [k' h7].
    omega.
  simpl.
  omega.
  specialize (h2 hf).
  red in hf. red in hf. simpl in hf.
  unfold pot_pred in hf. 
  destruct n as [n h8]. 
  simpl in h2, hf. 
  simpl in h3.
  destruct (lt_eq_lt_dec 1 n) as [h10 | h10].
  destruct h10 as [h10 | h10].
  specialize (h2 h10).
  terml h2.
  gterml.
  fold x0.
  fold x in h2.
  assert (he:x = x0).
    unfold x, x0.
    f_equal. apply proj1_sig_injective; simpl.
    apply proj1_sig_injective; simpl.
    reflexivity. 
  rewrite <- he at 1.
  rewrite h2.  
  f_equal.
  apply proj1_sig_injective; simpl.
  reflexivity.
  subst.
  assert (h10: m = 1). omega. subst.
  inversion h4. inversion H0.
  omega.
  clear h2.
  destruct h3 as [h3 h5].
  specialize (h1  (exist _ _ h4)). simpl in h1.
  specialize (h1 h5).
  rewrite h1.
  f_equal.
  apply proj1_sig_injective; auto.
Qed.

(*Wyckoff*)
Definition pos_odds_fun_unite_finite_evens'
           {T:Type} (p:pos_odds_t->T) (n:pos_odds_t)
           (p':{m:pos_evens_t | pet_le m (pot_pred n)}->T) :
  seg_one_t (proj1_sig n) -> T :=
  proj1_sig (constructive_definite_description
               _ (pos_odds_fun_unite_finite_evens_ex' p n p')).

(*Wyckoff*)
Lemma pos_odds_fun_unite_finite_evens_compat' : 
  forall {T:Type} (p:pos_odds_t->T) (n:pos_odds_t)
         (p':{m:pos_evens_t | pet_le m (pot_pred n)}->T),
    let f := pos_odds_fun_unite_finite_evens' p n p' in
    (forall (o:pos_odds_t) (pf:proj1_sig o <= proj1_sig n),
       p o = f (exist (fun m => 1 <= m <= (proj1_sig n)) _
                      (conj (odd_ge_1 _ (proj2_sig o)) pf))) /\
    (forall (e:pos_evens_t) (pf:pet_le e (pot_pred n))
            (pflt: 1 < proj1_sig n),
       p' (exist (fun m => pet_le m (pot_pred n))  _ pf) =
       f (exist (fun m => 1 <= m <= proj1_sig n) _
                (conj (pet_ge_1 e) (pet_le_pot_pred _ _ pflt pf)))).
  unfold pos_odds_fun_unite_finite_evens'; intros;
  destruct constructive_definite_description; auto.
Qed.


(*Wyckoff*)
Definition pos_evens_fun_unite_finite_odds_aux 
           {T:Type} (p:pos_evens_t->T) (n:pos_odds_t)
           (p':{m:pos_odds_t | pot_le m n}->T) :
  seg_one_t (proj1_sig n) -> T.
intro m. 
destruct m as [m h1]. 
destruct (even_odd_dec m) as [h2 | h2].
assert (h3:0 < m).
  destruct h1 as [h1 h3].
  red. assumption.
pose (exist (fun a => 0 < a /\ even a) _ (conj h3 h2)) as m'.
refine (p m').
pose (exist _ _ h2) as m'.
assert (h4:pot_le m' n).
  red. red. unfold m'. simpl. destruct h1; auto.
refine (p' (exist (fun a => pot_le a n) _ h4)).
Defined.

(*Wyckoff*)
Lemma pos_evens_fun_unite_finite_odds_ex :
  forall {T:Type} (p:pos_evens_t->T) (n:pos_odds_t)
         (p':{m:pos_odds_t | pot_le m n}->T),
  exists! f:seg_one_t (proj1_sig n) -> T,
    (forall (o:pos_odds_t) (pf:proj1_sig o <= proj1_sig  n),
       p' (exist (fun m => pot_le m n)  _ pf) =
       f (exist (fun m => 1 <= m <= proj1_sig n) _ (conj (odd_ge_1 _ (proj2_sig o)) pf))) /\
    (forall (e:pos_evens_t) (pf:proj1_sig e <= proj1_sig n),
       p e = f (exist (fun m => 1 <= m <= (proj1_sig n)) _ (conj (pet_ge_1 e) pf))).
  intros T p n p'.
  exists (pos_evens_fun_unite_finite_odds_aux p n p').
  red. split. split.
  intros o h1.
  unfold pos_evens_fun_unite_finite_odds_aux.
  simpl.
  destruct o as [o h3].
  destruct even_odd_dec as [h2 | h2].
  pose proof (not_even_and_odd _ h2 h3).
  contradiction.
  f_equal.
  apply proj1_sig_injective; simpl.
  apply proj1_sig_injective; auto.
  intros e h1.
  destruct e as [e h3].
  unfold pos_evens_fun_unite_finite_odds_aux.
  simpl.
  destruct even_odd_dec as [h2 | h2].
  f_equal.
  apply proj1_sig_injective; simpl.
  reflexivity.
  destruct h3 as [h3 h3'].
  pose proof (not_even_and_odd _ h3' h2).
  contradiction.
  intros f h1.
  destruct h1 as [h1 h2].
  unfold pos_evens_fun_unite_finite_odds_aux.
  apply functional_extensionality.
  intro m. destruct m as [m h3].
  destruct h3 as [h3 h4].
  destruct even_odd_dec as [h5 | h5].

  assert (h6:0 < m).
    red. assumption.  
  specialize (h2 (exist (fun a => 0 < a /\ even a) _ (conj h6 h5))).
  simpl in h2.
  specialize (h2 h4).
  assert (h7:p (exist _ _ (conj h6 h5)) = p (exist _ _ (conj h3 h5))).
    f_equal. apply proj1_sig_injective; auto.
  rewrite h7 in h2.
  rewrite h2 at 1.
  f_equal.
  apply proj1_sig_injective; auto.
  specialize (h1 (exist _ _ h5)).
  simpl in h1.
  specialize (h1 h4).
  terml h1.
  gterml.
  fold x in h1; fold x0.
  assert (h6:x = x0).
    unfold x, x0.
    f_equal.
  rewrite h6 in h1.
  rewrite h1.
  f_equal.
  apply proj1_sig_injective; auto.
Qed.


(*Wyckoff*)
Definition pos_evens_fun_unite_finite_odds
           {T:Type} (p:pos_evens_t->T) (n:pos_odds_t)
           (p':{m:pos_odds_t | pot_le m n}->T) : 
  seg_one_t (proj1_sig n) -> T :=
  proj1_sig (constructive_definite_description
               _
               (pos_evens_fun_unite_finite_odds_ex p n p')).


(*Wyckoff*)
Lemma pos_evens_fun_unite_finite_odds_compat :
  forall {T:Type} (p:pos_evens_t->T) (n:pos_odds_t)
         (p':{m:pos_odds_t | pot_le m n}->T),
    let f := pos_evens_fun_unite_finite_odds p n p' in
    (forall (o:pos_odds_t) (pf:proj1_sig o <= proj1_sig  n),
       p' (exist (fun m => pot_le m n)  _ pf) =
       f (exist (fun m => 1 <= m <= proj1_sig n) _ (conj (odd_ge_1 _ (proj2_sig o)) pf))) /\
    (forall (e:pos_evens_t) (pf:proj1_sig e <= proj1_sig n),
       p e = f (exist (fun m => 1 <= m <= (proj1_sig n)) _
                      (conj (pet_ge_1 e) pf))).
  unfold pos_evens_fun_unite_finite_odds;
  intros; destruct constructive_definite_description; auto.
Qed.


(*Wyckoff*)
Definition pos_evens_fun_unite_finite_odds_aux'
           {T:Type} (p:pos_evens_t->T) (n:pos_evens_t)
           (p':{m:pos_odds_t | pot_le m (pet_pred n)}->T) : seg_one_t (proj1_sig n) -> T.
intro m.
destruct m as [m h1].
destruct (even_odd_dec m) as [h2 | h2].
assert (h3:0 < m). red. destruct h1; auto.
pose (exist (fun a => 0 < a /\ even a) _ (conj h3 h2)) as m'.
refine (p m').
pose (exist _ _ h2) as m'. 
assert (h3:pot_le m' (pet_pred n)).
  red. red. unfold m'. simpl.
  apply le_proj1_sig_pet_pred; auto.
  destruct h1; auto.
refine (p' (exist (fun a => pot_le a (pet_pred n)) _ h3)).
Defined.


(*Wyckoff*)
Lemma pos_evens_fun_unite_finite_odds_ex'  :
  forall {T:Type} (p:pos_evens_t->T) (n:pos_evens_t)
         (p':{m:pos_odds_t | pot_le m (pet_pred n)}->T),
  exists! f:seg_one_t (proj1_sig n) -> T,
    (forall (o:pos_odds_t) (pf:proj1_sig o <= proj1_sig n),
       p' (exist (fun m => pot_le m (pet_pred n))  _
                 (le_pot_le_pet_pred _ _ pf)) =
       f (exist (fun m => 1 <= m <= proj1_sig n) _
                (conj (odd_ge_1 _ (proj2_sig o)) pf))) /\
    (forall (e:pos_evens_t) (pf:proj1_sig e <= proj1_sig n),
       p e = f (exist (fun m => 1 <= m <= (proj1_sig n)) _ (conj (pet_ge_1 e) pf))).
  intros T p n p'.
  exists (pos_evens_fun_unite_finite_odds_aux' p n p').
  red. split. split.
  intros o h1.
  unfold pos_evens_fun_unite_finite_odds_aux'.
  destruct o as [o h3].
  destruct even_odd_dec as [h2 | h2].
  simpl in h2.
  pose proof (not_even_and_odd _ h2 h3).
  contradiction.
  f_equal. apply proj1_sig_injective; simpl.
  apply proj1_sig_injective; auto.
  intros e h1.
  destruct e as [e h4].
  unfold pos_evens_fun_unite_finite_odds_aux'.
  destruct even_odd_dec as [h2 | h2].
  simpl.
  f_equal.
  apply proj1_sig_injective; simpl.
  reflexivity.
  simpl in h2.
  destruct h4 as [h4 h5].
  pose proof (not_even_and_odd _ h5 h2).
  contradiction.
  intros f h1.
  destruct h1 as [h1 h2].
  apply functional_extensionality.
  intro m.
  destruct m as [m h3].
  unfold pos_evens_fun_unite_finite_odds_aux'.
  destruct even_odd_dec as [h4 | h4].
  destruct h3 as [h3 h5].
  assert (h6:0 < m).
    red. assumption.
  specialize (h2 (exist _ _ (conj h6 h4))).
  simpl in h2.
  specialize (h2 h5).
  terml h2.
  gterml.
  fold x in h2.
  fold x0.
  assert (h7:x = x0).
    unfold x, x0.
    f_equal.
    apply proj1_sig_injective; auto.
  rewrite h7 in h2.
  rewrite h2 at 1.
  f_equal.
  apply proj1_sig_injective; auto.
  specialize (h1 (exist _ _ h4)). simpl in h1.
  destruct h3 as [h3 h5].
  specialize (h1 h5).
  gterml.
  terml h1.
  assert (h6:x = x0).
    unfold x, x0.
    f_equal; apply proj1_sig_injective; simpl; auto.
  unfold x, x0 in h6.
  rewrite h6, h1; f_equal; apply proj1_sig_injective; auto.
Qed.


(*Wyckoff*)
Definition pos_evens_fun_unite_finite_odds'
  {T:Type} (p:pos_evens_t->T) (n:pos_evens_t)
  (p':{m:pos_odds_t | pot_le m (pet_pred n)}->T) :=
  proj1_sig (constructive_definite_description
               _
               (pos_evens_fun_unite_finite_odds_ex' p n p')).

(*Wyckoff*)
Lemma pos_evens_fun_unite_finite_odds_compat' :
  forall {T:Type} (p:pos_evens_t->T) (n:pos_evens_t)
         (p':{m:pos_odds_t | pot_le m (pet_pred n)}->T),
    let f:= pos_evens_fun_unite_finite_odds' p n p' in
    (forall (o:pos_odds_t) (pf:proj1_sig o <= proj1_sig n),
       p' (exist (fun m => pot_le m (pet_pred n))
                 _ (le_pot_le_pet_pred _ _ pf)) =
       f (exist (fun m => 1 <= m <= proj1_sig n)
                _ (conj (odd_ge_1 _ (proj2_sig o)) pf))) /\
    (forall (e:pos_evens_t) (pf:proj1_sig e <= proj1_sig n),
       p e = f (exist (fun m => 1 <= m <= (proj1_sig n))
                      _ (conj (pet_ge_1 e) pf))).
  unfold pos_evens_fun_unite_finite_odds'; intros;
  destruct constructive_definite_description; auto.
Qed.
End Unite.


End RandomFacts.


Notation "pr ||-> x" := (sig_fun_app (fst pr) (snd pr) x) (at level 20).

Section RandomFacts'.

  

(*Wyckoff*)
Lemma incl_im_im_left_inverse_compose :
  forall {T U V:Type} (f:T->U) (pff:injective f) (A:Ensemble T)
         (deft:T) (g:T->V),
    Included
      (Im (Im A f)
          (fun u : U => g ((left_inverse f pff A, deft) ||-> u)))
      (Im A g).
intros T U V f h1 A deft g.
red. intros y h2. rewrite im_im in h2.
destruct h2 as [x h2]. subst.
apply Im_intro with x; auto.
simpl. unfold sig_fun_app.
lr_match_goal'.
f_equal. 
erewrite f_equal_dep. 
rewrite left_inverse_compat.
reflexivity.
reflexivity.
contradict hr.
apply Im_intro with x; auto.

Grab Existential Variables.
assumption.
Qed.


(*Wyckoff*)
Lemma in_eq_rect : 
  forall {T U:Type} (pf:T=U) (A:Ensemble T) (x:U),
    In (eq_rect _ (fun V=>Ensemble V) A _ pf) x <->
    In A (eq_rect_r id x pf).
intros; subst; simpl; tauto.
Qed.


(*Wyckoff*)
Lemma in_eq_rect_r : 
  forall {T U:Type} (pf:T=U) (A:Ensemble U) (x:T),
    In (eq_rect_r (fun V=>Ensemble V) A pf) x <->
    In A (eq_rect _ id x _ pf).
intros; subst; simpl; tauto.
Qed.


(*Wyckoff*)
Definition max_set_nat (E:Ensemble nat) (pf:Finite E)
           (pfi:Inhabited E) :=
   max_set_sortable E pf pfi lt lt_sortable.


(*Wyckoff*)           
Lemma max_set_nat_compat : 
  forall (A:Ensemble nat) (pff:Finite A) (pfi:Inhabited A),
    let n := max_set_nat A pff pfi in
      In A n /\
        (forall m:nat,
           In (Subtract A n) m ->
           m < n).
intros A h1 h2 n.
unfold n, max_set_nat, max_set_sortable.
destruct constructive_definite_description as [n' h3].
simpl.
assumption.
Qed.         


(*Wyckoff*)
Lemma max_set_nat_functional :
  forall (A A':Ensemble nat) (pff:Finite A) (pff':Finite A')
         (pfi:Inhabited A) (pfi':Inhabited A'),
    A = A' ->
    max_set_nat A  pff  pfi =
    max_set_nat A' pff' pfi'.
  intros A A' h1 h1' h2 h2' ?.
  subst.
  pose proof (proof_irrelevance _ h1 h1').
  pose proof (proof_irrelevance _ h2 h2').
  subst; auto.
Qed.

(*Wyckoff*)
Lemma max_set_nat_sing :
  forall (a:nat),
    max_set_nat (Singleton a) (Singleton_is_finite _ _) (inh_sing _) =
    a.
  intro a.
  pose proof (max_set_nat_compat
                (Singleton a)
                (Singleton_is_finite _ _) (inh_sing _)) as h1.
  simpl in h1.
  destruct h1 as [h1 h2].
  rewrite in_sing_iff in h1; auto.
Qed.
 

(*Wyckoff*)
Lemma lt_max_set_nat_add :
  forall (A:Ensemble nat) (pf:Finite A) (pfi:Inhabited A)
         (m:nat),
    let pf' := Add_preserves_Finite _ A m pf in
    let pfi' := inh_add A m in
    m = max_set_nat (Add A m) pf' pfi' ->
    ~Inn A m ->
    forall (n:nat),
      Inn A n -> n < m.
  intros A h1 h2 m h3 h4 h5 h6 n h7.
  pose proof (max_set_nat_compat (Add A m) h3 h4) as h8.
  destruct h8 as [h8 h9]. 
  specialize (h9 n).
  inversion h8 as [a hm | a hm].
  rewrite h5 in h6.
  contradiction.
  rewrite in_sing_iff in hm.
  rewrite hm.
  apply h9.
  constructor.
  left; auto.
  intro h10; destruct h10.
  rewrite h5 in h6.
  contradiction.
Qed.


(*Wyckoff*)
Definition max_set_nat_sig_ex : 
  forall {P:nat->Prop} (A:Ensemble {m:nat | P m}),
     Finite A ->
     Inhabited A -> 
     exists! n:{m:nat | P m},
       In A n /\
       (forall r:{m:nat | P m},
          In A r -> (proj1_sig r) <= (proj1_sig n)).
  intros P A h1 h2.
  rewrite <- finite_im_proj1_sig_iff in h1.
  rewrite <- inh_im_proj1_sig_iff in h2.
  pose proof (max_set_nat_compat _ h1 h2) as h3.
  simpl in h3.
  destruct h3 as [h3 h4]. 
  inversion h3 as [n h6 q h7]. subst.
  exists n.
  red. split.
  split; auto.
  intros r h8.
  destruct (nat_eq_dec (proj1_sig r) (proj1_sig n)) as [h9 | h9].
  rewrite h9.
  auto with arith.
  specialize (h4 (proj1_sig r)).
  hfirst h4.
  constructor.
  apply Im_intro with r; auto.
  
  rewrite in_sing_iff.
  rewrite <- h7 in h9.
  apply neq_sym; auto.
  specialize (h4 hf).
  rewrite h7 in h4.
  auto with arith.
  intros m' h8.
  destruct h8 as [h8 h9].
  specialize (h4 (proj1_sig m')).
  specialize (h9 n h6).
  apply le_lt_eq_dec in h9.
  destruct h9 as [h9 | h9].
  hfirst h4.
  constructor.
  apply Im_intro with m'; auto.
  rewrite in_sing_iff.
  intro  h10.
  rewrite <- h7 in h9 at 1.
  omega.
  specialize (h4  hf).
  rewrite h7 in h4.
  pose proof (lt_trans _ _ _ h4 h9) as h10.
  apply lt_irrefl in h10; contradiction.
  apply proj1_sig_injective in h9; auto.
Qed.


(*Wyckoff*)
Definition max_set_nat_sig {P:nat->Prop} (A:Ensemble {m:nat | P m})
           (pff:Finite A) (pfi:Inhabited A) :=
  proj1_sig (constructive_definite_description
               _
               (max_set_nat_sig_ex A pff pfi)).
 
(*Wyckoff*)
Definition max_set_nat_sig_compat :
  forall {P:nat->Prop} (A:Ensemble {m:nat | P m})
         (pff:Finite A) (pfi:Inhabited A),
    let n := max_set_nat_sig A pff pfi in
    In A n /\ (forall r:{m:nat | P m},
                 In A r ->
                 (proj1_sig r) <= (proj1_sig n)).
  intros; unfold n, max_set_nat_sig; intros;
  destruct constructive_definite_description; auto.
Qed.

(*Wyckoff*)
Definition min_set_nat (E:Ensemble nat) (pf:Finite E)
           (pfi:Inhabited E) :=
   min_set_sortable E pf pfi lt lt_sortable.

(*Wyckoff*)
Lemma min_set_nat_functional :
  forall (E E':Ensemble nat)
         (pf:Finite E)
         (pf':Finite E')
         (pfi:Inhabited E)
         (pfi':Inhabited E'),
    E = E' ->
    min_set_nat E pf pfi =
    min_set_nat E' pf' pfi'.
  intros; subst.
  f_equal; apply proof_irrelevance.
Qed.


(*Wyckoff*)
Lemma min_set_nat_compat :
  forall (E:Ensemble nat) (pf:Finite E) (pfi:Inhabited E),
    let m := min_set_nat E pf pfi in
      In E m /\
    (forall n:nat,
       In (Subtract E m) n ->
       m < n).
  intros E h1 h2. simpl.
  unfold min_set_nat.
  pose proof (min_set_sortable_compat  E h1 h2 lt lt_sortable) as h3.
  simpl in h3.
  destruct h3 as [h3 h4].
  split; auto.
Qed.


(*Wyckoff*)
Lemma min_set_nat_add :
  forall (E:Ensemble nat) (pff:Finite E) (pfi:Inhabited E) (m:nat),
    min_set_nat (Add E m) (Add_preserves_Finite _ _ _ pff) (inh_add E m) =
    min (min_set_nat E pff pfi) m.
  intros E h1 h2 m.
  unfold min_set_nat.
  pose proof (min_set_sortable_add
                E m
                (Add_preserves_Finite _ _ _ h1)
                (inh_add _ _)
                lt (sort_intro _ tso_lt lt_trans lt_irrefl)) as h3.
  destruct h3 as [h3 h4].
  specialize (h3 h2).
  simpl in h3.
  unfold lt_sortable. rewrite h3.
  unfold minR.    
  lr_if_goal; fold hlr; destruct hlr as [[ha | ha] | ha].
  rewrite min_l.
  f_equal.
  apply proof_irrelevance.
  erewrite min_set_sortable_functional; auto.
  apply lt_le_weak in ha.
  apply ha; auto.
  rewrite min_r.
  rewrite <- ha.
  apply min_set_sortable_functional; auto.
  erewrite min_set_sortable_functional in ha; auto.
  rewrite <- ha.
  apply le_refl.
  rewrite min_r; auto with arith.
  erewrite min_set_sortable_functional in ha; auto.
  apply lt_le_weak in ha.
  apply ha.
Qed.


(*Wyckoff*)
Lemma min_set_nat_0 :
  forall (E:Ensemble nat) (pff:Finite E) (pfi:Inhabited E),
    Inn E 0 ->
    min_set_nat E pff pfi = 0.
  intros E h1 h2 h3.
  destruct (zerop (min_set_nat E h1 h2)) as [h4 | h4]; auto.
  pose proof (min_set_nat_compat E h1 h2) as h5.
  simpl in h5.
  destruct h5 as [h5 h6].
  specialize (h6 0).
  hfirst h6.
  constructor; auto.
  rewrite in_sing_iff.
  omega.
  specialize (h6 hf).
  omega.
Qed.



(*Wyckoff*)
Lemma in_im_seg_sig :
  forall {T:Type} {D:Ensemble T} (f:nat->sig_set D) (n:nat)
         (x:sig_set (Im (seg n) (fun i => proj1_sig (f i)))),
    Inn D (proj1_sig x).
  intros T D f n x.
  destruct x as [x hx]; simpl.
  destruct hx as [x]; subst.
  apply proj2_sig.
Qed.


(*Wyckoff*)
Definition inv_im_seg_min
           {U:Type}
           (f:nat->U) (n:nat)
           (y:sig_set (Im (seg n) f)) :=
  min_set_nat (inv_im_seg (fun_to_sig_seg f n) y)
              (finite_inv_im_seg (fun_to_sig_seg f n) y)
              (inh_inv_im_sig_seg f n y).



(*Wyckoff*)
Lemma inv_im_seg_min_compat :
  forall {U:Type}
         (f:nat->U) (n:nat)
         (y:sig_set (Im (seg n) f)),
    f (inv_im_seg_min f n y) = proj1_sig y.
  intros U f n y.
  unfold inv_im_seg_min.
  pose proof (min_set_nat_compat (inv_im_seg (fun_to_sig_seg f n) y)
        (finite_inv_im_seg (fun_to_sig_seg f n) y) 
        (inh_inv_im_sig_seg f n y)) as h1.
  simpl in h1.
  destruct h1 as [h1 h2].
  destruct h1 as [h1].
  destruct h1 as [h1 h3].
  rewrite fun_to_sig_seg_compat in h3.
  rewrite <- h3 at 4.
  simpl; auto.
Qed.


(*Wyckoff*)
Lemma inv_im_seg_min_compat' :
  forall {U:Type}
         (f:nat->U) (n:nat)
         (y:sig_set (Im (seg n) f)),
    Inn (inv_im1 f (proj1_sig y)) (inv_im_seg_min f n y).
  intros U f n y.
  unfold inv_im1.
  constructor.
  apply inv_im_seg_min_compat.
Qed.



(*Wyckoff*)
Lemma inv_im_seg_min_sig_compat :
  forall {U:Type} {E:Ensemble U}
         (f:nat->sig_set E) (n:nat)
         (y:sig_set (Im (seg n) (proj1_sig_fun f))),
    f (inv_im_seg_min (proj1_sig_fun f) n y) =
    exist _ _ (in_proj1_sig_fun f _ _ (proj2_sig y)).
  intros U E f n y.
  pose proof (inv_im_seg_min_compat (proj1_sig_fun f) n y) as h1.
  rewrite proj1_sig_fun_compat in h1.
  apply proj1_sig_injective; auto.
Qed.


(*Wyckoff*)
Lemma lt_inv_im_seg_min :
  forall {U:Type}
         (f:nat->U) (n:nat)
         (y:sig_set (Im (seg n) f)),
   inv_im_seg_min f n y < n.
  intros U f n y.
  unfold inv_im_seg_min.
  pose proof (min_set_nat_compat
                (inv_im_seg (fun_to_sig_seg f n) y)
                (finite_inv_im_seg (fun_to_sig_seg f n) y)
                (inh_inv_im_sig_seg f n y)) as h1.
  simpl in h1.
  destruct h1 as [h1 h2].
  destruct h1 as [h1].
  destruct h1; auto.
Qed.

(*Wyckoff*)
Lemma in_inv_im_seg_min :
  forall {U:Type} (f:nat->U) (pf:surjective f) (n:nat)
         (y:sig_set (Im (seg n) f)),
    Inn (inv_im_seg (fun_to_sig_seg f n) y) (inv_im_seg_min f n y).
  intros; constructor.
  exists (lt_inv_im_seg_min _ _ _).
  destruct y as [y hy].
  destruct hy as [i hy]; subst.
  rewrite fun_to_sig_seg_compat.
  apply proj1_sig_injective; simpl.
  rewrite inv_im_seg_min_compat; simpl; auto.
Qed.


(*Wyckoff*)
Lemma le_inv_im_seg_min :
  forall {U:Type} (f:nat->U) (pf:surjective f) (i n:nat)
         (pf:Inn (seg n) i),
    inv_im_seg_min
      f n
      (exist _ _ (Im_intro _ _ (seg n) f i pf _ eq_refl))  <= i.
  intros U f h1 i n h2.
  destruct h2 as [h2].
  unfold inv_im_seg_min. 
  glel.
  redterm0 xle.
  redterm1 xle.
    pose proof (min_set_nat_compat _ c0 c) as h3.
  unfold c, c0 in h3; simpl in h3. 
  destruct h3 as [h3 h4].
  destruct (le_lt_dec xle i) as [h5 | h5].
  assumption.
  specialize (h4 i).
  hfirst h4.
  constructor.
  constructor.
  exists h2.
  rewrite fun_to_sig_seg_compat.
  apply proj1_sig_injective; auto.
  rewrite in_sing_iff.
  intro h6.
  unfold xle in h5.
  rewrite h6 in h5.
  apply lt_irrefl in h5; auto.
  specialize (h4 hf).
  omega.
Qed.


(*Wyckoff*)
Definition nat_sig_to_fun 
           {U V:Type}
           {D:Ensemble U} {E:Ensemble V}
           (f:nat->sig_set D)
           (g:nat->sig_set E)
           (n:nat) :
  sig_set (Im (seg n) (proj1_sig_fun f)) -> V :=
             fun x => proj1_sig
                        (g (inv_im_seg_min (proj1_sig_fun f) n x)).

           
(*Wyckoff*)
Lemma bij_seg :
  forall {T:Type} (A:Ensemble T),
    Finite A ->
    exists f:sig_set (seg (card_fun1 A)) ->
             sig_set A,
      bijective f.
intros T A h1.
induction h1 as [|A h1 h2 x h3].
rewrite card_fun1_empty.
exists (fun x:sig_set (seg 0) =>
          False_rect (sig_set (Empty_set T)) (nin_seg_0 _ (proj2_sig x))).
red. split.
red. intros x y h1.
destruct x as [x h2].
pose proof (nin_seg_0 _ h2). contradiction.
red.
intro y.
destruct y as [y h1].
contradict h1.
intro h1. contradiction. 
rewrite card_add_nin'; auto.
rewrite seg_Sn.
destruct h2 as [f h2].
red in h2.
destruct h2 as [h2a h2b].
exists (fun c:sig_set (Add (seg (card_fun1 A)) (card_fun1 A)) =>
        match (in_add_nin_dec _ _ (nin_seg _)  _ (proj2_sig c)) with
          | left P =>
            exist
              (fun a => In (Add A x) a) _
              (Add_intro1 _ _ x _ (proj2_sig (f (exist _ _ P))))
          | right _ => exist _ _ (Add_intro2 _ A x)
        end).
red. split.
red.
intros m n h0.
lr_match' h0.
rename hl into h4.
lr_match' h0.
rename hl into h6.
apply exist_injective in h0.
apply proj1_sig_injective in h0.
apply h2a in h0.
apply exist_injective in h0.
apply proj1_sig_injective in h0.
assumption.
apply exist_injective in h0.
subst.
contradict h3.
apply proj2_sig.
rename hr into h5.
lr_match' h0.
rename hl into h6.
apply exist_injective in h0.
subst.
contradict h3.
apply proj2_sig.
rename hr into h7.
rewrite <- h7 in h5.
apply proj1_sig_injective in h5.
assumption.
red. intro y.
destruct y as [y h4].
destruct h4 as [y h4 | y h5].
red in h2b.
specialize (h2b (exist _ _ h4)).
destruct h2b as [a h2b].
destruct a as [a h6].
exists (exist _ _ (Add_intro1 _ _ (card_fun1 A) _ h6)).
simpl.
lr_match_goal'.
rename hl into h7.
apply proj1_sig_injective; simpl.
pose proof (f_equal (@proj1_sig _ _) h2b) as h8.
simpl in h8.
subst.
f_equal. f_equal. apply proj1_sig_injective; auto.
apply proj1_sig_injective; simpl.
subst.
pose proof (nin_seg _ h6).
contradiction.
destruct h5.
exists (exist _ _ (Add_intro2 _ (seg (card_fun1 A)) (card_fun1 A))).
simpl.
lr_match_goal'.
rename hl into h4.
apply proj1_sig_injective; simpl.
pose proof (nin_seg _ h4).
contradiction.
apply proj1_sig_injective; simpl.
reflexivity.
Qed.


(*Use is made of Full_set rather than T to exploit the previous
  theorem easier. See [SetUtilities2], for the version of this with 
  an arbitrary set instead of Full_set. *)
(*Wyckoff*)
Lemma bij_seg_inh_full_iff : 
  forall {T:Type} 
         (f:sig_set (seg (card_fun1 (Full_set T))) -> 
            sig_set (Full_set _)),
    bijective f ->
    forall (P:T->Prop),
      (Inhabited [x:T | P x] <-> 
       Inhabited [n : nat
        | exists pf : In (seg (card_fun1 (Full_set T))) n,
            P
              (proj1_sig
                 (f (exist _ _ pf)))]).
intros T f h1 P.
apply bijective_impl_invertible in h1.
pose (proj1_sig (function_inverse _ h1)) as f'.
split.
intro h2.
destruct h2 as [x h2].
destruct h2 as [h2].
apply Inhabited_intro with (proj1_sig (f' (exist _ _ (Full_intro _ x)))).
constructor.
exists (proj2_sig _).
pose proof (proj2_sig (function_inverse f h1)) as h3.
simpl in h3.
destruct h3 as [h3 h4].
gterm0.
assert (h5:c = x).
  unfold c.
  rewrite <- unfold_sig.
  unfold f'.
  rewrite h4.
  simpl. 
  reflexivity.
unfold c in h5.
rewrite h5 at 1.
assumption.

intro h2.
destruct h2 as [n h2].
destruct h2 as [h2].
destruct h2 as [h2 h3].
apply Inhabited_intro with (proj1_sig (f (exist _ _ h2))).
constructor.
assumption.
Qed.


Definition Im_iter {T:Type} (f:T->T) (x0:T) (n:nat): Ensemble T :=
  Im (seg n) (fun m:nat => iter_uo f x0 m).

Lemma im_iter_zero : 
  forall {T:Type} (f:T->T) (x0:T),
    Im_iter f x0 0 = Empty_set _.
intros T f x0.
unfold Im_iter.
rewrite seg_O.
rewrite image_empty.
reflexivity.
Qed.

Lemma im_iter_Sn : 
  forall {T:Type} (f:T->T) (x0:T) (n:nat),
    Im_iter f x0 (S n) = Add (Im_iter f x0 n) (iter_uo f x0 n).
intros T f x0 n.
unfold Im_iter.
rewrite seg_Sn.
rewrite Im_add.
reflexivity.
Qed.

Lemma finite_im_iter : 
  forall {T:Type} (f:T->T) (x0:T) (n:nat),
    Finite (Im_iter f x0 n).
intros T f x0 n.
unfold Im_iter.
apply finite_image.
apply finite_seg.
Qed.

Lemma im_iter_strict_order_Sn : 
  forall {T:Type} (f:T->T) (x0:T) (n:nat)
         (R:T->T->Prop),
    Strict_Order R -> 
    (forall x:T, R (f x) x) ->
    R (iter_uo f x0 (S n)) (iter_uo f x0 n).
intros T f x0 n R h1 h2.
apply h2.
Qed.

Lemma im_iter_strict_order_preservers_lt : 
  forall {T:Type} (f:T->T) (x0:T) (m n:nat),
    m < n ->
    forall (R:T->T->Prop),
      Strict_Order R -> 
      (forall x:T, R (f x) x) ->
      R (iter_uo f x0 n) (iter_uo f x0 m).
intros T f x0 m n. revert m.
induction n as [|n h1].
intros; omega.
intros m h2 R h3 h4.
red in h2.
apply le_S_n in h2.
apply le_lt_eq_dec in h2.
destruct h2 as [h5 | h6].
specialize (h1 m h5 R h3 h4). 
simpl.
pose proof (h4 (iter_uo f x0 n)) as h6.
destruct h3 as [h3a h3b].
eapply h3a. apply h6. apply h1.
subst.
apply h4.
Qed.



Lemma card_im_iter_strict_order : 
  forall {T:Type} (f:T->T) (x0:T) (n:nat)
         (R:T->T->Prop),
    Strict_Order R ->
    (forall x:T, R (f x) x) ->
    card_fun1 (Im_iter f x0 n) = n.
intros T f x0 n R h1 h2.
revert R h1 h2.
induction n as [|n h1].

intros R h1 h2.
rewrite im_iter_zero.
rewrite card_fun1_empty.
reflexivity.
intros R h2 h3.
specialize (h1 R h2 h3).
rewrite im_iter_Sn.
rewrite card_add_nin'.
rewrite h1. reflexivity.
apply finite_im_iter.
intro h4.
inversion h4 as [m h5 x h6].  
clear h1.
subst.
destruct h5 as [h5].
pose proof (im_iter_strict_order_preservers_lt f x0 m n h5 R h2 h3) as h8.
rewrite h6 in h8.
destruct h2 as [h2a h2b].
apply h2b in h8; auto.
Qed.


Lemma strict_order_decreasing_uo_infinite : 
  forall {T:Type},
    Inhabited (Full_set T) ->
    forall (f:T->T) (R:T->T->Prop),
      Strict_Order R ->
      (forall x:T, R (f x) x) ->
      ~Finite (Full_set T).
intros T h1 f R h2 h3 h4.
pose (card_fun1 (Full_set T)) as n.
destruct h1 as [x h1].
pose proof (card_im_iter_strict_order f x (S n) R h2 h3) as h5.
pose proof (incl_card_fun1 (Im_iter f x (S n)) _ h4) as h7.
hfirst h7.
red; intros; constructor.
specialize (h7 hf).
rewrite h5 in h7.
unfold n in h7.
omega.
Qed.




(*Wyckoff*)
Lemma le_prop_impl : 
  forall {T:Type} {R:T->T->Prop},
    Transitive R ->
    forall (t t':T),
      R t t' ->
      forall (x:T), (x = t \/ R x t) ->
                    (x = t' \/ R x t').
intros T R h1 t t' h2 x h3.
destruct h3 as [h3 | h4]; subst.
right; auto.
red in h1. specialize (h1 _ _ _ h4 h2).
right; auto.
Qed.


(*Wyckoff*)
Definition le_sig_conv_ex :  
  forall {T:Type} {R:T->T->Prop} {t t':T},
    Transitive R -> 
    forall (pf:R t t')
           (x':{x:T | t = x  \/ R x t}),
  exists ! x'':{x:T | t' = x \/ R x t'},
      proj1_sig x' = proj1_sig x''.
intros T R t t' h0 h1 x'.
destruct x' as [x' h2].
destruct h2 as [h2 | h3]. subst.
exists (exist (fun x:T => t' = x \/ R x t') _ (or_intror h1)).
red; simpl; split.
reflexivity.
intros x'' h2.
destruct x'' as [x'' h3]. subst.
simpl in h1. simpl.
apply proj1_sig_injective; auto.
red in h0. specialize (h0 _ _ _ h3 h1).
exists (exist (fun x:T => t' = x \/ R x t') _ (or_intror h0)).
red; split; simpl; auto.
intros x'' h4.
destruct x'' as [x'' h5]. simpl in h4.
subst.
apply proj1_sig_injective; auto.
Qed.


(*Wyckoff*)
Definition le_sig_conv {T:Type} {R:T->T->Prop} {t t':T}
           (pft:Transitive R) (pfr:R t t')
           (x':{x:T | t = x  \/ R x t}) : {x:T | t' = x \/ R x t'} :=
  proj1_sig (constructive_definite_description
               _
               (le_sig_conv_ex pft pfr x')).

(*Wyckoff*)
Lemma le_sig_conv_cond :
  forall {T:Type} {R:T->T->Prop} {t t':T}
         (pft:Transitive R) (pfr:R t t')
         (x':{x:T | t = x  \/ R x t}),
    let x'' := le_sig_conv pft pfr x' in
    proj1_sig x' = proj1_sig x''.
intros.
unfold x'', le_sig_conv.
destruct constructive_definite_description as [z h2].
simpl.
assumption.
Qed.


(*Wyckoff*)
Lemma overlapping_functions_constructive_definite_description : 
  forall {T U:Type} (R:T->T->Prop) (pft:Transitive R)
         (P:{x:T & ({y : T | y = x \/ R y x} -> U)}->Prop)
         (pf: forall x : T,
              exists ! f : {y : T | y = x \/ R y x} -> U,
                P (existT _ x f)),
    let psi := (fun x:T =>
                  proj1_sig
                    (constructive_definite_description _ (pf x))) in
    (forall (x y:T) (pfr:R x y)
         (f:{y0 : T | y0 = y \/ R y0 y}->U),
         let f' := restriction_sig_prop 
                     (fun t:T => t = x \/ R t x) 
                     (fun t:T => t = y \/ R t y)
                     (le_prop_impl pft _ _ pfr) f

         in 
         P (existT _ y f) -> P (existT _ x f')) ->
      forall (x y:T) (pfr:R x y),
         psi x (exist _ _ (or_introl eq_refl)) = 
         psi y (exist _ _ (or_intror pfr)).
intros T U R h0 P h1 psi hp x y h2. 
unfold psi. 
destruct constructive_definite_description as [f h3]. 
destruct constructive_definite_description as [g h4].
simpl.
specialize (hp x y h2 g h4).
pose proof (h1 x) as h5.
destruct h5 as [f' h5]. red in h5.
destruct h5 as [h5 h6].  
apply h6 in hp.  subst.  
apply h6 in h3.
subst.
unfold restriction_sig_prop.
f_equal.
apply proj1_sig_injective; auto.
Qed.


End RandomFacts'.


Arguments FamilyUnion [T] F _.
Arguments FamilyIntersection [T] F _.
Arguments IndexedUnion [It] [T] FI _.
Arguments IndexedIntersection [It] [T] FI _.
Arguments Union_associative [Xt] N M P.
Arguments assoc_prod_psa [T] N M P.
Arguments Union_commutative [Xt] N M. 
Arguments abs_sum_psa [T] N M. 
Arguments abs_prod_psa [T] N M.
Arguments comp_sum_psa [Xt] N. 
Arguments comp_prod_psa [Xt] N.

Section ArithSets.

(*Wyckoff*)
Definition lb_nat (A:Ensemble nat) (b:nat) :=
  forall n:nat, In A n -> b <= n.

(*Wyckoff*)
Lemma lb_nat_lt : 
  forall (A:Ensemble nat) (b:nat),
    lb_nat A b ->
    forall (b':nat), b' < b -> lb_nat A b'.
intros A b h1 b' h2.
red in h1. red.
intros n h3.
specialize (h1 n h3).
omega.
Qed.


(*Wyckoff*)
Definition glb_nat (A:Ensemble nat) (b:nat) :=
  lb_nat A b /\ forall b':nat, lb_nat A b' ->
  b' <= b.

(*Wyckoff*)
Definition nat_translate (n:nat) := [m:nat | m >= n].

(*Wyckoff*)
Lemma nat_translate_inj : 
  forall (m n:nat),
    nat_translate m = nat_translate n -> m = n.
intros m n h1.
destruct (lt_eq_lt_dec m n) as [h2 | h3].
destruct h2 as [h2 | h3]. 
assert (h3:m <= pred n).
  destruct (zerop n) as [h3 | h4]. subst. simpl. omega.
  omega.
assert (h4:In (nat_translate m) (pred n)). 
  constructor. omega.
rewrite  h1 in h4.
destruct h4 as [h4].
assert (h5:n = 0). omega. subst.
omega.
assumption.
assert (h3':n <= pred m).
  destruct (zerop m) as [h3' | h4']. subst. simpl. omega.
  omega.
assert (h4:In (nat_translate n) (pred m)). 
  constructor. omega.
rewrite <- h1 in h4.
destruct h4 as [h4].
assert (h5:n = 0). omega. subst.
omega.
Qed.


(*Wyckoff*)
Definition nat_translatable (S:Ensemble nat) :=
  exists n:nat, S = nat_translate n.

(*Wyckoff*)
Definition nat_translatable' (S:Ensemble nat) :=
  exists! n:nat, S = nat_translate n.

(*Wyckoff*)
Definition nat_translatable_endpoint (S:Ensemble nat) (pf:nat_translatable' S) :=
  proj1_sig (constructive_definite_description _ pf).

(*Wyckoff*)
Lemma nat_translatable_endpoint_compat :  
  forall (S:Ensemble nat) (pf:nat_translatable' S),
    S = nat_translate (nat_translatable_endpoint S pf).
intros S h1.
unfold nat_translatable_endpoint.
destruct constructive_definite_description as [n h2].
subst. simpl.
reflexivity.
Qed.


(*Wyckoff*)
Lemma nat_translatable_prime_compat : 
  forall S:Ensemble nat,
    nat_translatable S <-> nat_translatable' S.
intro S. 
split.
intro h1.
destruct h1 as [n h1]. subst.
red. exists n. red. split; auto.
intros n' h1.
apply nat_translate_inj in h1. assumption.
intro h1.
destruct h1 as [n h1].
red in h1. destruct h1 as [h1 h2].
subst.
red.
exists n.
reflexivity.
Qed.


(*Wyckoff*)
Lemma nat_translate_0 : nat_translate 0 = Full_set nat.
apply Extensionality_Ensembles.
red; split; red; intros n h1; try constructor; omega.
Qed.


(*Wyckoff*)
Lemma nat_translatable_iff : 
  forall A:Ensemble nat,
    Inhabited A ->
    (nat_translatable A <->
    (forall m n:nat,
      In A m -> m <= n -> In A n)).
intros A h1. split.
intros h2 m n h3 h4.
red in h2.
destruct h2 as [r h2]. subst.
destruct h3 as [h3].
constructor.
omega.
intro h2.
destruct h1 as [n h1].
red.
revert h2 h1. revert A.
induction n as [|n h1].
intros A h1 h2.
exists 0. rewrite nat_translate_0.
apply Extensionality_Ensembles; red; split; intros m h3; try constructor.
clear h3.
apply (h1 0 m h2). omega.
intros A h2 h3.
specialize (h1 A h2).
destruct (classic (In A n)) as [h4 | h5].
apply h1; auto.
exists (S n).
apply Extensionality_Ensembles.
red. split.
red. intros m h6.
constructor.
destruct (lt_eq_lt_dec m (S n)) as [h7 | h8].
destruct h7 as [h7 | h9].
assert (h8:m <= n). omega.
specialize (h2 m n h6 h8). contradiction. subst.
omega. 
omega.
red. intros m h6.
destruct h6 as [h6].
apply (h2 _ _ h3).
omega.
Qed.


(*Wyckoff*)
Definition nat_translatable_glb_compat : 
  forall (S:Ensemble nat) (pf:nat_translatable' S),
    glb_nat S (nat_translatable_endpoint S pf).
intros S h1.
unfold nat_translatable_endpoint.
destruct constructive_definite_description as [b h2]. subst.
simpl.
red. split.
red. intros n h2. destruct h2. auto with arith.
intros b' h2.
red in h2.
apply h2. constructor. 
auto with arith.
Qed.


(*Wyckoff*)
Lemma step_le_compat : 
  forall A:Ensemble nat,
    (forall m n:nat,
      In A m -> m <= n -> In A n) <->
    (forall n:nat,
      In A n -> In A (S n)).
intro A.
split.
intro h1.
intros n h2.
apply (h1 n (S n) h2). omega.
intro h1. 
intros m n. revert m h1. revert A.
induction n as [|n ih].
intros.
assert (m=0). omega. subst.
assumption.
intros A m h1 h2 h3.
specialize (ih A). 
pose proof (le_lt_eq_dec _ _ h3) as h4.
destruct h4 as [h4 | h5].
assert (h5:m <= n). omega. clear h4.
specialize (ih _ h1 h2 h5).
apply h1; auto. subst.
assumption.
Qed.


(*Wyckoff*)
Lemma nat_translatable_iff' : 
  forall A:Ensemble nat,
    Inhabited A ->
    (nat_translatable A <->
    (forall n:nat,
      In A n -> In A (S n))).
intros A h0. 
rewrite nat_translatable_iff; auto.
rewrite step_le_compat.
reflexivity.
Qed.


(*Wyckoff*)
Lemma nat_translatable_boundary' : 
  forall (P:nat->Prop),
    P 0 ->
    Inhabited [n:nat | ~ P n] ->
    nat_translatable [n:nat | ~ P n] ->
    exists ! n:nat, P n /\ ~ P (S n).
intros P h0 h1 h2. 
pose proof h2 as h2'.
rewrite nat_translatable_prime_compat in h2.
pose proof (nat_translatable_glb_compat _ h2) as h3.
rewrite nat_translatable_iff' in h2'.
destruct (zerop (nat_translatable_endpoint [n:nat | ~ P n] h2)) as [h4 | h5].
pose proof (nat_translatable_endpoint_compat [n:nat | ~ P n] h2) as h5.
rewrite h4 in h5.
rewrite nat_translate_0 in h5.
pose proof (Full_intro _ 0) as h6.
rewrite <- h5 in h6. destruct h6. contradiction.
exists (pred (nat_translatable_endpoint [n : nat | ~ P n] h2)).
red. split.
split.  
red in h3. destruct h3 as [h3 hl].
red in h3.
apply NNPP.
intro h6.
assert (h7:In [n0:nat | ~ P n0]
              (pred (nat_translatable_endpoint [n : nat | ~ P n] h2))).
  constructor. assumption.
specialize (h3 _ h7).
omega.
rewrite <- (S_pred _ _ h5).
  pose proof (nat_translatable_endpoint_compat [n:nat | ~ P n] h2) as he.
  assert (h6:In (nat_translate
                   (nat_translatable_endpoint
                      [n : nat | ~ P n] h2))
                (nat_translatable_endpoint [n : nat | ~ P n] h2)).
  constructor. auto with arith.
rewrite <- he in h6.
destruct h6 as [h6].
assumption.
rename h5 into hlt.
intros n h5. destruct h5 as [h5 h6]. 
apply S_inj. rewrite <- (S_pred _ _ hlt).
red in h3.
destruct h3 as [h3a h3b].
destruct (lt_eq_lt_dec
            (nat_translatable_endpoint [n0 : nat | ~ P n0] h2)
            (S n)) as [h7 | h8].
destruct h7 as [h7 | h8].

assert (h8:lb_nat [n0 : nat | ~ P n0] n).
  red. intros m h8.
  destruct h8 as [h8].
  destruct (lt_eq_lt_dec n m) as [h9 | h10].
  destruct h9 as [h9 | h11]. omega. subst. contradiction.
  assert (h11:In [n : nat | ~ P n] m).
    constructor; auto.
  rewrite <- step_le_compat in h2'.

  assert (h13:m <= n). omega.
  specialize (h2' m n h11 h13).
  destruct h2'. contradiction.
specialize (h3b n h8).
assert (h9: n = nat_translatable_endpoint [n0 : nat | ~ P n0] h2).
  omega.
subst.
contradict h5.
pose proof (nat_translatable_endpoint_compat [n:nat | ~ P n] h2) as h5. 
assert (h6':In (nat_translate
                  (nat_translatable_endpoint
                     [n : nat | ~ P n] h2))
               (nat_translatable_endpoint [n : nat | ~ P n] h2)).
  constructor. auto with arith.
rewrite <- h5 in h6'.
destruct h6'. assumption.
assumption.
assert (h9:In [n:nat | ~ P n] (S n)). constructor; auto.
pose proof (nat_translatable_endpoint_compat [n:nat | ~ P n] h2) as h5'.
rewrite h5' in h9.
destruct h9 as [h9].
omega.
assumption.
Qed.


End ArithSets.


(*Wyckoff*)
Definition odd_nat1t_fun_unite_finite_even 
           {T:Type} (p:odd_nat1t->T) (n:even_nat1t)
           (p':{m:even_nat1t | en1t_le m n}->T) : 
  seg_one_t (nat_of_en1t n) -> T.
intro m.
destruct m as [m h1].
destruct (even_odd_dec m) as [h2 | h2].
assert (h3:0 < m). 
  destruct h1. auto with sets.
pose (gt0_to_1t h3) as m'.
assert (h4: even1t m').
  red. unfold m'. 
  rewrite match_gt0_to_1t_even_iff.
  rewrite gt0_to_1t_compat.
  assumption.
pose (even_nat1t_intro _ h4) as m''.
pose proof (nat_of_en1t_le_compat m'' n) as h5.
assert (h6: nat_of_en1t m'' <= nat_of_en1t n).
  simpl. unfold m'. rewrite gt0_to_1t_compat.
  destruct h1; auto.
specialize (h5 h6).
refine (p' (exist (fun m => en1t_le m n) _ h5)).
assert (h3:0 < m).
destruct h1; auto with arith.
assert (h4:odd1t (gt0_to_1t h3)).
  assert (h5:~even m).
    intro h6. apply (not_even_and_odd m h6 h2); auto.
  rewrite (even_gt0_to_1t_iff m h3) in h5.
  destruct (even_odd_dec_1t (gt0_to_1t h3)) as [|h6]; try contradiction.
  assumption.
refine (p (odd_nat1t_intro _ h4)).
Defined.

(*Wyckoff*)
Lemma odd_nat1t_fun_unite_finite_even_compat : 
  forall {T:Type} (p:odd_nat1t->T) (n:even_nat1t)
         (p':{m:even_nat1t | en1t_le m n}->T),
    let f := odd_nat1t_fun_unite_finite_even p n p' in
      (forall (o:odd_nat1t) 
              (pf:nat_of_on1t o <= nat_of_en1t n),
              p o = f (exist (fun m => (1 <= m <= nat_of_en1t n))
                             _ (conj (le_1_odd1t o) pf))) /\
      (forall (e:even_nat1t)
              (pf:nat_of_en1t e <= nat_of_en1t n),
         p' (exist (fun m => en1t_le m n) _
                   (nat_of_en1t_le_compat _ _ pf)) = 
         f (exist (fun m => (1 <= m <= nat_of_en1t n))
                  _ (conj (en1t_ge_1 e) pf))).
intros T p n p' f. split.
intros o h1.
unfold f, odd_nat1t_fun_unite_finite_even.
destruct even_odd_dec as [h2 | h2].
destruct o as [o' h3]. simpl in h2.
red in h3.
destruct o' as [o'' o'].
simpl in h2. red in h3.
pose proof (not_even_and_odd _ h2 h3). contradiction.
f_equal.
destruct even_odd_dec_1t as [h3 | h3].
red in h3.
pose proof (iff1 (match_gt0_to_1t_even_iff _ (le_1_odd1t o)) h3) as h3'.
rewrite gt0_to_1t_compat in h3'.
destruct o as [o' o]. destruct o' as [o'' o'].
simpl in h3', h2.
pose proof (not_even_and_odd _ h3' h2). contradiction.  

destruct o as [o' o].
destruct o' as [o'' o'].
apply odd_nat1t_functional.
apply nat1t_functional; simpl.
unfold gt0_to_1t.
destruct constructive_definite_description as [a h5]; simpl.
destruct a; simpl in h5; auto.

intros e h1.
unfold f, odd_nat1t_fun_unite_finite_even.
destruct even_odd_dec as [h2 | h2].
f_equal.
apply proj1_sig_injective; simpl.
destruct e as [e' e]. destruct e' as [e'' e'].
apply even_nat1t_functional.
apply nat1t_functional.
unfold gt0_to_1t.
destruct Description.constructive_definite_description as [n' h3]; simpl.
destruct n'. simpl in h3; auto.
destruct e  as [e' h3]. destruct e' as [e'' e'].
simpl. simpl in h2.
red in h3. red in h3.
pose proof (not_even_and_odd _ h3 h2). contradiction.
Qed.


(*Wyckoff*)
Definition odd_nat1t_fun_unite_finite_even' 
           {T:Type} (p:odd_nat1t->T) (n:odd_nat1t)
           (p':{m:even_nat1t | en1t_le m (on1t_pred' n)}->T) : 
  seg_one_t (nat_of_on1t n) -> T.
intro m.
destruct m as [m h1].
destruct (even_odd_dec m) as [h2 | h2].
assert (h3:0 < m). 
  destruct h1. auto with sets.
pose (gt0_to_1t h3) as m'.
assert (h4: even1t m').
  red. unfold m'. 
  rewrite match_gt0_to_1t_even_iff.
  rewrite gt0_to_1t_compat.
  assumption.
pose (even_nat1t_intro _ h4) as m''. 
pose proof (nat_of_en1t_le_pred_compat m'' n) as h5.
assert (h6: nat_of_en1t m'' <= nat_of_on1t n ).
  unfold m''. simpl. unfold m'.
  rewrite gt0_to_1t_compat.
  destruct h1; auto.
specialize (h5 h6).
refine (p' (exist (fun m => en1t_le m (on1t_pred' n)) _ h5)).
refine (p _).
refine (odd_to_on1t _).
apply h2.
Defined.

(*Wyckoff*)
Lemma odd_nat1t_fun_unite_finite_even_compat' : 
  forall {T:Type} (p:odd_nat1t->T) (n:odd_nat1t)
         (p':{m:even_nat1t | en1t_le m (on1t_pred' n)}->T),
    let f := odd_nat1t_fun_unite_finite_even' p n p' in
      (forall (o:odd_nat1t) 
              (pf:nat_of_on1t o <= nat_of_on1t n),
              p o = f (exist (fun m => (1 <= m <= nat_of_on1t n))
                             _ (conj (le_1_odd1t o) pf))) /\
      (forall (e:even_nat1t)
              (pf:nat_of_en1t e <= nat_of_on1t n),
              p' (exist (fun m => en1t_le m (on1t_pred' n)) _ 
                         (nat_of_en1t_le_pred_compat _ _ pf)) = 
              f (exist (fun m => (1 <= m <= nat_of_on1t n))
                       _ (conj (en1t_ge_1 e) pf))).
intros T p n p' f. split.
intros o h1.
unfold f, odd_nat1t_fun_unite_finite_even'.
destruct even_odd_dec as [h2 | h2].
destruct o as [o' h3]. simpl in h2.
red in h3.
destruct o' as [o'' o'].
simpl in h2. red in h3.
pose proof (not_even_and_odd _ h2 h3). contradiction.
f_equal.
destruct o as [o' o].
destruct o' as [o'' o'].
apply odd_nat1t_functional.
unfold odd_to_on1t, odd_to_1t. simpl.
apply nat_of_1t_inj.
simpl.
rewrite gt0_to_1t_compat.
reflexivity.

intros e h1.
unfold f, odd_nat1t_fun_unite_finite_even'.
destruct even_odd_dec as [h2 | h2].
f_equal.
apply proj1_sig_injective; simpl.
destruct e as [e' e]. destruct e' as [e'' e']. 
apply even_nat1t_functional.
apply nat1t_functional.
unfold gt0_to_1t.
destruct constructive_definite_description as [n' h3].
simpl.
destruct n'; simpl in h3; auto.
destruct e  as [e' h3]. destruct e' as [e'' e'].
simpl; simpl in h2.
do 2 red in h3.
pose proof (not_even_and_odd _ h3 h2). contradiction.
Qed.


(*Wyckoff*)
Definition even_nat1t_fun_unite_finite_odd 
           {T:Type} (p:even_nat1t->T) (n:odd_nat1t)
           (p':{m:odd_nat1t | on1t_le m n}->T) : 
  seg_one_t (nat_of_on1t n) -> T.
intro m.
destruct m as [m h1].
destruct (even_odd_dec m) as [h2 | h2].
assert (h3:0 < m).
  destruct h1. auto with sets.
assert (h4:even1t (gt0_to_1t h3)).
  red.
  rewrite match_gt0_to_1t_even_iff.
  rewrite gt0_to_1t_compat.
  assumption.
refine (p (even_nat1t_intro _ h4)).
pose (odd_to_1t h2) as m'.
pose proof (odd_to_1t_compat h2) as h4.
pose (odd_nat1t_intro _ h4) as m''.
assert (h6:nat_of_on1t m'' <= nat_of_on1t n).
  unfold m''. simpl.
  unfold odd_to_1t.
  rewrite gt0_to_1t_compat.
  destruct h1; auto.
pose proof (nat_of_on1t_le_compat m'' n) as h5.
specialize (h5 h6).
refine (p' (exist (fun m => on1t_le m n) _ h5)).
Defined.

(*Wyckoff*)
Lemma even_nat1t_fun_unite_finite_odd_compat : 
  forall {T:Type} (p:even_nat1t->T) (n:odd_nat1t)
         (p':{m:odd_nat1t | on1t_le m n}->T),
    let f := even_nat1t_fun_unite_finite_odd p n p' in
      (forall (e:even_nat1t) 
              (pf:nat_of_en1t e <= nat_of_on1t n),
              p e = f (exist (fun m => (1 <= m <= nat_of_on1t n))
                             _ (conj (en1t_ge_1 e) pf))) /\
      (forall (o:odd_nat1t)
              (pf:nat_of_on1t o <= nat_of_on1t n),
              p' (exist (fun m => on1t_le m n) _ (nat_of_on1t_le_compat _ _ pf)) = 
              f (exist (fun m => (1 <= m <= nat_of_on1t n))
                       _ (conj (le_1_odd1t o) pf))).
intros T p n p' f. split.
intros e h1.
unfold f, even_nat1t_fun_unite_finite_odd.
destruct even_odd_dec as [h2 | h2].
f_equal.
destruct e as [e' e].
destruct e' as [e'' e'].
apply even_nat1t_functional. 
pose proof (nat1t_functional (nat1t_intro e')
                             (gt0_to_1t (en1t_ge_1 (even_nat1t_intro (nat1t_intro e') e)))) as h4.
simpl in h4.
hfirst h4.
  unfold gt0_to_1t.
  destruct Description.constructive_definite_description as [a h5].
  simpl.
  destruct a.
  simpl in h5. subst.
  reflexivity.
specialize (h4 hf).
assumption.

destruct e as [e' h3]. simpl in h2.
red in h3.
destruct e' as [e'' e'].
simpl in h2. red in h3.
pose proof (not_even_and_odd _ h3 h2). contradiction.

intros o h1.
unfold f, even_nat1t_fun_unite_finite_odd.
destruct even_odd_dec as [h2 | h2].
destruct o  as [o' h3]. destruct o' as [o'' o'].
simpl. simpl in h2.
red in h3. red in h3.
pose proof (not_even_and_odd _ h2 h3). contradiction.
f_equal.
apply proj1_sig_injective; simpl.
destruct o as [o' o]. destruct o' as [o'' o'].
apply odd_nat1t_functional.
pose proof (nat1t_functional (nat1t_intro o')
                             (odd_to_1t h2)) as h3.
apply h3. clear h3.
unfold odd_to_1t, gt0_to_1t.
destruct constructive_definite_description as [n' h3].
simpl.
destruct n'. simpl in h3.
subst.
reflexivity.
Qed.

(*Wyckoff*)
Definition even_nat1t_fun_unite_finite_odd'
           {T:Type} (p:even_nat1t->T) (n:even_nat1t)
           (p':{m:odd_nat1t | on1t_le m (en1t_pred n)}->T) : 
  seg_one_t (nat_of_en1t n) -> T.
intro m.
destruct m as [m h1].
destruct (even_odd_dec m) as [h2 | h2].
assert (h3:0 < m).
  destruct h1. auto with sets.
assert (h4:even1t (gt0_to_1t h3)).
  red.
  rewrite match_gt0_to_1t_even_iff.
  rewrite gt0_to_1t_compat.
  assumption.
refine (p (even_nat1t_intro _ h4)).
pose (odd_to_1t h2) as m'.
pose proof (odd_to_1t_compat h2) as h4.
pose (odd_nat1t_intro _ h4) as m''.
pose proof (nat_of_on1t_le_pred_compat m'' n) as h5.
assert (h6:nat_of_on1t m'' <= nat_of_en1t n).
  unfold m''. simpl.
  unfold odd_to_1t.
  rewrite gt0_to_1t_compat.
  destruct h1; auto.
specialize (h5 h6).
refine (p' (exist (fun m => on1t_le m (en1t_pred n)) _ h5)).
Defined.

(*Wyckoff*)
Lemma even_nat1t_fun_unite_finite_odd_compat' : 
  forall {T:Type} (p:even_nat1t->T) (n:even_nat1t)
         (p':{m:odd_nat1t | on1t_le m (en1t_pred n)}->T),
    let f := even_nat1t_fun_unite_finite_odd' p n p' in
      (forall (e:even_nat1t) 
              (pf:nat_of_en1t e <= nat_of_en1t n),
              p e = f (exist (fun m => (1 <= m <= nat_of_en1t n))
                             _ (conj (en1t_ge_1 e) pf))) /\
      (forall (o:odd_nat1t)
              (pf:nat_of_on1t o <= nat_of_en1t n),
              p' (exist (fun m => on1t_le m (en1t_pred n)) _ (nat_of_on1t_le_pred_compat _ _ pf)) = 
              f (exist (fun m => (1 <= m <= nat_of_en1t n))
                       _ (conj (le_1_odd1t o) pf))).
intros T p n p' f. split.
intros e h1.
unfold f, even_nat1t_fun_unite_finite_odd'.
destruct even_odd_dec as [h2 | h2].
f_equal.
destruct e as [e' e].
destruct e' as [e'' e'].
apply even_nat1t_functional.
apply nat1t_functional.
pose proof (nat1t_functional (nat1t_intro e')
                             (gt0_to_1t (en1t_ge_1 (even_nat1t_intro (nat1t_intro e') e)))) as h4.
hfirst h4.
unfold gt0_to_1t.
destruct Description.constructive_definite_description as [a h5].
simpl.
destruct a.
simpl in h5. subst.
reflexivity.
specialize (h4 hf).
assumption.

destruct e as [e' h3]. simpl in h2.
red in h3.
destruct e' as [e'' e'].
simpl in h2. red in h3.
pose proof (not_even_and_odd _ h3 h2). contradiction.

intros o h1.
unfold f, even_nat1t_fun_unite_finite_odd'.
destruct even_odd_dec as [h2 | h2].
destruct o  as [o' h3]. destruct o' as [o'' o'].
simpl. simpl in h2.
red in h3. red in h3.
pose proof (not_even_and_odd _ h2 h3). contradiction.
f_equal.
apply proj1_sig_injective; simpl.
destruct o as [o' o]. destruct o' as [o'' o'].
apply odd_nat1t_functional.
pose proof (nat1t_functional (nat1t_intro o')
                             (odd_to_1t h2)) as h3.
apply h3. clear h3.
unfold odd_to_1t, gt0_to_1t.
destruct constructive_definite_description as [n' h3].
simpl.
destruct n'. simpl in h3.
subst.
reflexivity.
Qed.

